/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    File        :   router_acces.c

    Date        :   02 04 94
    Version     :   1

    Auteur      :   BEAURIN oliver

    Description :   Create, delete and modify acces struct

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include MUT_H
#include MLO_H
#include MPH_H

#include "router_dpr.h"

/*---------------------------------------------------------\
                     globals variables
\---------------------------------------------------------*/
long            total_acc = 0l;

/*---------------------------------------------------------\
                     mkacces

Creation de la structure ACCESS
\---------------------------------------------------------*/
ACCESS         *mkacces (acc_next, acc_prev, pos, orient, signal, done)
ACCESS         *acc_next;
ACCESS         *acc_prev;
POSSIBILITY    *pos;
char            orient;
SIGNAL         *signal;
char            done;
{
    ACCESS         *ptacces;

    ptacces = (ACCESS *) mbkalloc (sizeof (ACCESS));

    ptacces->NEXT = acc_next;
    ptacces->PREV = acc_prev;
    ptacces->POS = pos;
    ptacces->ORIENT = orient;
    ptacces->SIG = signal;
    ptacces->DONE = done;

    total_acc++;

    return (ptacces);
}

/*---------------------------------------------------------\
                     del_acces

suprime proprement une structure acces de la liste
\---------------------------------------------------------*/
void            del_acces (noeud, acces)
NODE           *noeud;
ACCESS         *acces;

{
    /* transformation du chainage */
    if (acces->NEXT)
        acces->NEXT->PREV = acces->PREV;
    else
        noeud->DPR_RIGHT = acces->PREV;
    if (acces->PREV)
        acces->PREV->NEXT = acces->NEXT;
    else
        noeud->DPR_LEFT = acces->NEXT;

    total_acc--;

}

/*---------------------------------------------------------\
                     del_acces_from_sig
\---------------------------------------------------------*/
void del_acces_from_sig (noeud, ptsig)
NODE *noeud;
SIGNAL *ptsig;
{
    ACCESS         *acces1;
    ACCESS         *acces2;

    for( ; noeud->DPR_DWN; noeud = noeud->DPR_DWN );
    for( ; noeud; noeud = noeud->DPR_UP )
    {
        for (acces1 = noeud->DPR_LEFT; acces1; acces1 = acces2)
        {
            acces2 = acces1->NEXT;
            if ((!acces1->DONE) && (acces1->SIG == ptsig))
                del_acces (noeud, acces1);
        }
    }
}

/*---------------------------------------------------------\
                     del_acces_sup
\---------------------------------------------------------*/
void del_acces_sup ( ptnoeud, con, orient )
NODE *ptnoeud;
CONNECTOR *con;
char orient;
{
    NODE           *ptnoeud1;
    ACCESS         *acces1;
    ACCESS         *acces2;

    for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);
    for (; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        for (acces1 = ptnoeud1->DPR_LEFT; acces1; acces1 = acces2)
        {
            acces2 = acces1->NEXT;
            if ((!acces1->DONE) && (acces1->POS->CON == con) && (acces1->ORIENT == orient))
                del_acces (ptnoeud1, acces1);
        }
    }
}

/*---------------------------------------------------------\
                     supress_acces

suprime tous les acces UP et DWN latteraux
\---------------------------------------------------------*/
void            supress_acces (noeud, acces)
NODE           *noeud;
ACCESS         *acces;

{
    NODE           *noeud1 = NULL;
    CONNECTOR      *con1 = NULL;
    ACCESS         *acces2 = NULL;
    ACCESS         *acces3 = NULL;
    char            orient1 = 'X';

    /* l'acces durcis devient obstacle */
    con1 = acces->POS->CON;
    orient1 = acces->ORIENT;
    acces->DONE = TRUE;
    total_acc--;

    /* on suprime les autres acces du connecteur... */
    for (noeud1 = noeud; noeud1->DPR_DWN; noeud1 = noeud1->DPR_DWN);

    for (; noeud1; noeud1 = noeud1->DPR_UP)
    {
        for (acces2 = noeud1->DPR_LEFT; acces2; acces2 = acces3)
        {
            acces3 = acces2->NEXT;
            if ((!acces2->DONE) &&
                (acces2->ORIENT == orient1) &&
                (acces2->POS->CON == con1))
            {
                    del_acces (noeud1, acces2);
            }
        }
    }
}

