/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/********************************************************************
*                                                                   *
* Laboratoire MASI CAO-VLSI, UPMC, Copyright 1991 1992 1993 1994    *
*                                                                   *
* Software support Email : cao-vlsi@masi.ibp.fr                     *
*                                                                   *
* Authors : Zouina AKTOUF &                                         *
*           El Arabi RHOMARI &                                      *
*           Jose MARTINS DOS SANTOS                                 *
*                                                                   *
* Supervision & Modifications : Lotfi BEN AMMAR                     *
*                                                                   *
* Restructuration : Olivier BEAURIN                                 *
*                                                                   *
********************************************************************/

/****************************************************************************
*                      Fichier: router_load.c                               *
*                                                                           *
*       version : 3.0  date    : 17/03/92                                   *
*                                                                           *
*       version : 3.1  date    : 27/03/94                                   *
*                                                                           *
****************************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include MLO_H
#include MPH_H
#include GFP_H

#include "dpr.h"
#include "router_dpr.h"
#include "router_util.h"

/*---------------------------------------------------------\
                     globals variables
\---------------------------------------------------------*/
chain_list *unused_signals = NULL;

/*---------------------------------------------------------\
                     unifsegs
\---------------------------------------------------------*/
void            unifsegs (ptfigf_ph)
phfig_list     *ptfigf_ph;
{
	phseg_list     *ptseg1 = NULL;
	phseg_list     *ptseg2 = NULL;
	phseg_list     *ptsegtmp = NULL;

	for(ptseg1=ptfigf_ph->PHSEG; ptseg1; ptseg1=ptseg1->NEXT)
		if (ptseg1->LAYER == (char)TALU2)
		{
			if ((ptseg1->TYPE == UP) || (ptseg1->TYPE == DOWN))
			{
                fflush( stdout );
				fprintf(stderr, "Warning: Allow-metal2 segment in ");
				fprintf(stderr, "%s", ptfigf_ph->NAME);
				fprintf(stderr, " not parallel to ALU2 alims\n");

				continue;
			}

			for(ptseg2=ptseg1->NEXT; ptseg2; ptseg2=ptsegtmp)
			{
				ptsegtmp = ptseg2->NEXT;

				if (ptseg2->LAYER == (char)TALU2)
				{
					if ((ptseg2->TYPE == UP) || (ptseg2->TYPE == DOWN))
					{
                        fflush( stdout );
						fprintf(stderr, "Warning: Allow-metal2 segment in ");
						fprintf(stderr, "%s", ptfigf_ph->NAME);
						fprintf(stderr, " not parallel to ALU2 alims\n");

						continue;
					}

					if (ptseg2->Y1 == ptseg1->Y1)
						if (   (ptseg1->X1 <= ptseg2->X2)
							&& (ptseg2->X1 <= ptseg1->X2))
						{
							ptseg1->X1 = dpr_min(ptseg1->X1, ptseg2->X1);
							ptseg1->X2 = dpr_max(ptseg1->X2, ptseg2->X2);
							delphseg(ptfigf_ph, ptseg2);
						}
				}
			}
		}
}

/*---------------------------------------------------------\
                     setaccess
\---------------------------------------------------------*/
void            setaccess (ptfig_ph)
phfig_list     *ptfig_ph;

{
	chain_list     *ptcmodel;
	char           *ptmodel;
	char           *ref_con;
	char           *ref_ref;
	char           *ref_type;
	phfig_list     *ptfigf_ph;
	phseg_list     *ptseg;
	phref_list     *ptref;
	phref_list     *ptreftmp;
	ptype_list     *ptype;
	int             nberr = 0;

	ref_con = namealloc("ref_con");
	ref_ref = namealloc("ref_ref");

	for(ptcmodel=ptfig_ph->MODELCHAIN; ptcmodel; ptcmodel=ptcmodel->NEXT)
	{
		ptmodel   = (char*)ptcmodel->DATA;
		ptfigf_ph = getphfig(ptmodel, 'P');

        if( verbose_mode == VERBOSE_ALL )
        {
            fprintf( stdout, "      -> %s.\n", ptmodel );
            fflush( stdout );
        }

		unifsegs (ptfigf_ph);

		for(ptseg=ptfigf_ph->PHSEG; ptseg; ptseg=ptseg->NEXT)
		{
			if (       (ptseg->LAYER == (char)TALU2)
			    && (   (ptseg->TYPE  ==       RIGHT)
					|| (ptseg->TYPE  ==        LEFT)) )
			{
				if (   (ptseg->X1 == ptfigf_ph->XAB1)
					&& (ptseg->X2 == ptfigf_ph->XAB2))
					ref_type = namealloc("ref_a2rl");
				else
				{
					if (   (ptseg->X1 == ptfigf_ph->XAB1)
						&& (ptseg->X2  < ptfigf_ph->XAB2))
						ref_type = namealloc("ref_a2l");
					else
					{
						if (   (ptseg->X1  > ptfigf_ph->XAB1)
							&& (ptseg->X2 == ptfigf_ph->XAB2))
							ref_type = namealloc("ref_a2r");
						else
						{
                            fflush( stdout );
							fprintf(stderr, "Warning: Allow-metal2 segment ");
							fprintf(stderr, "in %s not valid\n",
												ptfigf_ph->NAME);
							break;
						}
					}
				}

				for(ptref=ptfigf_ph->PHREF; ptref; ptref=ptreftmp)
				{
					ptreftmp = ptref->NEXT;

					if (ptref->FIGNAME == ref_con)
					{
						if (   (ptref->YREF == ptseg->Y1)
							&& (ptref->XREF >= ptseg->X1)
							&& (ptref->XREF <= ptseg->X2))
							ptref->USER = addptype(ptref->USER,
												   (long)SETACCESS,
												   (void*)ref_type);
					}
					else
						if (   (ptref->FIGNAME != ref_ref)
							|| (strncmp(ptref->NAME,"nwell",5)))
							delphref(ptfigf_ph, ptref);
				}
			}
		}

		for(ptref=ptfigf_ph->PHREF; ptref; ptref=ptref->NEXT)
			if (ptref->FIGNAME == ref_con)
			{
				ptype = getptype(ptref->USER, (long)SETACCESS);

				if (ptype == NULL)
				{
                    fflush( stdout );
					fprintf(stderr, "Error: the virtual terminal ");
					fprintf(stderr, "%s in %s is not valid\n",
												  ptref->NAME,
											  ptfigf_ph->NAME);
					nberr++;
				}
			}
	}

	if (nberr != 0)
	{
        fflush( stdout );
		fprintf(stderr, "sorry, can not continue!!\n");
		exit (1);
	}
}


/*  Fonction  :  'concompare'.                                    */
/* -------------------------------------------------------------- */
/*  Verifie la coherence entre la liste des connecteurs logiques,
 *  argument "ptlocon", et celle des connecteurs physiques,
 *  argument "ptphcon".
 *      Les connecteurs physiques qui ont pu etre rattache a un
 *  connecteur logique sont marque en ajoutant au champs USER
 *  un ptype "SESAME" (dont la DATA vaut "YES").
 *      On ajoute au champs USER des connecteurs logiques la
 *  liste des connecteurs physiques qui lui sont associe.
 *      Chaque connecteur physique doit etre associe a un
 *  connecteur logique, et au moin un connecteur physique a
 *  chaque connecteur logique.
 *      Les alimentations sont une exeption, elles peuvent
 *  ne pas avoir de connecteur physoque asocie.
 */

long            concompare (ptphcon, ptlocon)
phcon_list     *ptphcon;
locon_list     *ptlocon;

{
	phcon_list     *ptcphcon;
	locon_list     *ptclocon;
	int             trouve;
	ptype_list     *ptype = NULL;
	chain_list     *ptchain = NULL;
	long            nberr = 0;

	for (ptclocon = ptlocon; ptclocon; ptclocon = ptclocon->NEXT)
	{
		trouve = 0;
        ptchain = NULL;
		for (ptcphcon = ptphcon; ptcphcon; ptcphcon = ptcphcon->NEXT)
		{
			if (ptcphcon->NAME == ptclocon->NAME)
			{
				trouve = 1;
				ptcphcon->USER = addptype (ptcphcon->USER, (long) SESAME, (void *) YES);
				ptchain = addchain (ptchain, (void *) ptcphcon);
			}
		}
		if (trouve != 1)
		{
			if (!isvdd (ptclocon->NAME) && !isvss (ptclocon->NAME))
            {
                fflush( stdout );
				fprintf (stderr, "Warning: No physical connector corresponding to %s\n",
					 ptclocon->NAME);
            }
		
		}
		else
	
			ptclocon->USER = addptype (ptclocon->USER, (long) SESAME, (void *) ptchain);
	}

	for (ptcphcon = ptphcon; ptcphcon; ptcphcon = ptcphcon->NEXT)
	{
		ptype = getptype (ptcphcon->USER, (long) SESAME);
		if (ptype == NULL)
		{
            fflush( stdout );
			fprintf (stderr, "Error: no logical connector corresponding to %s\n",
				 ptcphcon->NAME);
			nberr++;
		}
	}
	return (nberr);
}

/*---------------------------------------------------------\
                     inscompare
\---------------------------------------------------------*/
long            inscompare (ptphins, ptloins)
phins_list     *ptphins;
loins_list     *ptloins;

{
	phins_list     *ptcphins;
	loins_list     *ptcloins;
	ptype_list     *ptype;
	long            nberr = 0;

	for (ptcloins = ptloins; ptcloins; ptcloins = ptcloins->NEXT)
	{
		for (ptcphins = ptphins; ptcphins; ptcphins = ptcphins->NEXT)
		{
			if (ptcphins->INSNAME == ptcloins->INSNAME)
			{
				ptcphins->USER = addptype (ptcphins->USER, (long) SESAME, (void *) NULL);
				break;
			}
		}
		if (ptcphins == NULL)
		{
            fflush( stdout );
			fprintf (stderr, "Error: no physical instance corresponding to %s\n",
				 ptcloins->INSNAME);
			nberr++;
		}
	}

	for (ptcphins = ptphins; ptcphins; ptcphins = ptcphins->NEXT)
	{
		ptype = getptype (ptcphins->USER, (long) SESAME);
		if (ptype == NULL)
		{
            fflush( stdout );
			fprintf (stderr, "Error: no logical instance corresponding to %s\n",
				 ptcphins->INSNAME);
			nberr++;
		}
	}
	return (nberr);
}

/*---------------------------------------------------------\
                     sigcheck
\---------------------------------------------------------*/
long            sigcheck (ptfig_lo)
lofig_list     *ptfig_lo;

{
	long            nberr = 0;
	losig_list     *ptclosig, *ptlosig;
	locon_list     *ptlocon;
	ptype_list     *ptype = NULL;
	chain_list     *ptchain = NULL;
	long            nbcon = 0;
	long            nbconex = 0;

	ptlosig = ptfig_lo->LOSIG;
	for (ptclosig = ptlosig; ptclosig; ptclosig = ptclosig->NEXT)
	{
		ptype = getptype (ptclosig->USER, (long) LOFIGCHAIN);
		nbcon = 0;
		nbconex = 0;
		for (ptchain = (chain_list *) ptype->DATA; ptchain; ptchain = ptchain->NEXT)
		{
			ptlocon = (locon_list *) ptchain->DATA;
			if (ptlocon->TYPE == EXTERNAL)
			{
				nbconex++;
			}
			nbcon++;
		}
		if (nbconex > 1)
		{
            fflush( stdout );
			fprintf (stderr, "Error: signal %s connected to more than one external logical terminal!!\n",
				 getsigname (ptclosig));
			nberr++;
		}
		if (nbcon < 2)
		{
            unused_signals = addchain( unused_signals, ptclosig->INDEX );
            if( verbose_mode == VERBOSE_ALL )
            {
			     fprintf (stdout, "      -> Signal %s not connected to any other terminal.\n",
				          getsigname (ptclosig));
                fflush( stdout );
            }
		}
	}
	return (nberr);
}

/*---------------------------------------------------------\
                     sesame
\---------------------------------------------------------*/
void            sesame (ptfig_ph, ptfig_lo)
phfig_list     *ptfig_ph;
lofig_list     *ptfig_lo;

{
	phcon_list     *ptphcon;
	locon_list     *ptlocon;
	phins_list     *ptphins;
	loins_list     *ptloins;
	long            nberr = 0;

	if ((ptfig_ph->PHSEG != NULL) || (ptfig_ph->PHVIA != NULL) || (ptfig_ph->PHREF != NULL))
	{
        fflush( stdout );
		fprintf (stderr, "Error: segment, via and reference lists of figure must be NULL!!\n");
		exit (1);
	}

	lofigchain (ptfig_lo);
	ptphcon = ptfig_ph->PHCON;
	ptlocon = ptfig_lo->LOCON;
	ptphins = ptfig_ph->PHINS;
	ptloins = ptfig_lo->LOINS;
	nberr = concompare (ptphcon, ptlocon);
	nberr += inscompare (ptphins, ptloins);
	nberr += sigcheck (ptfig_lo);
	if (nberr != 0)
	{
        fflush( stdout );
		fprintf (stderr, "Sorry, can not continue because of %ld errors !!!\n", nberr);
		exit (1);
	}
}

/*---------------------------------------------------------\
                     addcol
\---------------------------------------------------------*/
chain_list     *addcol (pthead, type, x1, y1, x2, y2)
chain_list     *pthead;
char            type;
long            x1, y1, x2, y2;

{
	COLUMN         *ptnewcol;

	ptnewcol = (COLUMN *) mbkalloc (sizeof (COLUMN));
	ptnewcol->NAME = NULL;
	ptnewcol->TYPE = type;
	ptnewcol->X = x1;
	ptnewcol->Y = y1;
	ptnewcol->WIDTH = x2 - x1;
	ptnewcol->S_WIDTH = 0;
	ptnewcol->HEIGHT = y2 - y1;
	ptnewcol->S_HEIGHT = 0;
	ptnewcol->INDEX = 0;
	ptnewcol->SIG = NULL;
	if (type == 'C')
		ptnewcol->DF_CON = (chain_list *) addchannel (NULL, NULL, NULL, NULL, NULL);
	else
		ptnewcol->DF_CON = NULL;
	ptnewcol->CT_CON = NULL;
	ptnewcol->ALIMS = NULL;
	ptnewcol->INS = NULL;
	ptnewcol->VIA = NULL;
	ptnewcol->SEG = NULL;
	ptnewcol->ALLOW = NULL;
	ptnewcol->RDATA = NULL;
	ptnewcol->ALLOWUSED = NULL;
	ptnewcol->USER = NULL;

	return (addchain (pthead, (void *) ptnewcol));
}

/*---------------------------------------------------------\
                     mkcolumns
\---------------------------------------------------------*/
void            mkcolumns (ptfig_ph, ptfig)
phfig_list     *ptfig_ph;
FIGURE         *ptfig;

{
	phfig_list     *ptfigf_ph;
	phins_list     *ptins;
	COLUMN         *ptcol, *ptnextcol;
	chain_list     *ptccol, *ptsav, *pthead, *ptnext, *ptfree;
	long            xab2, xab1, yab2, yab1;
	long            x1col, x2col;
	long            y1ins, y2ins, x1ins, x2ins;
	long            x1next, x2next;
	long            xleft, xright;
	long            xmax = MINLONG, ymax = MINLONG, xmin = MAXLONG, ymin = MAXLONG;
	int             index = 0;
    int             finished;

	xab1 = ptfig_ph->XAB1;
	xab2 = ptfig_ph->XAB2;
	yab1 = ptfig_ph->YAB1;
	yab2 = ptfig_ph->YAB2;
	if (((xab2 - xab1) <= 0) || ((yab2 - yab1) <= 0))
	{
        fflush( stdout );
		fprintf (stderr, "Error: Not valid abutment box coordinates\n");
		exit (1);
	}
	xleft = xab1;
	xright = xab2;

	/* building  perephiric Right Column */
	pthead = addcol ((chain_list *) NULL, 'R', xright, yab1, xright, yab2);

	/* building  cell lines   */
	for (ptins = ptfig_ph->PHINS; ptins; ptins = ptins->NEXT)
	{
		ptfigf_ph = getphfig (ptins->FIGNAME, 'P');
		x1ins = ptins->XINS;
		x2ins = ptins->XINS + deltab_x (ptfigf_ph->XAB1, ptfigf_ph->YAB1, ptfigf_ph->XAB2,
					    ptfigf_ph->YAB2, ptins->TRANSF);
		y1ins = ptins->YINS;
		y2ins = ptins->YINS + deltab_y (ptfigf_ph->XAB1, ptfigf_ph->YAB1, ptfigf_ph->XAB2,
					    ptfigf_ph->YAB2, ptins->TRANSF);

		/* pour tester si l'abutment box englobe toutes les instances */
		xmin = dpr_min (xmin, x1ins);
		xmax = dpr_max (xmax, x2ins);
		ymin = dpr_min (ymin, y1ins);
		ymax = dpr_max (ymax, y2ins);

		for (ptccol = pthead; ptccol; ptccol = ptccol->NEXT)
		{
			ptcol = (COLUMN *) ptccol->DATA;
			x1col = ptcol->X;
			x2col = x1col + ptcol->WIDTH;
			if (x2col > x1ins)
				break;
			ptsav = ptccol;
		}
		if (x2ins <= x1col)	/* insert new column  */
		{
			if (ptccol == pthead)
			{
				pthead = addcol (ptccol, 'O', x1ins, yab1, x2ins, yab2);
			}
			else
			{
				ptsav->NEXT = addcol (ptccol, 'O', x1ins, yab1, x2ins, yab2);
			}
		}
        else
			/* modify the column  */
		{
			x1col = ((x1ins < x1col) ? x1ins : x1col);
			x2col = ((x2ins > x2col) ? x2ins : x2col);
			ptcol->X = x1col;
			ptcol->WIDTH = x2col - x1col;

            do {
                finished = TRUE;
                for (ptccol = pthead, ptnext = ptccol->NEXT;
                     ptnext;
                     ptccol = ptnext, ptnext = ptnext->NEXT)
                {
                    ptcol = (COLUMN *) ptccol->DATA;
                    x1col = ptcol->X;
                    x2col = x1col + ptcol->WIDTH;

                    ptnextcol = (COLUMN *) ptnext->DATA;
                    x1next = ptnextcol->X;
                    x2next = x1next + ptnextcol->WIDTH;

                    if( x1next < x2col )
                    {
                        x1col = ((x1next < x1col) ? x1next : x1col);
                        x2col = ((x2next > x2col) ? x2next : x2col);

                        ptcol->X = x1col;
                        ptcol->WIDTH = x2col - x1col;

                        ptccol->NEXT = ptnext->NEXT;
                        finished = FALSE;

                        ptnext->NEXT = NULL;
                        mbkfree(ptnext->DATA);
                        freechain(ptnext);
                        break;
                    }
                }
            } while( !finished );
		}
	}

	if ((xmin < xab1) || (xmax > xab2) || (ymin < yab1) || (ymax > yab2))
	{
        fflush( stdout );
		fprintf (stderr, "Error: All the instances must be inside the abutment box!!\n");
		exit (1);
	}

	/* building  perepherical  Left column  */
	pthead = addcol (pthead, 'L', xleft, yab1, xleft, yab2);

	/* building Channel columns  */
	for (ptccol = pthead; ptccol->NEXT; ptccol = ptsav)
	{
		ptsav = ptccol->NEXT;
		ptcol = (COLUMN *) ptccol->DATA;
		x1col = ptcol->X + ptcol->WIDTH;
		ptnextcol = (COLUMN *) ptsav->DATA;
		x2col = ptnextcol->X;
		ptccol->NEXT = addcol (ptsav, 'C', x1col, yab1, x2col, yab2);
	}

	/* indexing  lines  */
	for (ptccol = pthead; ptccol; ptccol = ptccol->NEXT)
	{
		ptcol = (COLUMN *) ptccol->DATA;
		ptcol->INDEX = ++index;
		if (ptcol->TYPE == 'C')
			ptcol->NAME = nameindex ("DPR_C", index);
		else if (ptcol->TYPE == 'O')
			ptcol->NAME = nameindex ("DPR_O", index);
		else if (ptcol->TYPE == 'R')
			ptcol->NAME = nameindex ("DPR_R", index);
		else if (ptcol->TYPE == 'L')
			ptcol->NAME = nameindex ("DPR_L", index);
		else
		{
            fflush( stdout );
			fprintf (stderr, "BUG: Column type not valid!!\n");
			exit (1);
		}

	}
	ptfig->COL = pthead;
}

/*---------------------------------------------------------\
                     addins
\---------------------------------------------------------*/
chain_list     *addins (pthead, insname, x, y, width, height, figname, transf)
chain_list     *pthead;
char           *insname;
char           *figname;
char            transf;
long            x, y, width, height;

{
	INSTANCE       *ptnewcell;

	ptnewcell = (INSTANCE *) mbkalloc (sizeof (INSTANCE));
	ptnewcell->INSNAME = insname;
	ptnewcell->FIGNAME = figname;
	ptnewcell->X = x;
	ptnewcell->Y = y;
	ptnewcell->WIDTH = width;
	ptnewcell->HEIGHT = height;
	ptnewcell->TRANSF = transf;
	ptnewcell->DF_CON = NULL;
	ptnewcell->CT_CON = NULL;
	ptnewcell->ALIMS = NULL;
	ptnewcell->RAPPELS = NULL;
	ptnewcell->ALLOW = NULL;
	ptnewcell->N_WELL = NULL;
	ptnewcell->USER = NULL;

	return (addchain (pthead, (void *) ptnewcell));
}

/*---------------------------------------------------------\
                     insertins
\---------------------------------------------------------*/
INSTANCE       *insertins (ptcol, insname, x, y, width, height, figname, transf)
COLUMN         *ptcol;
char            transf;
char           *insname;
char           *figname;
long            x, y, width, height;

{
	INSTANCE       *ptcell = NULL;
	chain_list     *ptsavccell = NULL, *ptccell = NULL;
	void           *ptnewcell = NULL;

	for (ptccell = ptcol->INS; ptccell; ptccell = ptccell->NEXT)
	{
		ptcell = (INSTANCE *) ptccell->DATA;
		if (ptcell->Y >= y)
			break;
		else if ((ptcell->Y + ptcell->HEIGHT) > y)
		{
            fflush( stdout );
			fprintf (stderr,"Error 1: X or Y overlapping between instances %s and %s\n"
				,insname, ptcell->INSNAME);
			exit (1);
		}
		ptsavccell = ptccell;
	}
	if (ptsavccell == NULL)
	{
		ptcol->INS = addins (ptcol->INS, insname, x, y, width, height, figname, transf);
		ptnewcell = ptcol->INS->DATA;
	}
	else
	{
		ptsavccell->NEXT = addins (ptsavccell->NEXT, insname, x, y, width,
					   height, figname, transf);
		ptnewcell = ptsavccell->NEXT->DATA;
	}
	for (; ptccell; ptccell = ptccell->NEXT)
	{
		ptcell = (INSTANCE *) ptccell->DATA;
		if (y != ptcell->Y)
		{
            break;
		}
	}

	for (; ptccell; ptccell = ptccell->NEXT)
	{
		ptcell = (INSTANCE *) ptccell->DATA;
		if ((y + height) > ptcell->Y)
		{
            fflush( stdout );
			fprintf (stderr, "Error 2: X or Y overlapping between instances %s and %s\n"
				 ,insname, ptcell->INSNAME);
			exit (1);
		}
		else if ((y + height) < ptcell->Y)
			break;
	}

	return ((INSTANCE *) ptnewcell);
}

/*---------------------------------------------------------\
                     dprcol
\---------------------------------------------------------*/
void            dprcol (ptfig_ph, ptfig_lo, ptroot)
phfig_list     *ptfig_ph;
lofig_list     *ptfig_lo;
FIGURE         *ptroot;

{
	COLUMN         *ptcol = NULL;
	INSTANCE       *ptcell = NULL;
	chain_list     *ptccol = NULL;
	phins_list     *ptins_ph = NULL;
	phfig_list     *ptfigf_ph = NULL;
	loins_list     *ptins_lo = NULL;
	long            xlins_col, xrins_col;
	long            xlins_pere, xrins_pere, yuins_pere, ydins_pere;
	long            xrcol_pere, xlcol_pere;

	/* scan  physical instances list        */
	for (ptins_ph = ptfig_ph->PHINS; ptins_ph; ptins_ph = ptins_ph->NEXT)
	{
		ptfigf_ph = getphfig (ptins_ph->FIGNAME, 'P');
		xlins_pere = ptins_ph->XINS;
		ydins_pere = ptins_ph->YINS;
		xrins_pere = xlins_pere + deltab_x (ptfigf_ph->XAB1, ptfigf_ph->YAB1,
			ptfigf_ph->XAB2, ptfigf_ph->YAB2, ptins_ph->TRANSF);
		yuins_pere = ydins_pere + deltab_y (ptfigf_ph->XAB1, ptfigf_ph->YAB1,
			ptfigf_ph->XAB2, ptfigf_ph->YAB2, ptins_ph->TRANSF);
		/* find corresponding line   */
		for (ptccol = ptroot->COL->NEXT->NEXT; ptccol->NEXT; ptccol = ptccol->NEXT->NEXT)
		{
			ptcol = (COLUMN *) ptccol->DATA;
			xlcol_pere = ptcol->X;
			xrcol_pere = xlcol_pere + ptcol->WIDTH;
			if ((xlcol_pere <= xlins_pere) && (xrcol_pere >= xrins_pere))
				break;
		}
		if (ptccol->NEXT == NULL)
		{
            fflush( stdout );
			fprintf (stderr, "BUG: there is a problem with mkcolumns and dprcol!!\n");
			exit (1);
		}

		xrins_col = xrins_pere - xlcol_pere;
		xlins_col = xlins_pere - xlcol_pere;
		/* add instance in line       */
		ptcell = insertins (ptcol, ptins_ph->INSNAME, xlins_col, ydins_pere
			     ,xrins_col - xlins_col, yuins_pere - ydins_pere
				    ,ptins_ph->FIGNAME, ptins_ph->TRANSF);

		/* On garde les modeles physique et logique de l'instance */
		ptcell->USER = addchain (ptcell->USER, (void *) ptfigf_ph);
		ptins_lo = getloins (ptfig_lo, ptins_ph->INSNAME);
		ptcell->USER = addchain (ptcell->USER, (void *) ptins_lo);
	}

}

/*---------------------------------------------------------\
                     loaddprfig
\---------------------------------------------------------*/
FIGURE         *loaddprfig (ptname)
char           *ptname;

{
	FIGURE         *ptroot = NULL;
	lofig_list     *ptfig_lo = NULL;
	phfig_list     *ptfig_ph = NULL;

    if( verbose_mode )
    {
        fprintf( stdout, "   o Loading physical and logical views.\n" );
        fflush( stdout );
    }

	/* PATCH JP : Mes fonctions de flattens.
	 */
	ptfig_lo = getlofig (ptname, 'A');
	UtHir_rFlattenCatalLoFig (ptfig_lo);
	ptfig_ph = getphfig (ptname, 'A');
	UtHir_rFlattenCatalPhFig (ptfig_ph);

    if( verbose_mode )
    {
        fprintf( stdout, "   o Setting access types.\n" );
    }
	setaccess (ptfig_ph);
    if( verbose_mode )
    {
        fprintf( stdout, "   o Checking logical and physical correspondance.\n" );
    }
	sesame (ptfig_ph, ptfig_lo);

	ptroot = (FIGURE *) mbkalloc (sizeof (FIGURE));
	ptroot->LIN = NULL;
	ptroot->COL = NULL;
	ptroot->SIG = NULL;
	ptroot->CON = NULL;
	ptroot->ALLOW = NULL;
	ptroot->ALIM = NULL;
	ptroot->GRID = NULL;

    if( verbose_mode )
    {
        fprintf( stdout, "   o Making DPR data structure.\n" );
    }
	mkcolumns (ptfig_ph, ptroot);
	dprcol (ptfig_ph, ptfig_lo, ptroot);
	sigconfils (ptroot);
	sigconpere (ptfig_lo, ptfig_ph, ptroot);

	/* Liberation de la memoire occupee par la structure MBK */
	/* apres avoir fini la construction de la structure DPR  */
	while (HEAD_PHFIG)
		delphfig (HEAD_PHFIG->NAME);
	while (HEAD_LOFIG)
		dellofig (HEAD_LOFIG->NAME);

	return (ptroot);
}
