/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/********************************************************************
*                                                                   *
* Laboratoire MASI CAO-VLSI, UPMC, Copyright 1991 1992 1993 1994    *
*                                                                   *
* Software support Email : cao-vlsi@masi.ibp.fr                     *
*                                                                   *
* Authors : Zouina AKTOUF &                                         *
*           El Arabi RHOMARI &                                      *
*           Jose MARTINS DOS SANTOS                                 *
*                                                                   *
* Supervision & Modifications : Lotfi BEN AMMAR                     *
*                                                                   *
* Restructuration : Olivier BEAURIN                                 *
*                                                                   *
********************************************************************/
/********************************************************************
*                                                                   *
*                              router_router.c                      *
*                                                                   *
* Version 3.0    28/03/92                                           *
* Version 3.1    16/03/94                                           *
*                                                                   *
* Main routing functions.                                           *
*                                                                   *
********************************************************************/

/*\
 *
 * Optimiser si possible le choix des traversees ( au niveau du nbre d'acces bouffe ).
 *
 * Optimiser le choix des acces : acces detruit min
 *
 * Reflechir a la notion d'ordre durant le routage.
 *
 * le test sur la non conectivite des acces lors de la construction de
 * la base de donnees en foireux.
 *
\*/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include MLO_H
#include MPH_H

#include "dpr.h"
#include "router_dpr.h"
#include "router_util.h"
#include "router_init_router.h"
#include "router_acces.h"
#include "router_phwork.h"

/*---------------------------------------------------------\
                     macros
\---------------------------------------------------------*/
#ifdef DEBUG
#  define PRINT_POINT { printf("."); fflush(stdout);} printf("\n");
#else
#  define PRINT_POINT ;
#endif

/*---------------------------------------------------------\
                     locals
\---------------------------------------------------------*/
chain_list *message_connectivity = NULL;    /* Bon, c'est moche je sais... mais c'est quand meme plus simple...*/

/*---------------------------------------------------------\
                     check_and_add_message
\---------------------------------------------------------*/
int check_and_add_message( name )
char *name;
{
    chain_list     *ptchain;

    for (ptchain = message_connectivity; ptchain; ptchain = ptchain->NEXT)
    {
        if ((char *)ptchain->DATA == name)
            return TRUE;
    }
    
    message_connectivity = addchain( message_connectivity, name );
    return FALSE;
} /* end of check_and_add_message */

/*---------------------------------------------------------\
                     in_con_chain
\---------------------------------------------------------*/
int             in_con_chain (chainlist, con)
chain_list     *chainlist;
CONNECTOR      *con;
{
    chain_list     *ptchain;

    for (ptchain = chainlist; ptchain; ptchain = ptchain->NEXT)
    {
        if ((CONNECTOR *)ptchain->DATA == con)
            return TRUE;
    }
    return FALSE;
}

/*---------------------------------------------------------\
                     durcir

fonction durcir() : Pour le traitement d'un connecteur
multiacce : -ajout du via.
            -ajout du segment.
            -insertin du con-sig dans le canal.
\---------------------------------------------------------*/
void            durcir (ptccolhead, ptacces)
chain_list     *ptccolhead;
ACCESS         *ptacces;

{
    chain_list     *ptccol_c = NULL;
    chain_list     *ptccol_o = NULL;
    chain_list     *ptcvia = NULL;
    chain_list     *ptcvia_save = NULL;
    COLUMN         *ptcol_o;
    COLUMN         *ptcol_c;
    SIGNAL         *ptsig = NULL;
    CHANNEL        *ptchannel;
    VIA            *ptvia = NULL;
    long            x = 0, y = 0, x1 = 0, x2 = 0;
    char            orient;
    char            type;

    ptccol_o = ptacces->POS->CON->COL;
    ptcol_o = (COLUMN *) ptccol_o->DATA;
    ptsig = ptacces->SIG;
    x = ptacces->POS->X;
    y = ptacces->POS->Y;
    orient = ptacces->ORIENT;
    type = ptacces->POS->TYPE;

    if (orient == EAST)
    {
        ptccol_c = ptccol_o->NEXT;
        ptcol_c = (COLUMN *) ptccol_c->DATA;
        x1 = x;
        x2 = ptcol_o->WIDTH;
    }
    else
    {
        ptccol_c = chain_pred (ptccolhead, ptccol_o);
        ptcol_c = (COLUMN *) ptccol_c->DATA;
        x1 = 0;
        x2 = x;
    }

    /* insertion et tri si necessaire selon y du via dans la colonne. */
    if (type == MULTIAX)
    {
        for (ptcvia = ptcol_o->VIA; ptcvia; ptcvia = ptcvia->NEXT)
        {
            ptvia = (VIA *) ptcvia->DATA;
            if (ptvia->Y > y)
                break;
            else if ((ptvia->X == x) && (ptvia->Y == y))
            {
                ptcvia_save = ptcvia;
                break;
            }
            else
                ptcvia_save = ptcvia;
        }
        if (ptcvia_save)
        {
            if ((ptvia->X != x) || (ptvia->Y != y))
                ptcvia_save->NEXT = addvia (ptcvia_save->NEXT, x, y, (char *) CONT_VIA);
        }
        else
            ptcol_o->VIA = addvia (ptcol_o->VIA, x, y, (char *) CONT_VIA);
    }

    /*
     * insertion du segment si necessaire, par trie selon y, dans la colonne.
     */
    if (x1 != x2)
        insertseg (ptcol_o, x1, y, x2, y, ALU2, WLAYER2, ptsig->NAME);

    /* insertion du connecteur-signal dans le canal correspondant */
    ptchannel = (CHANNEL *) ptcol_c->DF_CON->DATA;
    insertcon_inchan (ptchannel, orient, ptsig->INDEX, y);

}


/*---------------------------------------------------------\
                     liber_noeud
\---------------------------------------------------------*/
void            liber_noeud (noeud, acces, ptccol)
NODE           *noeud;
ACCESS         *acces;
chain_list     *ptccol;

{
    ACCESS         *acces1 = NULL;
    char            orient;

    orient = acces->ORIENT;
    if (orient == EAST)         /* c.a.d durcir vers la droite */
    {
        /* on supprime (ou on durcis) ce qui est plus a droite */
        for (acces1 = acces; acces1->NEXT;)
        {
            del_acces (noeud, acces1->NEXT);
        }

        /* on supprime ce qui est plus a gauche */
        for (acces1 = noeud->DPR_LEFT; acces1 != acces;)
        {
            acces1 = acces1->NEXT;
            /* test si meme orient sur un autre connecteur */
            if ((acces1->PREV->ORIENT == orient) && (acces1->PREV->SIG != acces->SIG))
            {
                /* meme ORIENT autre SIG */
                del_acces (noeud, acces1->PREV);
            }
        }
    }
    else
        /* c.a.d durcir vers la gauche */
    {
        /* on supprime (ou on durcis) ce qui est plus a gauche */
        for (acces1 = acces; acces1->PREV;)
        {
            del_acces (noeud, acces1->PREV);
        }

        /* on supprime ce qui est plus a droite */
        for (acces1 = noeud->DPR_RIGHT; acces1 != acces;)
        {
            /* test si meme signal sur un autre connecteur */
            acces1 = acces1->PREV;
            if ((acces1->NEXT->ORIENT == orient) && (acces1->NEXT->SIG != acces->SIG))
            {
                /* meme ORIENT autre SIG */
                del_acces (noeud, acces1->NEXT);
            }
        }
    }
    durcir (ptccol, acces);
    supress_acces (noeud, acces);
}

/*---------------------------------------------------------\
                     liber_noeud_in_op
\---------------------------------------------------------*/
void            liber_noeud_in_op (noeud, acces1, acces2)
NODE           *noeud;
ACCESS         *acces1;
ACCESS         *acces2;
{
    NODE           *ptnoeud;
    CONNECTOR      *con1;
    CONNECTOR      *con2;
    ACCESS         *acces;
    ACCESS         *acces1b;
    ACCESS         *acces2b;
    char           *sig_name;
    COLUMN         *ptcol;
    long            x1;
    long            x2;
    long            y1;
    long            y2;

    ptcol = (COLUMN *) acces1->POS->CON->COL->DATA;

    /* on supprime les extremites : A gauche */
    for (acces1b = acces1; acces1b->PREV;)
    {
        if (acces1b->PREV->ORIENT == EAST)
        {
            del_acces (noeud, acces1b->PREV);
        }
        else
            acces1b = acces1b->PREV;
    }
    /* on supprime les extremites : A droite */
    for (acces2b = acces2; acces2b->NEXT;)
    {
        if (acces2b->NEXT->ORIENT == WEST)
        {
            del_acces (noeud, acces2b->NEXT);
        }
        else
            acces2b = acces2b->NEXT;
    }

    x1 = acces1->POS->X;
    y1 = acces1->POS->Y;
    x2 = acces2->POS->X;
    y2 = acces2->POS->Y;
    sig_name = acces1->SIG->NAME;

    /* On fait ce qu'il faut entre les deux */
    add_via_in_op (ptcol, acces1->POS->X, acces1->POS->Y);
    add_via_in_op (ptcol, acces2->POS->X, acces2->POS->Y);
    for (acces = acces1->NEXT; acces != acces2; acces = acces->NEXT)
    {
        if (acces1->SIG != acces->SIG)
        {
            acces = acces->PREV;
            del_acces (noeud, acces->NEXT);
        }
        else
        {
            add_via_in_op (ptcol, acces->POS->X, acces->POS->Y);
            acces = acces->PREV;
            del_acces (noeud, acces->NEXT);
        }
    }

    con1 = acces1->POS->CON;
    con2 = acces2->POS->CON;
    supress_acces( noeud, acces1 );
    supress_acces( noeud, acces2 );
    del_acces (noeud, acces2);
    total_acc++;
    acces1->POS = (POSSIBILITY *)NULL;
    acces1->ORIENT = ORIENTX;
    acces1->SIG = (SIGNAL *)NULL;
    acces1->DONE = TRUE;

    insertseg (ptcol,
               x1, y1, x2, y2,
               ALU2,
               WLAYER2,
               sig_name);

    for( ptnoeud = noeud; ptnoeud->DPR_DWN; ptnoeud=ptnoeud->DPR_DWN );
    for( ;ptnoeud; ptnoeud=ptnoeud->DPR_UP )
        for (acces = ptnoeud->DPR_LEFT; acces; acces = acces->NEXT)
        {
           if( ( !acces->DONE ) && (acces->POS->CON == con2) )
                acces->POS->CON = con1;
        }

}

/*---------------------------------------------------------\
                     existence

permet de detecter et d'estimer le cout d'une
conection eventuelle d'un connecteur quelconque
pour le signal donne
\---------------------------------------------------------*/
ACCESS         *existence (ptnoeud, ptsig, orient, nbr_acces)
NODE           *ptnoeud;
SIGNAL         *ptsig;
char            orient;
long           *nbr_acces;

{
    ACCESS         *ptacces = NULL;
    ACCESS         *ptacces_save = NULL;

    *nbr_acces = 0;
    if (orient == EAST)
    {
        for (ptacces = ptnoeud->DPR_RIGHT; ptacces; ptacces = ptacces->PREV)
            if (!ptacces->DONE)
            {
                if ((ptsig == ptacces->SIG) && (ptacces->ORIENT == EAST))
                    break;
                else
                    (*nbr_acces)++;
            }
            else 
            {
                return NULL;
            }
        if (ptacces)
        {
            ptacces_save = ptacces;
            for (ptacces = ptacces->PREV; ptacces; ptacces = ptacces->PREV)
                if (ptacces->ORIENT == EAST)
                    (*nbr_acces)++;
            return ptacces_save;
        }
        else
            return NULL;
    }
    else
        /* orient==WEST */
    {
        for (ptacces = ptnoeud->DPR_LEFT; ptacces; ptacces = ptacces->NEXT)
            if (!ptacces->DONE)
            {
                if ((ptsig == ptacces->SIG) && (ptacces->ORIENT == WEST))
                    break;
                else
                    (*nbr_acces)++;
            }
            else 
            {
                return NULL;
            }
        if (ptacces)
        {
            ptacces_save = ptacces;
            for (ptacces = ptacces->NEXT; ptacces; ptacces = ptacces->NEXT)
                if (ptacces->ORIENT == WEST)
                    (*nbr_acces)++;
            return ptacces_save;
        }
        else
            return NULL;
    }
}

/*---------------------------------------------------------\
                     existence_in_op
\---------------------------------------------------------*/
ACCESS         *existence_in_op (ptnoeud, ptacces, ptsig)
NODE           *ptnoeud;
ACCESS         *ptacces;
SIGNAL         *ptsig;
{
    for (ptacces = ptacces->NEXT; ptacces; ptacces = ptacces->NEXT)
    {
        if (!ptacces->DONE)
        {
            if (ptsig == ptacces->SIG)
            {
                ptacces->ORIENT = WEST;
                return ptacces;
            }
            else
            {
                if(!exist_con(ptacces->POS->CON, ptnoeud))
                    break;
            }
        }
        else
            break;
    }

     return NULL;
}

/*---------------------------------------------------------\
                     put_obstacle

met une structure obstacle dans un noeud
\---------------------------------------------------------*/
void            put_obstacle (noeud, ptsig)
NODE           *noeud;
SIGNAL         *ptsig;

{
    /* destruction de tous les acces du noeud */
    while (noeud->DPR_LEFT)
    {
        del_acces (noeud, noeud->DPR_LEFT);
    }

    /* on place un unique obstacle dans le noeud */
    noeud->DPR_RIGHT = noeud->DPR_LEFT = mkacces ((ACCESS *) NULL,
                                                  (ACCESS *) NULL,
                                                  (POSSIBILITY *) NULL,
                                                  ORIENTX,
                                                  ptsig,
                                                  TRUE);
    total_acc--;
}

/*---------------------------------------------------------\
                     uniqorient

Fonction booleenne permettant de detecter si une
possibilite d'acces est devenu unique dans
l'orientation (EAST ou WEST)
\---------------------------------------------------------*/
int             uniqorient (noeud, acces)
NODE           *noeud;
ACCESS         *acces;
{
    NODE           *noeud1 = NULL;
    ACCESS         *acces1 = NULL;
    CONNECTOR      *con = NULL;
    char            orient;

    orient = acces->ORIENT;
    con = acces->POS->CON;

    for (noeud1 = noeud; noeud1->DPR_UP; noeud1 = noeud1->DPR_UP);

    for (; noeud1; noeud1 = noeud1->DPR_DWN)
    {
        if (noeud1 != noeud)
        {
            for (acces1 = noeud1->DPR_LEFT; acces1; acces1 = acces1->NEXT)
            {
                if ((!acces1->DONE) && (acces1->ORIENT == orient) && (acces1->POS->CON == con))
                    return (FALSE);
            }
        }
    }

    return (TRUE);
}

/*---------------------------------------------------------\
                     inhibuniq

Fonction booleenne permettant de detecter si une
eventuelle connexion inhibe un acces unique
\---------------------------------------------------------*/
int             inhibuniq (noeud, acces)
NODE           *noeud;
ACCESS         *acces;
{
    ACCESS         *acces1 = NULL;
    ACCESS         *acces2 = NULL;

    if (acces->ORIENT == WEST)
    {
        for (acces1 = acces->PREV; acces1; acces1 = acces1->PREV)
        {
            if( (!acces1->DONE) && (acces1->SIG != acces->SIG) )
            {
                if (!exist_con (acces1->POS->CON, noeud))
                    return TRUE;
                else if ( second_uniq( noeud, acces1, acces->POS->CON ) )
                    return TRUE;
            }
        }
        for (acces1 = acces->NEXT; acces1; acces1 = acces1->NEXT)
        {
            if(  (!acces1->DONE)
              && (acces1->SIG != acces->SIG)
              && (  (!exist_con (acces1->POS->CON, noeud))
                 || (second_uniq( noeud, acces1, acces->POS->CON ))))
            {
                for (acces2 = acces1->NEXT; acces2; acces2 = acces2->NEXT)
                {
                   if (!acces2->DONE)
                   {
                       if(  (acces2->POS->CON != acces1->POS->CON)
                         && (  (!exist_con (acces2->POS->CON, noeud))
                            || (second_uniq( noeud, acces2, acces->POS->CON ))))
                       {
                           return TRUE;
                       }
                   }
                   else
                       return TRUE;
                }
            }
        }
    }
    else
    {
        for (acces1 = acces->NEXT; acces1; acces1 = acces1->NEXT)
        {
            if( (!acces1->DONE) && (acces1->SIG != acces->SIG) )
            {
                if (!exist_con (acces1->POS->CON, noeud))
                    return TRUE;
                else if ( second_uniq( noeud, acces1, acces->POS->CON ) )
                    return TRUE;
            }
        }
        for (acces1 = acces->PREV; acces1; acces1 = acces1->PREV)
        {
            if(  (!acces1->DONE)
              && (acces1->SIG != acces->SIG)
              && (  (!exist_con (acces1->POS->CON, noeud))
                 || (second_uniq( noeud, acces1, acces->POS->CON ))))
            {
                for (acces2 = acces1->PREV; acces2; acces2 = acces2->PREV)
                {
                   if (!acces2->DONE)
                   {
                       if(  (acces2->POS->CON != acces1->POS->CON)
                         && (  (!exist_con (acces2->POS->CON, noeud))
                            || (second_uniq( noeud, acces2, acces->POS->CON ))))
                       {
                           return TRUE;
                       }
                   }
                   else
                       return TRUE;
                }
            }
        }
    }
    return FALSE;
}

/*---------------------------------------------------------\
                     inhibuniq_non_rec
\---------------------------------------------------------*/
int             inhibuniq_non_rec (noeud, acces, ptconsource)
NODE           *noeud;
ACCESS         *acces;
CONNECTOR      *ptconsource;
{
    ACCESS         *acces1 = NULL;
    ACCESS         *acces2 = NULL;

    if (acces->ORIENT == WEST)
    {
        for (acces1 = acces->PREV; acces1; acces1 = acces1->PREV)
        {
            if(  (!acces1->DONE) && (acces1->SIG != acces->SIG) && ( acces1->POS->CON != ptconsource ) )
            {
                if (!exist_con (acces1->POS->CON, noeud))
                    return TRUE;
            }
        }
        for (acces1 = acces->NEXT; acces1; acces1 = acces1->NEXT)
        {
            if(  (!acces1->DONE)
              && (acces1->SIG != acces->SIG)
              && (acces1->POS->CON != ptconsource )
              && (  (!exist_con (acces1->POS->CON, noeud))
                 || (second_uniq( noeud, acces1, acces->POS->CON ))))
            {
                for (acces2 = acces1->NEXT; acces2; acces2 = acces2->NEXT)
                {
                   if (!acces2->DONE)
                   {
                       if(  (acces2->POS->CON != acces1->POS->CON)
                         && (acces2->POS->CON != ptconsource )
                         && (  (!exist_con (acces2->POS->CON, noeud))
                            || (second_uniq( noeud, acces2, acces->POS->CON ))))
                       {
                           return TRUE;
                       }
                   }
                   else
                       return TRUE;
                }
            }
        }
    }
    else
    {
        for (acces1 = acces->NEXT; acces1; acces1 = acces1->NEXT)
        {
            if( (!acces1->DONE) && (acces1->SIG != acces->SIG) && (acces1->POS->CON != ptconsource ) )
            {
                if (!exist_con (acces1->POS->CON, noeud))
                    return TRUE;
            }
        }
        for (acces1 = acces->PREV; acces1; acces1 = acces1->PREV)
        {
            if(  (!acces1->DONE)
              && (acces1->SIG != acces->SIG)
              && (acces1->POS->CON != ptconsource )
              && (  (!exist_con (acces1->POS->CON, noeud))
                 || (second_uniq( noeud, acces1, acces->POS->CON ))))
            {
                for (acces2 = acces1->PREV; acces2; acces2 = acces2->PREV)
                {
                   if (!acces2->DONE)
                   {
                       if(  (acces2->POS->CON != acces1->POS->CON)
                         && (acces2->POS->CON != ptconsource )
                         && (  (!exist_con (acces2->POS->CON, noeud))
                            || (second_uniq( noeud, acces2, acces->POS->CON ))))
                       {
                           return TRUE;
                       }
                   }
                   else
                       return TRUE;
                }
            }
        }
    }
    return FALSE;
}
/*---------------------------------------------------------\
                     second_uniq
\---------------------------------------------------------*/
int second_uniq( ptnoeud, ptacces, ptconsource )
NODE           *ptnoeud;
ACCESS         *ptacces;
CONNECTOR      *ptconsource;
{
    NODE           *ptnoeud1 = NULL;
    ACCESS         *ptacces1 = NULL;
    CONNECTOR      *ptcon;
    int             retour = FALSE;

/*
    if ( ptacces->POS )
        ptcon = ptacces->POS->CON;
    else
        return FALSE;
*/

    if( (!ptnoeud->DPR_UP) && (!ptnoeud->DPR_DWN))
        return TRUE;

    if( ptnoeud->DPR_UP )
        for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_UP; ptnoeud1 = ptnoeud1->DPR_UP);
    else
        ptnoeud1 = ptnoeud->DPR_DWN;

    if( ptnoeud->DPR_UP )
        ptnoeud->DPR_UP->DPR_DWN = ptnoeud->DPR_DWN;
    if( ptnoeud->DPR_DWN )
        ptnoeud->DPR_DWN->DPR_UP = ptnoeud->DPR_UP;

    for (; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_DWN)
    {
        for (ptacces1 = ptnoeud1->DPR_LEFT; ptacces1; ptacces1 = ptacces1->NEXT)
        {
            if(  (!ptacces1->DONE)
              && (ptacces1->POS->CON == ptcon)
              && (inhibuniq_non_rec(ptnoeud1, ptacces1, ptconsource)))
            {
                retour = TRUE;
                break;
            }
        }
    }

    if( ptnoeud->DPR_UP )
        ptnoeud->DPR_UP->DPR_DWN = ptnoeud;
    if( ptnoeud->DPR_DWN )
        ptnoeud->DPR_DWN->DPR_UP = ptnoeud;

    return retour;
} /* fin de second_uniq */

/*---------------------------------------------------------\
                     exist_con
\---------------------------------------------------------*/
int             exist_con (ptcon, ptnoeud)
CONNECTOR      *ptcon;
NODE           *ptnoeud;

{
    NODE           *ptnoeud1 = NULL;
    ACCESS         *ptacces1 = NULL;

    /* recherche sur les noeuds verticaux */
    for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_UP; ptnoeud1 = ptnoeud1->DPR_UP);

    for (; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_DWN)
    {
        if (ptnoeud1 != ptnoeud)
        {
            for (ptacces1 = ptnoeud1->DPR_LEFT; ptacces1; ptacces1 = ptacces1->NEXT)
            {
                if ((ptacces1->POS) && (ptacces1->POS->CON == ptcon))
                    return (TRUE);
            }
        }
    }
    return FALSE;
}

/*---------------------------------------------------------\
                     test_allow
\---------------------------------------------------------*/
int             test_allow (noeud)
NODE           *noeud;

{
    ACCESS         *acces = NULL;

    for (acces = noeud->DPR_LEFT; acces; acces = acces->NEXT)
    {
        /* on refuse le passage s'il y a un obstacle */
        if (acces->DONE)
            return FALSE;

        /* on refuse le passage si on detruit une possibilite unique */
        if (!exist_con (acces->POS->CON, noeud))
            return FALSE;

        /*
         * on refuse le passage si on detruit une possibilite unique en
         * orientation
         */
        if (acces->ORIENT == WEST)
        {
            if ((uniqorient (noeud, acces)) &&
                (acces->SIG->MIN_COL != ((COLUMN *) acces->POS->CON->COL->DATA)->INDEX))
                return FALSE;
        }
        else
        {
            if ((uniqorient (noeud, acces)) &&
                ((acces->SIG->MIN_COL + acces->SIG->WIDTH) != ((COLUMN *) acces->POS->CON->COL->DATA)->INDEX))
                return FALSE;
        }
    
        /* on refuse le passage si on detruit une possibilite qui deviendra impossible */
        if (second_uniq ( noeud, acces, (CONNECTOR *)NULL))
            return FALSE;
    }
 
    return TRUE;
}

/*---------------------------------------------------------\
                     nb_coninlin

nb_coninlin() : retourne vrai si un
connecteur d'un signal existe dans la ligne
donnee en second argument.
\---------------------------------------------------------*/
int             nb_coninlin (ptcheadcon, ptclin)
chain_list     *ptcheadcon;
chain_list     *ptclin;

{
    for (; ptcheadcon; ptcheadcon = ptcheadcon->NEXT)
    {
        if (((CONNECTOR *) ptcheadcon->DATA)->LIN == ptclin)
            return TRUE;
    }

    return FALSE;
}

/*---------------------------------------------------------\
                     decroche

decroche() :renvoie TRUE si un signal admet
un connecteur unique dans la ligne
\---------------------------------------------------------*/
int             decroche (ptsig, ptclin)
SIGNAL         *ptsig;
chain_list     *ptclin;

{
    if (ptsig->WIDTH == 0)
        return TRUE;
    if (ptsig->HEIGHT != 0)
        return (nb_coninlin (ptsig->CON, ptclin) == 1);

    return FALSE;
}

/*---------------------------------------------------------\
                     scan2o_1
premier (et unique) parcours pour l'heuristique 2o
recherche du vis a vis
\---------------------------------------------------------*/
void            scan2o_1 (noeud, head_noeudmax, node_r, acces_r, node_l, acces_l)
NODE           *noeud;
chain_list     *head_noeudmax;
NODE          **node_r;
ACCESS        **acces_r;
NODE          **node_l;
ACCESS        **acces_l;

{
    NODE           *noeud1 = NULL, *noeud2 = NULL, *noeud3 = NULL, *noeud4 = NULL;
    ACCESS         *acces1 = NULL, *acces2 = NULL;
    SIGNAL         *ptsig = NULL;
    chain_list     *chain_noeudmax = NULL;
    long            cout;

    /* initialisation des resultats a FALSE */
    *node_r = NULL;
    *acces_r = NULL;
    *node_l = NULL;
    *acces_l = NULL;

    chain_noeudmax = head_noeudmax;
    for( noeud1 = noeud; noeud1; noeud1 = noeud1->DPR_UP )
    {
        chain_noeudmax->DATA = (void *)noeud1->NEXT;
        chain_noeudmax = chain_noeudmax->NEXT;
    }

    /* Parcours de gauche a droite */
    for (noeud1 = noeud; noeud1->NEXT; noeud1 = noeud1->NEXT)
    {
        chain_noeudmax = head_noeudmax;

        /* Parcours de bas en haut */
        for (noeud2 = noeud1; noeud2; noeud2 = noeud2->DPR_UP)
        {
            if( chain_noeudmax->DATA == (void *)noeud2 )
                chain_noeudmax->DATA = (void *)noeud2->NEXT;

            for (acces1 = noeud2->DPR_LEFT; acces1; acces1 = acces1->NEXT)
            {
                if ((acces1->ORIENT == EAST) && ( !acces1->DONE ) && (!inhibuniq (noeud2, acces1)))
                {
                    /*
                     * On a trouve un acces par la droite quandidat
                     */
                    ptsig = acces1->SIG;

                    for (noeud3 = noeud2; noeud3; noeud3 = noeud3->NEXT)
                    {
                        if (noeud3 == noeud2)
                            acces2 = existence_in_op (noeud2, acces1, ptsig);
                        else
                            acces2 = existence (noeud3, ptsig, WEST, &cout);
                        if (acces2)
                        {
                            if (!inhibuniq (noeud3, acces2))
                            {
                                for (noeud4 = noeud2; noeud4 != noeud3; noeud4 = noeud4->NEXT)
                                {
                                    if( sig_crosscol( noeud4, ptsig ) )
                                        break;
                                }
                                if( noeud4 != noeud3 )
                                    break;

                                *node_r = noeud3;
                                *acces_r = acces2;
                                *node_l = noeud2;
                                *acces_l = acces1;

                                return;
                            }
                            else
                            {
                                acces2 = NULL;
                                break;
                            }
                        }

                        if (noeud3 == noeud2)
                            continue;

                        if( chain_noeudmax->DATA == (void *)noeud3 )
                        {
                            if (!test_allow (noeud3))
                            {
                                break;
                            }
                            chain_noeudmax->DATA = (void *)noeud3->NEXT;
                        }
                    }
                }
            }
            chain_noeudmax = chain_noeudmax->NEXT;
        }
    }
}

/*---------------------------------------------------------\
                     sig_in_east
\---------------------------------------------------------*/
int sig_in_east( ptnoeud, ptsig )
NODE           *ptnoeud;
SIGNAL         *ptsig;
{
    ACCESS         *acces;

    for( ; ptnoeud->DPR_DWN; ptnoeud = ptnoeud->DPR_DWN );
    for( ; ptnoeud; ptnoeud = ptnoeud->DPR_UP )
    {
        for( acces = ptnoeud->DPR_LEFT; acces; acces = acces->NEXT )
        {
            if( ( acces->DONE ) && ( acces->SIG == ptsig ) )
            {
                if( ( acces->ORIENT == ORIENTX ) && ( !acces->PREV ) && ( !acces->NEXT ) )
                {
                    return TRUE;
                }
                if( acces->ORIENT == EAST )
                {
                    return TRUE;
                }
            }
        }
    }
    return FALSE;
} /* fin de sig_in_east */

/*---------------------------------------------------------\
                     sig_in_west
\---------------------------------------------------------*/
int sig_in_west( ptnoeud, ptsig )
NODE           *ptnoeud;
SIGNAL         *ptsig;
{
    ACCESS         *acces;

    for( ; ptnoeud->DPR_DWN; ptnoeud = ptnoeud->DPR_DWN );
    for( ; ptnoeud; ptnoeud = ptnoeud->DPR_UP )
    {
        for( acces = ptnoeud->DPR_LEFT; acces; acces = acces->NEXT )
        {
            if( ( acces->DONE ) && ( acces->SIG == ptsig ) )
            {
                if( ( acces->ORIENT == ORIENTX ) && ( !acces->PREV ) && ( !acces->NEXT ) )
                {
                    return TRUE;
                }
                if( acces->ORIENT == WEST )
                {
                    return TRUE;
                }
            }
        }
    }
    return FALSE;
} /* fin de sig_in_east */

/*---------------------------------------------------------\
                     sig_crosscol

cherche si un signal traverse une colonne OPERATOR
pour cela cherche si signal existe dans canal a gauche:
cote EAST et dans canal a droite: cote WEST
\---------------------------------------------------------*/
int             sig_crosscol (ptnoeud, ptsig)
NODE           *ptnoeud;
SIGNAL         *ptsig;

{
    ACCESS         *acces;
    chain_list     *chain_east = NULL;
    chain_list     *chain_west = NULL;
    chain_list     *ptchain;

    for( ; ptnoeud->DPR_DWN; ptnoeud = ptnoeud->DPR_DWN );
    for( ; ptnoeud; ptnoeud = ptnoeud->DPR_UP )
    {
        for( acces = ptnoeud->DPR_LEFT; acces; acces = acces->NEXT )
        {
            if( ( acces->DONE ) && ( acces->SIG == ptsig ) )
            {
                if( ( acces->ORIENT == ORIENTX ) && ( !acces->PREV ) && ( !acces->NEXT ) )
                {
                    freechain( chain_west );
                    freechain( chain_east );
                    return TRUE;
                }
                if( acces->ORIENT == EAST )
                {
                    chain_east = addchain( chain_east, (void *)acces->POS->CON );
                }
                if( acces->ORIENT == WEST )
                {
                    chain_west = addchain( chain_west, (void *)acces->POS->CON );
                }
            }
        }
    }

    for( ptchain = chain_east; ptchain; ptchain=ptchain->NEXT )
    {
        if( in_con_chain( chain_west, (CONNECTOR *)ptchain->DATA ) )
        {
            freechain( chain_west );
            freechain( chain_east );
            return TRUE;
        }
    }

    freechain( chain_west );
    freechain( chain_east );
    return FALSE;
}

/*---------------------------------------------------------\
                     heuristique_2o

recherche du vis a vis
\---------------------------------------------------------*/
int             heuristique_2o (ptccol_head, ptnoeud_head, for_speed_up_scan2o)
chain_list     *ptccol_head;
NODE           *ptnoeud_head;
chain_list     *for_speed_up_scan2o;

{
    NODE           *node_r = NULL;
    ACCESS         *acces_r = NULL;
    NODE           *node_l = NULL;
    ACCESS         *acces_l = NULL;
    chain_list     *ptccol_c = NULL;
    COLUMN         *ptcol_c = NULL;
    COLUMN         *ptcol_o = NULL;
    long            y = 0;
    int             traverse = FALSE;
    SIGNAL         *ptsig;
    long            index;

    scan2o_1 (ptnoeud_head, for_speed_up_scan2o, &node_r, &acces_r, &node_l, &acces_l );

    if ((node_l == node_r) && (acces_l != NULL))
    {
        liber_noeud_in_op (node_l, acces_l, acces_r);
        return (TRUE);
    }
    else if (acces_l)
    {
        ptccol_c = acces_l->POS->CON->COL->NEXT;
        ptcol_c = (COLUMN *) ptccol_c->DATA;

        liber_noeud (node_l, acces_l, ptccol_head);

        /* gestion des durcissements intermediaires */
        y = acces_r->POS->Y;
        ptsig = acces_r->SIG;
        index = acces_r->SIG->INDEX;
        for (node_l = node_l->NEXT; node_l != node_r; node_l = node_l->NEXT)
        {
            ptcol_o = (COLUMN *) ptccol_c->NEXT->DATA;
            traverse = sig_crosscol (node_l, ptsig);
            if (!traverse)
            {
                insertseg (ptcol_o, 0L, y, ptcol_o->WIDTH, y, ALU2, WLAYER2, acces_r->SIG->NAME);
                insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, WEST, index, y);
            }
            ptccol_c = ptccol_c->NEXT->NEXT;
            ptcol_c = (COLUMN *) ptccol_c->DATA;
            if (!traverse)
            {
                insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, EAST, index, y);
                /* modification de la grille symbolique */
                put_obstacle (node_l,ptsig);
            }
        }

        liber_noeud (node_r, acces_r, ptccol_head);
        return (TRUE);
    }
    return (FALSE);
}

/*---------------------------------------------------------\
                     scan_3_1

Parcours de  la grille symbolique; recherche du
connecteur 'qui decroche' dont l'acces est de
cout minimum.
\---------------------------------------------------------*/

void            scan3_1 (noeud, node, acces)
NODE           *noeud;
NODE          **node;
ACCESS        **acces;

{
    NODE           *noeud1 = NULL, *noeud2 = NULL;
    ACCESS         *acces1 = NULL, *acces2 = NULL;
    chain_list     *ptccon;
    long            cout_min = -1, cout_east = 0, cout_west = 0;
    COLUMN         *ptcol;
    CONNECTOR      *ptcon;
    SIGNAL         *ptsig;

    /* initialisation du resultat a FALSE */
    *node = NULL;
    *acces = NULL;

    /* Parcours de gauche a droite */
    for (noeud1 = noeud->NEXT; noeud1->NEXT; noeud1 = noeud1->NEXT)
    {
        /* Parcours de bas en haut */
        for (noeud2 = noeud1; noeud2; noeud2 = noeud2->DPR_UP)
        {
            /* Parcours des acces du noeud */
            for (acces1 = noeud2->DPR_LEFT; acces1; acces1 = acces1->NEXT)
            {
                /*
                 * recherche d'un acces pour un connecteur qui decroche
                 */
                if ((!acces1->DONE) && (decroche (acces1->SIG, acces1->POS->CON->LIN))
                    && (!inhibuniq (noeud2, acces1)))
                {
                    ptsig = acces1->SIG;
                    if (ptsig->WIDTH == 0)
                    {
                        ptcol = (COLUMN *) acces1->POS->CON->COL->DATA;
                        acces2 = existence (noeud2, ptsig, WEST, &cout_west);
                        if (acces2)
                        {
                            for (ptccon = ptsig->CON; ptccon; ptccon = ptccon->NEXT)
                            {
                                ptcon = (CONNECTOR *) ptccon->DATA;
                                if ((ptcon->TYPE == EXTERNAL) &&
                                    (((COLUMN *) ptcon->COL->DATA)->INDEX == ptcol->INDEX - 1))
                                {
                                    if ((cout_west < cout_min) || (cout_min == -1))
                                    {
                                        cout_min = cout_west;
                                        *node = noeud2;
                                        *acces = acces2;
                                        if (cout_min == 0)
                                            return;
                                    }
                                }
                            }
                        }
                        acces2 = existence (noeud2, ptsig, EAST, &cout_east);
                        if (acces2)
                        {
                            for (ptccon = ptsig->CON; ptccon; ptccon = ptccon->NEXT)
                            {
                                ptcon = (CONNECTOR *) ptccon->DATA;
                                if ((ptcon->TYPE == EXTERNAL) &&
                                    (((COLUMN *) ptcon->COL->DATA)->INDEX == ptcol->INDEX + 1))
                                {
                                    if ((cout_east < cout_min) || (cout_min == -1))
                                    {
                                        cout_min = cout_east;
                                        *node = noeud2;
                                        *acces = acces2;
                                        if (cout_min == 0)
                                            return;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return;
}

/*---------------------------------------------------------\
                     heuristique_3

recherche du connecteur qui decroche de cout
minimum sur toute la grille symbolique.
\---------------------------------------------------------*/
int             heuristique_3 (ptccol_head, ptnoeud_head)
chain_list     *ptccol_head;
NODE           *ptnoeud_head;

{
    NODE           *node;
    ACCESS         *acces;

    scan3_1 (ptnoeud_head, &node, &acces);
    if (acces)
    {
        liber_noeud (node, acces, ptccol_head);
        return (TRUE);
    }

    return (FALSE);
}

/*---------------------------------------------------------\
                     scan_allow
\---------------------------------------------------------*/
NODE           *scan_allow (ptnoeud)
NODE           *ptnoeud;

{
    NODE           *ptnoeud_up = NULL;
    NODE           *ptnoeud_dwn = NULL;

    if (test_allow (ptnoeud))
        return ptnoeud;

    ptnoeud_up = ptnoeud->DPR_UP;
    ptnoeud_dwn = ptnoeud->DPR_DWN;
    while (ptnoeud_up || ptnoeud_dwn)
    {
        if (ptnoeud_dwn)
        {
            if (test_allow (ptnoeud_dwn))
                return ptnoeud_dwn;
            ptnoeud_dwn = ptnoeud_dwn->DPR_DWN;
        }
        if (ptnoeud_up)
        {
            if (test_allow (ptnoeud_up))
                return ptnoeud_up;
            ptnoeud_up = ptnoeud_up->DPR_UP;
        }
    }
    return NULL;
}

/*---------------------------------------------------------\
                     diff_allows
\---------------------------------------------------------*/
void            diff_allows (ptnode_l, ptccol_l, ptpos_l, ptccol_r, ptsig, pttrack_head)
NODE           *ptnode_l;
chain_list     *ptccol_l;
POSSIBILITY    *ptpos_l;
chain_list     *ptccol_r;
SIGNAL         *ptsig;
TRACK          *pttrack_head;

{
    GABARIT        *ptgab = NULL;
    NODE           *ptnoeud1 = NULL;
    chain_list     *ptccol = NULL;
    COLUMN         *ptcol_o = NULL;
    NODE           *ptnoeud2 = NULL;
    chain_list     *ptccol_c = NULL;
    NODE           *ptnoeud_sav = NULL;
    COLUMN         *ptcol_c = NULL;
    TRACK          *pttrack = NULL;
    int             traverse = FALSE;

    ptgab = ((LINE *) ptpos_l->CON->LIN->DATA)->GAB;

    ptnoeud1 = ptnode_l;
    ptccol_c = ptccol_l->NEXT;
    for (ptccol = ptccol_l->NEXT->NEXT; ptccol != ptccol_r; ptccol = ptccol->NEXT->NEXT)
    {
        ptcol_c = (COLUMN *) ptccol_c->DATA;
        ptcol_o = (COLUMN *) ptccol->DATA;
        ptnoeud1 = ptnoeud1->NEXT;
        traverse = sig_crosscol (ptnoeud1, ptsig);

        if (!traverse)
        {
            ptnoeud2 = scan_allow (ptnoeud1);
            if (!ptnoeud2)
            {
                fflush(stdout);
                fprintf( stderr, "ERROR : can't go throught column %d for signal %s ( libere_null )\n", ptcol_o->INDEX, ptsig->NAME );
                ptccol_c = ptccol_c->NEXT->NEXT;
                continue;
            }
            ptnoeud_sav = ptnoeud2;
            /*
             * extraction de y en fonction du gabarit et de ptnoeud2
             */
            if (!pttrack_head)  /* c.a.d appel pour traiter un bit-slice */
                for (pttrack = ptgab->TRACK; ptnoeud2; pttrack = pttrack->NEXT,
                     ptnoeud2 = ptnoeud2->DPR_DWN);
            else
                /* c.a.d appel pour traiter la figure */
                for (pttrack = pttrack_head; ptnoeud2; pttrack = pttrack->NEXT,
                     ptnoeud2 = ptnoeud2->DPR_DWN);
            /* gestion du resultat physique */
            insertseg (ptcol_o, 0L, pttrack->Y, ptcol_o->WIDTH, pttrack->Y, ALU2, WLAYER2, ptsig->NAME);
            insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, WEST,
                   ((SIGNAL *) ptpos_l->CON->SIG->DATA)->INDEX, pttrack->Y);
        }
        ptccol_c = ptccol_c->NEXT->NEXT;
        ptcol_c = (COLUMN *) ptccol_c->DATA;
        if (!traverse)
        {
            insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, EAST,
                   ((SIGNAL *) ptpos_l->CON->SIG->DATA)->INDEX, pttrack->Y);
            /* modification de la grille symbolique */
            put_obstacle (ptnoeud_sav,ptsig);
            ptnoeud1 = ptnoeud_sav;
        }
    }
}

/*---------------------------------------------------------\
                     chercher_right
\---------------------------------------------------------*/
void            chercher_right (ptsig, node, node_r, acces_r)
SIGNAL         *ptsig;
NODE           *node;
NODE          **node_r;
ACCESS        **acces_r;

{
    NODE           *ptnoeud1;
    NODE           *ptnoeud2;
    ACCESS         *acces1;
    long            nbr_acces;

    *acces_r = NULL;
    *node_r = NULL;

    for (ptnoeud1 = node; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);

    for (; ptnoeud1; ptnoeud1 = ptnoeud1->NEXT)
    {
        if( sig_crosscol( ptnoeud1, ptsig ) )
            return;

        for (ptnoeud2 = ptnoeud1; ptnoeud2; ptnoeud2 = ptnoeud2->DPR_UP)
        {
            acces1 = existence (ptnoeud2, ptsig, WEST, &nbr_acces);
            if ((acces1) && (!inhibuniq (ptnoeud2, acces1)))
            {
                *acces_r = acces1;
                *node_r = ptnoeud2;
                return;
            }
			if( !scan_allow( ptnoeud2 ) )
            {
                return;
            }
        }
    }
}

/*---------------------------------------------------------\
                     looking_for_sig
\---------------------------------------------------------*/
void            looking_for_sig (ptsig, node, node_r, acces_r)
SIGNAL         *ptsig;
NODE           *node;
NODE          **node_r;
ACCESS        **acces_r;

{
    NODE           *ptnoeud1;
    NODE           *ptnoeud2;
    ACCESS         *acces1;

    for (ptnoeud1 = node; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);

    for ( ; ptnoeud1; ptnoeud1 = ptnoeud1->NEXT)
    {
        for (ptnoeud2 = ptnoeud1; ptnoeud2; ptnoeud2 = ptnoeud2->DPR_UP)
        {
            for( acces1=ptnoeud2->DPR_LEFT; acces1; acces1 = acces1->NEXT)
                if ((acces1->SIG == ptsig) && (acces1->POS))
                {
                    *acces_r = acces1;
                    *node_r = ptnoeud2;
                    return;
                }
        }
    }
    *acces_r = NULL;
    *node_r = NULL;
}

/*---------------------------------------------------------\
                     libere
\---------------------------------------------------------*/
char            libere (ptnoeud, ptacces, ptccol, ptsig, direction, pttrack_head)
NODE           *ptnoeud;
ACCESS         *ptacces;
chain_list     *ptccol;
SIGNAL         *ptsig;
char            direction;
TRACK          *pttrack_head;
{
    NODE           *ptnoeud1;
    NODE           *ptnoeud2;
    NODE           *ptnoeud3;
    ACCESS         *acces;
    long            nbr;
    chain_list     *go_east = NULL;
    chain_list     *go_west = NULL;
    chain_list     *ptchain;
    chain_list     *ptccol_c;
    chain_list     *ptccol_o;
    COLUMN         *ptcol_o;
    COLUMN         *ptcol_c;
    int             traverse;
    TRACK          *pttrack = NULL;
    char            Retour = FALSE;
    char            must_cross;

    /* 0 - est-ce qu'on traverse la colonne */
    ptccol_o = ptacces->POS->CON->COL;
    ptccol_c = chain_pred (ptccol, ptccol_o);
    traverse = sig_crosscol (ptnoeud, ptsig);

    /* 1 - on note les connecteurs allant deja a l'EST */
    /* 2 - on note les connecteurs allant deja a l'OUEST */
    for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);
    for (; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        for (acces = ptnoeud1->DPR_LEFT; acces; acces = acces->NEXT)
        {
           if( ( acces->DONE ) && ( acces->ORIENT == EAST ) && ( acces->SIG == ptsig ) )
           {
               go_east = addchain( go_east, (void *)acces->POS->CON );
           }
           else if( ( acces->DONE ) && ( acces->ORIENT == WEST ) && ( acces->SIG == ptsig ) )
           {
               go_west = addchain( go_west, (void *)acces->POS->CON );
           }
        }
    }

    /* 3 - recherche d'un acces EST non encore fait et possible */
    for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);
    for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        if( direction != WEST )
        {
            acces = existence (ptnoeud1, ptsig, EAST, &nbr);
            if ((acces) && (!inhibuniq (ptnoeud1, acces)) && (!in_con_chain (go_west, acces->POS->CON)))
            {
                go_east = addchain (go_east, (void *)acces->POS->CON);
                liber_noeud (ptnoeud1, acces, ptccol);
                Retour = TRUE;
            }
        }
        else
        {
            acces = existence (ptnoeud1, ptsig, WEST, &nbr);
            if ((acces) && (!inhibuniq (ptnoeud1, acces)) && (!in_con_chain (go_east, acces->POS->CON)))
            {
                go_west = addchain (go_west, (void *)acces->POS->CON);
                liber_noeud (ptnoeud1, acces, ptccol);
                Retour = TRUE;
            }
        }
    }
    /* 4 - recherche d'un acces OUEST non encore fait et possible */
    for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);
    for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        if( direction != WEST )
        {
            acces = existence (ptnoeud1, ptsig, WEST, &nbr);
            if ((acces) && (!inhibuniq (ptnoeud1, acces)) && (!in_con_chain (go_east, acces->POS->CON)))
            {
                go_west = addchain (go_west, (void *)acces->POS->CON);
                liber_noeud (ptnoeud1, acces, ptccol);
                Retour = TRUE;
            }
        }
        else
        {
            acces = existence (ptnoeud1, ptsig, EAST, &nbr);
            if ((acces) && (!inhibuniq (ptnoeud1, acces)) && (!in_con_chain (go_west, acces->POS->CON)))
            {
                go_east = addchain (go_east, (void *)acces->POS->CON);
                liber_noeud (ptnoeud1, acces, ptccol);
                Retour = TRUE;
            }
        }
    }

    /* 5 - on regarde si on doit traverser : ! un meme connecteur dans go_east et go_west */
    for( ptchain = go_east; ptchain; ptchain = ptchain->NEXT )
    {
        if( in_con_chain( go_west, (CONNECTOR *)ptchain->DATA) )
        {
            break;
        }
    }
    must_cross = ( direction != EAST ) && ( direction != WEST );
    if( ! must_cross )
    {
        if( direction == EAST )
            must_cross = (go_west) && (!ptchain);
        else
            must_cross = (go_east) && (!ptchain);
    }

    /* 6 - si je ne traverse pas encore la colonne et qu'on doit le faire, on continue le delire ... */
    if ((!traverse) && ( must_cross) )
    {
        /* 7 - recherche d'un acces WEST possible appartenant a go_east, si trouve le faire */
        /* 8 - recherche d'un acces EAST possible appartenant a go_west, si trouve le faire */
        for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);
        for ( ; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
        {
            acces = existence (ptnoeud1, ptsig, WEST, &nbr);
            if ((acces) && (!inhibuniq (ptnoeud1, acces)) && (in_con_chain (go_east, acces->POS->CON)))
            {
                Retour = TRUE;
                go_west = addchain (go_west, (void *)acces->POS->CON);
                liber_noeud (ptnoeud1, acces, ptccol);
                break;
            }
            acces = existence (ptnoeud1, ptsig, EAST, &nbr);
            if ((acces) && (!inhibuniq (ptnoeud1, acces)) && (in_con_chain (go_west, acces->POS->CON)))
            {
                Retour = TRUE;
                go_east = addchain (go_east, (void *)acces->POS->CON);
                liber_noeud (ptnoeud1, acces, ptccol);
                break;
            }
        }

        /* 9 - la c'est la merde, il faut trouver une piste ... */
        if (!ptnoeud1)
        {
            ptcol_o = (COLUMN *) ptccol_o->DATA;
            ptnoeud2 = scan_allow (ptnoeud);
            if (ptnoeud2)
            {
                ptnoeud3 = ptnoeud2;
                for (pttrack = pttrack_head; ptnoeud2; pttrack = pttrack->NEXT,
                     ptnoeud2 = ptnoeud2->DPR_DWN);
                /* gestion du resultat physique */
                insertseg (ptcol_o, 0L, pttrack->Y, ptcol_o->WIDTH, pttrack->Y, ALU2, WLAYER2, ptsig->NAME);
                put_obstacle (ptnoeud3,ptsig);
                ptcol_c = (COLUMN *) ptccol_c->DATA;
                insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, WEST, ptsig->INDEX, pttrack->Y);
                ptccol_c = ptccol_c->NEXT->NEXT;
                ptcol_c = (COLUMN *) ptccol_c->DATA;
                insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, EAST, ptsig->INDEX, pttrack->Y);
            }
            else
            {
                fflush( stdout );
                fprintf( stderr, "ERROR : can't go throught column for signal %s ( libere )\n", ptsig->NAME );
            }
        }
    }

    /* 10 - on vide les acces restant */
    for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);
    for (; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        do
        {
            acces = existence (ptnoeud1, ptsig, EAST, &nbr);
            if ((acces) && (in_con_chain (go_west, acces->POS->CON)))
            {
                del_acces (ptnoeud1, acces);
                Retour = TRUE;
            }
            else
               acces = NULL;
        } while( acces );
        do
        {
            acces = existence (ptnoeud1, ptsig, WEST, &nbr);
            if ((acces) && (in_con_chain (go_east, acces->POS->CON)))
            {
                del_acces (ptnoeud1, acces);
                Retour = TRUE;
            }
            else
               acces = NULL;
        } while( acces );
    }

    freechain( go_east );
    freechain( go_west );

    return Retour;
}


/*---------------------------------------------------------\
                     scan_not_done_connector
\---------------------------------------------------------*/
char scan_not_done_connector ( ptnoeud, ptsig, con )
NODE           *ptnoeud;
SIGNAL         *ptsig;
CONNECTOR      *con;
{
    ACCESS         *acces;

    for( ; ptnoeud->DPR_DWN; ptnoeud = ptnoeud->DPR_DWN );
    for( ; ptnoeud; ptnoeud = ptnoeud->DPR_UP )
    {
        for (acces = ptnoeud->DPR_LEFT; acces; acces = acces->NEXT)
        {
            if( (!acces->DONE)
             && (acces->SIG == ptsig)
             && (acces->POS->CON != con) )
            {
                return TRUE;
            }
        }
    }
    return FALSE;
}

/*---------------------------------------------------------\
                     libere_null_width
\---------------------------------------------------------*/
char libere_null_width ( ptnoeud, ptccol, ptsig, pttrack_head )
NODE           *ptnoeud;
chain_list     *ptccol;
SIGNAL         *ptsig;
TRACK          *pttrack_head;
{
    NODE           *ptnoeud1;
    NODE           *ptnoeud2;
    NODE           *ptnoeud3;
    ACCESS         *acces;
    long            nbr;
    TRACK          *pttrack = NULL;
    chain_list     *go_east = NULL;
    chain_list     *go_west = NULL;
    chain_list     *pos_west = NULL;
    chain_list     *pos_east = NULL;
    chain_list     *ptchain_head;
    chain_list     *ptchain1;
    chain_list     *ptchain2;
    chain_list     *ptccol_c;
    chain_list     *ptccol_o;
    COLUMN         *ptcol_o;
    COLUMN         *ptcol_c;
    char            in_one_dir = TRUE;
    char            found_commun;
    char            must_cross;
    char            Retour = FALSE;
    char            direction;
    char            other_direction;
    char            traverse;

    /* 0 - est-ce qu'on traverse la colonne */
    for (acces = ptnoeud->DPR_LEFT; acces->DONE; acces = acces->NEXT);
    ptccol_o = acces->POS->CON->COL;
    ptccol_c = chain_pred (ptccol, ptccol_o);
    traverse = sig_crosscol (ptnoeud, ptsig);

    /* 1 - on note les connecteurs allant deja a l'EST */
    /* 2 - on note les connecteurs allant deja a l'OUEST */
    for( ; ptnoeud->DPR_DWN; ptnoeud=ptnoeud->DPR_DWN );
    for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        for (acces = ptnoeud1->DPR_LEFT; acces; acces = acces->NEXT)
        {
           if( ( acces->DONE ) && ( acces->ORIENT == EAST ) && ( acces->SIG == ptsig ) )
           {
               go_east = addchain( go_east, (void *)acces->POS->CON );
           }
           else if( ( acces->DONE ) && ( acces->ORIENT == WEST ) && ( acces->SIG == ptsig ) )
           {
               go_west = addchain( go_west, (void *)acces->POS->CON );
           }
        }
    }

    /* 3 - si go_west seulement alors chercher a l'ouest */
    if( ( go_west ) && (!go_east) )
    {
        direction = WEST;
        other_direction = EAST;
    }
    else
    /* Sinon on essaye a l'est */
    {
        direction = EAST;
        other_direction = WEST;
    }

    for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        acces = existence (ptnoeud1, ptsig, direction, &nbr);
        if ((acces) && (!inhibuniq (ptnoeud1, acces)))
        {
            if( direction == EAST )
            {
                if( !in_con_chain( pos_east, acces->POS->CON) )
                    pos_east = addchain (pos_east, (void *)acces->POS->CON);
            }
            else
            {
                if( !in_con_chain( pos_west, acces->POS->CON) )
                    pos_west = addchain (pos_west, (void *)acces->POS->CON);
            }
        }
    }

    /* 4 - Reciproquement : regarder dans l'autre direction */
    for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        acces = existence (ptnoeud1, ptsig, other_direction, &nbr);

        if ((acces) && (!inhibuniq (ptnoeud1, acces)))
        {
            if( other_direction == EAST )
            {
                if( ( !in_con_chain( pos_east, acces->POS->CON) ) && ( !in_con_chain( go_west, acces->POS->CON) ) )
                    pos_east = addchain (pos_east, (void *)acces->POS->CON);
                if( !in_con_chain( pos_west, acces->POS->CON) )
                    in_one_dir = FALSE;
            }
            else
            {
                if( ( !in_con_chain( pos_west, acces->POS->CON) ) && ( !in_con_chain( go_east, acces->POS->CON) ) )
                    pos_west = addchain (pos_west, (void *)acces->POS->CON);
                if( !in_con_chain( pos_east, acces->POS->CON) )
                    in_one_dir = FALSE;
            }
        }
    }

    /* 5 - Est-ce que l'on a quelquechose a faire ? */
    if( (!pos_west) && (!pos_east ) )
    {
        for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
        {
            acces = existence (ptnoeud1, ptsig, EAST, &nbr);
            if( ( acces ) && ( !scan_not_done_connector (ptnoeud, ptsig, acces->POS->CON ) ) )
            {
                del_acces_sup( ptnoeud, acces->POS->CON, EAST );
                Retour = TRUE;
            }
            acces = existence (ptnoeud1, ptsig, WEST, &nbr);
            if( ( acces ) && ( !scan_not_done_connector (ptnoeud, ptsig, acces->POS->CON ) ) )
            {
                del_acces_sup( ptnoeud, acces->POS->CON, WEST );
                Retour = TRUE;
            }
        }
        return Retour;
    }

    if( (!go_east) && (!go_west) && (!pos_west) && (!pos_east->NEXT ) )
    {
        if( !scan_not_done_connector (ptnoeud, ptsig, (CONNECTOR *)pos_east->DATA))
        {
            del_acces_sup( ptnoeud, (CONNECTOR *) pos_east->DATA, EAST );
            Retour = TRUE;
        }
        freechain( pos_east );
        freechain( go_east );
        freechain( go_west );
        return Retour;
    }

    if( (!go_east) && (!go_west) && (!pos_east) && (!pos_west->NEXT ) )
    {
        if( !scan_not_done_connector (ptnoeud, ptsig, (CONNECTOR *)pos_west->DATA))
        {
            del_acces_sup( ptnoeud, (CONNECTOR *) pos_west->DATA, WEST );
            Retour = TRUE;
        }
        freechain( pos_west );
        freechain( go_east );
        freechain( go_west );
        return Retour;
    }

    if( (!go_east) && (!go_west) && (pos_east) && (pos_west) && (!pos_east->NEXT) && (!pos_west->NEXT ) && (pos_east->DATA == pos_west->DATA ) )
    {
        if( !scan_not_done_connector (ptnoeud, ptsig, (CONNECTOR *)pos_east->DATA))
        {
            del_acces_sup( ptnoeud, (CONNECTOR *) pos_west->DATA, WEST );
            del_acces_sup( ptnoeud, (CONNECTOR *) pos_east->DATA, EAST );
            Retour = TRUE;
        }
        freechain( pos_east );
        freechain( pos_west );
        freechain( go_east );
        freechain( go_west );
        return Retour;
    }

    /* 6 - Si la premiere direction etait bonne, on la fait */
    if( direction == WEST )
    {
        in_one_dir = (in_one_dir) && (!go_east);
    }
    else
    {
        in_one_dir = (in_one_dir) && (!go_west);
    }

    if( in_one_dir )
    {
        for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
        {
            acces = existence (ptnoeud1, ptsig, direction, &nbr);
            if ((acces) && (!inhibuniq (ptnoeud1, acces)))
            {
                Retour = TRUE;
                liber_noeud (ptnoeud1, acces, ptccol);
            }
        }
    }
    else
    /* sinon alors peut etre que dans l'autre direction ? */
    {
        if( direction == WEST )
        {
            ptchain_head = pos_east;
            in_one_dir = (!go_east);
        }
        else
        {
            ptchain_head = pos_west;
            in_one_dir = (!go_west);
        }

        for (ptnoeud1 = ptnoeud; (ptnoeud1) && (in_one_dir); ptnoeud1 = ptnoeud1->DPR_UP)
        {
            acces = existence (ptnoeud1, ptsig, direction, &nbr);

            if ((acces) && (!inhibuniq (ptnoeud1, acces)) && (!in_con_chain (ptchain_head, acces->POS->CON)))
            {
                in_one_dir = FALSE;
                break;
            }
        }
        /* Si c'est bon, on le fait ... */
        if( in_one_dir )
        {
            for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
            {
                acces = existence (ptnoeud1, ptsig, other_direction, &nbr);
                if ((acces) && (!inhibuniq (ptnoeud1, acces)))
                {
                    Retour = TRUE;
                    liber_noeud (ptnoeud1, acces, ptccol);
                }
            }
        }
        else
        /* la ca commence a etre la merde */
        {
            for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
            {
                acces = existence (ptnoeud1, ptsig, direction, &nbr);
                if ((acces) && (!inhibuniq (ptnoeud1, acces)))
                {
                    Retour = TRUE;
                    liber_noeud (ptnoeud1, acces, ptccol);
                }
            }
            for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
            {
                acces = existence (ptnoeud1, ptsig, other_direction, &nbr);
                if ((acces) && (!inhibuniq (ptnoeud1, acces)))
                {
                    Retour = TRUE;
                    liber_noeud (ptnoeud1, acces, ptccol);
                }
            }

            if(!traverse)
            {
                found_commun = FALSE;
                for( ptchain1 = pos_east; (ptchain1) && (!found_commun); ptchain1=ptchain1->NEXT )
                    for( ptchain2 = pos_west; ptchain2; ptchain2=ptchain2->NEXT )
                    {
                        if( ptchain1->DATA == ptchain2->DATA )
                        {
                            found_commun = TRUE;
                            break;
                        }
                    }

                 if( found_commun )
                 {
                     must_cross = TRUE;
                     for (ptnoeud1 = ptnoeud; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
                     {
                         acces = existence (ptnoeud1, ptsig, EAST, &nbr);
                         if ((acces) && (!inhibuniq (ptnoeud1, acces)))
                         {
                             Retour = TRUE;
                             liber_noeud (ptnoeud1, acces, ptccol);
                             if( must_cross )
                             {
                                 acces = existence (ptnoeud1, ptsig, WEST, &nbr);
                                 if ((acces) && (!inhibuniq (ptnoeud1, acces)))
                                 {
                                     liber_noeud (ptnoeud1, acces, ptccol);
                                     must_cross = FALSE;
                                 }
                             }
                         }
                         else
                         {
                             Retour = TRUE;
                             acces = existence (ptnoeud1, ptsig, WEST, &nbr);
                             if ((acces) && (!inhibuniq (ptnoeud1, acces)))
                             {
                                 liber_noeud (ptnoeud1, acces, ptccol);
                             }
                         }
                     }
                 }
                 else
                 {
                     ptcol_o = (COLUMN *) ptccol_o->DATA;
                     ptnoeud2 = scan_allow (ptnoeud);
                     if (ptnoeud2)
                     {
                         ptnoeud3 = ptnoeud2;
                         for (pttrack = pttrack_head; ptnoeud2; pttrack = pttrack->NEXT,
                              ptnoeud2 = ptnoeud2->DPR_DWN);
                         /* gestion du resultat physique */
                         insertseg (ptcol_o, 0L, pttrack->Y, ptcol_o->WIDTH, pttrack->Y, ALU2, WLAYER2, ptsig->NAME);
                         put_obstacle (ptnoeud3,ptsig);
                         ptcol_c = (COLUMN *) ptccol_c->DATA;
                         insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, WEST, ptsig->INDEX, pttrack->Y);
                         ptccol_c = ptccol_c->NEXT->NEXT;
                         ptcol_c = (COLUMN *) ptccol_c->DATA;
                         insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, EAST, ptsig->INDEX, pttrack->Y);
                     }
                     else
                    {
                        fflush( stdout );
                        fprintf( stderr, "ERROR : can't go throught column for signal %s ( libere_null )\n", ptsig->NAME );
                    }
                 }
             }
        }
    }

    for (ptnoeud1 = ptnoeud; ptnoeud1->DPR_DWN; ptnoeud1 = ptnoeud1->DPR_DWN);
    for (; ptnoeud1; ptnoeud1 = ptnoeud1->DPR_UP)
    {
        do
        {
            acces = existence (ptnoeud1, ptsig, EAST, &nbr);
            if ((acces) && ((in_con_chain (go_west, acces->POS->CON))|| (in_con_chain (pos_west, acces->POS->CON))))
            {
                del_acces (ptnoeud1, acces);
                Retour = TRUE;
            }
            else
               acces = NULL;
        } while( acces );
        do
        {
            acces = existence (ptnoeud1, ptsig, WEST, &nbr);
            if ((acces) && ((in_con_chain (go_east, acces->POS->CON))||(in_con_chain (pos_east, acces->POS->CON))))
            {
                del_acces (ptnoeud1, acces);
                Retour = TRUE;
            }
            else
               acces = NULL;
        } while( acces );
    }

    freechain( pos_west );
    freechain( pos_east );
    freechain( go_west );
    freechain( go_east );

    return Retour;
}

/*---------------------------------------------------------\
                     diff_line

diffusion des signaux (recherche de la continuite) sur
un slice unique; a ce stade il n'y a plus de vis a vis:
entre une extremite gauche et une autre droite
\---------------------------------------------------------*/
void            diff_line (ptccol_head, ptnoeud_head)
chain_list     *ptccol_head;
NODE           *ptnoeud_head;

{
    NODE           *ptnoeud1 = NULL;
    NODE           *ptnoeud2 = NULL;
    ACCESS         *acces1 = NULL;
    ACCESS         *acces_r = NULL;
    NODE           *node_r = NULL;

    for (ptnoeud1 = ptnoeud_head; ptnoeud1->NEXT; ptnoeud1 = ptnoeud1->NEXT)
    {
        for (ptnoeud2 = ptnoeud1; ptnoeud2; ptnoeud2 = ptnoeud2->DPR_UP)
        {
            for (acces1 = ptnoeud2->DPR_RIGHT; acces1; acces1 = acces1->PREV)
            {
                /* recherche du noeud de gauche */
                if ((!acces1->DONE) && (!inhibuniq (ptnoeud2, acces1)))
                {
                    if( (acces1->ORIENT == EAST) && (!sig_crosscol( ptnoeud2, acces1->SIG ) ) )
                    {
                        /*
                         * recherche du noeud de droite
                         */
                        chercher_right (acces1->SIG, ptnoeud2->NEXT, &node_r, &acces_r);
                        if (acces_r)
                        {
                            diff_allows (ptnoeud2,
                                         acces1->POS->CON->COL,
                                         acces1->POS,
                                         acces_r->POS->CON->COL,
                                         acces_r->SIG, (TRACK *) NULL);

                            liber_noeud (ptnoeud2, acces1, ptccol_head);
                            liber_noeud (node_r, acces_r, ptccol_head);

                            break;
                        }
                    }
                }
            }
        }
    }
}

/*---------------------------------------------------------\
                     sig_inchan

retourne TRUE si on trouve le signal dans le canal
sinon FALSE
\---------------------------------------------------------*/
int             sig_inchan (ptcol, ptsig)
COLUMN         *ptcol;
SIGNAL         *ptsig;
{
    long            index;
    CHANNEL        *ptchannel = NULL;
    ConnectorList  *ptcon_ch = NULL;

    index = ptsig->INDEX;
    ptchannel = (CHANNEL *) ptcol->DF_CON->DATA;

    for (ptcon_ch = ptchannel->NORTHCON; ptcon_ch; ptcon_ch = ptcon_ch->NextCon)
        if (ptcon_ch->ConName == index)
            return TRUE;

    for (ptcon_ch = ptchannel->SOUTHCON; ptcon_ch; ptcon_ch = ptcon_ch->NextCon)
        if (ptcon_ch->ConName == index)
            return TRUE;

    for (ptcon_ch = ptchannel->WESTCON; ptcon_ch; ptcon_ch = ptcon_ch->NextCon)
        if (ptcon_ch->ConName == index)
            return TRUE;

    for (ptcon_ch = ptchannel->EASTCON; ptcon_ch; ptcon_ch = ptcon_ch->NextCon)
        if (ptcon_ch->ConName == index)
            return TRUE;

    return FALSE;
}

/*---------------------------------------------------------\
                     goto_chan

ramene un signal jusqu'au premier canal de routage
ou il existe
\---------------------------------------------------------*/
void            goto_chan (ptnoeud_deb, ptclin_deb, ptsig, pttrack_head)
NODE           *ptnoeud_deb;
chain_list     *ptclin_deb;
SIGNAL         *ptsig;
TRACK          *pttrack_head;
{
    chain_list     *ptccol_c = NULL;
    COLUMN         *ptcol_c = NULL;
    chain_list     *ptccol_o = NULL;
    COLUMN         *ptcol_o = NULL;
    NODE           *ptnoeud1 = NULL;
    NODE           *ptnoeud2 = NULL;
    NODE           *ptnoeud3 = NULL;
    TRACK          *pttrack = NULL;

    ptnoeud1 = ptnoeud_deb->NEXT;
    for (ptccol_c = ptclin_deb->NEXT; ptccol_c->NEXT->NEXT; ptnoeud1 = ptnoeud1->NEXT)
    {
        ptcol_c = (COLUMN *) ptccol_c->DATA;
        if (sig_inchan (ptcol_c, ptsig))
            return;
        ptccol_o = ptccol_c->NEXT;
        ptcol_o = (COLUMN *) ptccol_o->DATA;
        if (!(ptnoeud2 = scan_allow (ptnoeud1)))
        {
            Warning (2, "Can not route through column:\n", ptcol_o->NAME);
            ptccol_c = ptccol_c->NEXT->NEXT;
            continue;
        }
        put_obstacle (ptnoeud2,ptsig);
        ptnoeud3 = ptnoeud2;

        /* extraction de l'ordonne du segment */
        for (pttrack = pttrack_head; ptnoeud2; pttrack = pttrack->NEXT, ptnoeud2 = ptnoeud2->DPR_DWN);

        insertseg (ptcol_o, 0L, pttrack->Y, ptcol_o->WIDTH, pttrack->Y, ALU2, WTRAN, ptsig->NAME);
        insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, WEST, ptsig->INDEX, pttrack->Y);
        ptccol_c = ptccol_c->NEXT->NEXT;
        ptcol_c = (COLUMN *) ptccol_c->DATA;
        insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, EAST, ptsig->INDEX, pttrack->Y);
        ptnoeud1 = ptnoeud3;
    }
}

/*---------------------------------------------------------\
                     back_chan

ramene un signal jusqu'au premier canal de routage
ou il existe
\---------------------------------------------------------*/
void            back_chan (ptnoeud_deb, ptclin_deb, ptsig, pttrack_head, ptccol_head)
NODE           *ptnoeud_deb;
chain_list     *ptclin_deb;
SIGNAL         *ptsig;
TRACK          *pttrack_head;
chain_list     *ptccol_head;

{
    chain_list     *ptccol_c = NULL;
    COLUMN         *ptcol_c = NULL;
    chain_list     *ptccol_o = NULL;
    COLUMN         *ptcol_o = NULL;
    NODE           *ptnoeud1 = NULL;
    NODE           *ptnoeud2 = NULL;
    NODE           *ptnoeud3 = NULL;
    TRACK          *pttrack = NULL;

    ptnoeud1 = ptnoeud_deb->PREV;
    for (ptccol_c = chain_pred (ptccol_head, ptclin_deb); ptccol_c; ptnoeud1 = ptnoeud1->PREV)
    {
        ptcol_c = (COLUMN *) ptccol_c->DATA;
        if (sig_inchan (ptcol_c, ptsig))
            return;
        ptccol_o = chain_pred (ptccol_head, ptccol_c);
        ptcol_o = (COLUMN *) ptccol_o->DATA;
        if (!(ptnoeud2 = scan_allow (ptnoeud1)))
        {
            Warning (2, "Can not route through column:\n", ptcol_o->NAME);
            ptccol_c = chain_pred (ptccol_head, chain_pred (ptccol_head, ptccol_c));
            continue;
        }

        put_obstacle (ptnoeud2,ptsig);
        ptnoeud3 = ptnoeud2;

        /* extraction de l'abscisse du segment */
        for (pttrack = pttrack_head; ptnoeud2; pttrack = pttrack->NEXT, ptnoeud2 = ptnoeud2->DPR_DWN);

        insertseg (ptcol_o, 0L, pttrack->Y, ptcol_o->WIDTH, pttrack->Y, ALU2, WTRAN, ptsig->NAME);
        insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, EAST, ptsig->INDEX, pttrack->Y);
        ptccol_c = chain_pred (ptccol_head, chain_pred (ptccol_head, ptccol_c));
        if (ptccol_c)
        {
            ptcol_c = (COLUMN *) ptccol_c->DATA;
            insertcon_inchan ((CHANNEL *) ptcol_c->DF_CON->DATA, WEST, ptsig->INDEX, pttrack->Y);
        }
        ptnoeud1 = ptnoeud3;
    }
}

/*---------------------------------------------------------\
                     diff_all

finition du routage pour tous les connecteurs
restants en travaillant sur la grille symbolique
du chemin de donnees en entier
\---------------------------------------------------------*/
TRACK          *diff_all (ptfig_dpr)
FIGURE         *ptfig_dpr;

{
    chain_list     *ptccolpos = NULL;
    COLUMN         *ptcolpos = NULL;
    chain_list     *ptclin1 = NULL;
    chain_list     *ptclin2 = NULL;
    LINE           *ptlin1 = NULL;
    LINE           *ptlin2 = NULL;
    NODE           *noeud2 = NULL;
    NODE           *noeud1 = NULL;
    NODE           *ptnoeud = NULL;
    ACCESS         *acces = NULL;
    SIGNAL         *ptsig = NULL;
    NODE           *node_r = NULL;
    ACCESS         *acces_r = NULL;
    GABARIT        *ptgab1 = NULL;
    TRACK          *pttrack1 = NULL;
    TRACK          *pttrack_head = NULL;
    TRACK          *pttrack_cour = NULL;
    int             loop;
    int             master_loop;
    int             nucol = 0;

    if( verbose_mode )
    {
        fprintf( stdout, "   o Multi-bit net symbolic routing.\n" );
        fflush( stdout );
    }

    /* creation du chainage des gabarits des differentes lignes */
    ptlin1 = (LINE *) (ptfig_dpr->LIN)->DATA;
    ptgab1 = ptlin1->GAB;
    pttrack_head = addtrack ((TRACK *) NULL, ptgab1->TRACK->Y);
    pttrack_cour = pttrack_head;
    for (ptclin1 = ptfig_dpr->LIN; ptclin1; ptclin1 = ptclin1->NEXT)
    {
        ptlin1 = (LINE *) ptclin1->DATA;
        ptgab1 = ptlin1->GAB;
        for (pttrack1 = ptgab1->TRACK->NEXT; pttrack1; pttrack1 = pttrack1->NEXT)
        {
            pttrack_cour->NEXT = addtrack (pttrack_cour->NEXT, pttrack1->Y);
            pttrack_cour = pttrack_cour->NEXT;
        }
    }

    /* chainage des grilles des differentes lignes */
    ptclin1 = ptfig_dpr->LIN;
    for (ptclin2 = ptclin1->NEXT; ptclin2; ptclin1 = ptclin2, ptclin2 = ptclin2->NEXT)
    {
        ptlin1 = (LINE *) ptclin1->DATA;
        ptlin2 = (LINE *) ptclin2->DATA;
        for (noeud1 = ptlin1->GRID; noeud1->DPR_UP; noeud1 = noeud1->DPR_UP);
        for (noeud2 = ptlin2->GRID; noeud1; noeud1 = noeud1->NEXT, noeud2 = noeud2->NEXT)
        {
            noeud1->DPR_UP = noeud2;
            noeud2->DPR_DWN = noeud1;
        }
    }

    if (total_acc == 0)
        return pttrack_head;

    /* parcours de la grille totale pour diffuser les signaux restants */
    ptclin1 = ptfig_dpr->LIN;
    ptlin1 = (LINE *) ptclin1->DATA;
    ptnoeud = ptlin1->GRID;

    master_loop=FALSE;
    for (noeud1 = ptnoeud; noeud1; noeud1 = master_loop ? noeud1 : noeud1->NEXT)
    {
        if( ( verbose_mode == VERBOSE_ALL ) && ( !master_loop ) )
        {
            if( nucol == 0 )
            {
                fprintf( stdout, "     -> Left connectors column.\n" );
                fflush( stdout );
                nucol++;
            }
            else if( noeud1->NEXT == NULL )
            {
                fprintf( stdout, "     -> Right connectors column.\n" );
                fflush( stdout );
                nucol++;
            }
            else
            {
                fprintf( stdout, "     -> Column %d.\n", nucol++ );
                fflush( stdout );
            }
        }
        master_loop=FALSE;
        for (noeud2 = noeud1; noeud2; noeud2 = noeud2->DPR_UP)
        {
            for (acces = noeud2->DPR_LEFT; acces; acces = loop ? noeud2->DPR_LEFT : acces->NEXT)
            {
                loop = FALSE;
                ptsig = acces->SIG;
                if( !acces->POS )
                    continue;
                ptccolpos = acces->POS->CON->COL;
                ptcolpos = (COLUMN *) ptccolpos->DATA;

                if( (!acces->DONE)
                  ||( ( ptsig->WIDTH )
                    &&( ( ( ptcolpos->INDEX == (ptsig->MIN_COL + ptsig->WIDTH) ) && ( !sig_in_west( noeud2, ptsig ) ) )
                      ||( ( ptcolpos->INDEX == (ptsig->MIN_COL)                ) && ( !sig_in_east( noeud2, ptsig ) ) )
                      ||( !sig_crosscol( noeud2, ptsig ) ) ) ) )
                {
                    if (!ptsig->WIDTH)
                    {
                        if (ptsig->TYPE == INTERNAL)
                        {
                            loop = libere_null_width ( noeud2, ptfig_dpr->COL, ptsig, pttrack_head );
                        }
                        else
                        {
                            del_acces_sup( noeud2, acces->POS->CON, acces->ORIENT );
                            loop = TRUE;
                        }
                    }
                    else
                    {
                        if( noeud2->NEXT )
                            looking_for_sig (ptsig, noeud2->NEXT, &node_r, &acces_r);
                        else
                            acces_r = NULL;
                        if (acces_r)
                        {
                            if (ptcolpos->INDEX == (ptsig->MIN_COL + ptsig->WIDTH))
                            {
                                fflush( stdout );
                                fprintf( stderr, "*** dpr bug *** : bug when looking for right acces(1).\n" );
                            }
                            diff_allows (noeud2, ptccolpos, acces->POS, acces_r->POS->CON->COL, acces_r->SIG, pttrack_head);
                            loop = libere (noeud2,
                                           acces,
                                           ptfig_dpr->COL,
                                           acces->SIG,
                                           (ptcolpos->INDEX == ptsig->MIN_COL)?EAST:ORIENTX,
                                           pttrack_head);
                        }
                        else
                        {
                            if (ptcolpos->INDEX != (ptsig->MIN_COL + ptsig->WIDTH))
                            {
                                if( !check_and_add_message( ptsig->NAME ) )
                                {
                                    fflush( stdout );
                                    fprintf( stderr, "*** dpr error *** : can't find right connection for %s.",
                                             ptsig->NAME );
                                    fprintf( stderr, " Check control/data connectivity.\n" );
                                }
                            }
                            else
                            {
                                loop = libere (noeud2, acces, ptfig_dpr->COL, acces->SIG, WEST, pttrack_head);
                            }
                        }
                    }
                }
                if( loop )
                    master_loop = TRUE;
            }
        }
    }
    return pttrack_head;
}

/*---------------------------------------------------------\
                     generobs
\---------------------------------------------------------*/
 void   generobs(ptnoeud,ptccol_head,ptlin)
 NODE *ptnoeud;
 chain_list *ptccol_head;
 LINE *ptlin;

{
    NODE *noeud1=NULL,*noeud2=NULL;
    ACCESS *acces1=NULL, *acces2=NULL,*acc_obs=NULL;
    long index=0,savindex=0,y=0,col=0,savcol=0;
    TRACK *pttrack=NULL;
    chain_list *ptcallow=NULL;
    SEGMENT    *ptallow=NULL;
    chain_list *ptccol=NULL;
    chain_list *ptcinscol=NULL;
    INSTANCE   *ptinscol=NULL;
    int putobs;

    /* parcours de gauche a droite */
    for (noeud1=ptnoeud->NEXT;noeud1->NEXT;noeud1=noeud1->NEXT,col++)
    {
        index=0;
        /* parcours de bas en haut */
        for (noeud2=noeud1;noeud2;noeud2=noeud2->DPR_UP,index++)
        {
            putobs = FALSE;
            if (noeud2->DPR_RIGHT)
            {
                /*la pile des acces est non vide */
                if (noeud2->DPR_RIGHT->ORIENT==WEST)
                {
                    /* on a trouve un obstacle a droite de pile d'acces */
                    acc_obs = mkacces(NULL,noeud2->DPR_RIGHT,NULL,ORIENTX,NULL,TRUE);
                    noeud2->DPR_RIGHT->NEXT = acc_obs;
                    noeud2->DPR_RIGHT=acc_obs;
                    total_acc--;
                    putobs = TRUE;
                }
                if (noeud2->DPR_LEFT->ORIENT==EAST)
                {
                    /* on a trouve un obstacle a gauche de pile d'acces */
                    acc_obs = mkacces(noeud2->DPR_LEFT,NULL,NULL,ORIENTX,NULL,TRUE);
                    noeud2->DPR_LEFT->PREV = acc_obs;
                    noeud2->DPR_LEFT = acc_obs;
                    total_acc--;
                    putobs = TRUE;
                }

                /* parcours des acces de droite a gauche */
                for (acces1=noeud2->DPR_RIGHT;acces1;acces1=acces1->PREV)
                {
                    if(acces1->ORIENT==WEST)
                    {
                        if((acces1->NEXT->POS)&&(acces1->POS->X!=acces1->NEXT->POS->X)
                         &&(acces1->POS->ACCES!=EAST_WEST)) /*deux ref !=*/
                        {
                            acc_obs = mkacces(acces1->NEXT,acces1,NULL,
                                              ORIENTX,NULL,TRUE);
                            acces1->NEXT->PREV = acc_obs;
                            acces1->NEXT = acc_obs;
                            total_acc--;
                            putobs = TRUE;
                        }
                    }
                    else if(acces1->ORIENT==EAST)
                    {
                        if((acces1->PREV->POS)&&(acces1->POS->X!=acces1->PREV->POS->X)
                         &&(acces1->POS->ACCES!=EAST_WEST)) /*deux ref !=*/
                        {
                            acc_obs = mkacces(acces1,acces1->PREV,NULL,
                                              ORIENTX,NULL,TRUE);
                            acces1->PREV->NEXT = acc_obs;
                            acces1->PREV = acc_obs;
                            total_acc--;
                            putobs = TRUE;
                        }
                    }
                }

                if ( putobs )
                {
                    for (acces1=noeud2->DPR_RIGHT;acces1;acces1=acces1->PREV)
                    {
                        if( acces1->ORIENT == ORIENTX )
                        {
                            break;
                        }
                        if( acces1->ORIENT == WEST )
                        {
                            del_acces( noeud2, acces1 );
                        }
                    }
                    for (acces2=noeud2->DPR_LEFT;acces2;acces2=acces2->NEXT)
                    {
                        if( acces2->ORIENT == ORIENTX )
                        {
                            break;
                        }
                        if( acces2->ORIENT == EAST )
                        {
                            del_acces( noeud2, acces2 );
                        }
                    }
                    while( acces2 != acces1 )
                    {
                        del_acces( noeud2, acces2 );
                        acces2 = acces2->NEXT;
                    }
                }
            }
         else /* c.a.d pile des acces vide, vraie transparence?? */
            {
            /* verification de la coherence pour chaque noeud ayant une pile vide */
            /* avec la liste des transparences de sa colonne.        */
            /* on dispose de l'index on recherche l'ordonnee reelle  */

            /*recherche de y de l'horizontale */
            savindex = index;
            for (pttrack=ptlin->GAB->TRACK->NEXT;index;
                 index--,pttrack=pttrack->NEXT);
            index = savindex;
            y = pttrack->Y;
            /*recherche de la colonne*/
            savcol = col;
            for (ptccol=ptccol_head->NEXT->NEXT;col;
                 col--,ptccol=ptccol->NEXT->NEXT);
            col = savcol;
            /*recherche de l'instance*/
            for (ptcinscol=((COLUMN *)ptccol->DATA)->INS;ptcinscol;
                 ptcinscol=ptcinscol->NEXT)
                {
                ptinscol = (INSTANCE *)ptcinscol->DATA;
                if ((ptinscol->Y<y)&&((ptinscol->Y+ptinscol->HEIGHT)>y))
                   {
                   /*recherche de la coherence par rapport a la liste*/
                   /*des transparences de l'instance*/
                   for (ptcallow=ptinscol->ALLOW;((ptcallow)&&(ptcallow->DATA));
                        ptcallow=ptcallow->NEXT)
                       {
                       if (((SEGMENT *)ptcallow->DATA)->Y1==y)
                          break;
                       }
                   if ((!ptcallow)||(!ptcallow->DATA))
                      /*c.a.d on ne trouve pas de transparence*/
                      {
                      /* on a detecte un obstacle */

                      acc_obs=mkacces(NULL,NULL,NULL,ORIENTX,NULL,TRUE);
                      noeud2->DPR_RIGHT=acc_obs;
                      noeud2->DPR_LEFT=acc_obs;
                      total_acc--;
                      }
                   break;
                   }
                if (ptinscol->Y>y)
                   break;
                }
            }
         }
     }
 }



/*---------------------------------------------------------\
                     dprrout

Corps du programme principal de routage global.
\---------------------------------------------------------*/
void            dprrout (ptfig)
FIGURE         *ptfig;
{
    chain_list     *ptclin = NULL;
    int             bitslice = 0;
    LINE           *ptlin = NULL;
    NODE           *ptnoeud = NULL;
    NODE           *noeud1 = NULL;
    NODE           *noeud2 = NULL;
    ACCESS         *acces = NULL;
    ACCESS         *acces1 = NULL;
    TRACK          *pttrack_head = NULL;
    TRACK          *pttrack = NULL;
    chain_list     *cyclic_case = NULL;
    chain_list     *for_speed_up_scan2o = NULL;

    /* creation de la grille symbolique de la figure pour eviter les */
    /* connexions mono-connecteur (la grille est dupliquee et les    */
    /* obstacles n'y sont pas representes!)                          */

    if( verbose_mode )
    {
        fprintf( stdout, "   o Global routing.\n" );
        fflush( stdout );
    }
    /* creation des grilles pour chaque bit-slice */
    for (ptclin = ptfig->LIN; ptclin; ptclin = ptclin->NEXT)
    {
        ptlin = (LINE *) ptclin->DATA;
        ptlin->GRID = grille (ptfig->COL, ptclin);
    }

    /* positionnement de ptfig->GRID */
    ptclin = ptfig->LIN;
    ptlin = (LINE *) ptclin->DATA;
    ptfig->GRID = addchain (ptfig->GRID, (void *) ptlin->GRID);

    if( verbose_mode )
    {
        fprintf( stdout, "   o Bit-slice symbolic routing.\n" );
        fflush( stdout );
    }
    /* routage sur grille symbolique ligne par ligne */
    for (ptclin = ptfig->LIN; ptclin; ptclin = ptclin->NEXT)
    {
        if( verbose_mode == VERBOSE_ALL )
        {
            fprintf( stdout, "     -> Bit-slice %d.\n", bitslice );
            fflush( stdout );
        }
        bitslice = bitslice + 1;

        ptlin = (LINE *) ptclin->DATA;
        ptnoeud = ptlin->GRID;

        generobs(ptnoeud,ptfig->COL,ptlin);

        /* initialisation de la chain list des noeud max */
        for_speed_up_scan2o = NULL;
        for( noeud1 = ptnoeud; noeud1; noeud1 = noeud1->DPR_UP )
        {
            for_speed_up_scan2o = addchain( for_speed_up_scan2o, (void *) NULL );
        }
        for_speed_up_scan2o = reverse( for_speed_up_scan2o );

        while (heuristique_2o (ptfig->COL, ptnoeud, for_speed_up_scan2o))
            PRINT_POINT

        freechain( for_speed_up_scan2o );

        diff_line (ptfig->COL, ptnoeud);
        while (heuristique_3 (ptfig->COL, ptnoeud));
    }

    /* chainage des grilles et routage multi-bit si necessaire */
    pttrack_head = diff_all (ptfig);

    /* pour voir les acces qui restent et enlever ceux qui sont attendus */
    if (total_acc)
    {
        for (noeud1 = ((LINE *) ptfig->LIN->DATA)->GRID; noeud1; noeud1 = noeud1->NEXT)
            for (noeud2 = noeud1; noeud2; noeud2 = noeud2->DPR_UP)
                for (acces = noeud2->DPR_LEFT; acces; acces = acces1)
                {
                    acces1 = acces->NEXT;
                    if(!acces->DONE)
                    {
                        if (!in_con_chain( cyclic_case, acces->POS->CON ))
                        {
                            cyclic_case = addchain( cyclic_case, (void *)acces->POS->CON );
                            fflush( stdout );
                            fprintf (stderr,"*** dpr warning *** : Routing signal %s not finished in col %d.\n",
                                    acces->SIG->NAME, (((COLUMN *)acces->POS->CON->COL->DATA)->INDEX - 2 ) / 2);
                        }
                        else
                            del_acces (noeud2, acces);
                    }
                }
    }
    freechain( cyclic_case );

    /* parcours de la grille pour recuperer les transparences globales */
    for (noeud1 = ((LINE *) ptfig->LIN->DATA)->GRID; noeud1; noeud1 = noeud1->DPR_UP)
    {
        for (noeud2 = noeud1; noeud2; noeud2 = noeud2->NEXT)
            if (noeud2->DPR_LEFT)       /* c.a.d obstacle */
                break;
        if (!noeud2)            /* c.a.d traverse la grille */
        {
            /* extraction de y de la transparence */
            noeud2 = noeud1;
            for (pttrack = pttrack_head; noeud2; pttrack = pttrack->NEXT, noeud2 = noeud2->DPR_DWN);
            ptfig->ALLOW = addseg (ptfig->ALLOW, "", 0L, pttrack->Y, 0L, pttrack->Y,
                                   WLAYER2, 'H', TALU2);
        }
    }

}
