/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_allowed.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/
#include  <stdio.h>
#include  <strings.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_support.h"
#include  "strecher_intv.h"
#include  "strecher_allowed.h"

#define  mc_ABS(value)  ( (value)<0 ? -(value) : (value) )

/*---------------------------------------------------------\
                     getcutline
\---------------------------------------------------------*/
long            getcutline(ptfig, a_strech_name, p_ycut)
char           *a_strech_name;
phfig_list     *ptfig;
long           *p_ycut;
{
    long            found;
    char           *strech_name;
    phref_list     *pref;

    found = FALSE;
    strech_name = namealloc(a_strech_name);

    for (pref = ptfig->PHREF; pref != (phref_list *) NULL;
         pref = pref->NEXT)
        if (strech_name == pref->NAME)
        {
            if (!found)
            {
                *p_ycut = pref->YREF;
                found = TRUE;
            }
            else
            {
                fflush( stdout );
                fprintf( stderr, "*** strecher warning *** : line cut reference dupplicated in %s.\n",
                         ptfig->NAME );
            }
        }

    if (!found)
    {
        fflush( stdout );
        fprintf( stderr, "*** strecher warning *** : line cut reference not found in %s.\n",
                 ptfig->NAME );
    }
    return (found);
} /* end of getcutline */

/*---------------------------------------------------------\
                     findcutline
\---------------------------------------------------------*/
long            findcutline(ptfig, p_ycut)
phfig_list     *ptfig;
long           *p_ycut;
{
    ptype_list     *ptuser;
    intv_list      *pintv;
    long            optim_cut, real_cut, ycut;

    if (ptfig == (phfig_list *) NULL)
        return (FALSE);
    else
    {
        if ((ptuser = get_ls_data(ptfig, C_SUPP_UALLOWED, FALSE))
                == (ptype_list *) NULL)
        {
            fflush( stdout );
            fprintf( stderr, "*** strecher warning *** : Unstrechable cellule %s.\n", ptfig->NAME );
            return (FALSE);
        }

        if (ptuser->DATA == (void *) NULL)
            return (FALSE);
    }

    real_cut = ptfig->YAB1;
    optim_cut = (ptfig->YAB2 - ptfig->YAB1) / 2;

    for (pintv = (intv_list *) ptuser->DATA;
         pintv != (intv_list *) NULL;
         pintv = pintv->NEXT)
        for (ycut = pintv->INF; ycut < pintv->SUP;
             ycut += SCALE_X)
        {
            if (mc_ABS(ycut - optim_cut)
                < mc_ABS(real_cut - optim_cut))
            {
                real_cut = ycut;
                continue;
            }

            *p_ycut = real_cut;
            return (TRUE);
        }

    *p_ycut = real_cut;
    return (TRUE);
} /* end of findcutline */


/*---------------------------------------------------------\
                     modify_ls_allowed
\---------------------------------------------------------*/
void            modify_ls_allowed(a_ptfig_ALLOWED, a_ptfig_MODIFY, must_cross)
phfig_list     *a_ptfig_ALLOWED, *a_ptfig_MODIFY;
int must_cross;
{
    phseg_list     *pseg;
    phvia_list     *pvia;
    phcon_list     *pcon;
    ptype_list     *ptuser;
    intv_list     **ppthead;
    static phfig_list *ptfig_ALLOWED, *ptfig_MODIFY;

    if ((ptfig_MODIFY = a_ptfig_MODIFY) == (phfig_list *) NULL)
        return;

    /* Recherche du pointeur sur la structure ALLOWED. */
    if ((ptfig_ALLOWED = a_ptfig_ALLOWED) == (phfig_list *) NULL)
        return;

    ptuser = get_ls_data(ptfig_ALLOWED, C_SUPP_UALLOWED, F_SUPP_UCREAT);

    /* Reinitialisation de la liste d'intervalles. */
    if (ptfig_ALLOWED == ptfig_MODIFY)
    {
        free_ls_intv((intv_list *) ptuser->DATA);
        ptuser->DATA = (void *) NULL;
    }

    if (ptuser->DATA == (void *) NULL)
        ptuser->DATA = (void *) addintv((intv_list *) NULL,
                                        ptfig_ALLOWED->YAB1,
                                        ptfig_ALLOWED->YAB2);

    /* ppthead, pointeur sur la tete de liste. */
    ppthead = (intv_list **) & (ptuser->DATA);

    /* Elimination du bas de l'abutment-box.           */
    /* ( normalement, redondant avec les connecteurs ) */
    *ppthead = modify_ls_intv(*ppthead,
                              ptfig_MODIFY->YAB1 - SCALE_X,
                              ptfig_MODIFY->YAB1);

    /* Prise en compte des segments. */
    for (pseg = ptfig_MODIFY->PHSEG; pseg != (phseg_list *) NULL; pseg = pseg->NEXT)
        switch (pseg->LAYER)
        {
        case NTRANS:;
        case PTRANS:
            if (mc_IsHor(pseg->TYPE))
                *ppthead = modify_ls_intv(*ppthead, pseg->Y1 - SX_XTRANS, pseg->Y1 + SX_XTRANS);
            else
                *ppthead = modify_ls_intv(*ppthead, pseg->Y1, pseg->Y2);
        case NDIF:;
        case PDIF:
            if (mc_IsHor(pseg->TYPE))
            {
                *ppthead = modify_ls_intv(*ppthead,
                                          pseg->Y1 - pseg->WIDTH / 2,
                                          pseg->Y1 + pseg->WIDTH / 2);
            }
            else
            {
                if( !must_cross )
                {
                    *ppthead = modify_ls_intv(*ppthead, pseg->Y1, pseg->Y2);
                }
            }
            break;
        }

    /* Prise en compte des VIAs. */
    for (pvia = ptfig_MODIFY->PHVIA; pvia != (phvia_list *) NULL; pvia = pvia->NEXT)
        switch (pvia->TYPE)
        {
        case CONT_DIF_N:;
        case CONT_DIF_P:
            *ppthead = modify_ls_intv(*ppthead,
                                      pvia->YVIA - SX_SIDE_DIFF / 2,
                                      pvia->YVIA + SX_SIDE_DIFF / 2);
            break;
        case C_X_N:;
        case C_X_P:
            break;
        }

    /* Prise en compte des connecteurs d'alimentation. */
    /* ( vdd et vss, orientes EST/OUEST en ALU2 )      */
    for (pcon = ptfig_MODIFY->PHCON; pcon != (phcon_list *) NULL; pcon = pcon->NEXT)
        switch (pcon->LAYER)
        {
        case ALU2:
            if (((pcon->ORIENT == WEST) || (pcon->ORIENT == EAST))
             && (isvdd(pcon->NAME) || isvss(pcon->NAME)))
            {
                *ppthead = modify_ls_intv(*ppthead, pcon->YCON - SCALE_X, pcon->YCON);
            }
            break;
        }
}

/*---------------------------------------------------------\
                     norm_allowed
\---------------------------------------------------------*/
void            norm_allowed(ptfig)
phfig_list     *ptfig;
{
#define    access_allowed    ( (intv_list*)ptuser->DATA )

    ptype_list     *ptuser;
    intv_list      *pintv;

    /* Recherche de la liste d'autorisation. */
    if ((ptuser = get_ls_data(ptfig, C_SUPP_UALLOWED, FALSE)) == (ptype_list *) NULL)
        return;

    /* Boucle de conversion. */
    for (pintv = access_allowed; pintv != (intv_list *) NULL; pintv = pintv->NEXT)
    {
        /* Seuillage des bornes. */
        pintv->INF = lambda_floor(pintv->INF + SCALE_X);
        pintv->SUP = lambda_floor(pintv->SUP);
    }

#undef    access_allowed
}
