/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_bics.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  <string.h>
#include  MPH_H
#include  MUT_H

#include  "dpr.h"
#include  "strecher_intv.h"
#include  "strecher_support.h"
#include  "strecher_dtdupl.h"
#include  "strecher_upper.h"
#include  "strecher_phcut.h"
#include  "strecher_bics.h"
#include  "strecher_strech.h"
#include  "strecher_phtalu.h"

/*---------------------------------------------------------\
                     strecher
\---------------------------------------------------------*/
char           *strecher(source, param)
char           *source;
para_strech_list *param;
{
    dtdupl_list    *ptwork;
    phcut_list     *cuts = NULL;
    phfig_list     *ptphfig;
    static int      first = 1;
    int             problem = FALSE;
    intv_list      *coupes_norm;
    intv_list      *coupes_force;
    long            X;
    long            DX = 1;
    long            SX = 1;
    long            ymin;
    long            ymax;
    long            delta;
    intv_list      *c;
    ptype_list     *ptuser;

    if (first)
    {
        first = 0;
        global_init();
    }

    ptphfig = getphfig(source, 'P');
    if (get_ls_data(ptphfig, C_SUPP_UALLOWED, FALSE) != NULL)
    {
        return getdtdupl_byphfig(ptphfig)->DUPNAME;
    }

    ptwork = allocdtdupl(source, NULL, F_TDUPL_BLOCK);

    coupes_force = make_work(ptwork, F_UPPER_PHCUT, TRUE );
    ptuser = get_ls_data(ptwork->PHFIG, C_SUPP_UALLOWED, FALSE);
    if( ptuser )
    {
        ptuser->DATA = NULL;
    }
    coupes_norm  = make_work(ptwork, F_UPPER_PHCUT, FALSE);

    if (debug_mode)
    {
        fprintf(stdout, "Lignes de coupes disponibles NORMALES : ");
        for (c = coupes_norm; c; c = c->NEXT)
        {
            fprintf(stdout, "[%ld,%ld]", c->INF / SCALE_X, c->SUP / SCALE_X);
        }
        fprintf(stdout, "\n");
        fprintf(stdout, "Lignes de coupes disponibles FORCEES : ");
        for (c = coupes_force; c; c = c->NEXT)
        {
            fprintf(stdout, "[%ld,%ld]", c->INF / SCALE_X, c->SUP / SCALE_X);
        }
        fprintf(stdout, "\n");
    }

    for (; param; param = param->NEXT)
    {
        ymax = param->ymax;
        ymin = param->ymin;
        delta = param->delta;

        if (ymin >= ymax)
        {
            fprintf(stderr, "*** strecher bugs *** : in function strecher, ymin >= ymax (%ld, %ld)\n",
                    ymin, ymax);
            exit(1);
        }

        for (X = ((ymax - ymin) / 2) + ymin, DX = SX = 1;
             (X <= ymax) && (X >= ymin);
             X += SX * DX, DX += 1, SX = -SX)
        {
            if (in_ls_intv(coupes_norm, X * SCALE_X, X * SCALE_X))
            {
                if (debug_mode)
                {
                    fprintf(stdout, "Ligne de coupe utilisee : %ld(%ld).\n", X, X * SCALE_X);
                }
                cuts = addophcut(cuts, X * SCALE_X, delta);
                break;
            }
        }
        if ((X > ymax) || (X < ymin))
        {
            ptuser = get_ls_data(ptwork->PHFIG, C_SUPP_USUPPLY, FALSE);
            if( ptuser )
            {
                if( is_out_of_supply( ptuser->DATA, ymin * SCALE_X, ymax * SCALE_X ) )
                {
                    continue;
                }
            }

            fprintf(stderr, "*** strecher warning *** : Unable to strech %s, trying to cross diffusion.\n",
                    ptwork->FIGNAME);
            ptuser = get_ls_data(ptwork->PHFIG, C_SUPP_UALLOWED, FALSE);
            ptuser->DATA = (void *)coupes_force;
            problem = TRUE;
            for (X = ((ymax - ymin) / 2) + ymin, DX = SX = 1;
                 (X <= ymax) && (X >= ymin);
                 X += SX * DX, DX += 1, SX = -SX)
            {
                if (in_ls_intv(coupes_force, X * SCALE_X, X * SCALE_X))
                {
                    if (debug_mode)
                    {
                        fprintf(stdout, "Ligne de coupe utilisee : %ld(%ld).\n", X, X * SCALE_X);
                    }
                    cuts = addophcut(cuts, X * SCALE_X, delta);
                    break;
                }
            }
            if ((X > ymax) || (X < ymin))
            {
                fprintf(stderr, "*** strecher error *** : Unable to strech %s : no cut line between %ld and %ld.\n",
                        ptwork->FIGNAME, ymin, ymax);
                exit(1);
            }
        }
    }
    if( cuts == NULL )
    {
        freedtdupl( );
        return NULL;
    }
    nameddtdupl(ptwork, "*", cuts);
    ptwork->PHCUT = cuts;
    make_work(ptwork, F_UPPER_PHDUP, problem);

    return ptwork->DUPNAME;
}
