/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*********************************************************************\

    Fichier     :    strecher_dtdupl.c

    Date        :    12 octobre 1995
    Version     :    2.0

    Auteur      :    CHAPUT Jean-Paul
    Modifie par :    BEAURIN Oliver

\*********************************************************************/

/*---------------------------------------------------------\
                     includes
\---------------------------------------------------------*/
#include  <stdio.h>
#include  <math.h>
#include  MPH_H
#include  MUT_H

#include  "strecher_support.h"
#include  "strecher_phcut.h"
#include  "strecher_dtdupl.h"

/*---------------------------------------------------------\
                     globals
\---------------------------------------------------------*/
dtdupl_list    *TABLE_DUPL = (dtdupl_list *) NULL;

/*---------------------------------------------------------\
                     freedtdupl
\---------------------------------------------------------*/
void freedtdupl( )
{
    dtdupl_list    *pDEL;

    pDEL = TABLE_DUPL;
    TABLE_DUPL = TABLE_DUPL->NEXT;

    mbkfree( pDEL );
} /* end of freedtdupl */

/*---------------------------------------------------------\
                     allocdtdupl
\---------------------------------------------------------*/
dtdupl_list    *allocdtdupl(figname, ptcut, flags)
char           *figname;        /* Nom du modele.    */
phcut_list     *ptcut;          /* Liste des coupes. */
long            flags;          /* Drapeaux d'etats. */
{
    dtdupl_list    *pNEW;

    /* Allocation et chainage du nouvel element. */
    pNEW = (dtdupl_list *) mbkalloc(sizeof(dtdupl_list));

    pNEW->PHFIG = (phfig_list *) NULL;
    pNEW->PHDUP = (phfig_list *) NULL;
    pNEW->PHCUT = ptcut;
    pNEW->FIGNAME = namealloc(figname);
    pNEW->DUPNAME = NULL;
    pNEW->FLAGS = flags | C_TDUPL_USER;
    pNEW->NEXT = TABLE_DUPL;
    TABLE_DUPL = pNEW;

    return (pNEW);
}

/*---------------------------------------------------------\
                     nameddtdupl
\---------------------------------------------------------*/
void            nameddtdupl(ptwork, dupname, ptcut)
dtdupl_list    *ptwork;
char           *dupname;        /* Nom du duplicata. */
phcut_list     *ptcut;          /* Liste des coupes. */
{
    phcut_list     *cut;
    char           *p_figname, *p_dupname, s_dupname[MBK_SZ_NAME];
    long            checksum;

    p_figname = namealloc(ptwork->FIGNAME);

    if (dupname[0] == '*' && dupname[1] == '\0')
    {
        checksum = 0;
        for (cut = ptcut; cut; cut = cut->NEXT)
        {
            checksum += (cut->YCUT * 7 / (SCALE_X)) + cut->PITCHNB;
        }
        sprintf(s_dupname, "%s_%ld", ptwork->FIGNAME, checksum);
        p_dupname = namealloc(s_dupname);

    }
    else
    {
        p_dupname = namealloc(dupname);
    }

    if (p_dupname == p_figname)
    {
        fflush( stdout );
        fprintf( stderr,
                 "*** strecher error *** : original and streched view name are identical(%s,%s).\n",
                 p_dupname,p_figname );
        exit( 1 );
    }

    ptwork->DUPNAME = p_dupname;
}

/*---------------------------------------------------------\
                     getdtdupl
\---------------------------------------------------------*/
dtdupl_list    *getdtdupl(figname, ptcut)
char           *figname;        /* Nom du modele.    */
phcut_list     *ptcut;          /* Liste des coupes. */
{
    char           *p_figname;
    dtdupl_list    *pdupl;

    p_figname = namealloc(figname);

    for (pdupl = TABLE_DUPL; pdupl != (dtdupl_list *) NULL;
         pdupl = pdupl->NEXT)
        if ((pdupl->PHFIG->NAME == p_figname)
            && (cmp_ls_phcut(pdupl->PHCUT, ptcut)))
            return (pdupl);

    return ((dtdupl_list *) NULL);
}

/*---------------------------------------------------------\
                     getdtdupl_byphfig
\---------------------------------------------------------*/
dtdupl_list    *getdtdupl_byphfig(ptfig)
phfig_list     *ptfig;
{
    dtdupl_list    *pdupl;

    for (pdupl = TABLE_DUPL; pdupl != (dtdupl_list *) NULL; pdupl = pdupl->NEXT)
        if (pdupl->PHFIG == ptfig)
            return (pdupl);

    return ((dtdupl_list *) NULL);
}

/*---------------------------------------------------------\
                     getdtdupl_byphdup
\---------------------------------------------------------*/
dtdupl_list    *getdtdupl_byphdup(ptfig)
phfig_list     *ptfig;
{
    dtdupl_list    *pdupl;

    for (pdupl = TABLE_DUPL; pdupl != (dtdupl_list *) NULL; pdupl = pdupl->NEXT)
        if (pdupl->PHDUP == ptfig)
            return (pdupl);

    return ((dtdupl_list *) NULL);
}
