/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 * 
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 * 
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 * 
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 * 
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



/*
 * %% \textrm{This File Looks Nicer with\textsf{ LGrind} and \LaTeXe.}       %%
 * %|                                                                        |%
 * %| $Id: FpLibGN.c,v 1.6 2000/03/08 18:42:47 jpc Exp $        |%
 * %|                                                                        |%
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |          Alliance CAD System 3.2b                                  | |%
 * %| |     F p G e n   P a c k a g e   1.15                               | |%
 * %| |                                                                    | |%
 * %| |  Copyright(c) 95-97, MASI, CAO-VLSI Team                           | |%
 * %| |                                                                    | |%
 * %| |  Author  :              Jean-Paul CHAPUT                           | |%
 * %| |  E-mail  : alliance-support@asim.lip6.fr                           | |%
 * %| | ================================================================== | |%
 * %| |  Module  :                   "FpLibGN.c"                           | |%
 * %| | ****************************************************************** | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 */


# include  <stdio.h>
# include  <string.h>
# include  <time.h>

# include  <genlib.h>

# include  "GenericGN.h"
# include  "SharedGN.h"


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                        Internal Constants                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

# define       IC_GfpGN_CellNBInv            4
# define       IC_GfpGN_CellNBBuff           3
# define       IC_GfpGN_CellNBNand2          2
# define       IC_GfpGN_CellNBNand3          2
# define       IC_GfpGN_CellNBAnd2           2
# define       IC_GfpGN_CellNBAnd3           2
# define       IC_GfpGN_CellNBNor2           2
# define       IC_GfpGN_CellNBNor3           2
# define       IC_GfpGN_CellNBOr2            2
# define       IC_GfpGN_CellNBOr3            2
# define       IC_GfpGN_CellNBXnor2          1
# define       IC_GfpGN_CellNBXor2           1
# define       IC_GfpGN_CellNBNand2mask      3
# define       IC_GfpGN_CellNBNor2mask       3
# define       IC_GfpGN_CellNBXnor2mask      3
# define       IC_GfpGN_CellNBConst          2
# define       IC_GfpGN_CellNBNbuse          2
# define       IC_GfpGN_CellNBBuse           2
# define       IC_GfpGN_CellNBNmux2cs        2
# define       IC_GfpGN_CellNBMux2cs         2
# define       IC_GfpGN_CellNBMux3cd         2
# define       IC_GfpGN_CellNBMux3cs         2
# define       IC_GfpGN_CellNBMux4cs         2
# define       IC_GfpGN_CellNBMux5cs         2
# define       IC_GfpGN_CellNBMux6cs         2
# define       IC_GfpGN_CellNBAdd2s         (7+1)
# define       IC_GfpGN_CellNBNpl1           2
# define       IC_GfpGN_CellNBPl1            2
# define       IC_GfpGN_CellNBPmemnw         3
# define       IC_GfpGN_CellNBPmemnwq        3
# define       IC_GfpGN_CellNBPmem           3
# define       IC_GfpGN_CellNBPmemr          3
# define       IC_GfpGN_CellNBPmemq          3
# define       IC_GfpGN_CellNBPmemt          9
# define       IC_GfpGN_CellNBPmemrt         8
# define       IC_GfpGN_CellNBPmemtq         8


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |               Internal Structures and Variables                    | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/* Name of the leaf cells belonging to all generators. */
static char *TablLeafCell[] = {
    /* Boolean generators leaf cells. */
         "n1",    "np1",   "ndrv", "ndrvp",
       "buff",    "drv",  "drvp",
        "na2",   "na2p",
        "na3",   "na3p",
         "a2",    "a2p",
         "a3",    "a3p",
        "no2",   "no2p",
        "no3",   "no3p",
         "o2",    "o2p",
         "o3",    "o3p",
       "nxr2",
        "xr2",
    /* Mask generators leaf cells. */
      "bna2x",   "na2x", "n1na2x",
      "bno2x", "n1no2x",   "no2x",
     "bnxr2x", "n1xr2x",  "nxr2x",
    /* Buse generators leaf cells. */
       "bnt1",    "nt1",
        "bt1",     "t1",
    /* Multiplexer generators leaf cells. */
    "nmx2sb0", "nmx2sc",
     "mx2sb0",  "mx2sc",
     "mx3db0",  "mx3dc",
     "mx3sb0",  "mx3sc",
     "mx4sb0",  "mx4sc",
     "mx5sb0",  "mx5sc",
     "mx6sb0",  "mx6sc",
    /* Adder generators leaf cells. */
      "nobuffer",   "bcsa",
          "csag",   "csa0",  "csap",   "csai", "csap0", "csai0",
    /* Register generators leaf cells. */
      "pl1b0", "npl1a0",
      "pl1b0",  "pl1a0",
      "pmsb0", "pdffb0", "pmsa0",
      "pmsb0", "pdffb0", "pmsa1",
      "pmsb2", "pdffb2", "pmsa0",
      "pmsb2", "pdffb2", "pmsa1",
      "pmsba", "pdffba", "pmsa9",
      "pmsb6", "pdffb6",
      "pmsg4",  "pmsh4", "pmsp4", "pmsi4", "pmsq4", "pmsj4",
      "pmsb6", "pdffb6", "pdffub6",
      "pmsg5",  "pmsh5", "pmsp5", "pmsi5", "pmsq5", "pmsj5",
      "pmsbe", "pdffbe",
      "pmsgd",  "pmshd", "pmspd", "pmsid", "pmsqd", "pmsjd",
    /* Constant generators leaf cells. */
       "zero",    "one"
    };
/* Table for Grey code conversion. */
static long  TablGreyCode[16] =
    {
      0x00, 0x01, 0x03, 0x02,
      0x06, 0x07, 0x05, 0x04,
      0x0C, 0x0D, 0x0F, 0x0D,
      0x0A, 0x0B, 0x09, 0x08
    };


/*
 *   %| /----------------------------------------------------------------\ |%
 *   %| |                       Macros Definitions                       | |%
 *   %| \----------------------------------------------------------------/ |%
 */

/* Macros for physical construction. */
# define  mc_IsEmptySlice(i)  (       ((i) <  GV.aSlice)    \
                               || (   ((i) >= GV.aSlice     \
                                             +GV.aWidthOp)  \
                                   && ((i) <  GV.aWidthDP)))
# define  Mc_IndexNorth(n,i)  \
            ((((GV.aWidthDP - GV.aWidthOp - GV.aSlice) % 2) * (n)) + (i))
# define  Mc_IndexBuffr(n,i)  (i)
/* Macros for various index computing. */
# define  mc_SymbolicHeight(model) (GENLIB_HEIGHT(model) / GV.SliceHeight)
# define  mc_IndexToSlice(index) \
            (((index) >= GV.aWidthDP) ? (index) : ((index) + GV.aSlice))
# define  mc_IndexToSymF(index) \
            ((GV.aWidthDP - mc_IndexToSlice((index))) % 2)
# define  mc_SliceToPower(slice) \
            (((slice) + GV.aWidthDP % 2) / 2)
# define  mc_SliceToPowerType(slice) \
            (((slice) + GV.aWidthDP) % 2)
# define  mc_SliceToPowerName(slice) \
            ((mc_SliceToPowerType(slice)) ? "vdd" : "vss")
# define  mc_IndexToPower(index) \
            (mc_SliceToPower(mc_IndexToSlice((index))))
# define  mc_IndexToPowerName(index) \
            (mc_SliceToPowerName(mc_IndexToSlice((index))))
# define  mc_SliceToBit(slice) \
            ((GV.aMSB0) ? (GV.aWidthDP - 1 - slice) : (slice))
# define  mc_IndexToBit(index) \
            (mc_SliceToBit(mc_IndexToSlice((index))))
/* Macro for the logarithmic algorithms. */
# define  mc_GreyCode(class,level,code) \
            ((TablGreyCode[(code)]<<((level)-1))&((1L<<((class)-3))-1))
# define  mc_GreyChar(class,level,code) \
            ((mc_GreyCode(class,level,code) < 10)              \
                ? (char)('0'+mc_GreyCode(class,level,code))    \
                : (char)('a'+mc_GreyCode(class,level,code)-10))
# define  mc_GateNB(n,layer)        \
            ((layer <= 1)           \
                ?  (((n) - 1) >> 2) \
                : ((((n) - 1) + (1 << (layer))) >> ((layer) + 1)))
# define  mc_GateModulo(n) \
            ((((n) - 1) % 4) + 1)
/* Macro for standard terminals names */
# define  mc_GenL_FULL_BUS(conname)  \
            (GENLIB_BUS((conname), mc_IndexToBit(GV.aWidthOp - 1), \
                            mc_IndexToBit(0)))
# define  mc_COPY_UP_CMA_ELM(inscon,insname,conname,index)  \
            (UtMPh_CopyUpCMA(                \
                (inscon),                    \
                (insname),                   \
                UtBas_PrintF("%s[%ld]",      \
                    (conname),               \
                    mc_IndexToBit(index))))
# define  mc_POWER_GENLIB_NAME(slice)  \
            (GENLIB_NAME("%s_%d", mc_SliceToPowerName(slice), \
                           mc_SliceToPower(slice)))
# define  mc_VDDE(conname,index)  \
            ((GV.aPower)          \
              ? UtBas_PrintF("%s => vdd"   , (conname))  \
              : UtBas_PrintF("%s => vdd_%d", (conname),  \
                                             mc_IndexToPower(index)))
# define  mc_VSSE(conname,index)  \
            ((GV.aPower)          \
              ? UtBas_PrintF("%s => vss"   , (conname))  \
              : UtBas_PrintF("%s => vss_%d", (conname),  \
                                           mc_IndexToPower(index)))


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                  Internal Functions Declarations                   | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

static char *FunctionModelData __FP(( long aSlice ));
static char *FunctionModelBuff __FP(( long aSlice ));
static void  CreateHoDPThru    __FP(( long aSlice ));
static void  CreatePower       __FP(( long aSlice ));


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |                                                                    | |%
 * %| |                     Functions Definitions                          | |%
 * %| |                                                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "FpGN_FunctionArgs()"                                | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Get all the arguments from the command line, according to the function.|%
 * %| Also performs checking to ensure correctness and coherency of these    |%
 * %| parameters.                                                            |%
 * %|                                                                        |%
 */

extern void  FpGN_FunctionArgs( ArgList )
  va_list  ArgList;
{
  long  ArgFlags;

  /* Select whose arguments are to be read. */
  switch( GV.aFunction) {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
      ArgFlags =  IF_GfpGN_WidthOp | IF_GfpGN_Slice | IF_GfpGN_Drive; break;
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
      ArgFlags = IF_GfpGN_WidthOp | IF_GfpGN_Slice | IF_GfpGN_Mask; break;
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNul:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      ArgFlags = IF_GfpGN_WidthOp | IF_GfpGN_Slice; break;
    default:
      ArgFlags = 0L;
  } /* End of switch : select arguments to be read. */
  GV.aWidthDP = va_arg( ArgList, long );
  /* Get and check the data-path wide. */
  if (   (GV.aWidthDP < C_GfpGN_LimitsMinWidthDP)
      || (GV.aWidthDP > C_GfpGN_LimitsMaxWidthDP)) {
    GfpUT_PrintError( GV.ModuleId, IE_GfpGN_WidthDP,
                      GV.aWidthDP, C_GfpGN_LimitsMinWidthDP - 1,
                                   C_GfpGN_LimitsMaxWidthDP + 1 );
  }
  /* Get and check the bus wide. */
  if (ArgFlags & IF_GfpGN_WidthOp) {
    GV.aWidthOp = va_arg(ArgList, long);
    if (   (GV.aWidthOp < 1)
        || (GV.aWidthOp > GV.aWidthDP))
      GfpUT_PrintError(
        GV.ModuleId, IE_GfpGN_WIDTH, ShGN_FunctionStr(), GV.aWidthOp);
  }
  else
    GV.aWidthOp = GV.aWidthDP;
  /* Get and check the LSB starting slice. */
  if (ArgFlags & IF_GfpGN_Slice) {
    GV.aSlice = va_arg(ArgList, long);
    if (   (GV.aSlice             < 0)
        || (GV.aSlice+GV.aWidthOp > GV.aWidthDP))
      GfpUT_PrintError(
        GV.ModuleId, IE_GfpGN_SLICE, ShGN_FunctionStr(), GV.aSlice);
  }
  else
    GV.aSlice = 0L;
  /* Get and check the output drive capacity. */
  if (ArgFlags & IF_GfpGN_Drive) {
    GV.aDrive = va_arg(ArgList, long);
    if (   (GV.aDrive < 1)
        || (GV.aDrive > ShGN_FunctionDrive()))
      GfpUT_PrintError(
        GV.ModuleId, IE_GfpGN_DRIVE, ShGN_FunctionStr(), GV.aDrive);
  }
  else
    GV.aDrive = 1L;
  /* Get the Mask (extended long pointer). */
  if (ArgFlags & IF_GfpGN_Mask)
    GV.aXlMask = va_arg(ArgList, long*);
  else
    GV.aXlMask = (long*)NULL;
  /* Get the bus index convention. */
  GV.aMSB0 = va_arg(ArgList, long);
  /* Get all the views asked. */
  GV.aViews |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewVHDL    : 0L;
  GV.aViews |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewNetlist : 0L;
  GV.aViews |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewLayout  : 0L;
  GV.aViews |= (va_arg(ArgList, long) != 0L) ? IF_GfpGN_ViewIcon    : 0L;
  GV.aPower  =  va_arg(ArgList, long);
  /* Check the operator's width. */
  switch(GV.aFunction) {
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
    case C_GfpGN_FunctionUnknow:
      break;
    case C_GfpGN_FunctionNul:
      if (   (GV.aWidthOp <  4)
          || (GV.aWidthOp > 32)) {
		GfpUT_PrintError(
          GV.ModuleId, IE_GfpGN_WidthOp,
          ShGN_FunctionStr(), GV.aWidthOp, 4 - 1, 32 + 1 );
      }
      if (GV.aSlice % 2) {
		GfpUT_PrintError(
          GV.ModuleId, IE_GfpGN_SliceEven, ShGN_FunctionStr(), GV.aSlice );
      }
      break;
  } /* End of switch : operator's width. */
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "FpGN_FunctionCells()"                               | |%
 * %| \--------------------------------------------------------------------/ |%
 * %|                                                                        |%
 * %| Return the offset of the first operator leaf cell, in the              |%
 * %| "TablLeafCell" table.                                                  |%
 * %|                                                                        |%
 */

extern long  FpGN_FunctionCells()
{
  long  LeafCellIndex;

  LeafCellIndex = 0;
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionNul:     LeafCellIndex += IC_GfpGN_CellNBConst;
    case C_GfpGN_FunctionConst:   LeafCellIndex += IC_GfpGN_CellNBPmemrt;
    case C_GfpGN_FunctionPdffrt:
      LeafCellIndex += IC_GfpGN_CellNBPmemt;
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
      LeafCellIndex += IC_GfpGN_CellNBPmemtq;
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:
      LeafCellIndex += IC_GfpGN_CellNBPmemr;
    case C_GfpGN_FunctionPdffr:
      LeafCellIndex += IC_GfpGN_CellNBPmem;
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPdff:
      LeafCellIndex += IC_GfpGN_CellNBPmemq;
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffq:
      LeafCellIndex += IC_GfpGN_CellNBPmemnw;
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPdffnw:
      LeafCellIndex += IC_GfpGN_CellNBPmemnwq;
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPdffnwq:   LeafCellIndex += IC_GfpGN_CellNBPl1;
    case C_GfpGN_FunctionPl1:       LeafCellIndex += IC_GfpGN_CellNBNpl1;
    case C_GfpGN_FunctionNpl1:      LeafCellIndex += IC_GfpGN_CellNBAdd2s;
    case C_GfpGN_FunctionAdd2s:     LeafCellIndex += IC_GfpGN_CellNBMux6cs;
    case C_GfpGN_FunctionMux6cs:    LeafCellIndex += IC_GfpGN_CellNBMux5cs;
    case C_GfpGN_FunctionMux5cs:    LeafCellIndex += IC_GfpGN_CellNBMux4cs;
    case C_GfpGN_FunctionMux4cs:    LeafCellIndex += IC_GfpGN_CellNBMux3cs;
    case C_GfpGN_FunctionMux3cs:    LeafCellIndex += IC_GfpGN_CellNBMux3cd;
    case C_GfpGN_FunctionMux3cd:    LeafCellIndex += IC_GfpGN_CellNBMux2cs;
    case C_GfpGN_FunctionMux2cs:    LeafCellIndex += IC_GfpGN_CellNBNmux2cs;
    case C_GfpGN_FunctionNmux2cs:   LeafCellIndex += IC_GfpGN_CellNBBuse;
    case C_GfpGN_FunctionBuse:      LeafCellIndex += IC_GfpGN_CellNBNbuse;
    case C_GfpGN_FunctionNbuse:     LeafCellIndex += IC_GfpGN_CellNBXnor2mask;
    case C_GfpGN_FunctionXnor2mask: LeafCellIndex += IC_GfpGN_CellNBNor2mask;
    case C_GfpGN_FunctionNor2mask:  LeafCellIndex += IC_GfpGN_CellNBNand2mask;
    case C_GfpGN_FunctionNand2mask: LeafCellIndex += IC_GfpGN_CellNBXnor2;
    case C_GfpGN_FunctionXor2:      LeafCellIndex += IC_GfpGN_CellNBXor2;
    case C_GfpGN_FunctionXnor2:     LeafCellIndex += IC_GfpGN_CellNBOr3;
    case C_GfpGN_FunctionOr3:       LeafCellIndex += IC_GfpGN_CellNBOr2;
    case C_GfpGN_FunctionOr2:       LeafCellIndex += IC_GfpGN_CellNBNor3;
    case C_GfpGN_FunctionNor3:      LeafCellIndex += IC_GfpGN_CellNBNor2;
    case C_GfpGN_FunctionNor2:      LeafCellIndex += IC_GfpGN_CellNBAnd3;
    case C_GfpGN_FunctionAnd3:      LeafCellIndex += IC_GfpGN_CellNBAnd2;
    case C_GfpGN_FunctionAnd2:      LeafCellIndex += IC_GfpGN_CellNBNand3;
    case C_GfpGN_FunctionNand3:     LeafCellIndex += IC_GfpGN_CellNBNand2;
    case C_GfpGN_FunctionNand2:     LeafCellIndex += IC_GfpGN_CellNBBuff;
    case C_GfpGN_FunctionBuff:      LeafCellIndex += IC_GfpGN_CellNBInv;
    case C_GfpGN_FunctionInv:       ;
  }
  return( GV.LeafCellIndex = LeafCellIndex );
}


/*  Function  :  'FunctionModelData'                                        */
/* ------------------------------------------------------------------------ */
/*  Compute the model name of a data leaf cell, according to his slice index.
 */

static char *FunctionModelData( aSlice )
  long  aSlice;
{
  char *LeafCellModelName;
  long  extraOffset;


  extraOffset = 0L;

  /* Select the model name. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + GV.aDrive - 1 ];
      break;

    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

      LeafCellModelName = TablLeafCell[
                            GV.LeafCellIndex + 1 + GfpUT_XlGetBit(
                                                     GV.aXlMask,
                                                     aSlice - GV.aSlice ) ];
      break;

    case C_GfpGN_FunctionConst:

      LeafCellModelName = TablLeafCell[
                            GV.LeafCellIndex + GfpUT_XlGetBit(
                                                 GV.aXlMask,
                                                 aSlice - GV.aSlice ) ];
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 2 ];
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdfftq: extraOffset = 1;
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionAdd2s:
      if (aSlice == GV.aSlice)
        LeafCellModelName = TablLeafCell[
                              GV.LeafCellIndex + 6 + extraOffset
                              + (GV.aWidthDP - aSlice) % 2 ];
      else
      {
        if (aSlice == GV.aSlice + GV.aWidthOp - 1)
          LeafCellModelName = TablLeafCell[
                                GV.LeafCellIndex + 2 + extraOffset
                                + (GV.aWidthDP - aSlice) % 2 ];
        else
          LeafCellModelName = TablLeafCell[
                                GV.LeafCellIndex + 4 + extraOffset
                                + (GV.aWidthDP - aSlice) % 2 ];
      }
      break;

    default:

      return( (char*)NULL );
  }

  return( LeafCellModelName );
}


/*
 * %| /--------------------------------------------------------------------\ |%
 * %| |  Function  :  "FpGN_HaveBuff()"                                    | |%
 * %| \--------------------------------------------------------------------/ |%
 */

extern long  FpGN_HaveBuff()
{
  switch(GV.aFunction) {
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:
      return( TRUE );
      break;
    default:
      break;
  }
  return( FALSE );
}


/*  Function  :  'FunctionModelBuff'                                        */
/* ------------------------------------------------------------------------ */
/*  Compute the model name of a buffer leaf cell, according to his slice
 *  index.
 */

static char *FunctionModelBuff( aSlice )
  long  aSlice;
{
  char *LeafCellModelName;


  /* Select the buffer model name.
   * (NULL if no need of a buffer)
   */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex ];
      break;

    case C_GfpGN_FunctionAdd2s:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 1 ];
      break;

    case C_GfpGN_FunctionPdffut:

      LeafCellModelName = TablLeafCell[ GV.LeafCellIndex + 2 ];
      break;

    default:

      LeafCellModelName = (char*)NULL;
      break;
  }

  return( LeafCellModelName );
}


/*  Function  :  'FpGN_CreateNetlist'                                       */
/* ------------------------------------------------------------------------ */
/*  Generation of the Netlist view, using  GENLIB.                           */

extern void  FpGN_CreateNetlist()
{
  char *RootCellModelName;
  char *DataCellModelName;
  char *DataCellInstaName;
  char *BuffCellModelName;
  char *BuffCellInstaName;

  long  i;


  /* Select the buffer model name.
   * (NULL if no need of a buffer)
   */
  RootCellModelName = FunctionModelBuff( GV.aWidthDP );
  if (RootCellModelName != (char*)NULL)
  {
    BuffCellModelName = UtBas_PrintF( "%s_f%c",
                                      RootCellModelName,
                                      UtTec_V.LibrarySuffix );
  }
  else
    BuffCellModelName = (char*)NULL;


   GENLIB_DEF_LOFIG( GV.aModelName );


  /* Controls terminals. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:

       GENLIB_LOCON( "msk", IN, "msk" );
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:

       GENLIB_LOCON( "sel", IN, "sel" );
      break;

    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:

       GENLIB_LOCON( "sel[1:0]", IN, "sel[1:0]" );
      break;

    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:

       GENLIB_LOCON( "sel[2:0]", IN, "sel[2:0]" );
      break;

    case C_GfpGN_FunctionAdd2s:

       GENLIB_LOCON( "cout", OUT, "cout" );
       GENLIB_LOCON( "over", OUT, "over" );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:

       GENLIB_LOCON( "ck", IN, "ck" );
      break;

    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:

       GENLIB_LOCON( "scin", IN,  "scin" );
       GENLIB_LOCON( "test", IN,  "test" );

    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:

       GENLIB_LOCON(  "wen", IN,   "wen" );

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:

       GENLIB_LOCON(  "ckm", IN,   "ckm" );
       GENLIB_LOCON(  "cks", IN,   "cks" );
      break;

    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdfftq:

       GENLIB_LOCON( "scin", IN,  "scin" );
       GENLIB_LOCON( "test", IN,  "test" );

    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffq:

       GENLIB_LOCON(  "wen", IN,   "wen" );

    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:

       GENLIB_LOCON(   "ck", IN,    "ck" );
      break;

    case C_GfpGN_FunctionPdffrt:

       GENLIB_LOCON( "scin", IN,  "scin" );
       GENLIB_LOCON( "test", IN,  "test" );

    case C_GfpGN_FunctionPdffr:

       GENLIB_LOCON( "reset", IN, "reset" );
       GENLIB_LOCON(   "wen", IN,   "wen" );
       GENLIB_LOCON(    "ck", IN,    "ck" );
      break;

    default:

      break;
  }

  /* Input data buses terminals. */
  switch( GV.aFunction )
  {
    case C_GfpGN_FunctionMux6cs:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "i5" ), IN, mc_GenL_FULL_BUS( "i5" ) );

    case C_GfpGN_FunctionMux5cs:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "i4" ), IN, mc_GenL_FULL_BUS( "i4" ) );

    case C_GfpGN_FunctionMux4cs:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "i3" ), IN, mc_GenL_FULL_BUS( "i3" ) );

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "i2" ), IN, mc_GenL_FULL_BUS( "i2" ) );

    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionOr2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionAdd2s:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "i1" ), IN, mc_GenL_FULL_BUS( "i1" ) );

    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "i0" ), IN, mc_GenL_FULL_BUS( "i0" ) );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:
    case C_GfpGN_FunctionPdffq:
    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:
    case C_GfpGN_FunctionPdfftq:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "d" ), IN, mc_GenL_FULL_BUS( "d" ) );
      break;

    case C_GfpGN_FunctionConst:

      break;
  }

  /* Output data buses terminals. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionAnd3:
    case C_GfpGN_FunctionAnd2:
    case C_GfpGN_FunctionOr3:
    case C_GfpGN_FunctionOr2:

       GENLIB_LOCON( mc_GenL_FULL_BUS(  "o" ),   OUT, mc_GenL_FULL_BUS(  "o" ) );

      if ( UtTec_Family() == C_UtTec_FamilyCMOS )
      {
         GENLIB_LOCON( mc_GenL_FULL_BUS( "no" ), INOUT, mc_GenL_FULL_BUS( "no" ) );
      }

      break;

    case C_GfpGN_FunctionNand3:
    case C_GfpGN_FunctionNand2:
    case C_GfpGN_FunctionNor3:
    case C_GfpGN_FunctionNor2:
    case C_GfpGN_FunctionXnor2:
    case C_GfpGN_FunctionXor2:
    case C_GfpGN_FunctionInv:
    case C_GfpGN_FunctionBuff:
    case C_GfpGN_FunctionNand2mask:
    case C_GfpGN_FunctionNor2mask:
    case C_GfpGN_FunctionXnor2mask:
    case C_GfpGN_FunctionConst:
    case C_GfpGN_FunctionNmux2cs:
    case C_GfpGN_FunctionMux2cs:
    case C_GfpGN_FunctionMux3cd:
    case C_GfpGN_FunctionMux3cs:
    case C_GfpGN_FunctionMux4cs:
    case C_GfpGN_FunctionMux5cs:
    case C_GfpGN_FunctionMux6cs:
    case C_GfpGN_FunctionAdd2s:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "o" ), OUT, mc_GenL_FULL_BUS( "o" ) );
      break;

    case C_GfpGN_FunctionNbuse:
    case C_GfpGN_FunctionBuse:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "o" ), TRISTATE, mc_GenL_FULL_BUS( "o" ) );
      break;

    case C_GfpGN_FunctionNpl1:
    case C_GfpGN_FunctionPl1:
    case C_GfpGN_FunctionPmsnwq:
    case C_GfpGN_FunctionPmsq:
    case C_GfpGN_FunctionPdffnwq:
    case C_GfpGN_FunctionPdffq:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "q" ), INOUT, mc_GenL_FULL_BUS( "q" ) );
      break;

    case C_GfpGN_FunctionPmstq:
    case C_GfpGN_FunctionPdfftq:

       GENLIB_LOCON( mc_GenL_FULL_BUS( "q" ), INOUT, mc_GenL_FULL_BUS( "q" ) );
      break;

    case C_GfpGN_FunctionPmsnw:
    case C_GfpGN_FunctionPms:
    case C_GfpGN_FunctionPmst:
    case C_GfpGN_FunctionPdffnw:
    case C_GfpGN_FunctionPdff:
    case C_GfpGN_FunctionPdffr:

       GENLIB_LOCON( mc_GenL_FULL_BUS(  "q" ),   OUT, mc_GenL_FULL_BUS(  "q" ) );
       GENLIB_LOCON( mc_GenL_FULL_BUS( "nq" ),   OUT, mc_GenL_FULL_BUS( "nq" ) );
      break;

    case C_GfpGN_FunctionPdfft:
    case C_GfpGN_FunctionPdffut:
    case C_GfpGN_FunctionPdffrt:

       GENLIB_LOCON( mc_GenL_FULL_BUS(  "q" ), INOUT, mc_GenL_FULL_BUS(  "q" ) );
       GENLIB_LOCON( mc_GenL_FULL_BUS( "nq" ),   OUT, mc_GenL_FULL_BUS( "nq" ) );
      break;
  }


  /* Power supplies terminals. */
  if ( GV.aPower )
  {
     GENLIB_LOCON( "vdd", IN, "vdd" );
     GENLIB_LOCON( "vss", IN, "vss" );
  }
  else
  {
    for( i = 0 ;
         i <=   GV.aWidthDP
              + ((BuffCellModelName != (char*)NULL) ? 2 : 0) ;
         i++ )
       GENLIB_LOCON( mc_POWER_GENLIB_NAME( i ), IN, mc_POWER_GENLIB_NAME( i ) );
  }


  /* Loop for all the instances of the data-path column. */
  for( i = 0 ; i < GV.aWidthOp ; i++ )
  {
    /* Build leaf cell instance and model name. */
    DataCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          ShGN_FunctionStr(),
                          mc_IndexToSlice( i ) );
    DataCellModelName = UtBas_PrintF(
                          "%s_f%c" ,
                          FunctionModelData( mc_IndexToSlice( i ) ),
                          UtTec_V.LibrarySuffix );


    /* Leaf cell instanciation. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionNand3:
      case C_GfpGN_FunctionNor3:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionAnd3:
      case C_GfpGN_FunctionOr3:

        if ( UtTec_Family() == C_UtTec_FamilyCMOS )
        {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "no => no[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;

      case C_GfpGN_FunctionNand2 :
      case C_GfpGN_FunctionNor2  :
      case C_GfpGN_FunctionXnor2 :
      case C_GfpGN_FunctionXor2  :

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionAnd2:
      case C_GfpGN_FunctionOr2:

        if ( UtTec_Family() == C_UtTec_FamilyCMOS )
        {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "no => no[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;

      case C_GfpGN_FunctionInv:
      case C_GfpGN_FunctionBuff:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
                        "i1 => i1",
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionConst:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
          UtBas_PrintF( "o => o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "en  => s1",
                      "enb => s0",
          UtBas_PrintF( "i => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s1i0 => s1",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o =>  o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux3cd:
  
        GENLIB_LOINS(
          DataCellModelName,
          DataCellInstaName,
                        "s2",
                        "s1",
                        "s0",
          UtBas_PrintF( "i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);
  
        break;
  
      case C_GfpGN_FunctionMux3cs:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s1i0 => s1",
                      "s2i0 => s2",
                      "s3i0 => s3",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux4cs:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s0i1 => s0",
                      "s1i0 => s2",
                      "s1i1 => s3",
                      "s2i0 => s4",
                      "s3i0 => s5",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i3 => i3[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux5cs:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s0i1 => s0",
                      "s1i0 => s2",
                      "s1i1 => s3",
                      "s2i0 => s4",
                      "s3i0 => s5",
                      "s4i0 => s6",
                      "s5i0 => s7",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i3 => i3[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i4 => i4[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionMux6cs:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                      "s0i0 => s0",
                      "s0i1 => s1",
                      "s0i2 => s2",
                      "s1i0 => s3",
                      "s1i1 => s4",
                      "s1i2 => s5",
                      "s2i0 => s6",
                      "s3i0 => s7",
                      "s4i0 => s8",
                      "s5i0 => s9",
          UtBas_PrintF( "i0 => i0[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i1 => i1[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i2 => i2[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i3 => i3[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i4 => i4[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "i5 => i5[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "o => o[%ld]",  mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionAdd2s:

        if (i == 0)
        {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
            UtBas_PrintF(    "a =>  i1[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(    "b =>  i0[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(    "s =>   o[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF( "cout => cry%ld"  , mc_IndexToBit( i ) ),
                           "cin => cin",
            mc_VDDE( "vdd", i ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
            GENLIB_LOINSE(
              DataCellModelName,
              DataCellInstaName,
              UtBas_PrintF(   "a =>  i1[%ld]", mc_IndexToBit( i     ) ),
              UtBas_PrintF(   "b =>  i0[%ld]", mc_IndexToBit( i     ) ),
              UtBas_PrintF( "cin => cry%ld"  , mc_IndexToBit( i - 1 ) ),
              UtBas_PrintF(   "s =>   o[%ld]", mc_IndexToBit( i     ) ),
              (i == GV.aWidthOp)
                ?               "cout => cout"
                : UtBas_PrintF( "cout => cry%ld", mc_IndexToBit( i ) ),
              mc_VDDE( "vdd", i ),
              mc_VSSE( "vss", i + 1 ),
              0L);
        }

        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                       "ck => wmd",
          UtBas_PrintF( "i => d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF( "o => q[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdff:
        GENLIB_LOINSE(DataCellModelName,
          DataCellInstaName,
                       "s0i0 => s0i0",
                       "s1i0 => s1i0",
                        "wmd => wmd",
                        "ws  => ws" ,
          UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i     ),
          mc_VSSE( "vss", i + 1 ),
          0L);
        break;
      case C_GfpGN_FunctionPdffr:
        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                       "s0i0 => s0i0",
                       "s1i0 => s1i0",
                        "wmr => wmr",
                        "wmd => wmd",
                        "ws  => ws" ,
          UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i     ),
          mc_VSSE( "vss", i + 1 ),
          0L);
        break;
      case C_GfpGN_FunctionPmsnwq:
      case C_GfpGN_FunctionPmsq:
      case C_GfpGN_FunctionPdffnwq:
      case C_GfpGN_FunctionPdffq:

        GENLIB_LOINSE(
          DataCellModelName,
          DataCellInstaName,
                        "wmd => wmd",
                        "ws  => ws" ,
          UtBas_PrintF( "din => d[%ld]", mc_IndexToBit( i ) ),
          UtBas_PrintF(   "q => q[%ld]", mc_IndexToBit( i ) ),
          mc_VDDE( "vdd", i     ),
          mc_VSSE( "vss", i + 1 ),
          0L);

        break;

      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPdffut:
        if (i == 0) {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
                         "s0i0 => s0i0",
                         "s1i0 => s1i0",
                         "s2i0 => s2i0",
                         "s3i0 => s3i0",
                          "wmd => wmd",
                          "ws  => ws",
                          "tin => scin",
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        } else {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
                         "s0i0 => s0i0",
                         "s1i0 => s1i0",
                         "s2i0 => s2i0",
                         "s3i0 => s3i0",
                          "wmd => wmd",
                          "ws  => ws",
                          "tio => scin",
            UtBas_PrintF( "tin =>  q[%ld]", mc_IndexToBit( i - 1 ) ),
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i     ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        break;
      case C_GfpGN_FunctionPdffrt:
        if (i == 0) {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
                         "s0i0 => s0i0",
                         "s1i0 => s1i0",
                         "s2i0 => s2i0",
                         "s3i0 => s3i0",
                          "wmr => wmr",
                          "wmd => wmd",
                          "ws  => ws",
                          "tin => scin",
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        } else {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
                         "s0i0 => s0i0",
                         "s1i0 => s1i0",
                         "s2i0 => s2i0",
                         "s3i0 => s3i0",
                          "wmr => wmr",
                          "wmd => wmd",
                          "ws  => ws",
                          "tio => scin",
            UtBas_PrintF( "tin =>  q[%ld]", mc_IndexToBit( i - 1 ) ),
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(  "nq => nq[%ld]", mc_IndexToBit( i     ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        break;
      case C_GfpGN_FunctionPmstq:
      case C_GfpGN_FunctionPdfftq:

        if (i == 0)
        {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
                          "tin => scin",
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }
        else
        {
          GENLIB_LOINSE(
            DataCellModelName,
            DataCellInstaName,
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
                          "tio => scin",
            UtBas_PrintF( "tin =>  q[%ld]", mc_IndexToBit( i - 1 ) ),
            UtBas_PrintF( "din =>  d[%ld]", mc_IndexToBit( i     ) ),
            UtBas_PrintF(   "q =>  q[%ld]", mc_IndexToBit( i     ) ),
            mc_VDDE( "vdd", i     ),
            mc_VSSE( "vss", i + 1 ),
            0L);
        }

        break;
    }
  }


  /* Instantiation of the buffer cell, if needed. */
  if (BuffCellModelName != (char*)NULL)
  {
    /* Build leaf cell instance and model name. */
    BuffCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          ShGN_FunctionStr(),
                          mc_IndexToSlice( GV.aWidthDP ) );

    /* Buffer leaf cell instanciation. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                     "i => msk",
                     "s => i1",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "i   => sel",
                   "en  => s1",
                   "enb => s0",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel0 => sel",
                  "s0i0 => s0",
                  "s1i0 => s1",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux3cd:

        GENLIB_LOINS(
          BuffCellModelName,
          BuffCellInstaName,
                   "sel[1]",
                   "sel[0]",
                    "s2",
                    "s1",
                    "s0",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux3cs:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s1i0 => s1",
                  "s2i0 => s2",
                  "s3i0 => s3",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux4cs:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s1i0 => s2",
                  "s1i1 => s3",
                  "s2i0 => s4",
                  "s3i0 => s5",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux5cs:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel2 => sel[2]",
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s1i0 => s2",
                  "s1i1 => s3",
                  "s2i0 => s4",
                  "s3i0 => s5",
                  "s4i0 => s6",
                  "s5i0 => s7",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionMux6cs:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                  "sel2 => sel[2]",
                  "sel1 => sel[1]",
                  "sel0 => sel[0]",
                  "s0i0 => s0",
                  "s0i1 => s1",
                  "s0i2 => s2",
                  "s1i0 => s3",
                  "s1i1 => s4",
                  "s1i2 => s5",
                  "s2i0 => s6",
                  "s3i0 => s7",
                  "s4i0 => s8",
                  "s5i0 => s9",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionAdd2s:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
          UtBas_PrintF( "cin => cry%ld"  , mc_IndexToBit( GV.aWidthOp - 1 ) ),
                       "cout => cout",
                        "ovr => over",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                     "i => ck" ,
                     "s => wmd",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPmsnwq:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "ckm => ckm",
                   "cks => cks",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPmsq:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "wen => wen",
                  "s0i0 => s0i0",
                  "s1i0 => s1i0",
                   "ckm => ckm",
                   "cks => cks",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPmstq:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                         "scin => scin",
          UtBas_PrintF( "scout => q[%ld]",
                        mc_IndexToBit(   GV.aWidthOp - 1 ) ),
                         "test => test",
                          "wen => wen",
                          "ckm => ckm",
                          "cks => cks",
                          "wmt => wmt",
                          "wmd => wmd",
                          "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffq:
        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "wen => wen",
                   "ck  => ck ",
                  "s0i0 => s0i0",
                  "s1i0 => s1i0",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);
        break;
      case C_GfpGN_FunctionPdffr:
        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                 "reset => reset",
                   "wen => wen",
                    "ck => ck ",
                  "s0i0 => s0i0",
                  "s1i0 => s1i0",
                   "wmr => wmr",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);
        break;
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdffnwq:

        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                   "ck  => ck ",
                   "wmd => wmd",
                   "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;

      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPdffut:
      case C_GfpGN_FunctionPdfftq:
        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                         "scin => scin",
          UtBas_PrintF( "scout => q[%ld]",
                        mc_IndexToBit( GV.aWidthOp - 1 ) ),
                         "test => test",
                          "wen => wen",
                          "ck  => ck",
                         "s0i0 => s0i0",
                         "s1i0 => s1i0",
                         "s2i0 => s2i0",
                         "s3i0 => s3i0",
                          "wmd => wmd",
                          "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);

        break;
      case C_GfpGN_FunctionPdffrt:
        GENLIB_LOINSE(
          BuffCellModelName,
          BuffCellInstaName,
                         "scin => scin",
          UtBas_PrintF( "scout => q[%ld]",
                        mc_IndexToBit( GV.aWidthOp - 1 ) ),
                         "test => test",
                        "reset => reset",
                          "wen => wen",
                          "ck  => ck",
                         "s0i0 => s0i0",
                         "s1i0 => s1i0",
                         "s2i0 => s2i0",
                         "s3i0 => s3i0",
                          "wmr => wmr",
                          "wmd => wmd",
                          "ws  => ws",
          mc_VDDE( "vdd",  GV.aWidthDP     ),
          mc_VSSE( "vss",  GV.aWidthDP + 1 ),
          mc_VSSE( "vss1", GV.aWidthDP + 2 ),
          0L);
        break;
    }
  }


  GENLIB_SAVE_LOFIG();
}


/*  Function  :  'FpGN_CreateNetlistComp'                                   */
/* ------------------------------------------------------------------------ */
/*  Generation of the comparator Netlist view, using  GENLIB. This function
 *  is separated from "CreateNetlist" because the don't have intrinsics
 *  regularities.
 */

extern void  FpGN_CreateNetlistComp()
{
  long  PowerIndex;
  long    InsSlice;
  long    InsSymF;
  long  ModelPref;
  long  FeedThrough, GreyCode;
  long      GateNB, GateModulo, GateTrunc, Offset, Step, i, Class, Level;

  char *ModelName;
  char   *InsName;
  char   *z16Name;
  char    *z8Name;
  char    *z4Name;


   GENLIB_DEF_LOFIG(GV.aModelName);


  /* Input data buses terminals. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionNul:
       GENLIB_LOCON(mc_GenL_FULL_BUS("i0"), IN, mc_GenL_FULL_BUS("i0"));
      break;
  }


  /* Output data buses terminals. */
  switch(GV.aFunction)
  {
    case C_GfpGN_FunctionNul:
       GENLIB_LOCON("o",   OUT, "o");
      break;
  }


  /* Power supplies terminals. */
  if (GV.aPower)
  {
     GENLIB_LOCON("vdd", IN, "vdd");
     GENLIB_LOCON("vss", IN, "vss");
  }
  else
    for(InsSlice = 0 ; InsSlice <= GV.aWidthDP ; InsSlice++)
       GENLIB_LOCON(mc_POWER_GENLIB_NAME(InsSlice), IN, mc_POWER_GENLIB_NAME(InsSlice));


  if (GV.aWidthOp < 5)
  {
    InsSlice  = mc_IndexToSlice( 0 );
    InsSymF   = mc_IndexToSymF(  0 );

    /* Build leaf cell model name. */
    ModelName = UtBas_PrintF("z%dt0w0%c_f%c",
                             GV.aWidthOp,
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    /* Build leaf cell instance name. */
    InsName = UtBas_PrintF("z%dt0w0_%ld",
                           GV.aWidthOp,
                           InsSlice);

    switch(GV.aWidthOp)
    {
      case 4:

        GENLIB_LOINSE(
          ModelName,
            InsName,
          UtBas_PrintF( "i3   => i0[%d]", mc_IndexToBit( 3 )),
          UtBas_PrintF( "i2   => i0[%d]", mc_IndexToBit( 2 )),
          UtBas_PrintF( "i1   => i0[%d]", mc_IndexToBit( 1 )),
          UtBas_PrintF( "i0   => i0[%d]", mc_IndexToBit( 0 )),
                        "zero =>  o",
          mc_VDDE("vdd"  , 0),
          mc_VSSE("vss"  , 1),
          mc_VDDE("vdd_1", 2),
          mc_VSSE("vss_1", 3),
          (InsSymF) ? mc_VDDE("vdd_2", 4)
                    : mc_VSSE("vss_2", 4),
          0L);

        break;
    }
  }
  else
  {
    /* Compute the class of the operator, i.e. the number of logicals
     * levels, with a logarithmic tree.
     */
    Class = ShGN_llog2(GV.aWidthOp - 1) + 1;

    /* Compute the first level number of gate. */
    GateNB     = mc_GateNB(    GV.aWidthOp, 1);
    GateModulo = mc_GateModulo(GV.aWidthOp);

    /* Two first levels of the operator. */
    for(i = 0 ; i < GateNB ; i++)
    {
      InsSlice = mc_IndexToSlice(4 * i);
      InsSymF  = mc_IndexToSymF( 4 * i);

      ModelName = UtBas_PrintF("nz4l0w%c%c_f%c",
                               mc_GreyChar(Class,1,i),
                               (InsSymF) ? 'i' : 'p',
                               UtTec_V.LibrarySuffix);
      InsName   = UtBas_PrintF("nz4l0w%c_%ld",
                               mc_GreyChar(Class,1,i), InsSlice);


      switch(mc_GreyCode(Class,1,i))
      {
        case 0x00:

          GENLIB_LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]", mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;

        case 0x01:

          GENLIB_LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]",   mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]",   mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]",   mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]",   mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i     ),
            UtBas_PrintF( "z8   =>  z8_%d" , i >> 1),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;

        case 0x02:

          GENLIB_LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]",   mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]",   mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]",   mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]",   mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i     ),
            UtBas_PrintF( "z16  => z16_%d" , i >> 2),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;

        case 0x03:

          GENLIB_LOINSE(
            ModelName,
              InsName,
            UtBas_PrintF( "i3   =>  i0[%d]", mc_IndexToBit(4 * i + 3)),
            UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
            UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
            UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
            UtBas_PrintF( "z4   =>  z4_%d" , i),
            UtBas_PrintF( "z8   =>  z8_%d" , i >> 1),
            UtBas_PrintF( "z16  => z16_%d" , i >> 2),
            mc_VDDE("vdd"  , 4 * i + 0),
            mc_VSSE("vss"  , 4 * i + 1),
            mc_VDDE("vdd_1", 4 * i + 2),
            mc_VSSE("vss_1", 4 * i + 3),
            0L);

          break;
      }
    }

    GateTrunc   = 4 * i;
    FeedThrough = FALSE;

    ModelPref = mc_GreyCode(Class, 1, i) & 1L;
    ModelName = UtBas_PrintF("%s%ldt0w%c%c_f%c",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    InsSlice  = mc_IndexToSlice(4 * i);
    InsSymF   = mc_IndexToSymF( 4 * i);
    InsName   = UtBas_PrintF("%s%ldt0w%c_%ld",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             InsSlice);

    /* Instanciation of the last gate of the first level.
     * (the model of this gate is different from the previous)
     */
    switch(GreyCode = mc_GreyCode(Class,1,i))
    {
      case 0x00:
      case 0x01:
      case 0x02:

        z4Name = UtBas_PrintF( "z%d", 4 << GreyCode );

        switch(GateModulo)
        {
          case 1:

            if (!ModelPref)
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
                UtBas_PrintF( "%s => i0[%d]", z4Name, mc_IndexToBit(4 * i)),
                mc_VDDE("vdd", 4 * i + 0),
                mc_VSSE("vss", 4 * i + 1),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
                UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i)),
                UtBas_PrintF( "%s => %s_%d", z4Name, z4Name, i >> GreyCode),
                mc_VDDE("vdd", 4 * i + 0),
                mc_VSSE("vss", 4 * i + 1),
                0L);
            }

            break;

          case 2:

            GENLIB_LOINSE(
              ModelName,
                InsName,
              UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
              UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
              UtBas_PrintF( "%s   =>  %s_%d" , z4Name, z4Name, i >> GreyCode),
              mc_VDDE("vdd"  , 4 * i + 0),
              mc_VSSE("vss"  , 4 * i + 1),
              (InsSymF) ? mc_VDDE("vdd_1", 4 * i + 2)
                        : mc_VSSE("vss_1", 4 * i + 2),
              0L);

            break;

          case 3:

            GENLIB_LOINSE(
              ModelName,
                InsName,
              UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
              UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
              UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
              UtBas_PrintF( "%s   =>  %s_%d" , z4Name, z4Name, i >> GreyCode),
              mc_VDDE("vdd"  , 4 * i + 0),
              mc_VSSE("vss"  , 4 * i + 1),
              mc_VDDE("vdd_1", 4 * i + 2),
              mc_VSSE("vss_1", 4 * i + 3),
              0L);

            break;

          case 4:

            GENLIB_LOINSE(
              ModelName,
                InsName,
              UtBas_PrintF( "i3   =>  i0[%d]", mc_IndexToBit(4 * i + 3)),
              UtBas_PrintF( "i2   =>  i0[%d]", mc_IndexToBit(4 * i + 2)),
              UtBas_PrintF( "i1   =>  i0[%d]", mc_IndexToBit(4 * i + 1)),
              UtBas_PrintF( "i0   =>  i0[%d]", mc_IndexToBit(4 * i    )),
              UtBas_PrintF( "%s   =>  %s_%d" , z4Name, z4Name, i >> GreyCode),
              mc_VDDE("vdd"  , 4 * i + 0),
              mc_VSSE("vss"  , 4 * i + 1),
              mc_VDDE("vdd_1", 4 * i + 2),
              mc_VSSE("vss_1", 4 * i + 3),
              (InsSymF) ? mc_VDDE("vdd_2", 4 * i + 4)
                        : mc_VSSE("vss_2", 4 * i + 4),
              0L);

            break;
        }

        break;

      case 0x03:

        z16Name = UtBas_PrintF( "z%d", 2 <<  GreyCode      );
         z8Name = UtBas_PrintF( "z%d", 2 << (GreyCode - 1) );
         z4Name = UtBas_PrintF( "z%d", 2 << (GreyCode - 2) );

        switch(GateModulo)
        {
          case 1:

            if (ModelPref)
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "%s => i0[%d]",  z4Name, mc_IndexToBit(4 * i)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd", 4 * i + 0),
        mc_VSSE("vss", 4 * i + 1),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i)),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd", 4 * i + 0),
        mc_VSSE("vss", 4 * i + 1),
                0L);
            }

            break;

          case 2:

            if (ModelPref)
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i1 => i0[%d]",          mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "%s => i0[%d]",  z4Name, mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        (InsSymF) ? mc_VDDE("vdd_1", 4 * i + 2)
                  : mc_VSSE("vss_1", 4 * i + 2),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i1 => i0[%d]", mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        (InsSymF) ? mc_VDDE("vdd_1", 4 * i + 2)
                  : mc_VSSE("vss_1", 4 * i + 2),
                0L);
            }

            break;

          case 3:

            if (ModelPref)
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i2 => i0[%d]",          mc_IndexToBit(4 * i + 2)),
        UtBas_PrintF( "i1 => i0[%d]",          mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "%s => i0[%d]",  z4Name, mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        mc_VDDE("vdd_1", 4 * i + 2),
        mc_VSSE("vss_1", 4 * i + 3),
                0L);

              FeedThrough = TRUE;
            }
            else
            {
              GENLIB_LOINSE(
                ModelName,
                  InsName,
        UtBas_PrintF( "i2 => i0[%d]", mc_IndexToBit(4 * i + 2)),
        UtBas_PrintF( "i1 => i0[%d]", mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        mc_VDDE("vdd_1", 4 * i + 2),
        mc_VSSE("vss_1", 4 * i + 3),
                0L);
            }

            break;

          case 4:

            GENLIB_LOINSE(
              ModelName,
                InsName,
        UtBas_PrintF( "i3 => i0[%d]", mc_IndexToBit(4 * i + 3)),
        UtBas_PrintF( "i2 => i0[%d]", mc_IndexToBit(4 * i + 2)),
        UtBas_PrintF( "i1 => i0[%d]", mc_IndexToBit(4 * i + 1)),
        UtBas_PrintF( "i0 => i0[%d]", mc_IndexToBit(4 * i    )),
        UtBas_PrintF( "%s => %s_%d" ,  z4Name,  z4Name, i >> (GreyCode - 3)),
        UtBas_PrintF( "%s => %s_%d" ,  z8Name,  z8Name, i >> (GreyCode - 2)),
        UtBas_PrintF( "%s => %s_%d" , z16Name, z16Name, i >> (GreyCode - 1)),
        mc_VDDE("vdd"  , 4 * i + 0),
        mc_VSSE("vss"  , 4 * i + 1),
        mc_VDDE("vdd_1", 4 * i + 2),
        mc_VSSE("vss_1", 4 * i + 3),
        (InsSymF) ? mc_VDDE("vdd_2", 4 * i + 4)
                  : mc_VSSE("vss_2", 4 * i + 4),
              0L);

            break;
        }

        break;
    }

    /* Process levels from 3 to upper. */
    for(Level = 2 ; (GateNB = mc_GateNB(GV.aWidthOp, Level)) ; Level++)
    {
      char *PinZero, *PinI3, *PinZ2X, *PinZX_up, *PinZX_down;


      /* Compute for the curent level :
       *   - Step   = Number of slices between two instances.
       *   - Offset = Offset of the first instance.
       */
      Step   = (1L << (Level + 1));
      Offset = (1L <<  Level) - 1;

      for(i = 0 ; i < GateNB ; i++)
      {
        InsSlice = mc_IndexToSlice(Step * i + Offset);
        InsSymF  = mc_IndexToSymF( Step * i + Offset);

        if (Class == Level + 1)
        {
          /* Special cells of the last level. */
          ModelName = UtBas_PrintF("%s2t%ldw0%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);
          InsName   = UtBas_PrintF("%s2t%ldw0_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   InsSlice);

          /* The output of the cells is the final output. */
          PinZero = "zero => o";
        }
        else
        {
          ModelName = UtBas_PrintF("%s2l%ldw%c%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);
          InsName   = UtBas_PrintF("%s2l%ldw%c_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   InsSlice);

          PinZero   = UtBas_PrintF("z%d => z%d_%d", Step, Step, i);
        }


        /* Terminal names assignments. */
        PinI3    = UtBas_PrintF(
                     "i3     =>  i0[%d]",
                     mc_IndexToBit(Step * i + Offset));
        PinZ2X  = UtBas_PrintF(
                    "z%d     => z%d_%d",
                    Step << 1,
                    Step << 1,
                       i >> 1);
        PinZX_up = UtBas_PrintF(
                     "z%d_up => z%d_%d",
                     Step >> 1,
                     Step >> 1,
                        i << 1);

        if ((GateTrunc - 1 == Step * i + Offset) && (FeedThrough))
        {
          PinZX_down = UtBas_PrintF(
                         "z%d_down => i0[%d]",
                         Step >> 1,
                         mc_IndexToBit(Step * i + Offset + 1));
        }
        else
        {
          PinZX_down = UtBas_PrintF(
                         "z%d_down => z%d_%d",
                         Step >> 1,
                         Step >> 1,
                         (i << 1) + 1);
        }


        switch(mc_GreyCode(Class,Level,i))
        {
          case 0x00:

            GENLIB_LOINSE(
              ModelName,
                InsName,
              PinI3,
              PinZX_down,
              PinZX_up,
              PinZero,
              mc_VDDE("vdd", Step * i + Offset    ),
              mc_VSSE("vss", Step * i + Offset + 1),
              0L);

            break;

          case 0x02:

              GENLIB_LOINSE(
               ModelName,
                 InsName,
               PinI3,
               PinZX_down,
               PinZX_up,
               PinZero,
               PinZ2X,
               mc_VDDE("vdd", Step * i + Offset    ),
               mc_VSSE("vss", Step * i + Offset + 1),
               0L);

            break;
        }
      }
    }
  }

  GENLIB_SAVE_LOFIG();
}


/*  Function  :  'CreateHoDPThru'                                           */
/* ------------------------------------------------------------------------ */

static void  CreateHoDPThru( aSlice )
  long  aSlice;
{
  long  X, i;
  char *sName="Transparency";


  X =   aSlice * UtTec_V.SliceHeight
      + (mc_SliceToPowerType( aSlice ) ? UtTec_V.ThruTop : UtTec_V.ThruBot);

  for( i = 0; i < UtTec_V.ThruNB; i++ )
  {
    GENLIB_PHSEG( UtTec_V.ThruLayer,
           UtTec_V.ThruWide,
           sName,
                              0, X,
           GENLIB_WIDTH(GV.aModelName), X );

    X += UtTec_V.ThruPitch;
  }
}


/*  Function  :  'CreatePower'                                              */
/* ------------------------------------------------------------------------ */
/*  Add a power line at the bottom of the slice given by "aSlice".          */

static void  CreatePower( aSlice )
  long  aSlice;
{
  long  PowerIndex;
  char *PowerName;


  GENLIB_PHSEG( UtTec_V.PowerLayer,
         mc_SliceToPowerType( aSlice ) ? UtTec_V.PowerWideVDD
                                       : UtTec_V.PowerWideVSS,
         mc_POWER_GENLIB_NAME( aSlice ),
         0,
         aSlice * UtTec_V.SliceHeight,
         GENLIB_WIDTH( GV.aModelName ),
         aSlice * UtTec_V.SliceHeight );

  GENLIB_PHCON( UtTec_V.PowerLayer,
         mc_SliceToPowerType( aSlice ) ? UtTec_V.PowerWideVDD
                                       : UtTec_V.PowerWideVSS,
         mc_POWER_GENLIB_NAME( aSlice ),
         WEST,
         0,
         aSlice * UtTec_V.SliceHeight );

  GENLIB_PHCON( UtTec_V.PowerLayer,
         mc_SliceToPowerType( aSlice ) ? UtTec_V.PowerWideVDD
                                       : UtTec_V.PowerWideVSS,
         mc_POWER_GENLIB_NAME( aSlice ),
         EAST,
         GENLIB_WIDTH( GV.aModelName ),
         aSlice * UtTec_V.SliceHeight );
}



/*  Function  :  'FpGN_CreateLayout'                                        */
/* ------------------------------------------------------------------------ */
/*  Generation of the Layout view, using  GENLIB.                            */

extern void  FpGN_CreateLayout()
{
  char *RootCellModelName;
  char *DataCellModelName;
  char *BuffCellModelName;
  char *DataCellInstaName;
  char *BuffCellInstaName;

  long  i, WireCTRL, InsSlice;


  /* Select the buffer model name.
   * (NULL if no need of a buffer)
   */
  RootCellModelName = FunctionModelBuff( GV.aWidthDP );
  if (RootCellModelName != (char*)NULL)
  {
    BuffCellModelName = UtBas_PrintF( "%s_f%c",
                                      RootCellModelName,
                                      UtTec_V.LibrarySuffix );
  }
  else
    BuffCellModelName = (char*)NULL;


  /* Check if the buffer is contiguous to the data-cells. */
  WireCTRL = (GV.aWidthDP == GV.aSlice + GV.aWidthOp) ? FALSE : TRUE;


  GENLIB_DEF_PHFIG(GV.aModelName);


  /* Loop for all the instances of the data-path column. */
  for( i = 0 ; i < GV.aWidthOp ; i++ )
  {
    /* Build leaf cell instance and model name. */
    DataCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          ShGN_FunctionStr(),
                          mc_IndexToSlice( i ) );
    DataCellModelName = UtBas_PrintF(
                          "%s_f%c" ,
                          FunctionModelData( mc_IndexToSlice( i ) ),
                          UtTec_V.LibrarySuffix );

    GENLIB_PLACE( DataCellModelName,
           DataCellInstaName,
              mc_IndexToSymF(  i ) ? SYM_Y : NOSYM,
           0, mc_IndexToSlice( i ) * UtTec_V.SliceHeight);
  }


  /* Instantiation of the buffer cell, if needed. */
  if (BuffCellModelName != (char*)NULL)
  {
    /* Build leaf cell instance name. */
    BuffCellInstaName = UtBas_PrintF( "%s_%ld",
                                      ShGN_FunctionStr(),
                                      mc_IndexToSlice( GV.aWidthDP ) );

    GENLIB_PLACE(BuffCellModelName,
          BuffCellInstaName,
          NOSYM,
          0, mc_IndexToSlice( GV.aWidthDP ) * UtTec_V.SliceHeight);
  }


  /* Set the size of the abutment box.
   */
  GENLIB_DEF_AB( 0,        - GV.aSlice   * UtTec_V.SliceHeight,
          0, (BuffCellModelName != (char*)NULL)
               ? 0
               : ((  GV.aWidthDP
                   - GV.aWidthOp
                   - GV.aSlice  ) * UtTec_V.SliceHeight ) );


  /* Loop for all the data slices of the data-path.
   * Copy up virtual terminals, nwell references and feed through.
   */
  for( i = 0 ; i < GV.aWidthOp ; i++ )
  {
    /* Build leaf cell instance and model name. */
    DataCellInstaName = UtBas_PrintF(
                          "%s_%ld",
                          ShGN_FunctionStr(),
                          mc_IndexToSlice( i ) );

    UtMPh_CopyUpHoDPThru( DataCellInstaName );
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:

        UtMPh_CopyUpNwell( DataCellInstaName );
        break;
    }

    /* Copy input CMAs from the instances. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionMux6cs:

        mc_COPY_UP_CMA_ELM( "i5", DataCellInstaName, "i5", i );

      case C_GfpGN_FunctionMux5cs:

        mc_COPY_UP_CMA_ELM( "i4", DataCellInstaName, "i4", i );

      case C_GfpGN_FunctionMux4cs:

        mc_COPY_UP_CMA_ELM( "i3", DataCellInstaName, "i3", i );

      case C_GfpGN_FunctionNand3:
      case C_GfpGN_FunctionNor3:
      case C_GfpGN_FunctionAnd3:
      case C_GfpGN_FunctionOr3:
      case C_GfpGN_FunctionMux3cd:
      case C_GfpGN_FunctionMux3cs:

        mc_COPY_UP_CMA_ELM( "i2", DataCellInstaName, "i2", i );

      case C_GfpGN_FunctionNand2:
      case C_GfpGN_FunctionNor2:
      case C_GfpGN_FunctionXnor2:
      case C_GfpGN_FunctionXor2:
      case C_GfpGN_FunctionAnd2:
      case C_GfpGN_FunctionOr2:
      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:

        mc_COPY_UP_CMA_ELM( "i1", DataCellInstaName, "i1", i );

      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        mc_COPY_UP_CMA_ELM( "i0", DataCellInstaName, "i0", i );
        break;

      case C_GfpGN_FunctionInv:
      case C_GfpGN_FunctionBuff:
      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        mc_COPY_UP_CMA_ELM( "i", DataCellInstaName, "i0", i );
        break;

      case C_GfpGN_FunctionConst:

        break;

      case C_GfpGN_FunctionAdd2s:

        mc_COPY_UP_CMA_ELM( "b", DataCellInstaName, "i0", i );
        mc_COPY_UP_CMA_ELM( "a", DataCellInstaName, "i1", i );
        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        mc_COPY_UP_CMA_ELM( "i", DataCellInstaName, "d", i );
        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffr:
      case C_GfpGN_FunctionPmsnwq:
      case C_GfpGN_FunctionPmsq:
      case C_GfpGN_FunctionPdffnwq:
      case C_GfpGN_FunctionPdffq:
      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPmstq:
      case C_GfpGN_FunctionPdffut:
      case C_GfpGN_FunctionPdffrt:
      case C_GfpGN_FunctionPdfftq:

        mc_COPY_UP_CMA_ELM( "din", DataCellInstaName, "d", i );
        break;
    }

    /* Copy output CMAs from the instances. */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionAnd3:
      case C_GfpGN_FunctionOr3:
      case C_GfpGN_FunctionAnd2:
      case C_GfpGN_FunctionOr2:

        if ( UtTec_Family() == C_UtTec_FamilyCMOS )
        {
          mc_COPY_UP_CMA_ELM( "no", DataCellInstaName, "no", i );
        }

      case C_GfpGN_FunctionNand3:
      case C_GfpGN_FunctionNor3:
      case C_GfpGN_FunctionNand2:
      case C_GfpGN_FunctionNor2:
      case C_GfpGN_FunctionXnor2:
      case C_GfpGN_FunctionXor2:
      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:
      case C_GfpGN_FunctionInv:
      case C_GfpGN_FunctionBuff:
      case C_GfpGN_FunctionConst:
      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:
      case C_GfpGN_FunctionNmux2cs:
      case C_GfpGN_FunctionMux2cs:
      case C_GfpGN_FunctionMux3cd:
      case C_GfpGN_FunctionMux3cs:
      case C_GfpGN_FunctionMux4cs:
      case C_GfpGN_FunctionMux5cs:
      case C_GfpGN_FunctionMux6cs:

        mc_COPY_UP_CMA_ELM( "o", DataCellInstaName, "o", i );
        break;

      case C_GfpGN_FunctionAdd2s:

        mc_COPY_UP_CMA_ELM( "s", DataCellInstaName, "o", i );
        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:

        mc_COPY_UP_CMA_ELM( "o", DataCellInstaName, "q", i );
        break;

      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffr:
      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPdffut:
      case C_GfpGN_FunctionPdffrt:

        mc_COPY_UP_CMA_ELM( "nq", DataCellInstaName, "nq", i );

      case C_GfpGN_FunctionPmsnwq:
      case C_GfpGN_FunctionPmsq:
      case C_GfpGN_FunctionPdffnwq:
      case C_GfpGN_FunctionPdffq:
      case C_GfpGN_FunctionPmstq:
      case C_GfpGN_FunctionPdfftq:

        mc_COPY_UP_CMA_ELM( "q", DataCellInstaName, "q", i );
        break;
    }
  }


  /* Copy up virtual terminals, references and controls terminals
   * from the buffer model. Add wires to connect buffer to
   * data cells if there is a gap.
   */
  if (BuffCellModelName != (char*)NULL)
  {
    /* Build leaf cell name. */
    BuffCellInstaName = UtBas_PrintF( "%s_%ld", ShGN_FunctionStr(), GV.aWidthDP );

    UtMPh_CopyUpHoDPThru( BuffCellInstaName );
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:

        UtMPh_CopyUpNwell( BuffCellInstaName );
        break;
    }

    /* Connect the internal output of the buffer to the
     * data control terminal.
     */
    switch( GV.aFunction )
    {
      case C_GfpGN_FunctionNand2mask:
      case C_GfpGN_FunctionNor2mask:
      case C_GfpGN_FunctionXnor2mask:

        if (WireCTRL)
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "i1", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName,  "s", 0 );
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "i", BuffCellInstaName, "msk" );
        break;

      case C_GfpGN_FunctionNbuse:
      case C_GfpGN_FunctionBuse:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "enb", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "enb", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "en", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "en", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "i", BuffCellInstaName, "sel" );
        break;

      case C_GfpGN_FunctionMux6cs:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s0i2", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s0i2", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s1i2", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s1i2", 0 );
        }

      case C_GfpGN_FunctionMux5cs:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s5i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s5i0", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s4i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s4i0", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "sel2", BuffCellInstaName, "sel[2]" );

      case C_GfpGN_FunctionMux4cs:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s0i1", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s0i0", Mc_IndexBuffr( 2, 0 ) );
        }

        GENLIB_WIRE1( ALU1, 1,
               DataCellInstaName, "s1i1", Mc_IndexNorth( 1, 0 ),
               BuffCellInstaName, "s1i1", Mc_IndexBuffr( 1, 0 ) );

      case C_GfpGN_FunctionMux3cs:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s3i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s3i0", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s2i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s2i0", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "sel1", BuffCellInstaName, "sel[1]" );

      case C_GfpGN_FunctionMux2cs:
      case C_GfpGN_FunctionNmux2cs:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s1i0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s1i0", Mc_IndexBuffr( 1, 0 ) );

          if (   (GV.aFunction == C_GfpGN_FunctionMux4cs)
              || (GV.aFunction == C_GfpGN_FunctionMux5cs))
          {
            GENLIB_WIRE1( ALU1, 1,
                   DataCellInstaName, "s0i0", Mc_IndexNorth( 1, 0 ),
                   BuffCellInstaName, "s0i0", Mc_IndexBuffr( 2, 1 ) );
          }
          else
          {
            GENLIB_WIRE1( ALU1, 1,
                   DataCellInstaName, "s0i0", Mc_IndexNorth( 1, 0 ),
                   BuffCellInstaName, "s0i0", Mc_IndexBuffr( 1, 0 ) );
          }
        }
        /* Set up the external control(s) terminal(s). */
        if (   (GV.aFunction == C_GfpGN_FunctionMux2cs)
            || (GV.aFunction == C_GfpGN_FunctionNmux2cs))
          GENLIB_COPY_UP_CON_FACE( NORTH, "sel0", BuffCellInstaName, "sel" );
        else
          GENLIB_COPY_UP_CON_FACE( NORTH, "sel0", BuffCellInstaName, "sel[0]" );
        break;

      case C_GfpGN_FunctionMux3cd:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s2", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s2", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s1", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s1", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "s0", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "s0", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "sel1", BuffCellInstaName, "sel[1]" );
        GENLIB_COPY_UP_CON_FACE( NORTH, "sel0", BuffCellInstaName, "sel[0]" );
        break;

      case C_GfpGN_FunctionAdd2s:

        if (WireCTRL)
        {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "cin", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "cin", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "cout", 0,
                 BuffCellInstaName, "cout", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "cout", BuffCellInstaName, "cout" );
        GENLIB_COPY_UP_CON_FACE( NORTH,  "ovr", BuffCellInstaName, "over" );
        break;

      case C_GfpGN_FunctionNpl1:
      case C_GfpGN_FunctionPl1:
        if (WireCTRL)
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "ck", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName,  "s", 0 );
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "i", BuffCellInstaName, "ck" );
        break;
      case C_GfpGN_FunctionPms:
      case C_GfpGN_FunctionPmsq:
        GENLIB_COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );
        GENLIB_WIRE1( ALU1, 1, DataCellInstaName, "s1i0", Mc_IndexNorth( 1, 0 ),
                        BuffCellInstaName, "s1i0", 0 );
        GENLIB_WIRE1( ALU1, 1, DataCellInstaName, "s0i0", Mc_IndexNorth( 1, 0 ),
                        BuffCellInstaName, "s0i0", 0 );
      case C_GfpGN_FunctionPmsnw:
      case C_GfpGN_FunctionPmsnwq:
        if (WireCTRL) {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "wmd", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmd", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "ws", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "ws", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "ckm", BuffCellInstaName, "ckm" );
        GENLIB_COPY_UP_CON_FACE( NORTH, "cks", BuffCellInstaName, "cks" );
        break;
      case C_GfpGN_FunctionPmst:
      case C_GfpGN_FunctionPmstq:
        GENLIB_COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );
        if (WireCTRL) {
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,  "s1i0", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "s1i0", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,  "s0i0", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "s0i0", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,   "wmd", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,   "wmd", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,    "ws", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,    "ws", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,   "tio", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "scin", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,     "q", 0,
                          BuffCellInstaName, "scout", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH,  "scin", BuffCellInstaName,  "scin" );
        GENLIB_COPY_UP_CON_FACE( NORTH, "scout", BuffCellInstaName,
          UtBas_PrintF( "q[%d]", mc_IndexToBit( GV.aWidthOp - 1 ) ) );
        GENLIB_COPY_UP_CON_FACE( NORTH,  "test", BuffCellInstaName,  "test" );
        GENLIB_COPY_UP_CON_FACE( NORTH,   "ckm", BuffCellInstaName,   "ckm" );
        GENLIB_COPY_UP_CON_FACE( NORTH,   "cks", BuffCellInstaName,   "cks" );
        break;
      case C_GfpGN_FunctionPdffr:
        if (WireCTRL)
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName, "wmr", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName, "wmr", 0 );
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "reset", BuffCellInstaName, "reset" );
      case C_GfpGN_FunctionPdff:
      case C_GfpGN_FunctionPdffq:
        GENLIB_COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );
        GENLIB_WIRE1( ALU1, 1, DataCellInstaName, "s1i0", Mc_IndexNorth( 1, 0 ),
                        BuffCellInstaName, "s1i0", 0 );
        GENLIB_WIRE1( ALU1, 1, DataCellInstaName, "s0i0", Mc_IndexNorth( 1, 0 ),
                        BuffCellInstaName, "s0i0", 0 );
      case C_GfpGN_FunctionPdffnw:
      case C_GfpGN_FunctionPdffnwq:
        if (WireCTRL) {
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "wmd", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmd", 0 );
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "ws", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "ws", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "ck", BuffCellInstaName, "ck" );
        break;
      case C_GfpGN_FunctionPdffrt:
        if (WireCTRL)
          GENLIB_WIRE1( ALU1, 1,
                 DataCellInstaName, "wmr", Mc_IndexNorth( 1, 0 ),
                 BuffCellInstaName, "wmr", 0 );
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH, "reset", BuffCellInstaName, "reset" );
      case C_GfpGN_FunctionPdfft:
      case C_GfpGN_FunctionPdffut:
      case C_GfpGN_FunctionPdfftq:
        GENLIB_COPY_UP_CON_FACE( NORTH, "wen", BuffCellInstaName, "wen" );
        if (WireCTRL) {
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,  "s3i0", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "s3i0", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,  "s2i0", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "s2i0", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,  "s1i0", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "s1i0", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,  "s0i0", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "s0i0", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,   "wmd", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,   "wmd", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,    "ws", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,    "ws", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,   "tio", Mc_IndexNorth( 1, 0 ),
                          BuffCellInstaName,  "scin", 0 );
          GENLIB_WIRE1( ALU1, 1, DataCellInstaName,     "q", 0,
                          BuffCellInstaName, "scout", 0 );
        }
        /* Set up the external control(s) terminal(s). */
        GENLIB_COPY_UP_CON_FACE( NORTH,  "scin", BuffCellInstaName,  "scin" );
        GENLIB_COPY_UP_CON_FACE( NORTH, "scout", BuffCellInstaName,
          UtBas_PrintF( "q[%d]", mc_IndexToBit( GV.aWidthOp - 1 ) ) );
        GENLIB_COPY_UP_CON_FACE( NORTH,  "test", BuffCellInstaName,  "test" );
        GENLIB_COPY_UP_CON_FACE( NORTH,    "ck", BuffCellInstaName,    "ck" );
        break;
    }
  }


  /* Addition of power supplies connectors and wires.
   */
  for(InsSlice  = 0;
      InsSlice <=   GV.aWidthDP
                  + ( (BuffCellModelName != (char*)NULL) ? 2 : 0 );
      InsSlice++)
  {
    CreatePower( InsSlice );

    if (mc_IsEmptySlice( InsSlice ))
      CreateHoDPThru( InsSlice );
  }


  GENLIB_SAVE_PHFIG();
}

/*  Function  :  'FpGN_CreateLayoutComp'                                    */
/* ------------------------------------------------------------------------ */
/*  Generation of the Layout view for heterogeneous operators, using  GENLIB.
 */

extern void  FpGN_CreateLayoutComp()
{
  long  PowerIndex;
  long    BusIndex;
  long    InsSlice;
  long    InsSymF;
  long  ModelPref;

  char *ModelName;
  char   *InsName;
  char    *z4Name;

  long  GateNB, GateModulo, Offset, Step, i, Class, Level;


  if (GV.aPower)
    GfpUT_PrintWarng(GV.ModuleId,
                     IW_GfpGN_PowerMerge,
                     ShGN_FunctionStr(),
                     GV.aModelName);


  GENLIB_DEF_PHFIG(GV.aModelName);


  /* Case of the one cell comparators (wide < 5). */
  if (GV.aWidthOp < 5)
  {
    InsSlice  = mc_IndexToSlice( 0 );
    InsSymF   = mc_IndexToSymF(  0 );

    /* Build leaf cell model name. */
    ModelName = UtBas_PrintF("z%dt0w0%c_f%c",
                             GV.aWidthOp,
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    /* Build leaf cell instance name. */
    InsName = UtBas_PrintF("z%dt0w0_%ld",
                           GV.aWidthOp,
                           InsSlice);


    GENLIB_PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
          0, InsSlice * UtTec_V.SliceHeight);

    /* Set the size of the abutment box.
     */
    GENLIB_DEF_AB(0,  - GV.aSlice    * UtTec_V.SliceHeight,
           0, (  GV.aWidthDP
               - GV.aWidthOp
               - GV.aSlice  ) * UtTec_V.SliceHeight);


    UtMPh_CopyUpHoDPThru(InsName);
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:
  
        UtMPh_CopyUpNwell(InsName);
        break;
    }
  
    UtMPh_CopyUpCMA("zero", InsName, "o");

    switch(GV.aWidthOp)
    {
      case 4:

        UtMPh_CopyUpCMA("i3", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(3)));
  
      case 3:
  
        UtMPh_CopyUpCMA("i2", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(2)));
  
      case 2:
  
        UtMPh_CopyUpCMA("i1", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(1)));
  
      case 1:
  
        UtMPh_CopyUpCMA("i0", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(0)));
  
        break;
    }
  }
  else
  {
    /* Compute the class of the operator, i.e. the number of logicals
     * levels, with a logarithmic tree.
     */
    Class = ShGN_llog2(GV.aWidthOp - 1) + 1;

    /* Compute the first level number of gate. */
    GateNB     = mc_GateNB(    GV.aWidthOp, 1);
    GateModulo = mc_GateModulo(GV.aWidthOp);

    /* Place cells of the two first levels of the operator. */
    for(i = 0 ; i < GateNB ; i++)
    {
      InsSlice  = mc_IndexToSlice(4 * i);
      InsSymF   = mc_IndexToSymF( 4 * i);

      /* Build leaf cell model name. */
      ModelName = UtBas_PrintF("nz4l0w%c%c_f%c",
                               mc_GreyChar(Class,1,i),
                               (InsSymF) ? 'i' : 'p',
                               UtTec_V.LibrarySuffix);

      /* Build leaf cell instance name. */
      InsName = UtBas_PrintF("nz4l0w%c_%ld",
                             mc_GreyChar(Class,1,i),
                             InsSlice);

      GENLIB_PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
            0, InsSlice * UtTec_V.SliceHeight);
    }



    /* Compute the model and instance name
     * of last cell of the first level.
     */
    ModelPref = mc_GreyCode(Class, 1, i) & 1;
    ModelName = UtBas_PrintF("%s%ldt0w%c%c_f%c",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             (InsSymF) ? 'i' : 'p',
                             UtTec_V.LibrarySuffix);

    InsSymF   = mc_IndexToSymF( 4 * i);
    InsSlice  = mc_IndexToSlice(4 * i);
    InsName   = UtBas_PrintF("%s%ldt0w%c_%ld",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             InsSlice);

    /* Place last cell of the first level. */
    GENLIB_PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
          0, InsSlice * UtTec_V.SliceHeight);


    /* Place cells of the levels from 3 to upper. */
    for(Level = 2 ; (GateNB = mc_GateNB(GV.aWidthOp, Level)) ; Level++)
    {
      /* Compute for the curent level :
       *   - Step   = Number of slices between two instances.
       *   - Offset = Offset of the first instance.
       */
      Step   = (1L << (Level + 1));
      Offset = (1L <<  Level) - 1;

      for( i=0 ; i<GateNB ; i++ )
      {
        InsSlice = mc_IndexToSlice(Step * i + Offset);
        InsSymF  = mc_IndexToSymF( Step * i + Offset);

        if (Class == Level + 1)
          ModelName = UtBas_PrintF("%s2t%ldw0%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);
        else
          ModelName = UtBas_PrintF("%s2l%ldw%c%c_f%c",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   (InsSymF) ? 'p' : 'i',
                                   UtTec_V.LibrarySuffix);

        if (Class == Level + 1)
          InsName   = UtBas_PrintF("%s2t%ldw0_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   InsSlice);
        else
          InsName   = UtBas_PrintF("%s2l%ldw%c_%ld",
                                   (Level%2) ? "nz" : "z",
                                    Level - 1,
                                   mc_GreyChar(Class,Level,i),
                                   InsSlice);

        GENLIB_PLACE(ModelName, InsName, (InsSymF) ? SYM_Y : NOSYM,
              0, InsSlice*UtTec_V.SliceHeight);
      }
    }

    /* Set the size of the abutment box.
     */
    GENLIB_DEF_AB(0,  - GV.aSlice    * UtTec_V.SliceHeight,
           0, (  GV.aWidthDP
               - GV.aWidthOp
               - GV.aSlice  ) * UtTec_V.SliceHeight);

    /* Compute the first level number of gate. */
    GateNB     = mc_GateNB(    GV.aWidthOp, 1);
    GateModulo = mc_GateModulo(GV.aWidthOp);

    for( i=0 ; i < GateNB ; i++)
    {
      InsSlice = mc_IndexToSlice(4*i);

      /* Build leaf cell instance name. */
      InsName = UtBas_PrintF("nz4l0w%c_%ld",
                             mc_GreyChar(Class,1,i),
                             InsSlice);

      UtMPh_CopyUpHoDPThru(InsName);
      switch( UtTec_Family() )
      {
        case C_UtTec_FamilyAsGa:

          break;

        case C_UtTec_FamilyCMOS:

          UtMPh_CopyUpNwell( InsName );
          break;
      }

      UtMPh_CopyUpCMA("i0", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(4*i  )));
      UtMPh_CopyUpCMA("i1", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(4*i+1)));
      UtMPh_CopyUpCMA("i2", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(4*i+2)));
    }


    ModelPref = mc_GreyCode(Class, 1, i) & 1L;
    InsSlice  = mc_IndexToSlice(4*i);
    InsName   = UtBas_PrintF("%s%ldt0w%c_%ld",
                             (ModelPref) ? "z" : "nz",
                             GateModulo,
                             mc_GreyChar(Class,1,i),
                             InsSlice);
  
    UtMPh_CopyUpHoDPThru(InsName);
    switch( UtTec_Family() )
    {
      case C_UtTec_FamilyAsGa:

        break;

      case C_UtTec_FamilyCMOS:
  
        UtMPh_CopyUpNwell(InsName);
        break;
    }
  
    switch(GateModulo)
    {
      case 4:

        UtMPh_CopyUpCMA("i3", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(4*i+3)));
  
      case 3:
  
        UtMPh_CopyUpCMA("i2", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(4*i+2)));
  
      case 2:
  
        UtMPh_CopyUpCMA("i1", InsName, GENLIB_NAME("i0[%d]", mc_IndexToBit(4*i+1)));
  
      case 1:
  
        switch(mc_GreyCode(Class,1,i))
        {
          case 0x03:

            z4Name = (GateModulo  < 4) ?  "z4" : "i0"; break;

          case 0x02:

            z4Name = (GateModulo == 1) ? "z16" : "i0"; break;

          case 0x01:

            z4Name = "i0"; break;

          case 0x00:

            z4Name = (GateModulo == 1)  ? "z4" : "i0"; break;
        }

        UtMPh_CopyUpCMA(z4Name,
                        InsName,
                        GENLIB_NAME("i0[%d]",
                        mc_IndexToBit(4*i)));
  
        break;
    }


    for(Level = 2 ; (GateNB = mc_GateNB(GV.aWidthOp, Level)) ; Level++)
    {
      /* Compute for the curent level :
       *   - Step   = Number of slices between two instances.
       *   - Offset = Offset of the first instance.
       */
      Step   = (1L << (Level + 1));
      Offset = (1L <<  Level) - 1;

      for(i = 0 ; i < GateNB ; i++)
      {
        InsSlice = mc_IndexToSlice(Step * i + Offset);

        if (Class == Level + 1)
        {
          InsName = UtBas_PrintF("%s2t%ldw0_%ld",
                                 (Level%2) ? "nz" : "z",
                                  Level - 1,
                                 InsSlice);
          UtMPh_CopyUpCMA("zero", InsName, "o");
        }
        else
          InsName = UtBas_PrintF("%s2l%ldw%c_%ld",
                                 (Level%2) ? "nz" : "z",
                                  Level - 1,
                                 mc_GreyChar(Class,Level,i),
                                 InsSlice);

        UtMPh_CopyUpHoDPThru(InsName);
        switch( UtTec_Family() )
        {
          case C_UtTec_FamilyAsGa:

            break;

          case C_UtTec_FamilyCMOS:

            UtMPh_CopyUpNwell( InsName );
            break;
        }

        UtMPh_CopyUpCMA("i3", InsName,
                        GENLIB_NAME("i0[%d]", mc_IndexToBit(Step*i+Offset)));
      }
    }
  }


  /* Addition of power supplies connectors and wires. */
  for(InsSlice = 0 ; InsSlice <= GV.aWidthDP ; InsSlice++)
  {
    CreatePower(InsSlice);
    if (mc_IsEmptySlice(InsSlice))
      CreateHoDPThru(InsSlice);
  }


  GENLIB_SAVE_PHFIG();
}
