/*
 * $Author: ac $
 * $Id: bflcube.c,v 1.2 2000/03/16 11:01:11 ac Exp $
 * $Date: 2000/03/16 11:01:11 $
 */
/*
 * $Log: bflcube.c,v $
 * Revision 1.2  2000/03/16 11:01:11  ac
 * fflush de stdout avant les message d'erreur.
 * Ajout de bflcubeiszero.
 *
 * Revision 1.1  1999/10/21 12:51:50  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <assert.h>
#include BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

#define BFLCUBEWORD(cube, bit)	(cube->data[(bit / sizeof(unsigned long))])
#define BFLCUBEPOS(bit)	((bit) % sizeof(unsigned long))
#define BFLCUBEBIT(word, pos)	(((word) & (1 << (pos))) != 0)

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

bflcube_t *
bflcubenew(bflsize_t size)
{
	bflcube_t *cube = bflallocate(bflcube_t, 1);
	cube->size = size;
	cube->word = size / sizeof(unsigned long) + (size % sizeof(unsigned long) != 0);
	cube->data = (unsigned long *)bflcalloc(cube->word, sizeof(unsigned long));
	return cube;
}

bflcube_t *
bflcubecopy(bflcube_t *cube)
{
	int i;
	bflcube_t *copy;

	copy = bflallocate(bflcube_t, 1);
	copy->size = cube->size;
	copy->word = cube->size;
	copy->data = (unsigned long *)bflcalloc(copy->word, sizeof(unsigned long));
	for (i = 0; i < copy->size; i++)
		copy->data[i] = cube->data[i];
	return copy;
}

void
bflcubefree(bflcube_t *cube)
{
	bfldestroy(cube->data);
	bfldestroy(cube);
}

bflcube_t *
bflcubeor(bflcube_t *res, bflcube_t *x, bflcube_t *y)
{
	int i;
	
	assert(res->size == x->size);
	assert(res->size == y->size);
	for (i = 0; i < res->word; i++)
		res->data[i] = x->data[i] | y->data[i];
	return res;
}

bflcube_t *
bflcubeand(bflcube_t *res, bflcube_t *x, bflcube_t *y)
{
	int i;
	
	assert(res->size == x->size);
	assert(res->size == y->size);
	for (i = 0; i < res->word; i++)
		res->data[i] = x->data[i] & y->data[i];
	return res;
}

bflcube_t *
bflcubenot(bflcube_t *res, bflcube_t *x)
{
	int i;
	
	assert(res->size == x->size);
	for (i = 0; i < res->word; i++)
		res->data[i] = ~x->data[i];
	return res;
}
		
void
bflcubeset(bflcube_t *cube, bflsize_t bit)
{
	unsigned long *word;
	
	assert(bit >= 0 && bit < cube->size);
	word = &BFLCUBEWORD(cube, bit);
	*word = *word | (1 << BFLCUBEPOS(bit));
}

void
bflcubeclr(bflcube_t *cube, bflsize_t bit)
{
	unsigned long *word;
	
	assert(bit >= 0 && bit < cube->size);
	word = &BFLCUBEWORD(cube, bit);
	*word = *word & ~(1 << BFLCUBEPOS(bit));
}

int
bflcubebit(bflcube_t *cube, bflsize_t bit)
{
	assert(bit >= 0 && bit < cube->size);
	return BFLCUBEBIT(BFLCUBEWORD(cube, bit), BFLCUBEPOS(bit));
}

int
bflcubecmp(bflcube_t *x, bflcube_t *y)
{
	int i;

	assert(x->size == y->size);
	for (i = 0; i < x->word; i++)
		if (x->data[i] != y->data[i])
			return y->data[i] - x->data[i];
	return 0;
}

void
bflcubeprint(FILE *fp, bflcube_t *cube)
{
	int bit;

	for (bit = 0; bit < cube->size; bit++)
		fprintf(fp, "%d", BFLCUBEBIT(BFLCUBEWORD(cube, bit), BFLCUBEPOS(bit)));
	fprintf(fp, "\n");
}

int
bflcubeiszero(bflcube_t *cube)
{
	int i;

	for (i = 0; i < cube->word; i++)
		if (cube->data[i] != 0)
			return 1;
	return 0;
}
