/*
 * $Author: ac $
 * $Id: bflmap.c,v 1.2 1999/10/21 12:51:51 ac Exp $
 * $Date: 1999/10/21 12:51:51 $
 */
/*
 * $Log: bflmap.c,v $
 * Revision 1.2  1999/10/21 12:51:51  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:37  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include BFL_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

#define NIL		&__red__black__sentinel

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

static bflmapnode_t __red__black__sentinel =
{NIL, NIL, 0, bflblack, 0};

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

void
bflmaprotateleft(map, node)
    bflmap_t *map;
    bflmapnode_t *node;
{
    bflmapnode_t *parent = node->right;

/* establish node->right link */
    node->right = parent->left;
    if (parent->left != NIL)
	parent->left->parent = node;

/* establish parent->parent link */
    parent->parent = node->parent;
    if (node == map->header->parent)
	map->header->parent = parent;
    else if (node == node->parent->left)
	node->parent->left = parent;
    else
	node->parent->right = parent;

/* link node and parent */
    parent->left = node;
    node->parent = parent;
}

void
bflmaprotateright(map, node)
    bflmap_t *map;
    bflmapnode_t *node;
{
    bflmapnode_t *parent = node->left;

/* establish node->left link */
    node->left = parent->right;
    if (parent->right != NIL)
	parent->right->parent = node;

/* establish parent->parent link */
    parent->parent = node->parent;
    if (node == map->header->parent)
	map->header->parent = parent;
    else if (node == node->parent->right)
	node->parent->right = parent;
    else
	node->parent->left = parent;

/* link node and parent */
    parent->right = node;
    node->parent = parent;
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

bflmap_t *
bflmapnew(compar, dup)
    bflcompar_fn compar;
    int dup;
{
    bflmap_t *map;

    map = (bflmap_t *) bflmemalloc(sizeof(bflmap_t));
    map->duplicate = dup;
    map->compar = compar;
    map->size = 0;
    map->header = (bflmapnode_t *) bflmemalloc(sizeof(bflmapnode_t));
    map->header->color = bflred;
    map->header->parent = NIL;
    map->header->left = map->header;
    map->header->right = map->header;
    return map;
}

void
bflmapfree(map)
    bflmap_t *map;
{
    bflmapclean(map);
    bflmemfree(map->header, sizeof(bflmapnode_t));
    bflmemfree(map, sizeof(bflmap_t));
}

bflmap_t *
bflmapcopy(map)
    bflmap_t *map;
{
    bflmap_t *copy = bflmapnew(map->compar, map->duplicate);
    bflmapnode_t *first = bflmaprbegin(map);
    bflmapnode_t *last = bflmaprend(map);

    while (first != last) {
	bflmapinsert(copy, bflmapkey(first), bflmapdata(first));
	first = bflmapprev(first);
    }
    return copy;
}

void
bflmapclean(map)
    bflmap_t *map;
{
    bflmapnode_t *first = bflmaprbegin(map);
    bflmapnode_t *last = bflmaprend(map);

    while (first != last) {
	bflmapnode_t *next = bflmapnext(first);	

	bflmaperase(map, first);
	first = next;
    }
}

bflmapnode_t *
bflmapinsert(map, key, data)
    bflmap_t *map;
    bflptr_t key;
    bflptr_t data;
{
    bflmapnode_t *current;
    bflmapnode_t *parent;
    bflmapnode_t *node;

/* find where node belongs */
    parent = map->header;
    current = bflmaproot(map);
    while (current != NIL) {
	int comp = map->compar(current->key, key);

	if (!(map->duplicate) && comp == 0) {
	    current->data = data;
	    return (current);
	}
	parent = current;
	current = (comp < 0) ? current->left : current->right;
    }

    map->size++;
    node = (bflmapnode_t *) bflmemalloc(sizeof(bflmapnode_t));
    node->key = key;
    node->data = data;
    node->left = NIL;
    node->right = NIL;
    node->parent = 0;
    node->color = bflred;

    if (parent == map->header || map->compar(parent->key, key) < 0) {
	parent->left = node;
	if (parent == map->header) {
	    map->header->parent = node;
	    map->header->right = node;
	} else if (parent == map->header->left)		/* leftmost */
	    map->header->left = node;
    } else {
	parent->right = node;
	if (parent == map->header->right)	/* rightmost */
	    map->header->right = node;
    }
    node->parent = parent;
    current = node;
/* check bflred-bflblack properties */
    while (current != map->header->parent && current->parent->color == bflred) {
    /* we have a violation */
	if (current->parent == current->parent->parent->left) {
	    parent = current->parent->parent->right;
	    if (parent->color == bflred) {
	    /* uncle is bflred */
		current->parent->color = bflblack;
		parent->color = bflblack;
		current->parent->parent->color = bflred;
		current = current->parent->parent;
	    } else {
	    /* uncle is bflblack */
		if (current == current->parent->right) {
		/* make current a left child */
		    current = current->parent;
		    bflmaprotateleft(map, current);
		}
	    /* recolor and rotate */
		current->parent->color = bflblack;
		current->parent->parent->color = bflred;
		bflmaprotateright(map, current->parent->parent);
	    }
	} else {
	/* mirror image of above code */
	    parent = current->parent->parent->left;
	    if (parent->color == bflred) {
	    /* uncle is bflred */
		current->parent->color = bflblack;
		parent->color = bflblack;
		current->parent->parent->color = bflred;
		current = current->parent->parent;
	    } else {
	    /* uncle is bflblack */
		if (current == current->parent->left) {
		    current = current->parent;
		    bflmaprotateright(map, current);
		}
		current->parent->color = bflblack;
		current->parent->parent->color = bflred;
		bflmaprotateleft(map, current->parent->parent);
	    }
	}
    }
    map->header->parent->color = bflblack;	/* root->color */
    return node;
}

void
bflmaperase(bflmap_t * map, bflmapnode_t * node)
{
    bflmapnode_t *y = node;
    bflmapnode_t *x;

    if (y->left == NIL)
	x = y->right;
    else if (y->right == NIL)
	x = y->left;
    else {
	y = y->right;
	while (y->left != NIL)
	    y = y->left;
	x = y->right;
    }
    if (y != node) {		/* relink y in place of node */
	node->left->parent = y;
	y->left = node->left;
	if (y != node->right) {
	    x->parent = y->parent;	/* possibly x == NIL */
	    y->parent->left = x;	/* y must be a left child */
	    y->right = node->right;
	    node->right->parent = y;
	} else
	    x->parent = y;	/* needed in case x == NIL */
	if (map->header->parent == node)
	    map->header->parent = y;
	else if (node->parent->left == node)
	    node->parent->left = y;
	else
	    node->parent->right = y;
	y->parent = node->parent;
	bflswap(y->color, node->color);
	bflswap(y, node);
    /* 
     *y points to node to be actually deleted,
     *node points to old node's former successor
     */
    } else {			/* y == node */
	x->parent = y->parent;	/* possibly x == NIL */
	if (map->header->parent == node)
	    map->header->parent = x;
	else if (node->parent->left == node)
	    node->parent->left = x;
	else
	    node->parent->right = x;
	if (map->header->left == node) {	/* leftmost */
	    if (node->right == NIL)
	    /* node->left must be NIL also */
		map->header->left = node->parent;
	/* makes root == header if node == root */
	    else
		map->header->left = bflmapmin(x);
	}
	if (map->header->right == node) {	/* rightmost */
	    if (node->left == NIL)
	    /* node->right must be NIL also */
		map->header->right = node->parent;
	/* makes root == header if node == root */
	    else		/* x == node->left */
		map->header->right = bflmapmax(x);
	}
    }
    if (y->color != bflred) {
	while (x != map->header->parent && x->color == bflblack)
	    if (x == x->parent->left) {
		bflmapnode_t *w = x->parent->right;

		if (w->color == bflred) {
		    w->color = bflblack;
		    x->parent->color = bflred;
		    bflmaprotateleft(map, x->parent);
		    w = x->parent->right;
		}
		if (w->left->color == bflblack && w->right->color == bflblack) {
		    w->color = bflred;
		    x = x->parent;
		} else {
		    if (w->right->color == bflblack) {
			w->left->color = bflblack;
			w->color = bflred;
			bflmaprotateright(map, w);
			w = x->parent->right;
		    }
		    w->color = x->parent->color;
		    x->parent->color = bflblack;
		    w->right->color = bflblack;
		    bflmaprotateleft(map, x->parent);
		    break;
		}
	    } else {
	    /* 
	     * same as then clause with "right" and
	     * "left" exchanged
	     */
		bflmapnode_t *w = x->parent->left;

		if (w->color == bflred) {
		    w->color = bflblack;
		    x->parent->color = bflred;
		    bflmaprotateright(map, x->parent);
		    w = x->parent->left;
		}
		if (w->right->color == bflblack && w->left->color == bflblack) {
		    w->color = bflred;
		    x = x->parent;
		} else {
		    if (w->left->color == bflblack) {
			w->right->color = bflblack;
			w->color = bflred;
			bflmaprotateright(map, w);
			w = x->parent->left;
		    }
		    w->color = x->parent->color;
		    x->parent->color = bflblack;
		    w->left->color = bflblack;
		    bflmaprotateright(map, x->parent);
		    break;
		}
	    }
	x->color = bflblack;
    }
    bflmemfree(y, sizeof(bflmapnode_t));
    map->size--;
}

bflmapnode_t *
bflmapfind(map, key)
    bflmap_t *map;
    bflptr_t key;
{
    bflmapnode_t *current;

    current = bflmaproot(map);
    while (current != NIL) {
	int comp = map->compar(current->key, key);

	if (comp == 0)
	    return (current);
	else
	    current = (comp < 0) ? current->left : current->right;
    }
    return ((bflmapnode_t *) 0);
}

bflmapnode_t *
bflmapmin(node)
    bflmapnode_t *node;
{
    while (bflmapleft(node) != NIL)
	node = bflmapleft(node);
    return node;
}

bflmapnode_t *
bflmapmax(node)
    bflmapnode_t *node;
{
    while (bflmapright(node) != NIL)
	node = bflmapright(node);
    return node;
}

bflmapnode_t *
bflmapnext(node)
    bflmapnode_t *node;
{
    if (node->right != NIL) {
	node = node->right;
	while (node->left != NIL)
	    node = node->left;
    } else {
	bflmapnode_t *parent = node->parent;

	while (node == parent->right) {
	    node = parent;
	    parent = parent->parent;
	}
	if (node->right != parent)	/* necessary because of rightmost */
	    node = parent;
    }
    return node;
}

bflmapnode_t *
bflmapprev(bflmapnode_t * node)
{
    if (node->color == bflred && node->parent->parent == node)
    /* check for header */
	node = node->right;	/* return rightmost */
    else if (node->left != NIL) {
	bflmapnode_t *parent = node->left;

	while (parent->right != NIL)
	    parent = parent->right;
	node = parent;
    } else {
	bflmapnode_t *parent = node->parent;

	while (node == parent->left) {
	    node = parent;
	    parent = parent->parent;
	}
	node = parent;
    }
    return node;
}
