/*
 * $Author: ac $
 * $Id: testlist.c,v 1.2 1999/10/21 12:51:54 ac Exp $
 * $Date: 1999/10/21 12:51:54 $
 */
/*
 * $Log: testlist.c,v $
 * Revision 1.2  1999/10/21 12:51:54  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:38  alliance
 * fpmap import
 *
 */

#include <stdio.h>
#include BFL_H

static void
list_print(name, list)
    char *name;
    bfllist_t *list;
{
    bfllistnode_t *first = bfllistbegin(list);
    bfllistnode_t *last = bfllistend(list);

    printf("%s (", name);
    while (first != last) {
	printf("%s ", (char *) bfllistdata(first));
	first = bfllistnext(first);
    }
    printf(")\n");
}

void
test_merge()
{
    bfllist_t *list1 = bfllistnew();
    bfllist_t *list2 = bfllistnew();

    printf("\ntest_merge\n");

    bfllistpushback(list1, "tata");
    bfllistpushback(list1, "titi");

    bfllistpushback(list2, "tete");
    bfllistpushback(list2, "toto");
    bfllistpushback(list2, "tutu");

    list_print("list1", list1);
    list_print("list2", list2);

    bfllistmerge(list1, list2, bflstrcmp);

    list_print("list1", list1);
    list_print("list2", list2);

    bfllistfree(list1);
    bfllistfree(list2);
}

void
test_sort()
{
    bfllist_t *list = bfllistnew();

    printf("\ntest_sort\n");

    bfllistpushback(list, "tata");
    bfllistpushback(list, "titi");
    bfllistpushback(list, "tete");
    bfllistpushback(list, "tutu");
    bfllistpushback(list, "toto");

    list_print("initial: ", list);
    bfllistsort(list, bflstrcmp);
    list_print("sorted : ", list);
    bfllistfree(list);

}

int
main(int argc, char **argv)
{
    bfllist_t *list = bfllistnew();
    bfllistnode_t *first;
    bfllistnode_t *last;

    bfllistpushfront(list, "titi");
    bfllistpushfront(list, "tete");
    bfllistpushfront(list, "tata");
    bfllistpushback(list, "toto");
    bfllistpushback(list, "tutu");

    list_print("list", list);

    first = bfllistrbegin(list);
    last = bfllistrend(list);
    while (first != last) {
	first = bfllistprev(first);
	printf("%s ", (char *) bfllistdata(first));
    }
    putchar('\n');

    printf("bfllistfind( list, bflstrcmp, \"titi\") = %s\n",
	   (char *) bfllistdata(bfllistfind(list, bflstrcmp, "titi")));

    puts("Removing \"tete\"");
    bfllistremove(list, bflstrcmp, "tete");

    printf("bfllistfront( list ) = %s\n",
	   (char *) bfllistfront(list));
    puts("bfllistpopfront( list )");
    bfllistpopfront(list);
    printf("bfllistfront( list ) = %s\n",
	   (char *) bfllistfront(list));

    printf("bfllistback( list )  = %s\n",
	   (char *) bfllistback(list));
    puts("bfllistpopback( list )");
    bfllistpopback(list);
    printf("bfllistback( list )  = %s\n",
	   (char *) bfllistback(list));

    list_print("list", list);

    printf("reverse:");
    first = bfllistrbegin(list);
    last = bfllistrend(list);
    while (first != last) {
	first = bfllistprev(first);
	printf(" %s", (char *) bfllistdata(first));
    }
    putchar('\n');

    bfllistclean(list);

    bfllistpushfront(list, "titi");
    bfllistpushfront(list, "tete");
    bfllistpushfront(list, "tata");
    bfllistpushback(list, "tutu");
    bfllistpushback(list, "toto");

    printf("before reverse\n");
    list_print("list", list);
    bfllistreverse(list);
    printf("after reverse\n");
    list_print("list", list);

    bfllistfree(list);

    test_merge();
    test_sort();

    return 0;
}
