/*
 * $Author: ac $
 * $Id: butbdd.c,v 1.2 1999/10/21 12:51:59 ac Exp $
 * $Date: 1999/10/21 12:51:59 $
 */
/*
 * $Log: butbdd.c,v $
 * Revision 1.2  1999/10/21 12:51:59  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 * Revision 1.1.1.1  1998/10/21 16:32:39  alliance
 * fpmap import
 *
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include  MUT_H
#include  BFL_H
#include  AUT_H
#include  ABL_H
#include  BDD_H
#include  ABE_H
#include  BUT_H

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

static chain_list	*ButHeadName		= (chain_list *) 0;
static befig_list	*ButBehFigure		= (befig_list *) 0;
static authtable	*ButHashOrder	= (authtable  *)0;

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

#define BUT_ORDER( Flag )	((Flag) & BUT_BDD_ORDER)
#define BUT_FLATTEN( Flag )	((Flag) & BUT_BDD_FLATTEN)

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

static bddnode *
but_expr2bdd __P(( ablexpr *Expr ));

static bddnode *
but_abl2bdd __P(( char *Name, ablexpr *Expr ));

static void
but_mbddtot __P(( befig_list *BehFigure ));

static void
but_orderexprbdd __P(( ablexpr *Expr ));

static void
but_orderablbdd __P(( char *Name, ablexpr *Expr ));

static long
but_sortcompare __P(( authelem *ValueArray, long Index1, long Index2 ));

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                              but_expr2bdd                             */
/*----------------------------------------------------------------------*/

static int
but_issavedaux( chain_list *SavedAux, char *Name )
{
    while ( SavedAux != 0 )
	if ( SavedAux->DATA == Name )
	    return 1;
	else
	    SavedAux = SavedAux->NEXT;

    return 0;
}

static bddnode *
but_expr2bdd( chain_list *Expr )
{
    bddnode *BddNode;
    bddnode *BddFirst;
    char *AtomValue;
    beaux_list *ScanAux;
    chain_list *ScanChain;
    long Oper;
    int Negative;

    if ( Expr == (ablexpr *) 0 )
	bflerror(1, "error: null expression");

    if (ABL_ATOM(Expr)) {
	AtomValue = ABL_ATOM_VALUE(Expr);

	if (AtomValue == ABL_ATOM_NAME_ONE)
	    return BddLocalSystem->ONE;

	if ( AtomValue == ABL_ATOM_NAME_ZERO || AtomValue == ABL_ATOM_NAME_DC )
	    return BddLocalSystem->ZERO;

	BddNode = searchbddcircuitin((bddcircuit *) 0, AtomValue);

	if ( BddNode != (bddnode *) 0 )
	    return incbddrefext(BddNode);

	for (ScanAux = ButBehFigure->BEAUX; ScanAux != (beaux_list *) 0;
	     ScanAux = ScanAux->NEXT)
	    if (ScanAux->NAME == AtomValue)
		break;

	if (ScanAux == (beaux_list *) 0)
	    bflerror(1, "error: unknown atom `%s'", AtomValue);

	for (ScanChain = ButHeadName; ScanChain != (chain_list *) 0;
	     ScanChain = ScanChain->NEXT)
	    if (ScanChain->DATA == AtomValue)
		bflerror(1, "error: combinatory loop on `%s'", AtomValue);

	if (ScanAux->NODE == (void *) 0) {
	    ButHeadName = addchain(ButHeadName, AtomValue);
	    ScanAux->NODE = but_expr2bdd(ScanAux->ABL);
	    ButHeadName = delchain(ButHeadName, ButHeadName);
	}
	return incbddrefext(ScanAux->NODE);
    }
    
    Oper = ABL_OPER(Expr);
    Expr = ABL_CDR(Expr);

    if (Oper == ABL_NOT) {
	if (ABL_CDR(Expr) != (chain_list *) 0)
	    bflerror(1, "error: illegal use of operator %ld", Oper);

	BddFirst = but_expr2bdd(ABL_CAR(Expr));
	BddNode = applybddnodenot((bddsystem *) 0, decbddrefext(BddFirst));

	return (BddNode);
    }
    
    if (isablunaryoper(Oper) || (ABL_CDR(Expr) == (chain_list *) 0))
	bflerror(1, "error: illegal use of operator %ld", Oper);

    if ((getabloperpolar(Oper) == ABL_POLAR_POSITIVE) ||
	(ABL_CDDR(Expr) == (chain_list *) 0)) {
	Negative = 0;
    } else {
	Negative = 1;
	Oper = getablopernot(Oper);
    }

    BddFirst = but_expr2bdd(ABL_CAR(Expr));

    while ((Expr = ABL_CDR(Expr)) != (chain_list *) 0) {
	BddNode = but_expr2bdd(ABL_CAR(Expr));
	BddFirst = applybddnode((bddsystem *) 0, Oper, decbddrefext(BddFirst),
						       decbddrefext(BddNode));
    }

    if (Negative)
	BddFirst = applybddnodenot((bddsystem *) 0, decbddrefext(BddFirst));

    return (BddFirst);
}

/*----------------------------------------------------------------------*/
/*                              but_abl2bdd                              */
/*----------------------------------------------------------------------*/

static bddnode *
but_abl2bdd( char *Name, chain_list *Expr )
{
    bddnode *BddNode;

    if (Name != (char *) 0)
	ButHeadName = addchain(ButHeadName, Name);

    BddNode = but_expr2bdd(Expr);

    if (Name != (char *) 0) {
	ButHeadName = delchain(ButHeadName, ButHeadName);

	BddNode = addbddcircuitout((bddcircuit *) 0, Name, BddNode);
    }
    return (BddNode);
}

/*----------------------------------------------------------------------*/
/*                              but_mbddtot                              */
/*----------------------------------------------------------------------*/

static void
but_mbddtot( befig_list *BehFigure )
{
    struct beaux *BehAux;
    struct beout *BehOut;
    struct bebus *BehBus;
    struct bebux *BehBux;
    struct bereg *BehReg;
    struct biabl *BiAbl;
    struct binode *BiNode;

    ButBehFigure = BehFigure;
    ButHeadName = (chain_list *) 0;

    printf( "   - Making Bdd ...\n" );

    for ( BehAux = BehFigure->BEAUX; BehAux != NULL; BehAux = BehAux->NEXT )
	if (BehAux->NODE == NULL)
	    BehAux->NODE = but_abl2bdd(BehAux->NAME, BehAux->ABL);
    
    for ( BehOut = BehFigure->BEOUT; BehOut != NULL; BehOut = BehOut->NEXT )
	if (BehOut->NODE == NULL)
	    BehOut->NODE = but_abl2bdd(BehOut->NAME, BehOut->ABL);

    for ( BehBus = BehFigure->BEBUS; BehBus != NULL; BehBus = BehBus->NEXT ) {
	BiAbl  = BehBus->BIABL;
	BiNode = BehBus->BINODE;

	while ( BiAbl != NULL ) {
	    if ( BiAbl->CNDABL != (ablexpr *)0 )
		BiNode->CNDNODE = but_abl2bdd((char *) 0, BiAbl->CNDABL);
	    if ( BiAbl->VALABL != (ablexpr *)0 )
		BiNode->VALNODE = but_abl2bdd(BehBus->NAME, BiAbl->VALABL);

	    BiAbl  = BiAbl->NEXT;
	    BiNode = BiNode->NEXT;
	}
    }

    for ( BehBux = BehFigure->BEBUX; BehBux != NULL; BehBux = BehBux->NEXT ) {
	BiAbl  = BehBux->BIABL;
	BiNode = BehBux->BINODE;

	while ( BiAbl != NULL ) {
	    if ( BiAbl->CNDABL != (ablexpr *)0 )
		BiNode->CNDNODE = but_abl2bdd((char *) 0, BiAbl->CNDABL);
	    if ( BiAbl->VALABL != (ablexpr *)0 )
		BiNode->VALNODE = but_abl2bdd(BehBux->NAME, BiAbl->VALABL);

	    BiAbl  = BiAbl->NEXT;
	    BiNode = BiNode->NEXT;
	}
    }

    for ( BehReg = BehFigure->BEREG; BehReg != NULL; BehReg = BehReg->NEXT ) {
	BiAbl  = BehReg->BIABL;
	BiNode = BehReg->BINODE;

	while ( BiAbl != NULL ) {
	    if ( BiAbl->CNDABL != (ablexpr *)0 )
		BiNode->CNDNODE = but_abl2bdd((char *) 0, BiAbl->CNDABL);
	    if ( BiAbl->VALABL != (ablexpr *)0 )
		BiNode->VALNODE = but_abl2bdd(BehReg->NAME, BiAbl->VALABL);

	    BiAbl  = BiAbl->NEXT;
	    BiNode = BiNode->NEXT;
	}
    }
}

/*----------------------------------------------------------------------*/
/*                            but_orderexprbdd                           */
/*----------------------------------------------------------------------*/

static void
but_orderexprbdd( chain_list *Expr )
{
    char       *AtomValue;
    beaux_list *ScanAux;
    chain_list *ScanChain;
    authelem   *Element;
    long        Oper;

    if ( Expr == (chain_list *)0 )
	bflerror( 1, "error: null expression" );

    if ( ABL_ATOM( Expr ) ) {
	AtomValue = ABL_ATOM_VALUE( Expr );

/*
 * if (AtomValue == 0)
 *	bflerror( 1, "error: null atom", AtomValue );
 */
	
        if ( ( AtomValue == ABL_ATOM_NAME_ONE  ) ||
             ( AtomValue == ABL_ATOM_NAME_ZERO ) ||
             ( AtomValue == ABL_ATOM_NAME_DC   ) ) {
          return;
        }

        Element = searchauthelem( ButHashOrder, AtomValue );

        if ( Element != (authelem *)0 ) {
	    Element->VALUE++;
 	    return;
        }

        for ( ScanAux = ButBehFigure->BEAUX; ScanAux != 0; ScanAux = ScanAux->NEXT )
            if ( ScanAux->NAME == AtomValue )
	       break;

        if ( ScanAux == (beaux_list *)0 )
   	    bflerror( 1, "error: unknown atom `%s'", AtomValue );

        for ( ScanChain = ButHeadName; ScanChain != 0; ScanChain = ScanChain->NEXT )
  	    if ( ScanChain->DATA == AtomValue )
	          bflerror( 1, "error: combinatory loop on `%s'", AtomValue );

	ButHeadName   = addchain( ButHeadName, AtomValue );
	but_orderexprbdd( ScanAux->ABL );
	ButHeadName   = delchain( ButHeadName, ButHeadName );

	return;
    }

    Oper = ABL_OPER( Expr );
    Expr = ABL_CDR( Expr );

    if ( Oper == ABL_NOT ) {
	if ( ABL_CDR( Expr ) != (chain_list *)0 )
	    bflerror( 1, "error: illegal use of operator %ld", Oper );

	but_orderexprbdd( ABL_CAR( Expr ) );
	return;
    }

    if ( isablunaryoper( Oper ) || ( ABL_CDR( Expr ) == (chain_list *)0 ) )
	bflerror( 1, "error: illegal use of operator %ld", Oper );

    but_orderexprbdd( ABL_CAR( Expr ) );

    while ( ( Expr = ABL_CDR( Expr ) ) != (chain_list *)0 )
	but_orderexprbdd( ABL_CAR( Expr ) );
}

/*----------------------------------------------------------------------*/
/*                            but_orderablbdd                            */
/*----------------------------------------------------------------------*/

static void
but_orderablbdd( char *Name, chain_list *Expr )
{
    if ( Name != (char *)0 )
	ButHeadName = addchain( ButHeadName, Name );

    but_orderexprbdd( Expr );

    if ( Name != (char *)0 )
	ButHeadName = delchain( ButHeadName, ButHeadName );
}

/*----------------------------------------------------------------------*/
/*                            but_sortcompare                            */
/*----------------------------------------------------------------------*/

static long
but_sortcompare( ValueArray, Index1, Index2 )
    authelem *ValueArray;
    long      Index1;
    long      Index2;
{
    return( ValueArray[ Index1 ].VALUE - ValueArray[ Index2 ].VALUE );
}
/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

void
but_makebdd( befig_list *BehFigure, int Flag, chain_list *SavedAux )
{
    bddsystem  *BddSystem;
    berin_list *BehRin;
    beaux_list *BehAux;
    beout_list *BehOut;
    bebus_list *BehBus;
    bebux_list *BehBux;
    bereg_list *BehReg;
    biabl_list *BiAbl;
    long        NumberInput;
    long        NumberOutput;
    authelem   *Element;
    authelem   *ValueArray;
    long       *IndexArray;
    long        SizeArray;
    long        Index;
    bddindex    BddIndex;


    ButBehFigure = BehFigure;
    ButHeadName  = (chain_list *)0;

    NumberInput  = 1;
    NumberOutput = 1;
    
    for ( BehAux = BehFigure->BEAUX; BehAux != NULL; BehAux = BehAux->NEXT ) {
	if ( BUT_FLATTEN( Flag ) && ( ! but_issavedaux( SavedAux, BehAux->NAME ) ) )
	    BehFigure->BERIN = beh_rmvberin(BehFigure->BERIN, BehAux->NAME );
	NumberOutput += 1;
    }

    for ( BehRin = BehFigure->BERIN; BehRin != NULL; BehRin = BehRin->NEXT )
	NumberInput += 1;

    for ( BehOut = BehFigure->BEOUT; BehOut != NULL; BehOut = BehOut->NEXT )
	NumberOutput += 1;

    for ( BehBus = BehFigure->BEBUS; BehBus != NULL; BehBus = BehBus->NEXT )
	NumberOutput += 1;

    BddSystem         = createbddsystem( 100, 1000, 100, 5000000 );
    BehFigure->CIRCUI = createbddcircuit( BehFigure->NAME, NumberInput,
	    				  NumberOutput, BddSystem );

    /*
     * Calcul de l'ordre initial si BUT_ORDER( Flag ) != 0, sinon
     * on met les variables dans l'ordre de la declaration.
     */

    
    if ( BUT_ORDER( Flag ) == 0 ) {
        ButBehFigure->BERIN = (berin_list *)reverse( (chain_list *)ButBehFigure->BERIN );
        for ( BehRin = ButBehFigure->BERIN; BehRin != NULL; BehRin = BehRin->NEXT ) {
	    BddIndex = BddLocalCircuit->NUMBER_NAME_IN + BDD_INDEX_MIN;
    
	    addbddcircuitin( (bddcircuit *)0, BehRin->NAME, BddIndex,
			     BDD_IN_MODE_IMPOSE | BDD_IN_MODE_LAST );
        }
  
        ButBehFigure->BERIN = (berin_list *)reverse( (chain_list *)ButBehFigure->BERIN );
  
        return;
    } else {
	printf( "   - Running the abl ordonancer ...\n" );
      
	ButHashOrder = createauthtable( NumberInput << 1 );
    
	for ( BehRin = ButBehFigure->BERIN; BehRin != NULL; BehRin = BehRin->NEXT )
	    addauthelem( ButHashOrder, BehRin->NAME, 0 );
  
	for ( BehAux = ButBehFigure->BEAUX; BehAux != NULL; BehAux = BehAux->NEXT )
            if ( !BUT_FLATTEN( Flag ) || but_issavedaux( SavedAux, BehAux->NAME ) )
		if (BehAux->ABL != 0) {
#ifdef DEBUG
		    bflwarning("  --> orderabl = %s", BehAux->NAME);
#endif
                    but_orderablbdd( BehAux->NAME, BehAux->ABL );
		}
	for ( BehOut = ButBehFigure->BEOUT; BehOut != NULL; BehOut = BehOut->NEXT )
	    if (BehOut->ABL != 0) {
#ifdef DEBUG
		bflwarning("  --> orderabl = %s", BehOut->NAME);
#endif
		but_orderablbdd( BehOut->NAME, BehOut->ABL );
	    }
    
	for ( BehBus = ButBehFigure->BEBUS; BehBus != NULL; BehBus = BehBus->NEXT )
	    for ( BiAbl = BehBus->BIABL; BiAbl != NULL; BiAbl = BiAbl->NEXT ) {
		if ( BiAbl->VALABL != 0 )
		    but_orderablbdd( (char *)0, BiAbl->VALABL );
		if ( BiAbl->CNDABL != 0 )
		    but_orderablbdd( (char *)0   , BiAbl->CNDABL );
	    }
  
	for ( BehBux = ButBehFigure->BEBUX; BehBux != NULL; BehBux = BehBux->NEXT )
	    for ( BiAbl = BehBux->BIABL; BiAbl != NULL; BiAbl  = BiAbl->NEXT ) {
		if ( BiAbl->VALABL != 0 )
		    but_orderablbdd( (char *)0, BiAbl->VALABL );
		if ( BiAbl->CNDABL != 0 )
		    but_orderablbdd( (char *)0, BiAbl->CNDABL );
	    }

	for ( BehReg = ButBehFigure->BEREG; BehReg != NULL; BehReg = BehReg->NEXT )
	    for ( BiAbl  = BehReg->BIABL; BiAbl != NULL; BiAbl  = BiAbl->NEXT ) {
		if ( BiAbl->VALABL != 0 )
		    but_orderablbdd( (char *)0, BiAbl->VALABL );
		if ( BiAbl->CNDABL != 0 )
		    but_orderablbdd( (char *)0, BiAbl->CNDABL );
	    }
   
	SizeArray  = ButHashOrder->TABLE_SIZE;
	ValueArray = ButHashOrder->TABLE;
	IndexArray = (long *)autallocblock( sizeof( long ) * SizeArray );
   
	sortautarray( ValueArray, IndexArray, SizeArray, but_sortcompare );
   
	for ( Index = 0; Index < SizeArray; Index++ ) {
	    Element = &ValueArray[ IndexArray[ Index ] ];
   
	    if ( Element->KEY != AUT_HASH_KEY_EMPTY ) {
		BddIndex = BddLocalCircuit->NUMBER_NAME_IN + BDD_INDEX_MIN;

		addbddcircuitin( (bddcircuit *)0, Element->KEY, BddIndex,
				 BDD_IN_MODE_IMPOSE | BDD_IN_MODE_LAST );

# ifdef DEBUG
		fprintf( stdout, "Var: %s Order: %ld\n", Element->KEY, Element->VALUE );
# endif
	    }
        }

	autfreeblock( (char *)IndexArray );
	destroyauthtable( ButHashOrder );
    }

    /*
     * Construction du bdd pour chaque ABL
     */
    but_mbddtot( BehFigure );

    BehAux = BehFigure->BEAUX;

    while ( BehAux != 0 ) {
	beaux_list *Next = BehAux->NEXT;
	
	if ( BUT_FLATTEN( Flag ) && ( ! but_issavedaux( SavedAux, BehAux->NAME ) ) ) {
	    if ( BehAux->NODE != 0 ) {
		decbddrefext( BehAux->NODE );
		delbddcircuitout( 0, BehAux->NAME );
	    }
	    BehFigure->BEAUX = beh_rmvbeaux( BehFigure->BEAUX, BehAux->NAME, 'Y' );
	}
	BehAux = Next;
    }

    sweepbddsystem( 0 );
}
