/*
 * $Author: ac $
 * $Id: virtex.c,v 1.1 1999/10/21 12:52:09 ac Exp $
 * $Date: 1999/10/21 12:52:09 $
 */
/*
 * $Log: virtex.c,v $
 * Revision 1.1  1999/10/21 12:52:09  ac
 * Nouvelle version avec les librairies bdd104, abe, abv, abt.
 * Ajout de fpm_network pour accelerer l'optimisation.
 *
 */

/*
 * fpga mapping for a generic virtex architecture. The CLB is supposed to
 * be a k-Lut folowed by a flip-flop.
 *
 *  ce  ------------------+
 *  din ------------+     |
 *         +------+ |     | +-----+
 *  in0    |      | | +-+ +-+     |
 *  in1    |      | +-|\|___|     |_______ xq
 *         |      +-+-|/|   |     |
 *  ink    |      | | +-+   |     |
 *         |      | |       +--^--+
 *         +------+ +----------|---------- x
 *  ck  -----------------------+
 */

/*----------------------------------------------------------------------*/
/*                               Includes                               */
/*----------------------------------------------------------------------*/

#include <assert.h>
#include  MUT_H
#include  BFL_H
#include  AUT_H
#include  ABL_H
#include  BDD_H
#include  ABE_H
#include "fpm_network.h"
#include "fpm_user.h"
#include "fpm_load.h"
#include "fpm_mapp.h"
#include "fpm_util.h"
#include "virtex.h"

/*----------------------------------------------------------------------*/
/*                        Constant declarations                         */
/*----------------------------------------------------------------------*/

/* #define PRINT_DC */

/*----------------------------------------------------------------------*/
/*                          Type declarations                           */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Stucture declarations                         */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                        Variable declarations                         */
/*----------------------------------------------------------------------*/

static int		 virtex_num_clb;
static authtable	*virtex_depth_hash;

/*----------------------------------------------------------------------*/
/*                          Macro declarations                          */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                      Static function prototypes                      */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                    Definition of static functions                    */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*                         Local Cost Functions                         */
/*----------------------------------------------------------------------*/

static chain_list *
virtex_cost_cut(bddnode * Node, long * Depth, long * Count)
{
	bddnode *NodeInv;

	*Depth += 1;
	*Count += 1;

	VIRTEX_SET_MAPPED(Node->MARK);
	addauthelem(virtex_depth_hash, (char *)Node, *Depth);

	NodeInv = applybddnodenot(0, Node);
	decbddrefext(NodeInv);
	VIRTEX_SET_MAPPED(NodeInv->MARK);
	VIRTEX_SET_INVERTED(NodeInv->MARK);

	return addchain((chain_list *) 0, Node);
}

static chain_list *
virtex_cost_rec(bddnode * Node, long * Count, long * Depth)
{
	authelem     *HashElem;
	bddnode	     *NodeInv;
	bddnode      *NodeAnd;
	bddnode      *NodeLow;
	bddnode      *NodeHigh;
	chain_list   *SupportLow;
	chain_list   *SupportHigh;
	long          DepthLow;
	long          DepthHigh;
	long          LengthHigh;
	long          LengthLow;

	if (Node->INDEX < 2) {
		*Depth = 0;
		return 0;
	}
	if (Node->LOW->INDEX < 2 && Node->HIGH->INDEX < 2) {
		*Depth = 0;
		return addchain((chain_list *) 0, ((void *) (int) Node->INDEX));
	}
	if (VIRTEX_IS_MAPPED(Node->MARK)) {
		if (VIRTEX_IS_INVERTED(Node->MARK))
			Node = decbddrefext(applybddnodenot(0, Node));

		assert(HashElem = searchauthelem(virtex_depth_hash, (char *)Node));
		*Depth = (long) HashElem->VALUE;
		return addchain((chain_list *) 0, Node);
	}
#if 1
	NodeAnd = applybddnode(0, ABL_AND, Node->HIGH, Node->LOW);
	decbddrefext(NodeAnd);

	if (NodeAnd == Node->LOW) {
		NodeLow = Node->LOW;
		NodeHigh = simpbddnodedcon(0, Node->HIGH, Node->LOW);
		decbddrefext(NodeHigh);
	} else if (NodeAnd == Node->HIGH) {
		NodeLow = simpbddnodedcon(0, Node->LOW, Node->HIGH);
		decbddrefext(NodeLow);
		NodeHigh = Node->HIGH;
	} else {
		NodeLow = Node->LOW;
		NodeHigh = Node->HIGH;
	}
#else
	NodeLow = Node->LOW;
	NodeHigh = Node->HIGH;
#endif
	SupportLow = virtex_cost_rec(NodeLow, Count, &DepthLow);
	SupportHigh = virtex_cost_rec(NodeHigh, Count, &DepthHigh);

	while (1) {
		int SupportSize;
		chain_list *Result;

		SupportSize = fpm_support_size(SupportLow, SupportHigh);

		if (SupportSize < VIRTEX_LUT_SIZE - 1) {
			Result = fpm_merge_chain(SupportLow, SupportHigh);
			*Depth = FP_MAX(DepthLow, DepthHigh);
			return addchain(Result, ((void *) (int) Node->INDEX));
		} else if (SupportSize == VIRTEX_LUT_SIZE - 1) {
			freechain(SupportLow);
			freechain(SupportHigh);

			*Depth = 1 + FP_MAX(DepthLow, DepthHigh);
			*Count += 1;

			Node->MARK |= VIRTEX_MAPPED;
			addauthelem(virtex_depth_hash, (char *)Node, *Depth);

			NodeInv = applybddnodenot(0, Node);
			decbddrefext(NodeInv);
			NodeInv->MARK |= (VIRTEX_MAPPED | VIRTEX_INVERTED);

			return addchain((chain_list *) 0, Node);
		} else {
			LengthLow  = fpm_length_chain(SupportLow);
			LengthHigh = fpm_length_chain(SupportHigh);

			if (LengthLow > LengthHigh) {
				freechain(SupportLow);
				SupportLow = virtex_cost_cut(NodeLow, &DepthLow, Count);
			} else {
				freechain(SupportHigh);
				SupportHigh = virtex_cost_cut(NodeHigh, &DepthHigh, Count);
			}
		}		/* switch */
	}			/* while(1) */
}

/*----------------------------------------------------------------------*/
/*                        Local Mapping Function                        */
/*----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/*      ablexpr *virtex_make_expr( Atom, ExprLow, ExprHigh, Inclus )      */
/* -------------------------------------------------------------------- */
/*	Reconstruit l'expression associee a un noeud bdd en fonction de	*/
/* ses fils et des simplification par `don't care' eventuelle.		*/
/*----------------------------------------------------------------------*/

static ablexpr *
virtex_make_expr( ablexpr *Atom, ablexpr *ExprLow, ablexpr *ExprHigh, int	Inclus )
{
	ablexpr *Expr = (ablexpr *)0;

	switch( Inclus ) {
	case 0:	
		/* pas d'inclusion:
		 *  Expr = (or (and ExprHigh Atom) (and ExprLow (not Atom)))
		 */
		Expr = createablbinexpr( ABL_OR,
				createablbinexpr( ABL_AND,
						ExprHigh,
						dupablexpr( Atom ) ),
				createablbinexpr( ABL_AND,
						ExprLow,
						createablnotexpr( Atom ) ) );	
		break;
	
	case 1:
		/* ExprLow est inclus dans ExprHigh
		 *  Expr = (or (and ExprHigh Atom) ExprLow)
		 */
		Expr = createablbinexpr( ABL_OR,
				createablbinexpr( ABL_AND, ExprHigh, Atom ),
				ExprLow );
		break;
	
	case 2:	
		/* ExprHigh est inclus dans ExprLow
		 *  Expr = (or ExprHigh (and ExprLow (not Atom)))
		 */
		Expr = createablbinexpr( ABL_OR,
				ExprHigh, 
				createablbinexpr( ABL_AND,
						ExprLow,
						createablnotexpr( Atom ) ) );	
	break;
	
	default:
		bflerror(1, "virtex_make_expr: error: invalid flag (%d)", Inclus );

	}
	return Expr;
}

static ablexpr *
virtex_get_abl( bddindex Index )
{
	bddnamein *NameIn    = FpmUser->NETWORK->NAME_IN;
	bddindex  *NameIndex = FpmUser->NETWORK->INDEX_IN;

	return convertbddindexabl( 0, NameIn, NameIndex, Index );
}

/*----------------------------------------------------------------------*/
/*          chain_list *virtex_mapp_int( Node, Expr, CreateAux )          */
/*----------------------------------------------------------------------*/

static chain_list *
virtex_mapp_int( bddnode *Node, ablexpr **Expr, int CreateAux )
{
    bddnode	*NodeAnd;
    bddnode	*NodeLow;
    bddnode	*NodeHigh;
    int		 Inclus;
    ablexpr	*ExprLow;
    ablexpr	*ExprHigh;
    chain_list	*SupportLow;
    chain_list	*SupportHigh;
    long	 LengthHigh;
    long	 LengthLow;
    
	if ( Node->INDEX < 2 ) {
		if (Node->INDEX == BDD_INDEX_ZERO) {
			*Expr = createablatomzero();
			return (chain_list *)0;
		} else {
			*Expr = createablatomone();
			return (chain_list *)0;
		}
	}
    
	if ( Node->LOW->INDEX < 2 && Node->HIGH->INDEX < 2 ) {
		*Expr = virtex_get_abl( Node->INDEX );
		if ( Node->LOW->INDEX == BDD_INDEX_ONE )
			*Expr = createablnotexpr( *Expr );
		return( addchain( 0, ((void *)(int)Node->INDEX) ) );
	}

	/* simplifications eventuelles avant l'appel recursif sur les fils.
	 * Inclus est mis a jour :
	 *     Inclus = 0 pas d'inclusion (par defaut)
	 *     Inclus = 1 low est inclus dans high
	 *     Inclus = 2 high est inclus dans low
	 */

	NodeAnd = applybddnode( 0, ABL_AND, Node->HIGH, Node->LOW );
	decbddrefext( NodeAnd );

	if ( NodeAnd == Node->LOW ) {
		Inclus = 1;
		NodeLow  = Node->LOW;
		NodeHigh = simpbddnodedcon( 0, Node->HIGH, Node->LOW );
	} else if ( NodeAnd == Node->HIGH ) {
		Inclus = 2;
		NodeLow  = simpbddnodedcon( 0, Node->LOW, Node->HIGH );
		NodeHigh = Node->HIGH;
	} else {
		Inclus = 0;
		NodeLow  = Node->LOW;
		NodeHigh = Node->HIGH;
	}

	SupportLow  = virtex_mapp_int( NodeLow , &ExprLow , 1 );
	SupportHigh = virtex_mapp_int( NodeHigh, &ExprHigh, 1 );

	while (1) {
		int         SupportSize;
		chain_list *Result;

		SupportSize = fpm_support_size( SupportLow, SupportHigh );

		if ( SupportSize < VIRTEX_LUT_SIZE - 1 ) {
			Result = fpm_merge_chain( SupportLow, SupportHigh );
			*Expr = virtex_make_expr( virtex_get_abl( Node->INDEX ), ExprLow, ExprHigh, Inclus );
			return addchain( Result, ((void *)(int)Node->INDEX) );
		} else if ( SupportSize == VIRTEX_LUT_SIZE - 1 ) {
			freechain( SupportLow );
			freechain( SupportHigh );
			*Expr = virtex_make_expr( virtex_get_abl( Node->INDEX ), ExprLow, ExprHigh, Inclus );
			if ( CreateAux == 1 ) {
				Node =  fpm_mapp_node( Node, Expr );
				virtex_num_clb++;
				Result = addchain( 0, ((void *)(int)Node->INDEX) );
			} else {
				Result = addchain( 0, ((void *)(int)Node->INDEX) );
			}
			return Result;
		} else {

			LengthLow  = fpm_length_chain( SupportLow  );
			LengthHigh = fpm_length_chain( SupportHigh );

			if ( LengthLow > LengthHigh ) {
				freechain( SupportLow );
				NodeLow = fpm_mapp_node( NodeLow, &ExprLow );
				virtex_num_clb++;
				SupportLow = addchain( 0, ((void *)(int)NodeLow->INDEX) );
			} else {
				freechain( SupportHigh );
				NodeHigh = fpm_mapp_node( NodeHigh, &ExprHigh );
				virtex_num_clb++;
				SupportHigh = addchain( 0, ((void *)(int)NodeHigh->INDEX) ); 
			}
		}
	}				/* while(1) */
	/* not reached */
	return 0;
}

/*----------------------------------------------------------------------*/
/*                   Definition of exported functions                   */
/*----------------------------------------------------------------------*/

void
virtex_init()
{

	printf("   + Initializing Virtex local structures...\n");
	virtex_depth_hash = createauthtable(4096);
}

void
virtex_done()
{
	printf("   + Destroying Virtex local structures...\n");
	destroyauthtable(virtex_depth_hash);
}

void
virtex_check()
{
	befig_list *BehFigure;
	bereg_list *BehReg;
	bebus_list *BehBus;
	bebux_list *BehBux;
	ptype_list *ClockType;
	int IobMode;
	int Error;

	BehFigure = FpmUser->BEFIG;
	IobMode   = FpmUser->IOB_MODE;
	Error     = 0;

	if (BehFigure == (befig_list *) 0)
		bflerror(1, "virtex_check: error: figure is empty");

	for (BehReg = BehFigure->BEREG; BehReg != 0; BehReg = BehReg->NEXT) {
		if (BehReg->BIABL->NEXT != 0) {
			Error = 1;
			bflwarning("virtex_check: error: multiple drivers on register `%s'",
				   BehReg->NAME);
		}
		ClockType = getptype(BehReg->BIABL->USER, FP_CLOCK_TYPE);
		if ((ClockType == 0) ||
		    (ClockType->DATA == (void *) FP_LATCH_AH) ||
		    (ClockType->DATA == (void *) FP_LATCH_AL)) {
			Error = 1;
			bflwarning("virtex_check: error: register `%s' is a latch", BehReg->NAME);
		}
	}

	for (BehBus = BehFigure->BEBUS; BehBus != 0; BehBus = BehBus->NEXT) {
		if (IobMode == 1 && BehBus->BIABL->NEXT != 0) {
			Error = 1;
			bflwarning("virtex_check: error: multiple drivers on port `%s'",
				   BehBus->NAME);
		}
		if (BehBus->TYPE == 'W') {
			Error = 1;
			bflwarning("virtex_check: error: unsupported type (wor_bit) on port `%s'",
				   BehBus->NAME);
		}
	}
	for (BehBux = BehFigure->BEBUX; BehBux != 0; BehBux = BehBux->NEXT) {
		if (BehBux->TYPE == 'W') {
			Error = 1;
			bflwarning("virtex_check: error: unsupported type (wor_bit) on signal `%s'",
				   BehBux->NAME);
		}
	}
	if (Error != 0)
		bflerror(1, "virtex_check: aborting ...");
}

void
virtex_cost(fpm_expr_t *Expr, long *Count, long *Depth)
{
	bddnode *Node = Expr->BDD;
	chain_list *Support = virtex_cost_rec(Node, Count, Depth);
	if ((Support != NULL) &&
	    (Support->NEXT != NULL || VIRTEX_IS_INVERTED(Node->MARK))) {

		bddnode *NodeInv = decbddrefext(applybddnodenot(0, Node));

		*Depth += 1;
		*Count += 1;

		VIRTEX_SET_MAPPED(Node->MARK);
		addauthelem(virtex_depth_hash, (char *)Node, *Depth);

		VIRTEX_SET_MAPPED(NodeInv->MARK);
		VIRTEX_SET_INVERTED(NodeInv->MARK);
	}
	freechain(Support);
}

void
virtex_clear(fpm_atom_t * Atom)
{
	bddvar MaxVar;
	bddindexnode *IndexNode;
	bddindex *VarToIndex;
	bddvar Var;
	bddhnodetable *NodeTable;
	long TableSize;
	long HashIndex;
	bddnode *HashNode;

	MaxVar = FpmUser->NETWORK->BDD_SYSTEM->NUMBER_VAR;
	IndexNode = FpmUser->NETWORK->BDD_SYSTEM->INDEX_NODE;
	VarToIndex = FpmUser->NETWORK->BDD_SYSTEM->VAR_TO_INDEX;

	for (Var = 0; Var < MaxVar; ++Var) {
		NodeTable = IndexNode[VarToIndex[Var]];
		TableSize = NodeTable->TABLE_SIZE;
		for (HashIndex = 0; HashIndex < TableSize; HashIndex++) {
			HashNode = NodeTable->TABLE[HashIndex];
			if ( (HashNode != BDD_HASH_NODE_EMPTY  ) &&
			     (HashNode != BDD_HASH_NODE_DELETED) ) {
				HashNode->MARK = 0;
			}
		}
	}
	resetauthtable(virtex_depth_hash);
}

ablexpr *
virtex_mapp( bddnode *Node, int CreateAux )
{
	chain_list *Support;
	ablexpr	   *Expr;

	Support = virtex_mapp_int( Node, &Expr, CreateAux );
	freechain( Support );

	if ( ! ABL_ATOM( Expr ) ) {
		if ( CreateAux == 1 )
        		fpm_mapp_node( Node, &Expr );
		virtex_num_clb++;
	}
	return simpablexpr( Expr );
}

void
virtex_report(void)
{
	printf("   + Final number of Lut: %d\n", virtex_num_clb);
}

void
virtex_save()
{
}
