#include "PPlacement.h"
#include <unistd.h>
#include <algorithm>
extern "C"{
#include MUT_H
#include MPH_H
#include MPU_H
#include MLO_H
#include MLU_H
#include "iocheader.h"
}
#include "PMove.h"
#include "PConstants.h"
#include "PCommon.h"

struct CompPInsByWidth {
    bool operator()(PIns* Ins1, PIns* Ins2) {
	return (Ins1->GetWidth() > Ins2->GetWidth());
    }
};

void
PPlacement::Init(lofig* fig, int NbRows)
{
    lofigchain(fig);
    rflattenlofig(fig, YES, YES);
    _fig = fig;
    _dx = _dy = 0;
    // construction des tables
    PNetMap pnetmap;
    PInsMap pinsmap;
    PConMap pconmap;
    PFixedMap fixedmap;
    PLoconMap ploconmap;

    unsigned fixedcpt = 0;
    if (_prePlace)
    {
	rflattenphfig(_prePlaceFig, YES, YES);
	for (phins* pins = _prePlaceFig->PHINS; pins; pins = pins->NEXT)
	{
	    fixedmap[pins->INSNAME] = 1;
	    ++fixedcpt;
	}
	_fixedInss.reserve(fixedcpt);

	_dx = _prePlaceFig->XAB1;
	_dy = _prePlaceFig->YAB1;

	for (phins* pins = _prePlaceFig->PHINS; pins; pins = pins->NEXT)
	{
	    // on verifie que la loins existe en memoire
	    loins* ins = getloins(_fig, pins->INSNAME);
	    // verification de la coherence figure logique / figure physique
	    if (ins->FIGNAME != pins->FIGNAME)
	    {
		cerr << "ERROR: The logical figure and the physical figure are not coherent:" << endl
		     << "Logical Instance: " << (char*)ins->INSNAME << " -> logical model: "<< (char*)ins->FIGNAME << endl
		     << "Physical Instance: " << (char*)pins->INSNAME << " -> physical model: "<< (char*)pins->FIGNAME << endl;
		exit(1);
	    }
	    int sigcpt = 0;
	    set<losig*> inssigset;
	    for (locon* con = ins->LOCON ; con ; con = con->NEXT)
	    {
		losig* sig = con->SIG;
		set<losig*>::iterator lsit = inssigset.find(sig);
		if (lsit == inssigset.end())
		{
		    if (!IsSpecialNet(sig))
		    {
			++sigcpt;
			inssigset.insert(sig);
		    }
		}
	    }
	    _fixedInss.push_back(PFixedIns(ins, pins, sigcpt, _dx , _dy));
	    pinsmap[ins->INSNAME] = &_fixedInss.back();
	}
    }
    
    unsigned nbins = 0;
    unsigned instoplacecpt = 0;
    unsigned insfixedcpt = 0;
    PFixedMap::const_iterator fit;
    for (loins* ins = fig->LOINS; ins; ins = ins->NEXT)
    {
	++nbins;
	fit = fixedmap.find(ins->INSNAME);
	if (fit == fixedmap.end())
	    ++instoplacecpt;
	else
	    ++insfixedcpt;
    }
    if (_verbose)
    {
    cout << "Number total of instances is ....  " << nbins << endl;
    cout << "Number of instances to place is ....  " << instoplacecpt << endl;
    cout << "Number of instances already placed is ....  " << insfixedcpt << endl;
    }
    
    _toPlaceInss.reserve(instoplacecpt);
    for (loins* ins = fig->LOINS; ins; ins = ins->NEXT)
    {
	PFixedMap::iterator fit = fixedmap.find(ins->INSNAME);
	if (fit == fixedmap.end())
	{
	    int sigcpt = 0;
	    // on evite de compter deux fois le meme net pour
	    // chaque instance.....
	    set<losig*> inssigset;
	    for (locon* con = ins->LOCON ; con ; con = con->NEXT)
	    {
		losig* sig = con->SIG;
		set<losig*>::iterator lsit = inssigset.find(sig);
		if (lsit == inssigset.end())
		{
		    if (!IsSpecialNet(sig))
		    {
			++sigcpt;
			inssigset.insert(sig);
		    }
		}
	    }
	    phfig* phmodel = getphfig(ins->FIGNAME, '0');
	    int height = (int)((double)(phmodel->YAB2 - phmodel->YAB1) / PITCH + 0.5);
	    if (height != ROWHEIGHT)
	    {
		cerr << "All cells must have the same height: i cannot place the instance: "
		     << ins->INSNAME << endl;
		exit(1);
	    }
	    _toPlaceInss.push_back(PToPlaceIns(ins, sigcpt));
	    pinsmap[ins->INSNAME] = &_toPlaceInss.back();
	}
    }
    _nInsToPlace = _toPlaceInss.size();

    // Traitement des connecteurs ....
    if (_placeCons)
    {
	unsigned concpt = 0;
	for (locon* con = fig->LOCON; con; con = con->NEXT)
	{
	    losig* sig = con->SIG;
	    if (!IsSpecialNet(sig))
		++concpt;
	}
	_cons.reserve(concpt);

	for (locon* con = fig->LOCON; con; con = con->NEXT)
	{
	    losig* sig = con->SIG;
	    if (!IsSpecialNet(sig))
	    {
		_cons.push_back(PCon(con));
		pconmap[con->NAME] = &_cons.back();
	    }
	}

    }
    
    for (locon* con = fig->LOCON; con; con = con->NEXT)
      ploconmap[con->NAME] = con;
    
    // traitement de l'emplacement des connecteurs
    // creation des PCon si fichier ioc
    if (_iocFile)
    {
	ParseIocFile(ploconmap);
	for (PCons::iterator pcon = _cons.begin() ; pcon != _cons.end(); pcon++)
	  pconmap[pcon->GetLocon()->NAME] = pcon;
    }

    unsigned sigcpt = 0;
    for (losig* sig = fig->LOSIG; sig; sig = sig->NEXT)
    {
	if (!IsSpecialNet(sig))
		++sigcpt;
    }
    
    _nets.reserve(sigcpt);
    sigcpt = 0;
    for (losig* sig = fig->LOSIG; sig; sig = sig->NEXT)
    {
	if (!IsSpecialNet(sig)) 
	{
	    set<loins*> siginsset;
	    int totreatinscpt = 0;
	    int elemcpt = 0;
	    for (chain_list* it = (chain_list *)(getptype(sig->USER, (long)LOFIGCHAIN)->DATA);
		    it;
		    it = it->NEXT)
	    {
		locon_list* con = (locon_list*)(it->DATA);
		if (con->TYPE == EXTERNAL)
		{
		    if (_placeCons || _iocFile)
		    {
		      PConMap::iterator cit = pconmap.find(con->NAME);
		      if (cit != pconmap.end())
			++totreatinscpt;
		    }
		}
		else
		{
		    loins* ins = (loins *)con->ROOT;
		    set<loins*>::iterator liit = siginsset.find(ins);
		    if (liit == siginsset.end())
		    {
			++totreatinscpt;
			siginsset.insert(ins);
		    }
		}
		++elemcpt;
	    }
	    if (elemcpt < 2)
	    {
		cerr << " o OCP Warning: strange net detected ...: " << endl;
		cerr << " o Net:" << endl;
		cerr << " o ";
		if (!sig->NAMECHAIN)
		    cerr << "(noname) Index: " << sig->INDEX;
		else
		{
		    struct chain* cit = sig->NAMECHAIN;

		    if (!cit->NEXT)
			cerr << (char*)cit->DATA;
		    else
			for (struct chain* cit = sig->NAMECHAIN; cit; cit = cit->NEXT)
			    cerr << (char*)cit->DATA << " ";
		}
		cerr << endl;
		if (elemcpt == 0)
		    cerr << " o is not connected ......" << endl;
		if (elemcpt == 1)
		{
		    cerr << " o is connected to only one element: ";
		    chain_list* it = (chain_list *)(getptype(sig->USER, (long)LOFIGCHAIN)->DATA);
		    locon_list* con = (locon_list*)(it->DATA);
		    if (con->TYPE == EXTERNAL)
		    {
			cerr << "connector " << con->NAME;
		    }
		    else
		    {
			loins* ins = (loins *)con->ROOT;
			cerr << "instance " << ins->INSNAME;
		    }
		    cerr << endl;
		}
		    
	    }
	    
	    _nets.push_back(PONet(sig, totreatinscpt));
	    pnetmap[sig->INDEX] = &_nets.back();
	}
	else {
       cerr << " o Special Net detected : " << getsigname(sig) << endl;
	}
    }

    unsigned vectcpt = 0;
    for (loins* ins = fig->LOINS ; ins ; ins = ins->NEXT)
    {
	set<losig*> inssigset;
	PInsMap::iterator iit = pinsmap.find(ins->INSNAME);
	if (iit == pinsmap.end())
	{
	    cerr << "ins " << ins->INSNAME << " is not present in data...." << endl;
	    exit(1);
	}
	PElem::PNets& netvector = iit->second->GetNets();
	vectcpt = 0;
	for (locon* con = ins->LOCON ; con ; con = con->NEXT)
	{
	    losig* sig = con->SIG;
	    set<losig*>::iterator lsit = inssigset.find(sig);
	    if (lsit == inssigset.end())
	    {
		if (!IsSpecialNet(sig))
		{
		    PNetMap::iterator nit = pnetmap.find(sig->INDEX);
		    if (nit == pnetmap.end())
		    {
			cerr << "net " << sig->INDEX << " is not present in data...." << endl;
			exit(1);
		    }
		    netvector[vectcpt++] = nit->second;
		    inssigset.insert(sig);
		}
	    }
	}
    }

    if (_placeCons || _iocFile)
    {
	for (locon* con = fig->LOCON; con; con = con->NEXT)
	{
	    losig* sig = con->SIG;
	    if (!IsSpecialNet(sig))
	    {
		PConMap::iterator cit = pconmap.find(con->NAME);
		if (cit != pconmap.end())
		{
		  PElem::PNets& netvector = cit->second->GetNets();
		  assert(netvector.size() == 1);
		  PNetMap::iterator nit = pnetmap.find(sig->INDEX);
		  if (nit == pnetmap.end())
		  {
		      cerr << "net " << sig->INDEX << " is not present in data...." << endl;
		      exit(1);
		  }
		  netvector[0] = nit->second;
		}
	    }
	}
    }

    for (losig* sig = fig->LOSIG ; sig ; sig=sig->NEXT)
    {
	set<loins*> siginsset;
	if (!IsSpecialNet(sig)) 
	{
	    PNetMap::iterator nit = pnetmap.find(sig->INDEX);
	    if (nit == pnetmap.end())
	    {
		cerr << "net " << sig->INDEX << " is not present in data...." << endl;
		exit(1);
	    }
	    PNet::PElems& elemsvector = nit->second->GetElems();
	    vectcpt = 0;
	    for (chain_list* it = (chain_list *)(getptype(sig->USER, (long)LOFIGCHAIN)->DATA);
		    it;
		    it = it->NEXT)
	    {
		locon_list* con = (locon_list*)(it->DATA);
		if (con->TYPE == INTERNAL)
		{
		    loins* ins = (loins *)con->ROOT;
		    set<loins*>::iterator liit = siginsset.find(ins);
		    if (liit == siginsset.end())
		    {
			PInsMap::iterator iit = pinsmap.find(ins->INSNAME);
			if (iit == pinsmap.end())
			{
			    cerr << "ins " << ins->INSNAME << " is not present in data...." << endl;
			    exit(1);
			}
			elemsvector[vectcpt++] = iit->second;
			siginsset.insert(ins);
		    }
		}
		else if (_placeCons || _iocFile)
		{
		    PConMap::iterator cit = pconmap.find(con->NAME);
		    if (cit != pconmap.end())
		    {
		      elemsvector[vectcpt++] = cit->second;
		    }
		}
		    
	    }
	    assert(elemsvector.size() == vectcpt);
	}

    }

    // getting NetList Caracteristics

    _sumToPlaceInssWidth = 0.0;
    _biggestToPlaceInsWidth = 0.0;
    double inswidth = 0.0;
    for (PToPlaceInss::const_iterator ins_it = _toPlaceInss.begin();
	    ins_it != _toPlaceInss.end(); ins_it++)
    {
        if (_biggestToPlaceInsWidth < (inswidth = ins_it->GetWidth()))
	    _biggestToPlaceInsWidth = inswidth;
        _sumToPlaceInssWidth += inswidth;
    }
    
    _binsMaxWidth = (double)(int)(2.0 * _biggestToPlaceInsWidth * (1.0 + _margin) + 0.5);
    _binsMinWidth = (double)(int)(_binsMaxWidth / 2);

    if (_prePlace)
    {
	InitPlaceWithPrePlace();
	if (_nInsToPlace > 0)
	    assert (_rows.begin() != _rows.end());
    }
    else
	if (_nInsToPlace > 0)
	    InitPlace(NbRows);

    // set positions of connectors in case of no ioc file
    if (_placeCons) GenerateConsPlacement();
    
    // set positions of connectors when ioc file is given
    if (_iocFile) SetPosIocFile(pconmap);
    
    if (_nInsToPlace==0)
	return;

    // sorting instances by their Width (biggest -> smallest)
    vector<PToPlaceIns*> ClassedInss;
    ClassedInss.reserve(_nInsToPlace);
    for (PToPlaceInss::iterator ifirst = _toPlaceInss.begin(); ifirst != _toPlaceInss.end(); ifirst++)
    {
        PToPlaceIns*   InsPt;
        ClassedInss.push_back(InsPt);
        ClassedInss.back() = &(*ifirst);
    }
    sort(ClassedInss.begin(), ClassedInss.end(), CompPInsByWidth());
    
    // Insertion of instances on rows
    vector<PToPlaceIns*>::iterator iptfirst = ClassedInss.begin();
    PRows::iterator rfirst = _rows.begin();
    PRow::PSubRows::iterator subrfirst = rfirst->GetSubRows().begin();
    
    PIns* InsInserting = *iptfirst;
    while(1) {
        int NbOfBins;
        int BinNumber;
            
        if (iptfirst == ClassedInss.end())
        {
            if (_verbose) cout << "Initial Placement Computing ... done. " << endl;
            break;
            // end of insertion
        }

	if (subrfirst == rfirst->GetSubRows().end())
	{
	    ++rfirst;
	    subrfirst = rfirst->GetSubRows().begin();
	}
	
        if (rfirst == _rows.end()) 
        {
            if (InsInserting != *iptfirst)
            {
                InsInserting = *iptfirst;
                rfirst = _rows.begin();
		subrfirst = rfirst->GetSubRows().begin();
            }
            else
            {
                // insertion of instances with respect of Bins margin
                // did not succed, inserting what's left.
                rfirst = _rows.begin();
		subrfirst = rfirst->GetSubRows().begin();
                NbOfBins = subrfirst->GetNBins();
                BinNumber = 0;
                while (iptfirst != ClassedInss.end()) {
                    if ((*iptfirst)->GetWidth() <=
                            subrfirst->GetWidth() - subrfirst->GetSize())
                    {
                        subrfirst->ForceIns(**(iptfirst++), BinNumber++);
                    }
                    else
                        BinNumber++;
                    if (BinNumber >= NbOfBins)
                    {
                        subrfirst++;
                        BinNumber = 0;
                    }
		    
		    if (subrfirst == rfirst->GetSubRows().end())
		    {
			rfirst++;
			subrfirst = rfirst->GetSubRows().begin();
		    }
		    
                    if (rfirst == _rows.end())
                    {
                        if (InsInserting == *iptfirst)
                        {
                            // impossible to succed
                            cerr << " o The is no enough free space, I need a bigger abutment box ......" << endl;
                            cerr << " o There are still "
                                << ClassedInss.end() - iptfirst 
                                << " Instances to place ..." << endl;
                            exit(1);
                        }
                        rfirst = _rows.begin();
			subrfirst = rfirst->GetSubRows().begin();
                        InsInserting = *iptfirst;
                    }
                }
            }
        }
        NbOfBins = subrfirst->GetNBins();
        BinNumber = 0;
        while ((iptfirst != ClassedInss.end())
                && (BinNumber < NbOfBins)
                && ((*iptfirst)->GetWidth() <= subrfirst->GetCapa() - subrfirst->GetSize()))
        {
            if (subrfirst->InsertIns(**iptfirst, BinNumber++))
	    {
                iptfirst++;
	    }
        }
	subrfirst++;
    }

}

// ======================================================================
// void SetPosIocFile(PConMap& pconmap)
// ======================================================================
void
PPlacement::SetPosIocFile(PConMap& pconmap)
{
  int topcount 	  = 0;
  int bottomcount = 0;
  int rightcount  = 0;
  int leftcount   = 0;
  int topspace	  = 0;
  int bottomspace = 0;
  int rightspace  = 0;
  int leftspace   = 0;
  int nbspacetop    = 0;
  int nbspacebottom = 0;
  int nbspaceright  = 0;
  int nbspaceleft   = 0;

  con_list* pttop    = NULL;
  con_list* ptbottom = NULL;
  con_list* ptright  = NULL;
  con_list* ptleft   = NULL;

  PPos position;
  
  int Width = (int)(BBox.GetMaxX() - BBox.GetMinX());
  int Height = (int)(BBox.GetMaxY() - BBox.GetMinY());
  if (_verbose) cout << "Width of the abutment box : " << Width << endl
    << "Height of the abutment box : " << Height << endl;

  con_list* tmpnext;

  for(con_list* tmpcon = _PtList ; tmpcon; tmpcon = tmpnext)
  {
    tmpnext = tmpcon->NEXT;
    switch (tmpcon->ORIENT)
    {
      case 'T' : tmpcon->NEXT = pttop;
		 pttop = tmpcon;
		 if (strcmp(tmpcon->NAME,"SPACE")==0)
		 {
		   topspace += tmpcon->VALUE;
		   nbspacetop++;
		 } else topcount++;
		 break;
		 
      case 'B' : tmpcon->NEXT = ptbottom;
		 ptbottom = tmpcon;
		 if (strcmp(tmpcon->NAME,"SPACE")==0)
		 {
		   bottomspace += tmpcon->VALUE;
		   nbspacebottom++;
		 } else bottomcount++;
		 break;
		 
      case 'R' : tmpcon->NEXT = ptright;
		 ptright = tmpcon;
		 if (strcmp(tmpcon->NAME,"SPACE")==0)
		 {
		   rightspace += tmpcon->VALUE;
		   nbspaceright++;
		 } else rightcount++;
		 break;
		 
      case 'L' : tmpcon->NEXT = ptleft;
		 ptleft = tmpcon;
		 if (strcmp(tmpcon->NAME,"SPACE")==0)
		 {
		   leftspace += tmpcon->VALUE;
		   nbspaceleft++;
		 } else leftcount++;
		 break;	 
    }
  }

  if (topcount != 0)
  {
    double pos = 0.0;
    if (topspace >= Width-topcount)
    {
      cerr << "ERROR : in ioc file : space too important" << endl;
      exit(1);
    }
    double conspace = ((double)(Width-topspace)/(double)(topcount-nbspacetop));
    pos += conspace / 2.0;

    if (_verbose) cout << "conspace : " << conspace << "1st connector : " << pos << endl;

    for(con_list* tmpcon = pttop ; tmpcon; tmpcon = tmpcon->NEXT)
    {
      if (strcmp(tmpcon->NAME,"SPACE") == 0)
      {
	pos += (double)tmpcon->VALUE;
	pos -= conspace;
      }
      else
      {
	if (_verbose) cout << "ajout de connecteur : " << tmpcon->NAME
	  << " x : " << (int)pos << " y : " << Height << endl;
	position.SetX((double)(int)pos);
	position.SetY(Height);
	// test a rajouter pour verifier si locon present ou pas!!!
	pconmap[tmpcon->NAME]->SetPos(position);
	pos += conspace;
      }
    }
  }

  if (bottomcount != 0)
  {
    double pos = 0.0;
    if (bottomspace >= Width-topcount)
    {
      cerr << "ERROR : in ioc file : space too important" << endl;
      exit(1);
    }
    double conspace = ((double)(Width-bottomspace)/
					(double)(bottomcount-nbspacebottom));
    pos += conspace / 2.0;

    if (_verbose) cout << "conspace : " << conspace << "1st connector : " << pos << endl;
    
    for(con_list* tmpcon = ptbottom ; tmpcon; tmpcon = tmpcon->NEXT)
    {
      if (strcmp(tmpcon->NAME,"SPACE") == 0)
      {
	pos += (double)tmpcon->VALUE;
	pos -= conspace;
      }
      else
      {
	if (_verbose) cout << "ajout de connecteur : " << tmpcon->NAME
	  << " x : " << (int)pos << " y : " << 0 << endl;
	position.SetX((double)(int)pos);
	position.SetY(0.0);
	// test a rajouter pour verifier si locon present ou pas!!!
	pconmap[tmpcon->NAME]->SetPos(position);
	pos += conspace;
      }
    }
  }

  if (leftcount != 0)
  {
    double pos = 0.0;
    if (leftspace >= Width-topcount)
    {
      cerr << "ERROR : in ioc file : space too important" << endl;
      exit(1);
    }
    double conspace = ((double)(Height-leftspace)/
					(double)(leftcount-nbspaceleft));
    pos += conspace / 2.0;

    for(con_list* tmpcon = ptleft ; tmpcon; tmpcon = tmpcon->NEXT)
    {
      if (strcmp(tmpcon->NAME,"SPACE") == 0)
      {
	pos += (double)tmpcon->VALUE;
	pos -= conspace;
      }
      else
      {
	if (_verbose) cout << "ajout de connecteur : " << tmpcon->NAME
	  << " x : " << 0 << " y : " << (int)pos << endl;
	position.SetX(0.0);
	position.SetY((double)(int)pos);
	// test a rajouter pour verifier si locon present ou pas!!!
	pconmap[tmpcon->NAME]->SetPos(position);
	pos += conspace;
      }
    }
  }

  if (rightcount != 0)
  {
    double pos = 0.0;
    if (rightspace >= Width-topcount)
    {
      cerr << "ERROR : in ioc file : space too important" << endl;
      exit(1);
    }
    double conspace = ((double)(Height-rightspace)/
					(double)(rightcount-nbspaceright));
    pos += conspace / 2.0;

    for(con_list* tmpcon = ptright ; tmpcon; tmpcon = tmpcon->NEXT)
    {
      if (strcmp(tmpcon->NAME,"SPACE") == 0)
      {
	pos += (double)tmpcon->VALUE;
	pos -= conspace;
      }
      else
      {
	if (_verbose) cout << "ajout de connecteur : " << tmpcon->NAME
	  << " x : " << Width << " y : " << (int)pos << endl;
	position.SetX(Width);
	position.SetY((double)(int)pos);
	// test a rajouter pour verifier si locon present ou pas!!!
	pconmap[tmpcon->NAME]->SetPos(position);
	pos += conspace;
      }
    }
  }

 
}
// ======================================================================
// void ParseIocFile(PLoconMap& ploconmap)
// ======================================================================

void
PPlacement::ParseIocFile(PLoconMap& ploconmap)
{
  PPos position;
  
  _PtList = iocparse(_iocFileName);
  
  int nbioc = 0;
  char orientation;
  
  for(con_list* tmpcon = _PtList ; tmpcon; tmpcon = tmpcon->NEXT)
    if (strcmp(tmpcon->NAME,"SPACE") != 0) nbioc++;

  _cons.reserve(nbioc);

  for(con_list* tmpcon = _PtList ; tmpcon ; tmpcon = tmpcon->NEXT)
  {
    if (strcmp(tmpcon->NAME,"SPACE") != 0)
    {
      switch (tmpcon->ORIENT)
      {
	case 'T' : orientation = NORTH;
		   break;
	case 'B' : orientation = SOUTH;
		   break;
	case 'L' : orientation = WEST;
		   break;
	case 'R' : orientation = EAST;
		   break;
	default: cerr << "Unknown Orientation for connector: "
		 << tmpcon->NAME << endl;
		 exit(1);
      }
	position.SetX(0);
	position.SetY(0);
	_cons.push_back(PCon(ploconmap[tmpcon->NAME], position, orientation));
    }
  }
  

}

// ======================================================================
// void Place(lofig* fig, int NbRows);
// ======================================================================
void
PPlacement::Place(lofig* fig, int NbRows)
{
    Init(fig, NbRows);
    if (_boolPlot)
    {
	PlotOnlyInstances("instances-init");
	PlotAll("init");
    }
    if (_nInsToPlace > 0)
    {
	PlaceGlobal();
	if (_boolPlot)
	{
	  PlotAll("afterglobal");
	  PlotOnlyBins("binsafterglobal");
	}
	PlaceFinal();
    }
}

ostream&
PPlacement::Print(ostream& os) const
{
    PToPlaceInss::const_iterator ifirst = _toPlaceInss.begin();
    PToPlaceInss::const_iterator ilast  = _toPlaceInss.end();
    while (ifirst != ilast) {
	const PToPlaceIns& ins = *ifirst++;
	os << " + ins: " << ins << endl;
	PIns::PNets::const_iterator first = ins.GetConstNets().begin();
	PIns::PNets::const_iterator last  = ins.GetConstNets().end();
	while (first != last)
	    os << "   - net: " << *first++ << endl;
    }

    PONets::const_iterator nfirst = _nets.begin();
    PONets::const_iterator nlast  = _nets.end();
    while (nfirst != nlast) {
	const PNet& net = *nfirst++;
	os << " + net: " << net << endl;
	PNet::PElems::const_iterator first = net.GetConstElems().begin();
	PNet::PElems::const_iterator last  = net.GetConstElems().end();
	while (first != last)
	    os << "   - ins: " << *first++ << endl;
    }

    return os;
}

// ======================================================================
// void PlotInstances(ofstream& out) const
// ======================================================================
void
PPlacement::PlotInstances(ofstream& out) const
{
  
  out << "#rows" << endl;
  for (PRows::const_iterator RowIt=_rows.begin(); RowIt!=_rows.end(); RowIt++)
  {
      RowIt->Plot(out);
  }
  out << "EOF" << endl;
  out << "#to place instances" << endl;
  for (PToPlaceInss::const_iterator insit = _toPlaceInss.begin();
	  insit != _toPlaceInss.end(); insit++)
  {
      insit->Plot(out);
  }
  out << "EOF" << endl;
  out << "#preplaced instances" << endl;
  for (PFixedInss::const_iterator insit = _fixedInss.begin();
	  insit != _fixedInss.end(); insit++)
  {
      insit->Plot(out);
  }
  out << "EOF" << endl;
}
// ======================================================================
// void PlotAll(char *output) const
// ======================================================================
void
PPlacement::PlotAll(char* output) const
{
  char* name = (char*)malloc((strlen(output) + 5)* sizeof(char*));
  name = strcpy(name, output);
  ofstream out(strcat(name, ".gpl"));
  
  out << "set noxtics\nset noytics\n"
    << "set noborder\nset nokey\nset title '" << _fig->NAME << "'\n"
    << "#set terminal postscript eps color solid\n#set output '"
    << output << ".ps'\n"
    << "set xrange[" << GetMinX() << ":" << GetMaxX() << "]\n"
    << "set yrange[" << GetMinY() << ":" << GetMaxY() << "]\n"
    << "plot [:][:][:][:] '-' w l, '-' w l 2, '-' w l 3, '-' w l 4\n"
    << GetMinX() << " " << GetMinY() << endl
    << GetMinX() << " " << GetMaxY() << endl
    << GetMaxX() << " " << GetMaxY() << endl
    << GetMaxX() << " " << GetMinY() << endl
    << GetMinX() << " " << GetMinY() << endl << endl;
  PlotInstances(out);
  
  out << "#nets" << endl;
  for (PONets::const_iterator NetIt=_nets.begin(); NetIt!=_nets.end(); NetIt++)
  {
      NetIt->Plot(out);
  }
  out << "EOF" << endl << "pause -1 'press any key'" << endl;
}

// ======================================================================
// void PlotFinal(char *output) const
// ======================================================================
void
PPlacement::PlotFinal(char* output) const
{
  char* name = (char*)malloc((strlen(output) + 5)* sizeof(char*));
  name = strcpy(name, output);
  ofstream out(strcat(name, ".gpl"));
  
  out << "set noxtics\nset noytics\n"
    << "set noborder\nset nokey\nset title '" << _fig->NAME << "'\n"
    << "#set terminal postscript eps color solid\n#set output '"
    << output << ".ps'\n"
    << "set xrange[" << GetMinX() << ":" << GetMaxX() << "]\n"
    << "set yrange[" << GetMinY() << ":" << GetMaxY() << "]\n"
    << "plot [:][:][:][:] '-' w l, '-' w l 2, '-' w l 3, '-' w l 4\n"
    << GetMinX() << " " << GetMinY() << endl
    << GetMinX() << " " << GetMaxY() << endl
    << GetMaxX() << " " << GetMaxY() << endl
    << GetMaxX() << " " << GetMinY() << endl
    << GetMinX() << " " << GetMinY() << endl << endl;

  out << "#preplaced instances" << endl;
  for (PFixedInss::const_iterator insit = _fixedInss.begin();
	  insit != _fixedInss.end(); insit++)
  {
      insit->Plot(out);
  }
  out << "EOF" << endl;

  out << "#subrows" << endl;
  for (PDetSubRows::const_iterator SRowIt = _detSubRows.begin() ;
	SRowIt != _detSubRows.end() ;
	SRowIt++)
  {
    SRowIt->Plot(out);
  }
  out << "EOF" << endl;

  out << "#instances" << endl;
  for (PDetSubRows::const_iterator SRowIt = _detSubRows.begin() ;
	SRowIt != _detSubRows.end() ;
	SRowIt++)
  {
    for (PDetSubRow::PDetInsVector::const_iterator InsIt = 
      					SRowIt->GetConstInssVector().begin() ;
    		InsIt != SRowIt->GetConstInssVector().end() ;
		InsIt++)
    {
      InsIt->Plot(out);
    }
  }
  out << "EOF" << endl;

  out << "#nets" << endl;
  for (PONets::const_iterator NetIt=_nets.begin(); NetIt!=_nets.end(); NetIt++)
  {
      NetIt->Plot(out);
  }
  out << "EOF" << endl << "pause -1 'press any key'" << endl;
}
// ======================================================================
// void PlotOnlyInstances(char* output) const
// ======================================================================
void
PPlacement::PlotOnlyInstances(char* output) const
{
  char* name = (char*)malloc((strlen(output) + 5)* sizeof(char*));
  name = strdup(output);
  ofstream out(strcat(name, ".gpl"));

  out << "set noxtics\nset noytics\n"
    << "set noborder\nset nokey\nset title '" << _fig->NAME << "'\n"
    << "#set terminal postscript eps color solid\n#set output '"
    << output << ".ps'\n";
    out << "set xrange[" << GetMinX() << ":" << GetMaxX() << "]\n"
    << "set yrange[" << GetMinY() << ":" << GetMaxY() << "]\n"
    << "plot [:][:][:] '-' w l, '-' w l 2, '-' w l 3\n"
    << GetMinX() << " " << GetMinY() << endl
    << GetMinX() << " " << GetMaxY() << endl
    << GetMaxX() << " " << GetMaxY() << endl
    << GetMaxX() << " " << GetMinY() << endl
    << GetMinX() << " " << GetMinY() << endl << endl;
  PlotInstances(out);
  out << "pause -1 'press any key'" << endl;
}
// ======================================================================
// void PlotOnlyBins(char* output) const
// ======================================================================
void
PPlacement::PlotOnlyBins(char* output) const
{  
  char* name = (char*)malloc((strlen(output) + 5)* sizeof(char*));
  name = strdup(output);
  ofstream out(strcat(name, ".gpl"));

  out << "set noxtics\nset noytics\n"
    << "set noborder\nset nokey\nset title '" << _fig->NAME << "'\n"
    << "#set terminal postscript eps color solid\n#set output '"
    << output << ".ps'\n";

  if (_totalMoves != 0)
  {
    for (PRows::const_iterator RowIt=_rows.begin(); RowIt!=_rows.end(); RowIt++)
    {
      RowIt->PlotLabel(out, _totalMoves);
    }
  }

  out << "plot [:][:] '-' w l 1, '-' w l 2" << endl << "#bins" << endl;
  for (PRows::const_iterator RowIt=_rows.begin(); RowIt!=_rows.end(); RowIt++)
  {
      RowIt->Plot(out);
  }
  out << "EOF" << endl;
  
  out << "#preplaced instances" << endl;
  for (PFixedInss::const_iterator insit = _fixedInss.begin();
	  insit != _fixedInss.end(); insit++)
  {
      insit->Plot(out);
  }
  out << "EOF" << endl;
  out << "pause -1 'press any key'" << endl;
}
// ======================================================================
// void PlotStat()
// ======================================================================
void
PPlacement::PlotStat()
{
  ofstream out("stat.gpl");

  out << "set title 'all stats'\n#set terminal postscript color solid\n"
    << "set output 'stat.ps'" << endl
    << "plot 'alldata.dat' using 1:2 title 'temperature' with lines,"
    << "'alldata.dat' using 1:3 title 'cost' with lines,"
    << "'alldata.dat' using 1:4 title 'rowcost' with lines,"
    << "'alldata.dat' using 1:5 title 'bincost' with lines,"
    << "'alldata.dat' using 1:6 title 'netcost' with lines,"
    << "'alldata.dat' using 1:7 title 'success ratio' with lines,"
    << "'alldata.dat' using 1:8 title 'dist' with lines,"
    << "'alldata.dat' using 1:9 title 'delta' with lines" << endl
    << "pause -1 'press any key'" << endl;
}

// ======================================================================
// void GenerateConsPlacement()
// private method
// ======================================================================
void
PPlacement::GenerateConsPlacement()
{
    //creation d'un placement aleatoire des connecteurs.
    unsigned faces[4];
    //faces north est sud ouest
    faces[0] = faces[1] = faces[2] = faces[3] = _cons.size() / 4;
    unsigned consrest = _cons.size()%4;

    unsigned faceidx = 0;
    while(consrest-- > 0)
	++faces[faceidx++];

    PCons::iterator cit = _cons.begin();

    double minx = BBox.GetMinX();
    double miny = BBox.GetMinY(); 
    double maxx = BBox.GetMaxX();
    double maxy = BBox.GetMaxY();

    double pos = 0.0;
    double interval = 0.0;
    
    //placement au nord.
    interval = ((maxx - minx) / (faces[0]));
    pos = minx + interval / 2;
    while(faces[0]--)
    {
	(cit)->SetOrient(NORTH);
	(cit++)->SetPos(PPos((double)(int)pos , maxy));
	pos += interval;
    }
    
    //placement a l'est.
    interval = ((maxy - miny) / (faces[1]));
    pos = miny + interval / 2;
    while(faces[1]--)
    {
	(cit)->SetOrient(EAST);
	(cit++)->SetPos(PPos((double)(int)maxx , pos));
	pos += interval;
    }
    
    //placement au sud.
    interval = ((maxx - minx) / (faces[2]));
    pos = minx + interval / 2;
    while(faces[2]--)
    {
	(cit)->SetOrient(SOUTH);
	(cit++)->SetPos(PPos((double)(int)pos , miny));
	pos += interval;
    }
    
    //placement a l'ouest.
    interval = ((maxy - miny) / (faces[3]));
    pos = miny + interval / 2;
    while(faces[3]--)
    {
	(cit)->SetOrient(WEST);
	(cit++)->SetPos(PPos((double)(int)minx , pos));
	pos += interval;
    }
}

// ======================================================================
// void InitPlace(int NbRows)
// private method
// ======================================================================
void
PPlacement::InitPlace(int NbRows)
{
    // no preplacement
    // we make a rectangle of rows with for each one subrow
    // of size = size of the row
    // composed of bins of constant size
    int NRows = NbRows;
    
    if (_verbose)
	cout << " o Computing Initial Placement ..." << endl;
    
    if (_verbose)
    {
      cout << "User Margin : " << _margin << endl;
    }
    
    double rowwidth = SquareShape(_margin, _sumToPlaceInssWidth,
	    _binsMinWidth, NRows);
    
    if (_verbose)
    {
      cout << "Number of Rows : " << NRows << endl;
    }
    
    // computing Effective Margin
    _rows.reserve(NRows);
    
    double Y = 0.0;
    unsigned rowidx = 0;
    bool RowOrientation = false;
    _rowZeroOrientation = RowOrientation;
    for (int i = 0; i < NRows; i++)
    {
	_rows.push_back(PRow(1));
        double XMin = 0.0;
	_rows.back().Init(Y, XMin, *this, RowOrientation);
	_rowsYMax[Y + ROWHEIGHT] = rowidx;
	_rowsYMinInv[Y] = rowidx;
	++rowidx;
	PSubRow& subrow = _rows.back()._subRows[0];
	subrow.Init(&(_rows.back()), Y, XMin, rowwidth, _margin, 
		_binsMaxWidth, _binsMinWidth);
	_rows.back()._subRowsXMax[rowwidth + XMin] = 0;
	_rows.back()._subRowsXMaxInv[rowwidth + XMin] = 0;
        Y += ROWHEIGHT;
	RowOrientation = !RowOrientation;
    }

    // Computing Placement BBox
    double MaxX = 0.0, MaxY = 0.0, RowMaxX, RowMaxY;
    for (PRows::iterator rfirst = _rows.begin(); rfirst != _rows.end(); rfirst++)
    {
        RowMaxX = rfirst->GetMaxX();
        RowMaxY = rfirst->GetMaxY();
        if (MaxX < RowMaxX)
            MaxX = RowMaxX;
        if (MaxY < RowMaxY)
            MaxY = RowMaxY;
    }
    BBox.SetMinX(0.0);
    BBox.SetMinY(0.0);
    BBox.SetMaxX(RowMaxX);
    BBox.SetMaxY(RowMaxY);
}

// ======================================================================
// void InitPlaceWithPrePlace()
// private method
// ======================================================================
void
PPlacement::InitPlaceWithPrePlace()
{
    // abutment box
    assert(_prePlaceFig);
    int Width = _prePlaceFig->XAB2 - _prePlaceFig->XAB1;
    int Height = _prePlaceFig->YAB2 - _prePlaceFig->YAB1;

    // VERIFICATION OF THE PREPLACEMENT
    // abutment box
    if (Width % PITCH != 0) {
      cerr << "ERROR : abutment box's width must be a multiple of 5 lambdas"
	<< endl;
      exit(1);
    }
    
    if (Height % (PITCH * ROWHEIGHT) != 0) {
      cerr << "ERROR : abutment box's height must be a multiple of 50 lambdas"
	<< endl;
      exit(1);
    }
    
    Width =(int)(Width / PITCH);
    Height = (int)(Height / (PITCH * ROWHEIGHT));

    PrePlaceTab tabpreplace(Height, PrePlaceRow(Width, false));

    // find the orientation of the first row.
    phins* refpins = _prePlaceFig->PHINS;

    int ycoord = (int)((refpins->YINS - _dy) / (ROWHEIGHT * PITCH));
    char transf = refpins->TRANSF;

    if (ycoord % 2 == 0)
    {
      if ((transf == NOSYM) || (transf == SYM_X))
	_rowZeroOrientation = true;
      else
	_rowZeroOrientation = false;
    } else {
      if ((transf == NOSYM) || (transf == SYM_X))
	_rowZeroOrientation = false;
      else
	_rowZeroOrientation = true;
    }

    // tests for each instance
    for (phins* pins = _prePlaceFig->PHINS; pins; pins = pins->NEXT)
    {
	phfig_list* phmodel = getphfig(pins->FIGNAME, '0');
	int pinswidth = phmodel->XAB2 - phmodel->XAB1;
	int pinsheight = phmodel->YAB2 - phmodel->YAB1;
	
	pinswidth = (int)(pinswidth / PITCH);			// largeur ramene au pitch
	pinsheight = (int)(pinsheight / PITCH);			// hauteur ramene au pitch
	int pinsheightrow = (int)(pinsheight / ROWHEIGHT);	// hauteur ramene a l'unite
								// (taille des lignes)
	int ypos = (int)((pins->YINS - _dy) /  PITCH);		// position en y ramene au pitch
	int xpos = (int)((pins->XINS - _dx) /PITCH);		// position en x ramene au pitch
	int ycoord = (int)(ypos / ROWHEIGHT);			// position en y ramene a l'unite

	if ((pins->YINS - _dy) % 50 != 0)
	{
	  cerr << "ERROR : in preplacement file : y position of "
	    << pins->INSNAME << " is incorrect" << endl;
	  exit (1);
	}

	if ((pins->XINS - _dx) % 5 != 0)
	{
	  cerr << "ERROR : in preplacement file : x position of "
	    << pins->INSNAME << " is incorrect" << endl;
	  exit (1);
	}

	if (   (pins->TRANSF == ROT_P) || (pins->TRANSF == ROT_M)
	    || (pins->TRANSF == SY_RP) || (pins->TRANSF == SY_RM) )
	{
	  cerr << "ERROR : " << pins->INSNAME << " : incorrect rotation" << endl;
	  exit(1);
	}
	
	// check if orientation of instance matches
	bool insOrient;
	if ((pins->TRANSF == NOSYM) || (pins->TRANSF == SYM_X))
	  insOrient = true;
	else
	  insOrient = false;

	if (   ((ycoord % 2 == 0) && (insOrient != _rowZeroOrientation))
	    || ((ycoord % 2 != 0) && (insOrient == _rowZeroOrientation)) )
	{
	  cerr << "ERROR : " << pins->INSNAME << " badly oriented" << endl
	       << "Incoherence with " << refpins->INSNAME << endl;
	  exit(1);
	}
	
	for (int yit = ycoord; yit < ycoord + pinsheightrow; yit++)
	{
	  for (int xit = xpos; xit < xpos + pinswidth; xit++)
	  {
	    if (   (xit > Width) || (yit > Height)
		|| (xit < 0    ) || (yit < 0     ) )
	    {
	      cerr << "ERROR : " << pins->INSNAME
		   << " out of the abutment box" << endl;
	      exit(1);
	    }
	    
	    if (tabpreplace[yit][xit] == false)
	    {
		tabpreplace[yit][xit] = true;
	    }
	    else{
	      cerr << "ERROR : " << pins->INSNAME << " badly placed .... There is already an instance at its position ...." << endl;
	      exit (1);
	    }
	  }
	}
    }

    // create rows and subrows
    if (_nInsToPlace > 0)
    {
	int nbrows = 0;
	for (int yit = 0; yit < Height; yit++)
	{
	  if (CheckCreateRow(tabpreplace, yit, Width) > 0) ++nbrows;
	}


	_rows.reserve(nbrows);

	bool orientation = _rowZeroOrientation;
	int numrows = 0;
	for (int yit = 0; yit < Height; yit++)
	{
	  int numsubrows = CheckCreateRow(tabpreplace, yit, Width);
	  if (numsubrows > 0)
	  {
	    _rows.push_back(PRow(*this, (double)(yit * ROWHEIGHT), Width, numsubrows, orientation));
	    _rowsYMax[(yit + 1) * ROWHEIGHT] = numrows;
	    _rowsYMinInv[yit * ROWHEIGHT] = numrows;
	    ++numrows;
	    CreateSubRows(&_rows.back(), tabpreplace, yit, numsubrows, Width);
	  }
	  orientation = !orientation;
	}
    }
    BBox.SetMinX(0.0);
    BBox.SetMinY(0.0);
    BBox.SetMaxX(Width);
    BBox.SetMaxY(Height * ROWHEIGHT);
}

// ======================================================================
// void CreateSubRows(PRow* row, PrePlaceTab& tabpreplace,int coordy, int nbsubrows, int Width)
// private method
// ======================================================================
void
PPlacement::CreateSubRows(PRow* row, PrePlaceTab& tabpreplace,
   					int coordy, int nbsubrows, int Width)
{
  int coordx = 0;
  int subrowscreated = 0;

  while (coordx < Width)
  {
    while (( coordx < Width  			  )
	&& ( tabpreplace[coordy][coordx] == true ))
      ++coordx;

    if (coordx >= Width)
	break;

    int xmin = coordx;
	
    while (( coordx < Width  			 )
	&& ( tabpreplace[coordy][coordx] == false ))
      ++coordx;

    int xmax = coordx;

    if ( (xmax - xmin) >= _binsMinWidth)
    {
	row->_subRows[subrowscreated].Init(row, (double)(coordy * ROWHEIGHT), (double)xmin,(double)xmax,
		(double)_margin, _binsMaxWidth, _binsMinWidth);
	row->_subRowsXMax[(double)xmax] = subrowscreated;
	++subrowscreated;
    }
    for (PRow::PSubRowsXMax::iterator srymit = row->_subRowsXMax.begin();
	    srymit != row->_subRowsXMax.end();
	    srymit++)
    {
	row->_subRowsXMaxInv[srymit->first] = srymit->second;
    }
  }
  assert(subrowscreated == nbsubrows);
}

// ======================================================================
// void CheckCreateRow(PrePlaceTab& tabpreplace, int coordy, int Width)
// private method
// ======================================================================
int
PPlacement::CheckCreateRow(PrePlaceTab& tabpreplace, int coordy, int Width)
{
  int coordx = 0;
  int nbsubrows = 0;
  
  while (coordx < Width)
  {
    while (  ( coordx < Width  			)
	  && ( tabpreplace[coordy][coordx] == true	))
      ++coordx;

    if (coordx >= Width)
	break;
	
    int xmin = coordx;

    while (  ( coordx < Width  			)
	  && ( tabpreplace[coordy][coordx] == false	))
      ++coordx;

    int xmax = coordx;

    if ( (xmax - xmin) >= _binsMinWidth)
	++nbsubrows;
  }

  return (nbsubrows);
}

// ======================================================================
// PRow& GetRow(const PRow* row, const double dist)
// private method
// ======================================================================
PRow&
PPlacement::GetRow(const PRow* row, const double dist)
{
    // returns randomly a row at maximum distance from yorig
    double bornesup, borneinf;

    
    if ((bornesup = (row->GetMinY() + row->GetHeight() / 2.0) + (double)(int)(dist * GetHeight() + 0.5) ) > GetMaxY())
        bornesup = GetMaxY();

    if ((borneinf = (row->GetMinY() + row->GetHeight() / 2.0) - (double)(int)(dist * GetHeight() + 0.5) ) < GetMinY())
        borneinf = GetMinY();

    PRowsYMax::iterator rinf = _rowsYMax.upper_bound(borneinf);
    PRowsYMax::iterator rsup = _rowsYMinInv.upper_bound(bornesup);

    unsigned randidx = rinf->second + (unsigned)((double)(rsup->second - rinf->second + 1) * (rand() / (RAND_MAX+1.0)));
    return _rows[randidx];
}
