/*
   ### -------------------------------------------------- ### 
   $Author: hcl $
   $Date: 2001/09/26 14:27:49 $
   $Log: ocrRouter.c,v $
   Revision 1.1  2001/09/26 14:27:49  hcl
   premier commit....

   ### -------------------------------------------------- ###
 */

#include <stdlib.h>
#include <string.h>
#include MUT_H
#include MLO_H
#include MLU_H
#include MPH_H
#include MPU_H

#include "ocr.h"
#include "ocrWRoutingDataBase.h"
#include "ocrWindow.h"
#include "ocrUtil.h"
#include "ocrPath.h"
#include "display.h"
#include "ocrNpoints.h"
#include "ocrGlobalRouting.h"

#include "ocrInitDataBase.h"
#include "ocrSignalUtil.h"
#include "getOption.h"
#include "ocrToPhFig.h"

static char *res_id =
  "$Id: ocrRouter.c,v 1.1 2001/09/26 14:27:49 hcl Exp $";

#define SCALE_X 100
#define PITCH 5

#define FINDNPATH findPathNPoints
#define FINDPATH findPathBetween2Points

ocrNaturalInt g_OptLayer = 5;
ocrNaturalInt g_OptWindow = 1;
ocrOption *g_pOption;

#define LEVEL (g_pOption->LEVEL)

/*
 * readPhFig : 
 *  1 - Lecture de la figure i_sPhFigName
 *  2 - Translation vers (0,0)
 */
phfig_list *
readPhFig (char *i_sPhFigName)
{
  phfig_list *l_pPhFig;
  phins_list *l_pInst;
  ocrNaturalInt l_uDeltaX;
  ocrNaturalInt l_uDeltaY;

  display (LEVEL, VERBOSE, "%s\n", "Read .ap ...");
  l_pPhFig = getphfig (i_sPhFigName, 'A');

//FIXME::pour l'instant pas de reentrance: donc flat jusqu'au CATAL

  rflattenphfig(l_pPhFig, YES, YES);

  if ((l_pPhFig->XAB1 != 0) || (l_pPhFig->YAB1 != 0))
    {
      display (LEVEL, VERBOSE, "  %s\n", "Translating .ap ...");

      l_uDeltaX = 0 - l_pPhFig->XAB1;
      l_uDeltaY = 0 - l_pPhFig->YAB1;

      // phins
      for (l_pInst = l_pPhFig->PHINS; l_pInst; l_pInst = l_pInst->NEXT)
        {
          l_pInst->XINS += l_uDeltaX;
          l_pInst->YINS += l_uDeltaY;
        }

      // phfig abutment box
      l_pPhFig->XAB1 += l_uDeltaX;
      l_pPhFig->XAB2 += l_uDeltaX;
      l_pPhFig->YAB1 += l_uDeltaY;
      l_pPhFig->YAB2 += l_uDeltaY;

    }

  return l_pPhFig;
}

ocrSignal *
findSignal (ocrRoutingDataBase * i_pDataBase, ocrNaturalInt i_uIndex)
{
  ocrNaturalInt i;

  if (i_uIndex == OCRNATURALINT_MAX)
    return NULL;

  for (i = 0; i < i_pDataBase->NB_GSIGNAL - 1; i++)
    {
      if (i_pDataBase->GSIGNAL[i]->INDEX == i_uIndex)
        return i_pDataBase->GSIGNAL[i];
    }
  return NULL;
}

/**
 * route les signaux dans l'ordre :
 *  1. i_pSignal
 *  2. les signaux qui ont ete supprimes par ripUp
 *
 *  return :
 *   1 OK
 *   0 ECHEC
 **/
ocrNaturalInt
reRoot (ocrRoutingDataBase * i_pDataBase,
        ocrRoutingParameters * i_pParam, ocrSignal * i_pSignal)
{
  ocrSignal *l_pSignal;
  ocrNaturalInt l_uLength;
  ocrNaturalInt l_uReturnValue = 1;
  chain_list *l_pSignalList;

//  puts ("reRoot");

  // routage du signal dsign par i_pSignal
  //  fprintf (stdout, "REROOT : Routage du signal %ld \n",
  //         i_pSignal->INDEX);
  l_uLength = FINDNPATH (i_pParam, i_pDataBase->GRID,
                         i_pSignal, i_pDataBase->WINDOWS[i_pSignal->WIN]);

  if (l_uLength != OCRNATURALINT_MAX)
    {
      // Succes du routage
      i_pDataBase->NB_ROUTED++;

      i_pSignal->ROUTED = 1;
    }
  else
    {
      l_uReturnValue = 0;
      //  puts ("Echec reRoot");

      i_pSignal->ROUTED = 2;
      deleteEquipotentielle (i_pParam, i_pDataBase->GRID, i_pSignal);
      i_pSignal->SEGMENT = NULL;

      i_pDataBase->NB_UNROUTED++;
      i_pSignal->PRIORITY = i_pDataBase->NB_IT;
    }

  // tant qu'il reste des signaux  re router
  l_pSignalList = i_pDataBase->RIPUP;
  while (l_pSignalList)
    {
      l_pSignal = (ocrSignal *) l_pSignalList->DATA;

//    fprintf (stdout, "REROOT : Routage du signal %ld \n",
      //             l_pSignal->INDEX);
      l_uLength = FINDNPATH (i_pParam, i_pDataBase->GRID,
                             l_pSignal, i_pDataBase->WINDOWS[i_pSignal->WIN]);

      if (l_uLength == OCRNATURALINT_MAX)
        {
          l_uReturnValue = 0;
          display (LEVEL, DEBUG, "Echec reRoot %ld\n", l_pSignal->INDEX);
          l_pSignal->ROUTED = 2;

          deleteEquipotentielle (i_pParam, i_pDataBase->GRID, l_pSignal);

          l_pSignal->SEGMENT = NULL;

          i_pDataBase->NB_UNROUTED++;
          l_pSignal->PRIORITY = i_pDataBase->NB_IT;
//        l_pSignal->PRIORITY ++;
        }
      else
        {
          i_pDataBase->NB_ROUTED++;
          l_pSignal->ROUTED = 1;
        }

      l_pSignalList = l_pSignalList->NEXT;
    }

  freechain (i_pDataBase->RIPUP);
  i_pDataBase->RIPUP = NULL;
  return l_uReturnValue;
}

/**
 * supprime les equipotentielles qui empeche le routage
 * du signal i_pSignal
 **/
void
ripUp (ocrRoutingDataBase * i_pDataBase,
       ocrRoutingParameters * i_pParam, ocrSignal * i_pSignal)
{
  ocrConnector *l_pCon;
  ocrVirtualConnector *l_pVirCon;
  ocrSignal *l_pSignal;

  // recherche du connecteur qui pose un problme
  for (l_pCon = i_pSignal->CON_LIST;
       l_pCon && l_pCon->CON; l_pCon = l_pCon->NEXT);

  if (l_pCon)
    {
      // parcours des connecteurs virtuels
      for (l_pVirCon = l_pCon->VIR_CON_LIST;
           l_pVirCon; l_pVirCon = (ocrVirtualConnector *) l_pVirCon->NEXT)
        {
          // recherche du signal qui gene
          l_pSignal = findSignal (i_pDataBase,
                                  getWSegmentCV (i_pDataBase->GRID,
                                                 l_pVirCon)->SIGNAL_INDEX);
          if ((l_pSignal) && (l_pSignal->PRIORITY == 0))
            {
              // ajoute le signal dans la liste  re-router.
              i_pDataBase->RIPUP = addchain (i_pDataBase->RIPUP,
                                             (void *) l_pSignal);

              // suppression de son equipotentielle
              //                        fprintf (stdout, "RIPUP : Suppression du signal %ld \n",
              //                                 l_pSignal->INDEX);

              deleteEquipotentielle (i_pParam, i_pDataBase->GRID, l_pSignal);
              l_pSignal->SEGMENT = NULL;
              l_pSignal->ROUTED = 0;
              i_pDataBase->NB_ROUTED--;
            }
        }
      // supression du signal lui meme
      //    fprintf (stdout, "RIPUP : Suppression du signal %ld \n",
      //             i_pSignal->INDEX);
      deleteEquipotentielle (i_pParam, i_pDataBase->GRID, i_pSignal);
      i_pSignal->SEGMENT = NULL;
      i_pSignal->ROUTED = 0;
    }
}

void
addViaForIntersection (ocrRoutingParameters * i_pParam,
                       ocrWSegment * i_pSegment, phfig_list * i_pPhFig)
{
  ocrWSegment *l_pSegment;

  if (i_pSegment->NEXT != NULL)
    {
      for (l_pSegment = i_pSegment->NEXT;
           l_pSegment; l_pSegment = l_pSegment->NEXT)
        {
          if ((i_pSegment->LAYER - 1 == l_pSegment->LAYER) ||
              (i_pSegment->LAYER + 1 == l_pSegment->LAYER))
            {
              if ((i_pSegment->OFFSET >= l_pSegment->P_MIN) &&
                  (i_pSegment->OFFSET <= l_pSegment->P_MAX) &&
                  (l_pSegment->OFFSET >= i_pSegment->P_MIN) &&
                  (l_pSegment->OFFSET <= i_pSegment->P_MAX))
                {
                  if (getWSegDirection (i_pParam, i_pSegment) ==
                      ocrHorizontal)
                    {
                      addViaToPhFig (i_pParam,
                                     l_pSegment->OFFSET,
                                     i_pSegment->OFFSET,
                                     MIN (i_pSegment->LAYER,
                                          l_pSegment->LAYER), i_pPhFig);
                    }
                  else
                    {
                      addViaToPhFig (i_pParam,
                                     i_pSegment->OFFSET,
                                     l_pSegment->OFFSET,
                                     MIN (i_pSegment->LAYER,
                                          l_pSegment->LAYER), i_pPhFig);
                    }
                }
            }
        }
    }
}

void
dumpSignalsToPhFig (ocrSignal * i_pSignal,
                    ocrRoutingParameters * i_pParam, phfig_list * i_pPhFig)
{
  ocrSignal *l_pSignal;
  ocrWSegment *l_pSegment;

  for (l_pSignal = i_pSignal; l_pSignal; l_pSignal = l_pSignal->NEXT)
    {
      for (l_pSegment = l_pSignal->SEGMENT;
           l_pSegment; l_pSegment = l_pSegment->NEXT)
        {
          addSegmentToPhFig (i_pParam, l_pSegment, i_pPhFig);
          addViaForIntersection (i_pParam, l_pSegment, i_pPhFig);
        }
    }
}

ocrNaturalInt
getNewSignal (ocrRoutingDataBase * i_pDataBase, ocrNaturalInt i_uOldSignal)
{
  static ocrNaturalInt l_uRangSignal;
  ocrNaturalInt l_uNewSignal;

//  display (LEVEL, DEBUG, "%s\n", "getNewSignal");

  if (g_pOption->ORDER & OCR_ODYN)
    {
      // Initialisation
      if (i_uOldSignal == OCRNATURALINT_MAX)
        {
          l_uRangSignal = 0;
          return 0;
        }

      // Signal Critique ?
      l_uNewSignal = isCriticalSignal (i_pDataBase, i_uOldSignal);
      if (l_uNewSignal == OCRNATURALINT_MAX)
        {
          // non. rang suivant si possible
          while (l_uRangSignal < i_pDataBase->NB_SIGNAL - 1)
            {
              l_uRangSignal++;
              if (i_pDataBase->SIGNAL[l_uRangSignal]->ROUTED == 0)
                {
                  display (LEVEL, DEBUG, "Signal %ld ", l_uRangSignal);
                  return l_uRangSignal;
                }
            }
          return OCRNATURALINT_MAX;
        }
      else
        {
          // oui signal critique; on ne modifie pas le rang !
          display (LEVEL, DEBUG, "Signal %ld critique \n", l_uNewSignal);
          return l_uNewSignal;
        }
    }
  else
    {
      l_uNewSignal = i_uOldSignal + 1;
      if (l_uNewSignal == i_pDataBase->NB_SIGNAL)
        return OCRNATURALINT_MAX;
      else
        return l_uNewSignal;
    }
}

ocrNaturalInt
routingWindow (ocrRoutingDataBase * i_pDataBase, phfig_list * i_pPhFig)
{
  ocrConnector *l_pCon;
  ocrWindow *l_pWindow;
  ocrSignal *l_pSignal;
  ocrVirtualConnector *l_pVirCon;
  ocrNaturalInt i, l_uWin;
  //ocrNaturalInt l_uNbf = sqrt (i_pDataBase->NB_F);


  i_pDataBase->NB_UNROUTED = 1;
  i_pDataBase->NB_IT = 0;
  i_pDataBase->NB_ROUTED = 0;

  // allocation de la grille
  display (LEVEL, DEBUG, "%s ", "Grid allocation ...");
  l_uWin = i_pDataBase->SIGNAL[0]->WIN;
  l_pWindow = i_pDataBase->WINDOWS[l_uWin];

  i_pDataBase->GRID = createWGrid (l_pWindow->XMAX - l_pWindow->XMIN + 1,
                                   l_pWindow->YMAX - l_pWindow->YMIN + 1,
                                   g_OptLayer);

  while (i_pDataBase->NB_UNROUTED != 0)
    {
      ocrNaturalInt length;

      i_pDataBase->NB_IT++;
      i_pDataBase->NB_UNROUTED = 0;
      i_pDataBase->NB_ROUTED = 0;


      // initialisation de la grille
      initWGrid (i_pDataBase->GRID, i_pDataBase->PARAM);
      display (LEVEL, DEBUG, "%s\n", "done");

      // Initialisation des TAG & ROUTED
      display (LEVEL, DEBUG, "%s\n", "Initialisation des TAG");
      for (i = 0; i < i_pDataBase->NB_SIGNAL; i++)
        {
          i_pDataBase->SIGNAL[i]->ROUTED = 0;

          for (l_pCon = i_pDataBase->SIGNAL[i]->CON_LIST;
               l_pCon; l_pCon = l_pCon->NEXT)
            {
              // suppression de la liste des segments
              i_pDataBase->SIGNAL[i]->SEGMENT = NULL;

              if (!IS_MARK_AS_LINKED (l_pCon))
                {
                  // on ne dtruit pas la position du connecteur facial associ
                  //             l_pCon->CON = NULL;
                }

              deleteVirtualConnectorList (i_pDataBase->SIGNAL[i]->VIA);
              i_pDataBase->SIGNAL[i]->VIA = NULL;

              for (l_pVirCon = l_pCon->VIR_CON_LIST;
                   l_pVirCon;
                   l_pVirCon = (ocrVirtualConnector *) l_pVirCon->NEXT)
                {
                  l_pVirCon->TAG = 0;
                  l_pVirCon->DIST = 0;
                }
            }
        }

      if (g_pOption->ORDER & OCR_RIPUP)
        display (LEVEL, VERBOSE, "%s\n", "Rip Up & Re Root");

      if (g_pOption->ORDER & OCR_ODYN)
        display (LEVEL, VERBOSE, "%s\n", "Ordonnancement dynamique");

      i = OCRNATURALINT_MAX;

      while ((i = getNewSignal (i_pDataBase, i)) != OCRNATURALINT_MAX)
        if (!g_pOption->PRIORITY ||
            (g_pOption->PRIORITY && i_pDataBase->SIGNAL[i]->PRIORITY))
          {
            l_pSignal = i_pDataBase->SIGNAL[i];

            if (l_pSignal->ROUTED != 0)
              exit (1);

            display (LEVEL, DEBUG,
                     "Routage %ld points, Index : %ld, Pri : %d :\n",
                     l_pSignal->NB_CON, l_pSignal->INDEX,
                     l_pSignal->PRIORITY);

            // Routage N points

            length = FINDNPATH (i_pDataBase->PARAM, i_pDataBase->GRID,
                                l_pSignal, l_pWindow);

            if (length != OCRNATURALINT_MAX)
              {
                l_pSignal->ROUTED = 1;
                i_pDataBase->NB_ROUTED++;
              }
            else
              {
                if (g_pOption->ORDER & OCR_RIPUP)
                  {
                    display (LEVEL, DEBUG,
                             "Rip up Sig %ld, %ld points, Index : %ld, Pri : %d :\n",
                             i, l_pSignal->NB_CON, l_pSignal->INDEX,
                             l_pSignal->PRIORITY);
/*
   display (LEVEL, INFO, "It = %ld Sig = %ld Echec = %ld\n",
   i_pDataBase->NB_IT,
   i_pDataBase->NB_ROUTED,
   i_pDataBase->NB_UNROUTED);
 */
                    ripUp (i_pDataBase, i_pDataBase->PARAM, l_pSignal);


                    reRoot (i_pDataBase, i_pDataBase->PARAM, l_pSignal);
                  }
/*
   display (LEVEL, INFO, "It = %ld Sig = %ld Echec = %ld\n\n",
   i_pDataBase->NB_IT,
   i_pDataBase->NB_ROUTED,
   i_pDataBase->NB_UNROUTED);
 */
                else
                  {
                    l_pSignal->ROUTED = 2;

                    i_pDataBase->NB_UNROUTED++;
                    l_pSignal->PRIORITY = i_pDataBase->NB_IT;
                  }

              }
          }

      display (LEVEL, VERBOSE, "It = %ld Sig = %ld Echec = %ld\n",
               i_pDataBase->NB_IT,
               i_pDataBase->NB_ROUTED, i_pDataBase->NB_UNROUTED);

      // Ordonnancement les + PRIORITAIRES en premier
      orderingSignals (i_pDataBase->SIGNAL,
//                       i_pDataBase->NB_SIGNAL, ORDER_PRIORITY_RANDOM);
                       i_pDataBase->NB_SIGNAL, ORDER_PRIORITY);

      {
        ocrNaturalInt PRIO_MAX = i_pDataBase->SIGNAL[0]->PRIORITY;
        if (PRIO_MAX > 8)
          for (i = 0; i < i_pDataBase->NB_SIGNAL - 1; i++)
            if (i_pDataBase->SIGNAL[i]->PRIORITY == PRIO_MAX)
              {
                if (i_pDataBase->SIGNAL[i]->INTEXT == INTERNAL)
                  display (LEVEL, VERBOSE,
                           "I Prio Max = %ld Sig = %ld Name = %s\n", PRIO_MAX,
                           i, i_pDataBase->SIGNAL[i]->NAME);
                else
                  display (LEVEL, VERBOSE,
                           "E Prio Max = %ld Sig = %ld Name = %s\n", PRIO_MAX,
                           i, i_pDataBase->SIGNAL[i]->NAME);
              }
      }

    }
  freeWGrid (i_pDataBase->GRID);
  i_pDataBase->GRID = NULL;

  display (LEVEL, INFO, "Itration = %ld\n", i_pDataBase->NB_IT);
  display (LEVEL, INFO, "Signaux = %ld\n", i_pDataBase->NB_ROUTED);
  display (LEVEL, INFO, "Echec = %ld\n", i_pDataBase->NB_UNROUTED);
  return 1;
}

void
dumpConToPhFig (ocrSignal * i_pSignal,
                ocrRoutingParameters * i_pParam, phfig_list * i_pPhFig)
{
  ocrConnector *l_pCon;
  ocrVirtualConnector *l_pVia;
  ocrSignal *l_pSignal;

  display (LEVEL, VERBOSE, "Dumping Connectors to PhFig ...\n");

  for (l_pSignal = i_pSignal; l_pSignal; l_pSignal = l_pSignal->NEXT)
    {
      // CON
      for (l_pCon = l_pSignal->CON_LIST; l_pCon; l_pCon = l_pCon->NEXT)
        {
          if (l_pCon->INTEXT == EXTERNAL)
            if (!IS_MARK_AS_FACE_CON (l_pCon))
              if (l_pCon->CON)
                addphcon (i_pPhFig,
                          l_pCon->CON->ORIENT,
                          l_pCon->NAME,
                          l_pCon->CON->X * 500,
                          l_pCon->CON->Y * 500, ALU2, 2 * SCALE_X);
        }

      // VIA & CON    
      for (l_pVia = l_pSignal->VIA;
           l_pVia; l_pVia = (ocrVirtualConnector *) l_pVia->NEXT)
        {
          addViaToPhFig (i_pParam, l_pVia->X, l_pVia->Y, -1, i_pPhFig);
        }
    }
}

void
linkWindow (ocrRoutingDataBase * i_pDataBase, ocrNaturalInt i_uWindow)
{
  ocrSignal *l_pSignal;

  display (LEVEL, INFO, "Link Window %ld\n", i_uWindow);
  for (l_pSignal = i_pDataBase->FSIGNAL[i_uWindow];
       l_pSignal; l_pSignal = l_pSignal->NEXT)
    linkSignal (i_pDataBase, l_pSignal);
}

/**
 * affiche l'aide
**/
void
dumpOption ()
{
  puts ("usage : ocr *.ap *.vst options");
  puts ("options :");
  puts ("  -v  :  verbose mode");
  puts ("  -d  :  debug mode");
  puts (" -l n :  n layers n>=2");
  puts (" -w n :  n windows");
  puts ("");

}

int
main (int argc, char **argv)
{
  phfig_list *l_pPhFig;
  lofig_list *l_pLoFig;
  ocrRoutingDataBase *l_pDataBase;
  char *l_sPhFigName;
  char *l_sLoFigName;

  if (argc < 2)
    {
      dumpOption ();
      exit (1);
    }

  // initialisation de MBK
  mbkenv ();

  l_sPhFigName = argv[1];
  l_sLoFigName = argv[2];

  // Option de la ligne de commande
  g_pOption = getOption (argc - 2, argv + 2);

  // Lecture de la vue physique
  l_pPhFig = readPhFig (l_sPhFigName);

  // Lecture de la vue logique
  l_pLoFig = getlofig (l_sLoFigName, 'A');
  
//FIXME::pas de reentrance
  rflattenlofig(l_pLoFig, YES, YES);

  display (LEVEL, VERBOSE, "%s\n", "Read .vst ...");

  l_pDataBase = initDataBase (l_pPhFig, l_pLoFig);

  // un peu de mnage.
//  delphfig (l_pPhFig->NAME);
  dellofig (l_pLoFig->NAME);

  // paramtres physiques de la grille 5*5 viacost=2 Layer paires = HORZ
  l_pDataBase->PARAM = setParameters (5, 5, 2, ocrHorizontal);

  {
    ocrNaturalInt i;

    createWindow (l_pDataBase);
//    dumpDataBase (l_pDataBase, stdout);
    translatingRealToWindows (l_pDataBase);
    globalRouting (l_pDataBase);

    // Routage de chaque Window
    for (i = 1; i <= l_pDataBase->NB_F; i++)
      {
        createWindowArray (l_pDataBase, i);
        display (LEVEL, INFO, "Routing Win %ld : %ld signaux ...\n",
                 i, l_pDataBase->NB_SIGNAL);
        orderingSignals (l_pDataBase->SIGNAL,
                         l_pDataBase->NB_SIGNAL, ORDER_PRIORITY_CON);

        // dumpDataBase (l_pDataBase,stdout);
        routingWindow (l_pDataBase, l_pPhFig);
        linkWindow (l_pDataBase, i);

        deleteWindowArray (l_pDataBase, i);
      }

    // dumpToDisk
    translatingWindowsToReal (l_pDataBase);
    display (LEVEL, VERBOSE, "Dumping Signals to PhFig ...\n", i);
    for (i = 1; i <= l_pDataBase->NB_F; i++)
      {
        dumpSignalsToPhFig (l_pDataBase->FSIGNAL[i],
                            l_pDataBase->PARAM, l_pPhFig);
        dumpConToPhFig (l_pDataBase->FSIGNAL[i],
                        l_pDataBase->PARAM, l_pPhFig);
      }

    dumpAlimToPhFig (l_pDataBase, l_pPhFig);

    dumpPhFigToDisk (l_pDataBase->NAME, l_pPhFig);
  }

  deleteDataBase (l_pDataBase);
  return 0;
}
