/*
   ### -------------------------------------------------- ### 
   $Author: hcl $
   $Date: 2001/09/26 14:27:51 $
   $Log: ocrToPhFig.c,v $
   Revision 1.1  2001/09/26 14:27:51  hcl
   premier commit....

   Revision 1.3  2000/02/26 00:24:09  root
   2 points OK

   Revision 1.2  2000/01/25 15:49:00  root
   *** empty log message ***

   ### -------------------------------------------------- ###
 */
#include MUT_H
#include MLO_H
#include MLU_H
#include MPH_H
#include MPU_H
#include <string.h>
#include "ocr.h"
#include "ocrUtil.h"
#include "ocrWRoutingDataBase.h"
#include "ocrWindow.h"
#include "mbk_tree.h"
#include "ocrWRoutingUtil.h"
#include "ocrToPhFig.h"

#define SCALE_X 	100
#define PITCH		5

static char *res_id = "$Id: ocrToPhFig.c,v 1.1 2001/09/26 14:27:51 hcl Exp $";


/* 
   ### --------------------------- ###
   ### Internal Functions
   ### --------------------------- ###
 */

/**
 * Retourne le VIA physique correspondant  la coordone z
**/
int
getPhVia (int i_uZ)
{
  switch (i_uZ)
    {
    case -1:
	 return CONT_VIA;
    case 0:
	 return CONT_VIA2;
    case 1:
	 return CONT_VIA3;
    case 2:
	 return CONT_VIA4;
    case 3:
	 return CONT_VIA5;
    case 4:
	 return CONT_VIA6;
    case 5:
	 return CONT_VIA7;
    }
  puts ("ERROR : getPhVia");
  exit (1);

  // pour viter un Warning ...  
  return 0;
}

/**
 * Retourne le LAYER physique correspondant  la coordone z
**/

ocrNaturalInt
getPhLayer (ocrNaturalInt i_uZ)
{
  switch (i_uZ)
    {
    case 0:
	 return ALU2;
    case 1:
	 return ALU3;
    case 2:
	 return ALU4;
    case 3:
	 return ALU5;
    case 4:
	 return ALU6;
    case 5:
	 return ALU7;
    case 6:
	 return ALU8;
    }

  puts ("ERROR : getPhLayer");
  exit (1);

  // pour viter un Warning ...  
  return 0;
}

/**
 * Ajoute un via dans la io_pPhFig
 * i_uLayer est le Layer de niveau le plus bas par rapport au via
**/
void
addViaToPhFig (ocrRoutingParameters * i_pParam,
			ocrNaturalInt i_uX,
			ocrNaturalInt i_uY,
			int i_nLayer,
			phfig_list * io_pPhFig)
{
#if 0
  addphvia (io_pPhFig,
		  getPhVia (i_nLayer),
		  i_uX * i_pParam->PITCH_H * SCALE_X,
		  i_uY * i_pParam->PITCH_H * SCALE_X);
  //printf ("Add via %d en %ld %ld\n",i_nLayer,i_uX,i_uY);
#else

  addphvia (io_pPhFig,
		  getPhVia (i_nLayer),
		  i_uX * i_pParam->PITCH_H * SCALE_X,
		  i_uY * i_pParam->PITCH_H * SCALE_X,
		  0,
		  0,
		  NULL);
#endif
}

/**
 * Ajoute un segment dans la io_pPhFig
**/

void
addSegmentToPhFig (ocrRoutingParameters * i_pParam,
			    ocrWSegment * i_pSegment,
			    phfig_list * io_pPhFig)
{
  char l_sName[7];

  sprintf (l_sName, "seg%ld", i_pSegment->SIGNAL_INDEX);

  if (getWSegDirection (i_pParam, i_pSegment) == ocrHorizontal)
    {
	 addphseg (io_pPhFig,
			 getPhLayer (i_pSegment->LAYER),
			 i_pParam->PITCH_V * 4 * SCALE_X / 10,
			 i_pSegment->P_MIN * i_pParam->PITCH_H * SCALE_X,
			 i_pSegment->OFFSET * i_pParam->PITCH_V * SCALE_X,
			 i_pSegment->P_MAX * i_pParam->PITCH_H * SCALE_X,
			 i_pSegment->OFFSET * i_pParam->PITCH_V * SCALE_X,
			 l_sName);
    }
  else
    {
	 addphseg (io_pPhFig,
			 getPhLayer (i_pSegment->LAYER),
			 i_pParam->PITCH_H * 2 * SCALE_X / 10,
			 i_pSegment->OFFSET * i_pParam->PITCH_H * SCALE_X,
			 i_pSegment->P_MIN * i_pParam->PITCH_V * SCALE_X,
			 i_pSegment->OFFSET * i_pParam->PITCH_H * SCALE_X,
			 i_pSegment->P_MAX * i_pParam->PITCH_V * SCALE_X,
			 l_sName);
    }
}


/**
 * Dump le signal en (x,y,z) en segment physique
**/

void
dumpSignal (ocrRoutingParameters * i_pParam,
		  ocrWRoutingGrid * i_pGrid,
		  phfig_list * io_pPhFig,
		  ocrNaturalInt x,
		  ocrNaturalInt y,
		  ocrNaturalInt z)
{

  ocrWSegment *l_pSegment;
  ocrWSegment *l_pSegmentVoisin;
  ocrNaturalInt l_uNoSignal;

  // segment concern
  l_pSegment = getWSegment (i_pGrid, x, y, z);

  // sauvegarde du numro du signal
  l_uNoSignal = l_pSegment->SIGNAL_INDEX;

// printf("dumpSignal %ld %ld %ld\n",x,y,z);
  //   printf("Index = %ld\n",l_uNoSignal);

  // quand on parcours un signal il devient libre
  // pour ne pas parcourir 2 fois le meme signal
  l_pSegment->SIGNAL_INDEX = WSEGMENT_FREE;

  if (getWSegDirection (i_pParam, l_pSegment) == ocrHorizontal)
    {
	 ocrNaturalInt i;

	 for (i = l_pSegment->P_MIN; i <= l_pSegment->P_MAX; i++)
	   {
		if (z > 0)
		  {
		    l_pSegmentVoisin = getWSegment (i_pGrid, i, y, z - 1);
		    // connexion avec LAYER - 1 ?
		    if (l_pSegmentVoisin /*&& (!l_pSegmentVoisin->TAG) */ )
			 if (l_pSegmentVoisin->SIGNAL_INDEX == l_uNoSignal)
			   {
				//l_pSegmentVoisin->TAG = 1;
				dumpSignal (i_pParam, i_pGrid, io_pPhFig, i, y, z - 1);
				// ajout d'un via
				addViaToPhFig (i_pParam, i, y, z - 1, io_pPhFig);
			   }
		  }
		if (z < i_pGrid->NB_OF_LAYERS - 1)
		  {
		    l_pSegmentVoisin = getWSegment (i_pGrid, i, y, z + 1);
		    // connexion avec LAYER + 1 ?
		    if (l_pSegmentVoisin /* && (!l_pSegmentVoisin->TAG) */ )
			 if (l_pSegmentVoisin->SIGNAL_INDEX == l_uNoSignal)
			   {
				//l_pSegmentVoisin->TAG = 1;
				dumpSignal (i_pParam, i_pGrid, io_pPhFig, i, y, z + 1);
				// ajout d'un via
				addViaToPhFig (i_pParam, i, y, z, io_pPhFig);
			   }
		  }
	   }
    }
  else
    {
	 ocrNaturalInt i;

	 for (i = l_pSegment->P_MIN; i <= l_pSegment->P_MAX; i++)
	   {
		if (z > 0)
		  {
		    l_pSegmentVoisin = getWSegment (i_pGrid, x, i, z - 1);
		    //         printf("  VERT %ld %ld %ld index = %ld TAG = %ld\n",
		    //             x,i,z-1,l_pSegmentVoisin->SIGNAL_INDEX,
		    //                  l_pSegmentVoisin->TAG);
		    if (l_pSegmentVoisin /*&& (!l_pSegmentVoisin->TAG) */ )
			 if (l_pSegmentVoisin->SIGNAL_INDEX == l_uNoSignal)
			   {
				//l_pSegmentVoisin->TAG = 1;
				dumpSignal (i_pParam, i_pGrid, io_pPhFig, x, i, z - 1);
				// ajout d'un via
				addViaToPhFig (i_pParam, x, i, z - 1, io_pPhFig);
			   }
		  }
		if (z < i_pGrid->NB_OF_LAYERS - 1)
		  {
		    l_pSegmentVoisin = getWSegment (i_pGrid, x, i, z + 1);
		    if (l_pSegmentVoisin /*&& (!l_pSegmentVoisin->TAG) */ )
			 if (l_pSegmentVoisin->SIGNAL_INDEX == l_uNoSignal)
			   {
				//l_pSegmentVoisin->TAG = 1;
				dumpSignal (i_pParam, i_pGrid, io_pPhFig, x, i, z + 1);
				// ajout d'un via
				addViaToPhFig (i_pParam, x, i, z, io_pPhFig);
			   }
		  }
	   }
    }

// ajoute le segment ainsi parcouru avec comme nom le numro du signal
  l_pSegment->SIGNAL_INDEX = l_uNoSignal;
  addSegmentToPhFig (i_pParam, l_pSegment, io_pPhFig);
  l_pSegment->SIGNAL_INDEX = WSEGMENT_FREE;
}

char *
indexSignal (char *i_sRacine, ocrNaturalInt i_uIndex)
{
  char l_sSignal[40];

  sprintf (l_sSignal, "%s %ld", i_sRacine, i_uIndex);
  return namealloc (l_sSignal);
}

void
dumpIntExtConToPhFig (ocrRoutingDataBase * i_pDataBase,
				  ocrRoutingParameters * i_pParam,
				  phfig_list * i_pPhFig)
{
  ocrNaturalInt i;
  ocrConnector *l_pCon;

  for (i = 0; i < i_pDataBase->NB_GSIGNAL; i++)
    {
	 for (l_pCon = i_pDataBase->GSIGNAL[i]->CON_LIST;
		 l_pCon;
		 l_pCon = l_pCon->NEXT)
	   {

		if (l_pCon->INTEXT == INTERNAL)
		  {
		    if (l_pCon->CON)
			 addViaToPhFig (i_pParam,
						 l_pCon->CON->X,
						 l_pCon->CON->Y,
						 -1,		// VIA ALU1<->ALU2
						  i_pPhFig);
		  }
		else
		  {
		    if (l_pCon->CON)
			 addphcon (i_pPhFig,
					 l_pCon->CON->ORIENT,
					 l_pCon->NAME,
					 l_pCon->CON->X * 500,
					 l_pCon->CON->Y * 500,
					 ALU2,
					 2 * SCALE_X);
		  }
	   }
    }
}

void
dumpAlimToPhFig (ocrRoutingDataBase * i_pDataBase,
			  phfig_list * i_pPhFig)
{
  ocrNaturalInt l_uIndex = 0;
  ocrNaturalInt i;

  puts ("Dumping Alim to PhFig ...");
  // Routage des alimentations ...
  // ajout des connecteurs vdd vss

  addphseg (i_pPhFig, ALU1, 600,
		  i_pDataBase->XAB1, i_pDataBase->YAB2 - 300,
		  i_pDataBase->XAB2, i_pDataBase->YAB2 - 300, NULL);

  addphseg (i_pPhFig, ALU1, 600,
		  i_pDataBase->XAB1, 300,
		  i_pDataBase->XAB2, 300, NULL);


  addphcon (i_pPhFig, WEST,
		  indexSignal ("vss", l_uIndex),
		  i_pDataBase->XAB1,
		  300,
		  ALU1, 600);

  addphcon (i_pPhFig, EAST,
		  indexSignal ("vss", l_uIndex),
		  i_pDataBase->XAB2,
		  300,
		  ALU1, 600);

  l_uIndex++;


  for (i = 2 * PITCH * 10 * SCALE_X; i < i_pDataBase->YAB2; i += 2 * PITCH * 10 * SCALE_X)
    {

	 addphseg (i_pPhFig, ALU1, 1200,
			 i_pDataBase->XAB1, i,
			 i_pDataBase->XAB2, i, namealloc ("vss"));

	 addphcon (i_pPhFig, WEST,
			 indexSignal ("vss", l_uIndex),
			 i_pDataBase->XAB1,
			 i - 300,
			 ALU1, 600);

	 addphcon (i_pPhFig, WEST,
			 indexSignal ("vss", l_uIndex),
			 i_pDataBase->XAB1,
			 i + 300,
			 ALU1, 600);

	 addphcon (i_pPhFig, EAST,
			 indexSignal ("vss", l_uIndex),
			 i_pDataBase->XAB2,
			 i - 300,
			 ALU1, 600);

	 addphcon (i_pPhFig, EAST,
			 indexSignal ("vss", l_uIndex),
			 i_pDataBase->XAB2,
			 i + 300,
			 ALU1, 600);

	 l_uIndex++;

    }

  for (i = PITCH * 10 * SCALE_X; i < i_pDataBase->YAB2; i += 2 * PITCH * 10 * SCALE_X)
    {

	 addphseg (i_pPhFig, ALU1, 1200,
			 i_pDataBase->XAB1, i,
			 i_pDataBase->XAB2, i, namealloc ("vdd"));

	 addphcon (i_pPhFig, WEST,
			 indexSignal ("vdd", l_uIndex),
			 i_pDataBase->XAB1,
			 i - 300,
			 ALU1, 600);

	 addphcon (i_pPhFig, WEST,
			 indexSignal ("vdd", l_uIndex),
			 i_pDataBase->XAB1,
			 i + 300,
			 ALU1, 600);

	 addphcon (i_pPhFig, EAST,
			 indexSignal ("vdd", l_uIndex),
			 i_pDataBase->XAB2,
			 i - 300,
			 ALU1, 600);

	 addphcon (i_pPhFig, EAST,
			 indexSignal ("vdd", l_uIndex),
			 i_pDataBase->XAB2,
			 i + 300,
			 ALU1, 600);


	 l_uIndex++;
    }

  addphcon (i_pPhFig, WEST,
		  indexSignal ("vss", l_uIndex),
		  i_pDataBase->XAB1,
		  i_pDataBase->YAB2 - 300,
		  ALU1, 600);

  addphcon (i_pPhFig, EAST,
		  indexSignal ("vss", l_uIndex),
		  i_pDataBase->XAB2,
		  i_pDataBase->YAB2 - 300,
		  ALU1, 600);
}

void
dumpPhFigToDisk (char *i_sName,
			  phfig_list * i_pPhFig)
{
  char l_sName[50];

  strcpy (l_sName, i_sName);
  strcat (l_sName, "_ocr");
  i_pPhFig->NAME = namealloc (l_sName);
  savephfig (i_pPhFig);
//	 viewphfig(i_pPhFig);

  printf ("Dumping PhFig to %s.ap ...\n", l_sName);
}

ocrNaturalInt
dumpGridToPhFig (ocrRoutingParameters * i_pParam,
			  ocrWRoutingGrid * i_pGrid,
			  phfig_list * i_pPhFig)
{
  ocrNaturalInt x, y, z;
  ocrNaturalInt n = 0;
  ocrWSegment *l_pSegment;

  puts ("dumping Grid in phfig ...");

  // parcours de la grille  la recherche des segments ...
  for (x = 0; x < i_pGrid->SIZE_H; x++)
    for (y = 0; y < i_pGrid->SIZE_V; y++)
	 for (z = 0; z < i_pGrid->NB_OF_LAYERS; z++)
	   {
		l_pSegment = getWSegment (i_pGrid, x, y, z);

		if (l_pSegment && (l_pSegment->SIGNAL_INDEX != WSEGMENT_FREE)
		    && (l_pSegment->SIGNAL_INDEX != WSEGMENT_OBSTACLE))
		  {
		    // parcours le signal pour le stocker dans i_pPhFig
		    dumpSignal (i_pParam, i_pGrid, i_pPhFig, x, y, z);
		    // incrmentation
		    n++;
		  }
	   }

  printf ("%ld signals\n", n);

  return 0;
}
