/*
   ### -------------------------------------------------- ### 
   $Author: hcl $
   $Date: 2001/09/26 14:27:56 $
   $Log: ocrSignalUtil.c,v $
   Revision 1.1  2001/09/26 14:27:56  hcl
   premier commit....


   ### -------------------------------------------------- ###
 */

#include <stdlib.h>
#include MUT_H
#include MLO_H
#include MLU_H
#include MPH_H
#include MPU_H
#include "ocr.h"
#include "ocrWRoutingDataBase.h"
#include "ocrWindow.h"
#include "ocrPath.h"
#include "display.h"

#include "ocrSignalUtil.h"
#include "ocrConnectorUtil.h"

static char *res_id = "$Id: ocrSignalUtil.c,v 1.1 2001/09/26 14:27:56 hcl Exp $";

extern ocrOption *g_pOption;

#define LEVEL (g_pOption->LEVEL)

/**
 * lie tous les connecteur d'un signal
**/
void
linkSignal (ocrRoutingDataBase * i_pDataBase, ocrSignal * i_pSignal)
{
  ocrConnector *l_pCon;

  for (l_pCon = i_pSignal->CON_LIST; l_pCon; l_pCon = l_pCon->NEXT)
    {
      linkConnector (i_pDataBase, i_pSignal->INDEX, l_pCon);
    }
}

/**
 * dtermine le nombre de signaux  router
**/
ocrNaturalInt getNumSignal (lofig_list * i_pLoFig)
{
  losig_list *l_pLoSig;
  ocrNaturalInt l_nNbSig = 0;

  // dtermine le nombre de signaux  router
  for (l_pLoSig = i_pLoFig->LOSIG; l_pLoSig; l_pLoSig = l_pLoSig->NEXT)
    l_nNbSig++;

  return l_nNbSig;
}

/**
 * Initialisation des signaux  router : tous sauf VDD et VSS
 * cration des signaux dans la bd
**/
void
initGlobalSignalList (ocrRoutingDataBase * i_pDataBase, 
					 lofig_list * i_pLoFig)
{
  losig_list *l_pLoSig;

  createGlobalSignalList (i_pDataBase, getNumSignal (i_pLoFig));

  for (l_pLoSig = i_pLoFig->LOSIG; l_pLoSig; l_pLoSig = l_pLoSig->NEXT)
    {
      if ((l_pLoSig->NAMECHAIN->DATA != namealloc ("vss")) &&
          (l_pLoSig->NAMECHAIN->DATA != namealloc ("vdd")))
        {
          ocrSignal *l_pSignal;

          l_pSignal = createSignal (l_pLoSig);
          addSignalGlobal (i_pDataBase, l_pSignal);
        }
    }
}

/*
 * FONCTIONS D'ORDONNANCEMENT DE SIGNAUX
 */

int
funcCon (const void *a, const void *b)
{
  return (*(ocrSignal **) b)->NB_CON - (*(ocrSignal **) a)->NB_CON;
}

int
funcRandom (const void *a, const void *b)
{
  return (((double) random () / RAND_MAX) - 0.5) < 0;
}

int
funcPriority (const void *a, const void *b)
{
  if ((*(ocrSignal **) b)->PRIORITY == (*(ocrSignal **) a)->PRIORITY)
    return 0;
  else
    return (*(ocrSignal **) b)->PRIORITY - (*(ocrSignal **) a)->PRIORITY;
}

int
funcPriorityRandom (const void *a, const void *b)
{
  if ((*(ocrSignal **) b)->PRIORITY == (*(ocrSignal **) a)->PRIORITY)
    return (((double) random () / RAND_MAX) - 0.5) < 0;
  else
    return (*(ocrSignal **) b)->PRIORITY - (*(ocrSignal **) a)->PRIORITY;
}

int
funcPriorityCon (const void *a, const void *b)
{
  if ((*(ocrSignal **) b)->PRIORITY == (*(ocrSignal **) a)->PRIORITY)
    return (*(ocrSignal **) b)->NB_CON - (*(ocrSignal **) a)->NB_CON;
  else
    return (*(ocrSignal **) b)->PRIORITY - (*(ocrSignal **) a)->PRIORITY;
}

/**
 * Dtermine si un signal possde tous sauf un de ses connecteurs
 * de pris. (Urgent  router)
 * recherche  partir du rang i_uRang.
 *
 * retourne le numro du signal
**/
ocrNaturalInt
isCriticalSignal (ocrRoutingDataBase * i_pDataBase, ocrNaturalInt i_uRang)
{
  ocrNaturalInt i;
  ocrConnector *l_pCon;

  for (i = i_uRang; i < i_pDataBase->NB_SIGNAL - 1; i++)
    {
      if (i_pDataBase->SIGNAL[i]->ROUTED == 0)
        for (l_pCon = i_pDataBase->SIGNAL[i]->CON_LIST;
             l_pCon; l_pCon = l_pCon->NEXT)
          {
            if (l_pCon->INTEXT == INTERNAL)
              if (isCriticalConnector (i_pDataBase->GRID, l_pCon))
                return i;
          }
    }
  return OCRNATURALINT_MAX;
}

void
orderingSignals (ocrSignal ** i_pSignal,
                 ocrNaturalInt i_uNbSig, ocrNaturalShort i_uType)
{
  switch (i_uType)
    {
    case ORDER_RANDOM:
      {
        display (LEVEL, VERBOSE, "%s\n", "Ordering : Random ...");
        qsort (i_pSignal, i_uNbSig, sizeof (ocrSignal *), funcRandom);
        break;
      }
    case ORDER_CON:
      {
        display (LEVEL, VERBOSE, "%s\n", "Ordering : Connectors ...");
        qsort (i_pSignal, i_uNbSig, sizeof (ocrSignal *), funcCon);
        break;
      }
    case ORDER_PRIORITY:
      {
        display (LEVEL, VERBOSE, "%s\n", "Ordering : Priority ...");
        qsort (i_pSignal, i_uNbSig, sizeof (ocrSignal *), funcPriority);
        break;
      }
    case ORDER_PRIORITY_RANDOM:
      {
        display (LEVEL, VERBOSE, "%s\n", "Ordering : Priority Random ...");
        qsort (i_pSignal, i_uNbSig, sizeof (ocrSignal *), funcPriorityRandom);
        break;
      }

    case ORDER_PRIORITY_CON:
      {
        display (LEVEL, VERBOSE, "%s\n",
                 "Ordering : Priority, Connectors ...");
        qsort (i_pSignal, i_uNbSig, sizeof (ocrSignal *), funcPriorityCon);
        break;
      }
    default:
      {
        puts ("ERROR : orderingSignals");
        exit (1);
      }
    }
}
