
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: cox109.h						*/
/* date		: May 23 1994						*/
/* version	: v109							*/
/* authors	: DUNOYER Julien					*/
/* description	: header for COX library				*/
/*									*/
/* ###--------------------------------------------------------------### */

#ifndef COX_COXDEF
#define COX_COXDEF

	/* ###------------------------------------------------------### */
	/*    defines							*/
	/* ###------------------------------------------------------### */

#define COX_OUT_SUFFIX    "cox"
#define COX_VAL_GENERATE  1
#define COX_VAL_PROPAGATE 2
#define COX_VAL_EXTRACTED 3

#define COX_FANFIELD      2

#define COX_NOT_EVALUATED 0
#define COX_EVALUATED     1

#define COX_LEVELIZED_EVALUATION  0x01
#define COX_PRIMINPUT_EVALUATION  0x02
#define COX_INDEPVAR_EVALUATION   0x04

#define COX_DEFAULT_EVALUATION    COX_INDEPVAR_EVALUATION

#define COX_NO_HEURISTIC          0x00
#define COX_DECREM_HEURISTIC      0x01
#define COX_DIVIDE_HEURISTIC      0x02
#define COX_BACKPACK_HEURISTIC    0x04
#define COX_PEDRAM_HEURISTIC      0x10
#define COX_SETH_HEURISTIC        0x20
#define COX_ROETHIG_HEURISTIC     0x40
#define COX_DUNOYER_HEURISTIC     0x80

#define COX_DEFAULT_HEURISTIC     COX_DECREM_HEURISTIC

#define COX_FIRST_LEVEL   0x00
#define COX_INF_LEVEL     0x01
#define COX_SUP_LEVEL     0x02

#define COX_TYP_MEMORY    0x00
#define COX_SAVE_MEMORY   0x01
#define COX_USE_MEMORY    0x02

#define COX_BASICABL_DEF 81
#define COX_SUPERABL_DEF 82
#define COX_OPERABL_DEF  84
#define COX_TERMABL_DEF  85


	/* ###------------------------------------------------------### */
	/*    global variables definition                               */
	/* ###------------------------------------------------------### */

extern unsigned int       COX_MAXDEPTH;
extern unsigned int       COX_MINDEPTH;

extern char               COX_EVALUATION;
extern char               COX_HEURISTIC;
extern char               COX_LEVEL_MODE;
extern char               COX_MEMORY_MODE;

extern struct coxval     *COX_INIT_VALUE;

extern char               COX_BACKPACK_NAME[256];

extern struct dct_entry **COX_DICTIONNARY;

	/* ###------------------------------------------------------### */
	/*    data structures						*/
	/* ###------------------------------------------------------### */

typedef struct coxval                   /* flag and BDD                 */
  {
  struct coxcol          *PROPAGATE ;    /* propagated color set         */
  struct coxcol          *GENERATE  ;    /* generated color set          */
  struct coxcol          *EXTRACTED ;    /* extracted color set          */
  float                   PROBABACK ;    /* propagation evaluated coeff  */
  char                    EVALUATED ;    /* flag for simulation handling */
  };

typedef struct coxseq			/* sequence of events		*/
  {
  struct coxseq  	*NEXT    ;	/* next sequence		*/
  char 			*NAME    ;	/* sequence's name		*/
  struct coxevt		*COXEVT  ;	/* list of events		*/
  char			 ERRFLG  ;	/* error flag			*/
  }
coxseq_list;

typedef struct coxevt			/* event			*/
  {
  struct coxevt   	*NEXT    ;	/* next event			*/
  char           	*NAME    ;	/* signal's name		*/
  struct coxval 	*VALUE   ;	/* event's value		*/
  void			*SIGNAL  ;	/* pointer on the signal	*/
  char			 TYPE    ;	/* SIG pointer type 		*/
  }
coxevt_list;


	/* ###------------------------------------------------------### */
	/*    functions							*/
	/* ###------------------------------------------------------### */

		/* ###--------------------------------------### */
		/*    cox_error					*/
		/* ###--------------------------------------### */

extern float              cox_ProbaError ();

		/* ###--------------------------------------### */
		/*    cox_util					*/
		/* ###--------------------------------------### */

extern char              *cox_avers ();
extern void               cox_warning ();
extern void               cox_toolbug ();
extern void               cox_message ();
extern int                cox_error ();

		/* ###--------------------------------------### */
		/*    cox_package				*/
		/* ###--------------------------------------### */

extern void              *cox_addvalue ();
extern void               cox_freevalue ();
extern int                cox_cmpvalue ();
extern int                cox_cpyvalue ();
extern int                cox_updvalue ();
extern int                cox_evalvalue ();
extern int                cox_resolvalue ();

		/* ###--------------------------------------### */
		/*    cox_sequences				*/
		/* ###--------------------------------------### */

extern struct gssseq     *cox_loadsequence  ();
extern void               cox_storesequence ();
extern void               cox_writesequence ();
extern void               cox_statistics    ();

		/* ###--------------------------------------### */
		/*    cox high level functions			*/
		/* ###--------------------------------------### */

extern struct dct_entry **cox_simulator ();

#endif
