
/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: cox_colorset.c					*/
/* date		: Dec 20 1994						*/
/* version	: v100							*/
/* authors	: DUNOYER Julien                 	 		*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <math.h>
#include MUT_H
#include CST_H
#include COX_H
#include "cox_colorset.h"

/* ###--------------------------------------------------------------### */
/* function	: coc_ChooseLevel					*/
/* description	: Choose a level between 2 (the max or the min)		*/
/* ###--------------------------------------------------------------### */

unsigned int   coc_ChooseLevel (level1, level2)

unsigned int   level1  ;
unsigned int   level2  ;

  {
  if (COX_LEVEL_MODE & COX_SUP_LEVEL)
    {
    if (level1 > level2)
      {
      return (level1);
      }
    else
      {
      return (level2);
      }
    }
  if (COX_LEVEL_MODE & COX_INF_LEVEL)
    {
    if (level1 < level2)
      {
      return (level1);
      }
    else
      {
      return (level2);
      }
    }

  return (level1);
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_GetMaxLevelInSet					*/
/* description	: Get the maximumlevel element in a set.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct coxcol *coc_GetMaxLevelInSet (set)

struct coxcol    *set     ;
  {
  struct coxcol    *set_pnt = set ;


  while (set != NULL)
    {
    if ( set->LEVEL > set_pnt->LEVEL )
      {
      set_pnt = set;
      }
    set = set->NEXT;
    }
  return (set_pnt);
  }


/* ###--------------------------------------------------------------### */
/* function	: coc_CreateFundamentalSet				*/
/* description	: Create a fundamental color set (without harmonics)	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct coxcol *coc_CreateFundamentalSet (element, level)

void          *element   ;
unsigned int   level         ;

  {
  struct coxcol  *set = NULL;

  set = (struct coxcol *) mbkalloc ( sizeof (struct coxcol) );
  set->FUNDAMENTAL = element;
  set->LEVEL       = level;
  set->NEXT        = NULL;

  return (set);
  }


/* ###--------------------------------------------------------------### */
/* function	: coc_CreateIntersectionSet				*/
/* description	: Create The Intersection between two pointers sets.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct coxcol  *coc_CreateIntersectionSet (set1, set2)

struct coxcol    *set1    ;
struct coxcol    *set2    ;

  {
  struct coxcol    *set_pnt = NULL ;

  while ((set1 != NULL) && (set2 != NULL))
    {
    if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) == 0 )
      {
      set_pnt = coc_AddSetElement(set_pnt,
                                  set1->FUNDAMENTAL,
                                  coc_ChooseLevel(set1->LEVEL, set2->LEVEL));

      set1 = set1->NEXT;
      set2 = set2->NEXT;
      }
    else
      {
      if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) < 0 )
        {
        set1 = set1->NEXT;
        }
      else
        {
        set2 = set2->NEXT;
        }
      }
    }
  return (set_pnt);
  }


/* ###--------------------------------------------------------------### */
/* function	: coc_UpdateSetLevels					*/
/* description	: Update a set with the levels of another set.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void coc_UpdateSetLevels (set1, set2)

struct coxcol    *set1     ;
struct coxcol    *set2     ;
  {
  struct coxcol    *set_pnt = NULL     ;

  while ((set1 != NULL) && (set2 != NULL))
    {
    if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) == 0 )
      {
      set1->LEVEL = set2->LEVEL;

      set1 = set1->NEXT;
      set2 = set2->NEXT;
      }
    else
      {
      if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) < 0 )
        {
        set1 = set1->NEXT;
        }
      else
        {
/*         fprintf(stderr,"ERREUR in coc_UpdateSetLevels\n"); */
        set2 = set2->NEXT;
        }
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_KeepPartialSet					*/
/* description	: Update a set to keep only the elements of another set	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void coc_KeepPartialSet (set1, set2)

struct coxcol    *set1     ;
struct coxcol    *set2     ;
  {
  struct coxcol    *set_pnt = NULL     ;

  while ((set1 != NULL) && (set2 != NULL))
    {
    if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) == 0 )
      {
      set1 = set1->NEXT;
      set2 = set2->NEXT;
      }
    else
      {
      if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) < 0 )
        {
        set1 = set1->NEXT;
        }
      else
        {
        set1->LEVEL = 0;
        set2 = set2->NEXT;
        }
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_CreateUnionSet					*/
/* description	: Add The Union between two pointers sets.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

/* Algo vraiment nul car on se retape toute la liste a chaque coup */

struct coxcol  *coc_CreateUnionSet (set1, set2)

struct coxcol    *set1     ;
struct coxcol    *set2     ;

  {
  struct coxcol    *set_pnt = NULL;

  while ((set1 != NULL) && (set2 != NULL))
    {
    if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) == 0 )
      {
      set_pnt = coc_AddSetElement(set_pnt,
                                  set1->FUNDAMENTAL,
                                  coc_ChooseLevel(set1->LEVEL, set2->LEVEL));

      set1 = set1->NEXT;
      set2 = set2->NEXT;
      }
    else
      {
      if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) < 0 )
        {
        set_pnt = coc_AddSetElement(set_pnt, set1->FUNDAMENTAL, set1->LEVEL);
        set1 = set1->NEXT;
        }
      else
        {
        set_pnt = coc_AddSetElement(set_pnt, set2->FUNDAMENTAL, set2->LEVEL);
        set2 = set2->NEXT;
        }
      }
    }

  while (set1 != NULL)
    {
    set_pnt = coc_AddSetElement(set_pnt, set1->FUNDAMENTAL, set1->LEVEL);
    set1 = set1->NEXT;
    }

  while (set2 != NULL)
    {
    set_pnt = coc_AddSetElement(set_pnt, set2->FUNDAMENTAL, set2->LEVEL);
    set2 = set2->NEXT;
    }

  return (set_pnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_CopySet						*/
/* description	: Dupplicate A complete Set.				*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct coxcol  *coc_CopySet (set)

struct coxcol    *set     ;

  {
  struct coxcol  *set_pnt  = NULL;
  struct coxcol  *sav_pnt  = NULL;

  while (set != NULL)
    {
    set_pnt  = coc_CreateFundamentalSet ( set->FUNDAMENTAL, set->LEVEL);

    set_pnt->NEXT  = sav_pnt;
    sav_pnt        = set_pnt;

    set = set->NEXT;
    }

  set_pnt = (struct coxcol *) reverse ((struct chain *)set_pnt);

  return (set_pnt);
  }


/* ###--------------------------------------------------------------### */
/* function	: coc_GetInSetElement					*/
/* description	: Get a pointer element in a pointer set.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct coxcol    *coc_GetInSetElement (set,element)

struct coxcol    *set     ;
void             *element ;

  {
  while ( set != NULL )
    {
    if ( COC_ELEMENT_COMPARE(element, set->FUNDAMENTAL) == 0 )
      {
      return (set);
      }
    set = set->NEXT;
    }

  return ( set );
  }


/* ###--------------------------------------------------------------### */
/* function	: coc_AddSetElement					*/
/* description	: Add a pointer element in a pointer set.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct coxcol  *coc_AddSetElement (set, element, level)

struct coxcol *set ;
void          *element ;
unsigned int   level   ;

  {
  struct coxcol  *set_pnt = set;
  struct coxcol  *sav_pnt = NULL;

  if (set == NULL)
    {
    set = coc_CreateFundamentalSet ( element, level);
    set->NEXT = NULL ;
    return (set);
    }

  while (set_pnt != NULL)
    {
    if ( COC_ELEMENT_COMPARE(element,set_pnt->FUNDAMENTAL) == 0 )
      {
      set_pnt->LEVEL = coc_ChooseLevel(level, set_pnt->LEVEL);
      return(set);
      }

    if ( COC_ELEMENT_COMPARE(element,set_pnt->FUNDAMENTAL) < 0 )
      {
      if (sav_pnt != NULL)
        {
        sav_pnt->NEXT = coc_CreateFundamentalSet ( element, level);
        sav_pnt->NEXT->NEXT = set_pnt ;
        }
      else
        {
        set= coc_CreateFundamentalSet ( element, level);
        set->NEXT = set_pnt ;
        }
      return (set);
      }

    sav_pnt = set_pnt;
    set_pnt = set_pnt->NEXT;
    }

  sav_pnt->NEXT = coc_CreateFundamentalSet ( element, level);
  sav_pnt->NEXT->NEXT = NULL ;
  return (set);
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_FreeSet						*/
/* description	: Free a complete set (fundamental & harmonics).	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void coc_FreeSet (set)

struct coxcol    *set     ;

  {
  struct coxcol  *sav_pnt = NULL;

  while ( set != NULL )
    {
    sav_pnt = set;
    set = set->NEXT;
    mbkfree (sav_pnt);
    }
  }


/* ###--------------------------------------------------------------### */
/* function	: coc_SetCompare					*/
/* description	: compare two sets.					*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

int  coc_SetCompare (set1, set2)

struct coxcol    *set1     ;
struct coxcol    *set2     ;

  {
  while ((set1 != NULL) && ( set2 != NULL))
    {
    if ( COC_ELEMENT_COMPARE(set1->FUNDAMENTAL, set2->FUNDAMENTAL) != 0 )
      return(1);
    else
      if (set1->LEVEL != set2->LEVEL)
        return(1);

    set1 = set1->NEXT;
    set2 = set2->NEXT;
    }

  if ((set1 != NULL) || (set2 != NULL))
      return(1);

  return (0);
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_Set2Cst						*/
/* description	: transform a color set in a chain_list like set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct chain *coc_Set2Cst (set)

struct coxcol    *set     ;

  {
  struct chain   *chn_pnt = NULL;

  while ( set != NULL )
    {
    chn_pnt = cst_AddSetElement (chn_pnt, set->FUNDAMENTAL);
    set = set->NEXT;
    }

  return (chn_pnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_Levelize						*/
/* description	: Decrementing all the colors levels.			*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

struct coxcol    *coc_Levelize (set)

struct coxcol    *set     ;

  {
  struct coxcol  *hed_pnt = set;
  struct coxcol  *set_pnt = set;
  struct coxcol  *sav_pnt = NULL;

  while ( set_pnt != NULL )
    {
    if ( set_pnt->LEVEL <= COX_MINDEPTH)
      {
      if ( sav_pnt != NULL )
        {
        sav_pnt->NEXT = set_pnt->NEXT;
        set_pnt->NEXT = NULL;
        coc_FreeSet (set_pnt);
        set_pnt = sav_pnt->NEXT;
        }
      else
        {
        hed_pnt = set_pnt->NEXT;
        set_pnt->NEXT = NULL;
        coc_FreeSet (set_pnt);
        set_pnt = hed_pnt;
        }
      }
    else
      {
      sav_pnt = set_pnt;
      set_pnt = set_pnt->NEXT;
      }
    }

  return (hed_pnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_DecremPropagate					*/
/* description	: Decrementing the colors levels of a propagated set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void coc_DecremPropagate (set_pnt, param)

struct coxcol    *set_pnt ;
unsigned int      param   ;

  {
  while ( set_pnt != NULL )
    {
    if ( set_pnt->LEVEL > COX_MINDEPTH)
      {
      set_pnt->LEVEL = set_pnt->LEVEL - param;
      }
    set_pnt = set_pnt->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_DividePropagate					*/
/* description	: Decrementing the colors levels of a propagated set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void coc_DividePropagate (set_pnt, param)

struct coxcol    *set_pnt ;
unsigned int      param   ;

  {
  while ( set_pnt != NULL )
    {
    if ( set_pnt->LEVEL > COX_MINDEPTH)
      {
      set_pnt->LEVEL = (unsigned int) (set_pnt->LEVEL / param);
      }
    set_pnt = set_pnt->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_BackPropagate					*/
/* description	: Decrementing the colors levels of a propagated set.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void coc_BackPropagate (set_pnt, param)

struct coxcol    *set_pnt ;
float             param   ;

  {
  while ( set_pnt != NULL )
    {
    if ( set_pnt->LEVEL > COX_MINDEPTH)
      {
      set_pnt->LEVEL = (unsigned int) fabs(set_pnt->LEVEL * param);
      }
    set_pnt = set_pnt->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: coc_BackExtracted					*/
/* description	: Decrementing the colors levels of a color.		*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void coc_BackExtracted (set_pnt, param, level1, level2)

struct coxcol    *set_pnt ;
float             param   ;
unsigned int      level1  ;
unsigned int      level2  ;

  {
/* printf("BackExtracted : lvl1 = %d, lvl2 = %d, fparam = %-4.4f\n",level1, level2,param); */
  set_pnt->LEVEL =(unsigned int) fabs(param * level1 * level2 / COX_MAXDEPTH);
  }
