
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: cox_sequences.c					*/
/* date		: Jun  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER julien					*/
/* description	: This file contains functions relative to the 		*/
/*		  interface with sequences :				*/
/*		  cox_writesequence , cox_loadsequence			*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#include MUT_H
#include MLO_H
#include LOG_H
#include "beh.h"
#include CST_H
#include GSS_H
#include COX_H
#include PRB_H
#include "cox_colorset.h"

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */

/* ###--------------------------------------------------------------### */
/* function	: cox_countlist						*/
/* description	: count the list of the signal				*/
/* called func.	: cox_toolbug, cox_error				*/
/* ###--------------------------------------------------------------### */

static int cox_countlist (list)

struct chain      *list ;

  {
  int count = 0; 

  while(list)
    {
    count ++;
    list = list->NEXT;
    }

  return (count);
  }


/* ###--------------------------------------------------------------### */
/* function	: cox_statistics					*/
/* description	: display statistics on correlations of the circuit	*/
/* called func.	: cox_toolbug, cox_error				*/
/* ###--------------------------------------------------------------### */

void cox_statistics (dict, ptlofig, head_befig)

struct dct_entry **dict;
struct lofig      *ptlofig ;
struct befig      *head_befig ;
  {
  struct locon      *ptlocon ;
  struct losig      *ptlosig ;
  struct loins      *ptloins ;
  struct befig      *ptbefig ;
  struct bemsg      *ptbemsg ;
  struct beaux      *ptbeaux ;
  struct bebux      *ptbebux ;
  struct bereg      *ptbereg ;
  char              *name ;
  char              *insname ;
  struct chain      *liste_generee ;
  struct chain      *liste_propagee ;
  struct chain      *liste_extraite ;
  int                tot_sig_nbr = 0;
  int                gen_sig_nbr = 0;
  int                pro_sig_nbr = 0;
  int                ext_sig_nbr = 0;
  int                tot_gen_nbr = 0;
  int                tot_pro_nbr = 0;
  int                tot_ext_nbr = 0;
 

  if ((ptlofig != NULL) && (head_befig != NULL))
    {
    ptlosig  = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      name = getsigname(ptlosig);

      liste_generee  = (struct chain *)
                       gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD);
      liste_propagee = (struct chain *)
                       gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD);
      liste_extraite = (struct chain *)
                       gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD);

      if (liste_generee)  gen_sig_nbr++;
      if (liste_propagee) pro_sig_nbr++;
      if (liste_extraite) ext_sig_nbr++;

      tot_gen_nbr += cox_countlist(liste_generee);
      tot_pro_nbr += cox_countlist(liste_propagee);
      tot_ext_nbr += cox_countlist(liste_extraite);

      tot_sig_nbr ++;
      ptlosig = ptlosig->NEXT;
      }
 
    ptloins = ptlofig->LOINS;
    while ( ptloins != NULL)
      {
      insname = ptloins->INSNAME;

      ptbefig = head_befig;

      while ((ptbefig != NULL) && (ptbefig->NAME != ptloins->FIGNAME))
        ptbefig = ptbefig->NEXT;

      ptbemsg  = ptbefig->BEMSG;
      while (ptbemsg != NULL)
        {
        name = concatname (insname, ptbemsg->LABEL);

        liste_generee  = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD);
        liste_propagee = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD);
        liste_extraite = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD);

        if (liste_generee)  gen_sig_nbr++;
        if (liste_propagee) pro_sig_nbr++;
        if (liste_extraite) ext_sig_nbr++;

        tot_gen_nbr += cox_countlist(liste_generee);
        tot_pro_nbr += cox_countlist(liste_propagee);
        tot_ext_nbr += cox_countlist(liste_extraite);

        tot_sig_nbr ++;
        ptbemsg  = ptbemsg->NEXT;
        }

 
      ptbeaux  = ptbefig->BEDLY;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        liste_generee  = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD);
        liste_propagee = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD);
        liste_extraite = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD);

        if (liste_generee)  gen_sig_nbr++;
        if (liste_propagee) pro_sig_nbr++;
        if (liste_extraite) ext_sig_nbr++;

        tot_gen_nbr += cox_countlist(liste_generee);
        tot_pro_nbr += cox_countlist(liste_propagee);
        tot_ext_nbr += cox_countlist(liste_extraite);

        tot_sig_nbr ++;
        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbeaux  = ptbefig->BEAUX;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        liste_generee  = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD);
        liste_propagee = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD);
        liste_extraite = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD);

        if (liste_generee)  gen_sig_nbr++;
        if (liste_propagee) pro_sig_nbr++;
        if (liste_extraite) ext_sig_nbr++;

        tot_gen_nbr += cox_countlist(liste_generee);
        tot_pro_nbr += cox_countlist(liste_propagee);
        tot_ext_nbr += cox_countlist(liste_extraite);

        tot_sig_nbr ++;
        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbebux  = ptbefig->BEBUX;
      while (ptbebux != NULL)
        {
        name = concatname (insname, ptbebux->NAME);

        liste_generee  = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD);
        liste_propagee = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD);
        liste_extraite = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD);

        if (liste_generee)  gen_sig_nbr++;
        if (liste_propagee) pro_sig_nbr++;
        if (liste_extraite) ext_sig_nbr++;

        tot_gen_nbr += cox_countlist(liste_generee);
        tot_pro_nbr += cox_countlist(liste_propagee);
        tot_ext_nbr += cox_countlist(liste_extraite);

        tot_sig_nbr ++;
        ptbebux  = ptbebux->NEXT;
        }
 
      ptbereg  = ptbefig->BEREG;
      while (ptbereg != NULL)
        {
        name = concatname (insname, ptbereg->NAME);

        liste_generee  = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD);
        liste_propagee = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD);
        liste_extraite = (struct chain*)
                     gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD);

        if (liste_generee)  gen_sig_nbr++;
        if (liste_propagee) pro_sig_nbr++;
        if (liste_extraite) ext_sig_nbr++;

        tot_gen_nbr += cox_countlist(liste_generee);
        tot_pro_nbr += cox_countlist(liste_propagee);
        tot_ext_nbr += cox_countlist(liste_extraite);

        tot_sig_nbr ++;
        ptbereg  = ptbereg->NEXT;
        }
      ptloins = ptloins->NEXT;
      }
    }

  printf("\n            total signal number = %d\n", tot_sig_nbr);
  printf("\n");

  printf(  "       generating signal number = %d\n", gen_sig_nbr);
  printf(  "ratio  generating signal        = %f\n",
         (float)((float)gen_sig_nbr / (float)tot_sig_nbr));
  printf("\n");

  printf(  "       propaging  signal number = %d\n", pro_sig_nbr);
  printf(  "ratio  propaging  signal        = %f\n",
         (float)((float)pro_sig_nbr / (float)tot_sig_nbr));
  printf("\n");

  printf(  "       extracting signal number = %d\n", ext_sig_nbr);
  printf(  "ratio  extracting signal        = %f\n",
         (float)((float)ext_sig_nbr / (float)tot_sig_nbr));
  printf("\n");

  printf(  "average propagation             = %f\n",
         (float) ((float)tot_pro_nbr / (float)pro_sig_nbr));
  printf("\n");
  printf(  "average extraction              = %f\n",
         (float) ((float)tot_ext_nbr / (float)ext_sig_nbr));
  printf("\n");
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_storesequence					*/
/* description	: store the result of the simulation cycle		*/
/* called func.	: cox_toolbug, cox_error				*/
/* ###--------------------------------------------------------------### */

void cox_storesequence (dict, ptlofig, head_befig)

struct dct_entry **dict;
struct lofig      *ptlofig ;
struct befig      *head_befig ;
  {
  struct losig      *ptlosig ;
  struct loins      *ptloins ;
  struct befig      *ptbefig ;
  char              *insname ;
  struct coxval     *value ;
  struct bemsg      *ptbemsg ;
  struct beaux      *ptbeaux ;
  struct bebux      *ptbebux ;
  struct bereg      *ptbereg ;
  char              *name ;
 

  if (ptlofig != NULL)
    {
    ptlosig  = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      name = getsigname(ptlosig);

      value = gss_getcurval(name);
  
      gss_addtab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD,
                 (struct chain*) coc_Set2Cst (value->PROPAGATE));

      gss_addtab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD,
                 (struct chain*) coc_Set2Cst (value->GENERATE ));

      gss_addtab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD,
                 (struct chain*) coc_Set2Cst (value->EXTRACTED));

      ptlosig = ptlosig->NEXT;
      }
 
    ptloins = ptlofig->LOINS;
    while ( ptloins != NULL)
      {
      insname = ptloins->INSNAME;

      ptbefig = head_befig;

      while ((ptbefig != NULL) && (ptbefig->NAME != ptloins->FIGNAME))
        ptbefig = ptbefig->NEXT;

      ptbemsg  = ptbefig->BEMSG;
      while (ptbemsg != NULL)
        {
        name = concatname (insname, ptbemsg->LABEL);

        value = gss_getcurval(name);
  
        gss_addtab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->PROPAGATE));

        gss_addtab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->GENERATE ));

        gss_addtab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->EXTRACTED));

        ptbemsg  = ptbemsg->NEXT;
        }

 
      ptbeaux  = ptbefig->BEDLY;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        value = gss_getcurval(name);

        gss_addtab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->PROPAGATE));

        gss_addtab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->GENERATE ));

        gss_addtab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->EXTRACTED));

        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbeaux  = ptbefig->BEAUX;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        value = gss_getcurval(name);

        gss_addtab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->PROPAGATE));

        gss_addtab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->GENERATE ));

        gss_addtab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->EXTRACTED));

        ptbeaux  = ptbeaux->NEXT;
        }

      ptbebux  = ptbefig->BEBUX;
      while (ptbebux != NULL)
        {
        name = concatname (insname, ptbebux->NAME);

        value = gss_getcurval(name);

        gss_addtab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->PROPAGATE));

        gss_addtab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->GENERATE ));

        gss_addtab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->EXTRACTED));

        ptbebux  = ptbebux->NEXT;
        }

      ptbereg  = ptbefig->BEREG;
      while (ptbereg != NULL)
        {
        name = concatname (insname, ptbereg->NAME);

        value = gss_getcurval(name);

        gss_addtab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->PROPAGATE));

        gss_addtab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->GENERATE ));

        gss_addtab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD,
                   (struct chain*) coc_Set2Cst (value->EXTRACTED));

        ptbereg  = ptbereg->NEXT;
        }
      ptloins = ptloins->NEXT;
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_writelist						*/
/* description	: write the list of the signal				*/
/* called func.	: cox_toolbug, cox_error				*/
/* ###--------------------------------------------------------------### */

static void cox_writelist (fp, list)

FILE              *fp   ;
struct chain      *list ;

  {

  while(list)
    {
    (void)fprintf(fp,"\t\t%s ,\n", list->DATA);
    list = list->NEXT;
    }
  (void)fprintf(fp,"\n");

  }


/* ###--------------------------------------------------------------### */
/* function	: cox_writesequence					*/
/* description	: write the result of the simulation cycle		*/
/* called func.	: cox_toolbug, cox_error				*/
/* ###--------------------------------------------------------------### */

void cox_writesequence (dict, ptlofig, head_befig, filename)

struct dct_entry **dict;
struct lofig      *ptlofig ;
struct befig      *head_befig ;
char              *filename;

{
FILE              *fp;
FILE              *fpplus;
time_t             clock;
struct locon      *ptlocon ;
struct losig      *ptlosig ;
struct loins      *ptloins ;
struct befig      *ptbefig ;
struct bemsg      *ptbemsg ;
struct beaux      *ptbeaux ;
struct bebux      *ptbebux ;
struct bereg      *ptbereg ;
char               printname[80] ;
char              *insname ;
char              *name;

fp = mbkfopen (filename, COX_OUT_SUFFIX, WRITE_TEXT);
fpplus = mbkfopen (filename, "co2", WRITE_TEXT);

if (fp == NULL)
  cox_error (107, NULL);

if (fpplus == NULL)
  cox_error (107, NULL);

else
  {
  time (&clock);

  if ((ptlofig != NULL) && (head_befig != NULL))
    {

    (void)fprintf (fp,"\t\tCox file generated from %s\n", ptlofig->NAME);
    (void)fprintf (fp,"\t\tdate : %s \n \n", ctime(&clock));

        /* ###------------------------------------------------------### */
        /*   For each signal write the values				*/
        /* ###------------------------------------------------------### */
 
    ptlocon  = ptlofig->LOCON;
    while (ptlocon != NULL)
      {
      ptlosig = ptlocon->SIG;

      name = getsigname (ptlosig);

      if (vectorindex(name) == -1)
        sprintf (printname, "%s", vectorradical(name));
      else
        sprintf (printname, "%s_%d", vectorradical(name), vectorindex(name));

      (void)fprintf(fp, "connector   %-15s :\n", printname);
      (void)fprintf(fpplus, "connector   %-15s :  ", printname);

      (void)fprintf(fp, "\t liste generee = \n" );
      cox_writelist
           (fp, gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD));

      (void)fprintf(fp, "\t liste propagee = \n" );
      cox_writelist
           (fp, gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD));

      (void)fprintf(fp, "\t liste extraite = \n" );
      cox_writelist
           (fp, gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD));
      if (gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD) != NULL)
        {
        (void)fprintf(fpplus, "\t liste extraite pleine \n" );
        }
      else
        {
        (void)fprintf(fpplus, "\t liste extraite vide \n" );
        }
      ptlocon  = ptlocon->NEXT;
      }

    ptlosig  = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      if (ptlosig->TYPE == 'I')
        {
        name = getsigname (ptlosig);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s_%d", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "signal   %-15s :\n", printname);
        (void)fprintf(fpplus, "signal   %-15s :  ", printname);

        (void)fprintf(fp, "\t liste generee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste propagee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste extraite = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD));
        if (gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD) != NULL)
          {
          (void)fprintf(fpplus, "\t liste extraite pleine \n" );
          }
        else
          {
          (void)fprintf(fpplus, "\t liste extraite vide \n" );
          }
        }
      ptlosig  = ptlosig->NEXT;
      }

    ptloins = ptlofig->LOINS;
    while ( ptloins != NULL)
      {
      insname = ptloins->INSNAME;

      ptbefig = head_befig;

      while ((ptbefig != NULL) && (ptbefig->NAME != ptloins->FIGNAME))
        ptbefig = ptbefig->NEXT;

      ptbemsg  = ptbefig->BEMSG;
      while (ptbemsg != NULL)
        {
        name = concatname(insname, ptbemsg->LABEL);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s_%d", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "message   %-15s :\n", printname);

        (void)fprintf(fp, "\t liste generee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste propagee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste extraite = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD));

        ptbemsg  = ptbemsg->NEXT;
        }

 
      ptbeaux  = ptbefig->BEDLY;
      while (ptbeaux != NULL)
        {
        name = concatname(insname, ptbeaux->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s_%d", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "delayed   %-15s :\n", printname);

        (void)fprintf(fp, "\t liste generee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste propagee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste extraite = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD));

        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbeaux  = ptbefig->BEAUX;
      while (ptbeaux != NULL)
        {
        name = concatname(insname, ptbeaux->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s_%d", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "auxiliary   %-15s :\n", printname);
        (void)fprintf(fpplus, "auxiliary   %-15s :  ", printname);

        (void)fprintf(fp, "\t liste generee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste propagee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste extraite = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD));
        if (gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD) != NULL)
          {
          (void)fprintf(fpplus, "\t liste extraite pleine \n" );
          }
        else
          {
          (void)fprintf(fpplus, "\t liste extraite vide \n" );
          }

        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbebux  = ptbefig->BEBUX;
      while (ptbebux != NULL)
        {
        name = concatname(insname, ptbebux->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s_%d", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "internal bus   %-15s :\n", printname);

        (void)fprintf(fp, "\t liste generee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste propagee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste extraite = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD));

        ptbebux  = ptbebux->NEXT;
        }
 
      ptbereg  = ptbefig->BEREG;
      while (ptbereg != NULL)
        {
        name = concatname(insname, ptbereg->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s_%d", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "registre   %-15s :\n", printname);

        (void)fprintf(fp, "\t liste generee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_GENERATE , GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste propagee = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_PROPAGATE, GSS_PNT0FIELD));

        (void)fprintf(fp, "\t liste extraite = \n" );
        cox_writelist
             (fp, gss_chktab(dict, name, COX_VAL_EXTRACTED, GSS_PNT0FIELD));

        ptbereg  = ptbereg->NEXT;
        }
      ptloins = ptloins->NEXT;
      }
    }
  fclose (fp);
  fclose (fpplus);
  }
}


/* ###--------------------------------------------------------------### */
/* function	: cox_getprobaback					*/
/* description	: return a float number : a backanotated probability	*/
/* called funct	: none							*/
/* ###--------------------------------------------------------------### */

static float cox_getprobaback (name)

char              *name;

  {
  struct prbseq *pt_prbseq  = NULL;
  struct prbevt *pt_prbevt  = NULL;

  if (!(COX_HEURISTIC & COX_BACKPACK_HEURISTIC))
    return (0.5);

  pt_prbseq = prb_getprbseq (COX_BACKPACK_NAME);

  pt_prbevt = pt_prbseq->PRBEVT;

  while ((pt_prbevt != NULL) && (pt_prbevt->NAME != name))
    pt_prbevt = pt_prbevt->NEXT;

  if (pt_prbevt != NULL)
    {
/* printf("Proba %-6.4f pour %s\n", pt_prbevt->VALUE->PROBVAL, name); */
    return (pt_prbevt->VALUE->PROBVAL);
    }
  else
    {
/* printf("Pas de Proba pour %s\n", name); */
    return (0.5);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_getfanout						*/
/* description	: return a number greater than 1 if the BERIN has	*/
/*		  multiple fan out (excluding assert statements)	*/
/* called funct	: none							*/
/* ###--------------------------------------------------------------### */

static unsigned int cox_getfanout (ptberin)

struct berin *ptberin;

  {
  unsigned int  fan_nbr = 0;
  struct chain *reference;

  if (ptberin != NULL)
    {
    reference = ptberin->OUT_REF;

    while (reference != NULL)
      {
      fan_nbr ++;
      reference = reference->NEXT;
      }

    reference = ptberin->BUS_REF;

    while (reference != NULL)
      {
      fan_nbr ++;
      reference = reference->NEXT;
      }

    reference = ptberin->AUX_REF;

    while (reference != NULL)
      {
      fan_nbr ++;
      reference = reference->NEXT;
      }

    reference = ptberin->BUX_REF;

    while (reference != NULL)
      {
      fan_nbr ++;
      reference = reference->NEXT;
      }

    reference = ptberin->REG_REF;

    while (reference != NULL)
      {
      fan_nbr ++;
      reference = reference->NEXT;
      }

    reference = ptberin->MSG_REF;

    while (reference != NULL)
      {
      fan_nbr ++;
      reference = reference->NEXT;
      }

    reference = ptberin->DLY_REF;

    while (reference != NULL)
      {
      fan_nbr ++;
      reference = reference->NEXT;
      }
    }

  return (fan_nbr);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_loadsequence					*/
/* description	: load a sequence with circuit fanout infos		*/
/*		  add it at the top of a sequence list			*/
/* called func.	: cox_error						*/
/* ###--------------------------------------------------------------### */

struct gssseq *cox_loadsequence (ptlofig, head_befig)

struct lofig      *ptlofig;
struct befig      *head_befig;
  {
  struct dct_entry **lcl_dic;
  unsigned int       fan_nbr = 0;
  struct locon      *ptlocon ;
  struct losig      *ptlosig ;
  struct loins      *ptloins ;
  struct befig      *ptbefig ;
  struct berin      *ptberin ;
  struct bereg      *ptbereg ;
  struct bebux      *ptbebux ;
  struct gssseq     *ptseq ;
  struct gssevt     *ptevt ;
  struct coxval     *value ;
  char              *name;

  ptseq = gss_addgssseq(NULL, NULL);
  ptevt = ptseq->EVENT;

COX_INIT_VALUE = (struct coxval *)cox_addvalue ();

if ((ptlofig != NULL) && (head_befig != NULL))
  {
  if (COX_EVALUATION & COX_PRIMINPUT_EVALUATION)
    {

        /* ###------------------------------------------------------### */
        /*    For each LOINS ... :                                      */
        /* ###------------------------------------------------------### */

    ptloins = ptlofig->LOINS;
    while (ptloins != NULL)
      {
        /* ###------------------------------------------------------### */
        /*    Scan the list of BEFIGs to find the model of the instance */
        /* ###------------------------------------------------------### */

      ptbefig = head_befig;
      while ((ptbefig != NULL) && (ptbefig->NAME != ptloins->FIGNAME))
        ptbefig = ptbefig->NEXT;

        /* ###------------------------------------------------------### */
        /*    For each BERIN ... :                                      */
        /* ###------------------------------------------------------### */

      ptberin = ptbefig->BERIN;
      while (ptberin != NULL)
        {
        ptbereg = ptbefig->BEREG;
        while ((ptbereg != NULL) && (ptbereg->NAME != ptberin->NAME))
          ptbereg = ptbereg->NEXT;

        if (ptbereg != NULL)
          {
          name = concatname (ptloins->INSNAME, ptberin->NAME);
          value = cox_addvalue();
          value->GENERATE  = coc_CreateFundamentalSet(name, COX_MAXDEPTH);
          value->EVALUATED = COX_EVALUATED;

          ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
          }

        ptbebux = ptbefig->BEBUX;
        while ((ptbebux != NULL) && (ptbebux->NAME != ptberin->NAME))
          ptbebux = ptbebux->NEXT;

        if (ptbebux != NULL)
          {
          name = concatname (ptloins->INSNAME, ptberin->NAME);
          value = cox_addvalue();
          value->GENERATE  = coc_CreateFundamentalSet(name, COX_MAXDEPTH);
          value->EVALUATED = COX_EVALUATED;

          ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
          }

        ptberin = ptberin->NEXT;
        }

      ptloins = ptloins->NEXT;
      }

        /* ###------------------------------------------------------### */
        /*    For each LOSIG BUS (pas pour l'instant)... :              */
        /*     or each INPUT LOCON                   :                  */
        /* ###------------------------------------------------------### */

    ptlosig = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      if (ptlosig->TYPE == EXTERNAL)
        {
        ptlocon = ptlofig->LOCON;
        while (ptlocon->SIG != ptlosig)
          {
          ptlocon = ptlocon->NEXT;
          }

          if ((ptlocon->DIRECTION == IN) ||
              (ptlocon->DIRECTION == INOUT) ||
              (ptlocon->DIRECTION == TRANSCV))
          {
          name = getsigname(ptlosig);
          value = cox_addvalue();
          value->GENERATE  = coc_CreateFundamentalSet(name, COX_MAXDEPTH);
          value->EVALUATED = COX_EVALUATED;
          }
        ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
        }
      ptlosig = ptlosig->NEXT;
      }

    ptseq->EVENT = ptevt;
    ptseq->ERRFLG = 0;
    gss_fretab (lcl_dic);
    }

  else if (COX_EVALUATION & COX_INDEPVAR_EVALUATION)
    {
    lcl_dic = gss_initab ();

        /* ###------------------------------------------------------### */
        /*    Fill the dictionnary with external & internal signals     */
        /* ###------------------------------------------------------### */

    ptlosig = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      gss_addtab (lcl_dic,ptlosig,NULL,COX_FANFIELD,0);
      ptlosig = ptlosig->NEXT;
      }

        /* ###------------------------------------------------------### */
        /*    For each LOINS ... :                                      */
        /* ###------------------------------------------------------### */

    ptloins = ptlofig->LOINS;
    while (ptloins != NULL)
      {
        /* ###------------------------------------------------------### */
        /*    Scan the list of BEFIGs to find the model of the instance */
        /* ###------------------------------------------------------### */

      ptbefig = head_befig;
      while ((ptbefig != NULL) && (ptbefig->NAME != ptloins->FIGNAME))
        ptbefig = ptbefig->NEXT;

        /* ###------------------------------------------------------### */
        /*    For each BERIN ... :                                      */
        /* ###------------------------------------------------------### */

      ptberin = ptbefig->BERIN;
      while (ptberin != NULL)
        {

        ptlocon = ptloins->LOCON;
        while ((ptlocon != NULL) && (ptlocon->NAME != ptberin->NAME))
          ptlocon = ptlocon->NEXT;

        if (ptlocon == NULL)
          {
          fan_nbr = cox_getfanout(ptberin);

          if ( fan_nbr > 1)
            {
            name = concatname (ptloins->INSNAME, ptberin->NAME);
            value = cox_addvalue();
            value->GENERATE  = coc_CreateFundamentalSet(name, COX_MAXDEPTH);

            value->PROBABACK = cox_getprobaback(name);

            if (COX_HEURISTIC & COX_BACKPACK_HEURISTIC)
              {
              value->GENERATE->LEVEL *=
                      sqrt(fabs(value->PROBABACK * value->PROBABACK
                                - value->PROBABACK));
              }

            ptbereg = ptbefig->BEREG;
            while ((ptbereg != NULL) && (ptbereg->NAME != ptberin->NAME))
              ptbereg = ptbereg->NEXT;

            if (ptbereg != NULL)
              {
              value->EVALUATED = COX_EVALUATED;
              }

            ptbebux = ptbefig->BEBUX;
            while ((ptbebux != NULL) && (ptbebux->NAME != ptberin->NAME))
              ptbebux = ptbebux->NEXT;

            if (ptbebux != NULL)
              {
              value->EVALUATED = COX_EVALUATED;
              }
            ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
            }

          else 
            {
            ptbereg = ptbefig->BEREG;
            while ((ptbereg != NULL) && (ptbereg->NAME != ptberin->NAME))
              ptbereg = ptbereg->NEXT;

            if (ptbereg != NULL)
              {
              name = concatname (ptloins->INSNAME, ptberin->NAME);
              value = cox_addvalue();
              value->GENERATE  = NULL;
              value->EVALUATED = COX_EVALUATED;

              value->PROBABACK = cox_getprobaback(name);

              ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
              }

            ptbebux = ptbefig->BEBUX;
            while ((ptbebux != NULL) && (ptbebux->NAME != ptberin->NAME))
              ptbebux = ptbebux->NEXT;

            if (ptbebux != NULL)
              {
              name = concatname (ptloins->INSNAME, ptberin->NAME);
              value = cox_addvalue();
              value->GENERATE  = NULL;
              value->EVALUATED = COX_EVALUATED;

              value->PROBABACK = cox_getprobaback(name);

              ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
              }
            }
          }

        else 
          {
          fan_nbr = gss_chktab (lcl_dic,ptlocon->SIG,NULL,COX_FANFIELD);

          if (fan_nbr <= 1)
            { 
            fan_nbr += cox_getfanout(ptberin);
            gss_addtab (lcl_dic, ptlocon->SIG, NULL, COX_FANFIELD, fan_nbr);
            }
          }
        ptberin = ptberin->NEXT;
        }

      ptloins = ptloins->NEXT;
      }

        /* ###------------------------------------------------------### */
        /*    For each LOSIG WITH MULTIPLE FANOUT... :                  */
        /*     or each INPUT LOCON                   :                  */
        /* ###------------------------------------------------------### */

    ptlosig = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      fan_nbr = gss_chktab (lcl_dic,ptlosig,NULL,COX_FANFIELD);
      if ( fan_nbr > 1)
        {
        name = getsigname(ptlosig);
        value = cox_addvalue();
        value->GENERATE  = coc_CreateFundamentalSet(name, COX_MAXDEPTH);

        value->PROBABACK = cox_getprobaback(name);

        if (COX_HEURISTIC & COX_BACKPACK_HEURISTIC)
          {
          value->GENERATE->LEVEL *=
                  sqrt(fabs(value->PROBABACK * value->PROBABACK
                            - value->PROBABACK));
          }

        if (ptlosig->TYPE == EXTERNAL)
          {
          ptlocon = ptlofig->LOCON;
          while (ptlocon->SIG != ptlosig)
            {
            ptlocon = ptlocon->NEXT;
            }

          if ((ptlocon->DIRECTION == IN) ||
              (ptlocon->DIRECTION == INOUT) ||
              (ptlocon->DIRECTION == TRANSCV))
            {
            value->EVALUATED = COX_EVALUATED;
            }
          }

        ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
        }
      else
        {
        if (ptlosig->TYPE == EXTERNAL)
          {
          ptlocon = ptlofig->LOCON;
          while (ptlocon->SIG != ptlosig)
            {
            ptlocon = ptlocon->NEXT;
            }

          if ((ptlocon->DIRECTION == IN) ||
              (ptlocon->DIRECTION == INOUT) ||
              (ptlocon->DIRECTION == TRANSCV))
            {
            name = getsigname(ptlosig);
            value = cox_addvalue();
            value->GENERATE  = NULL;
            value->EVALUATED = COX_EVALUATED;

            value->PROBABACK = cox_getprobaback(name);

            ptevt = gss_addgssevt(ptevt, name, GSS_UNKDEF, value);
            }
          }
        }
      ptlosig = ptlosig->NEXT;
      }

    ptseq->EVENT = ptevt;
    ptseq->ERRFLG = 0;
    gss_fretab (lcl_dic);
    }
  }
return (ptseq);
}
