
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: util.c						*/
/* date		: Feb 13 1993						*/
/* authors	: Julien DUNOYER					*/
/* version	: v1.0							*/
/* description	: This file contains some utility functions :		*/
/*		  cox_error, cox_toolbug, cox_message,			*/
/*                cox_warning, cox_avers				*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H

/* ###--------------------------------------------------------------### */
/* function	: cox_avers						*/
/* description	: return the current version				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char *cox_avers ()
  {
  return ("v1.00");
  }


/* ###--------------------------------------------------------------### */
/* function	: cox_warning						*/
/* description	: print a warning message corresponding to the argument	*/
/*		  code							*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void cox_warning (code,str1)

int   code;
char *str1;

  {
  fprintf (stderr,"Warning %d : ",code);
  switch (code)
    {
    case 1:
      fprintf (stderr,"cannot find register `%s`\n",str1);
      break;
    case 2:
      fprintf (stderr,"consistency checks will be disabled\n");
      break;
    case 3:
      fprintf (stderr,"signal `%s` never used\n",str1);
      break;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_toolbug						*/
/* description	: print a fatal error message corresponding to the	*/
/*		  argument code and exit				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void cox_toolbug (code,str1,str2,nbr1)

int   code;
char *str1;
char *str2;
int   nbr1;

  {
  (void) fprintf (stderr,"Fatal error %d executing `%s`: ", code,str1);
  switch (code)
    {
    case 1:
      (void) fprintf (stderr,"unknown operator\n");
      break;
    case 2:
      (void) fprintf (stderr,"cannot create empty atom\n");
      break;
    case 5:
      (void) fprintf (stderr,"impossible call : back_dependency list empty\n");
      break;
    case 7:
      (void) fprintf (stderr,"unknown type `%c` for IO `%s`\n",nbr1,str2);
      break;
    case 8:
      (void) fprintf (stderr,"illegal signal value : hex `%x`\n",nbr1);
      break;
    case 13 :
      (void) fprintf (stderr,"cannot find model of `%s`\n",str2);
      break;
    case 18:
      (void) fprintf (stderr,"unknown mode `%c` for port `%s`\n",nbr1,str2);
      break;
    case 22:
      (void) fprintf (stderr,"`%c` unknown resolution function\n",nbr1);
      break;
    case 23:
      (void) fprintf (stderr,"in instance %s : `%c` unknown signal type\n",str2,nbr1);
      break;
    }
  exit (1);
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_message						*/
/* description	: print a message corresponding to the argument code	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void cox_message (code, str1, nmb1)

int   code;
char *str1;
int   nmb1;

  {
  switch (code)
    {
    case 1:
      (void) fprintf (stdout,"\nLinking Description Structures...\n");
      break;
    case 2:
      (void) fprintf (stdout,"\nCreating Stimulis Sequence.......\n");
      break;
    case 3:
      (void) fprintf (stdout,"\nLinking Stimulis & Structures....\n");
      break;
    case 4:
      (void) fprintf (stdout,"\nProcessing Circuit Stability.....\n");
      break;
    case 5:
      (void) fprintf (stdout,"\nProcessing Symbolic Simulation...\n",nmb1);
      break;
    case 6:
      (void) fprintf (stdout,"\nCreating the result file ........\n");
      break;
    default:
      (void) fprintf (stderr,"cox_message : code %d unknown.\n",code);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: cox_error						*/
/* description	: print an error message depending on the argument code	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int cox_error (code,str1,dat,line)

int           code;
char         *str1;
float         dat ;
unsigned int  line;

  {
  if (code < 200)
    (void)fprintf (stderr,"Error %d :",code);

  if ((code > 200) && (code < 300))
    (void)fprintf (stderr,"Prb File Parse Error %d, line %d :", code, line);

  switch (code)
    {


    case 1:
      (void) fprintf (stderr, "combinatory loop : `%s`\n", str1);
      break;
    case 3:
      (void) fprintf (stderr, "cannot find terminal `%s`\n", str1);
      break;
    case 6:
      (void) fprintf (stderr, "cannot make abl of empty expression\n");
      break;
    case 7:
      (void) fprintf (stderr, "call on empty list\n");
      break;
    case 8:
      (void) fprintf (stderr, "missing extraction lists\n");
      break;
    case 9:
      (void) fprintf (stderr, "missing propagation lists\n");
      break;
    case 10:
      (void) fprintf (stderr, "missing propagation table\n");
      break;
    case 11:
      (void) fprintf (stderr, "missing extraction table\n");
      break;
    case 12:
      (void) fprintf (stderr, "missing propagation tree\n");
      break;
    case 13:
      (void) fprintf (stderr, "missing extraction list\n");
      break;


    case 100:
      (void) fprintf (stderr,"cannot find `%s`\n",str1);
      break;
    case 102:
      (void) fprintf (stderr,"instance and model of `%s` mismatch\n",str1);
      break;
    case 103:
      (void) fprintf (stderr,"incompatible input/output mode for `%s`\n",str1);
      break;
    case 104:
      (void) fprintf (stderr,"input/output `%s` doesn't exist\n",str1);
      break;
    case 105:
      (void) fprintf (stderr,"Possible combinatory loop. Simulation aborted\n");
      break;
    case 106:
      (void) fprintf (stderr,"Too many errors. Simulation interrupted\n");
      break;
    case 107:
      (void) fprintf (stderr,"Cannot open result file\n");
      break;
    case 109:
      (void) fprintf (stderr,"signal `%s` never assigned\n",str1);
      break;
    case 110:
      (void) fprintf (stderr,"signal `%s` not connected to any port\n",str1);
      break;
    case 111:
      (void) fprintf (stderr,"illegal connections on signal `%s`\n",str1);
      break;
    case 112:
      (void) fprintf (stderr,"multiple driver on guarded signal `%s`\n",str1);
      break;
    case 113:
      (void) fprintf (stderr,"expected value and simulation's result mismatch on `%s`\n",str1);
      break;
    case 114:
      (void) fprintf (stderr,"multiple driver on unguarded signal `%s`\n",str1);
      break;
    case 115:
      (void) fprintf (stderr,"save file doesn't match with the description\n");
      break;
    case 116:
      (void) fprintf (stderr,"cannot open file `%s`\n", str1);
      break;

    case 201:
      fprintf (stderr,"illegal value for a correlation :%f\n",dat);
      break;
    case 202:
      fprintf (stderr,"already declared :%s\n",str1);
      break;
    case 203:
      fprintf (stderr,"vector with bad direction :%s\n",str1);
      break;
    case 204:
      fprintf (stderr,"%s\n", str1);
      break;
    case 205:
      fprintf (stderr, "cannot open file `%s`\n", str1);
      break;
    case 206:
      fprintf (stderr,"erroneous file connot go further\n");
      break;
    case 207:
      fprintf (stderr,"\n");
      break;
    case 208:
      fprintf (stderr,"\n");
      break;
    case 209:
      fprintf (stderr,"syntax error\n");
      break;

    case 300:
      (void) fprintf (stderr,"\n	cannot continue further more.\n");
      (void) fprintf (stderr,"\n		Have a nice day...\n");
      break;

    default :
      fprintf (stderr,"unknown error\n");
      break;
    }

  return (1);
  }
