
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: gss109.h						*/
/* date		: May 07 1995						*/
/* version	: v109							*/
/* author	: DUNOYER Julien					*/
/* contents	: defines and structure definitions used in GSS library	*/
/*									*/
/* ###--------------------------------------------------------------### */

#ifndef GSS_GSSDEF
#define GSS_GSSDEF

	/* ###------------------------------------------------------### */
	/*    defines							*/
	/* ###------------------------------------------------------### */

#define GSS_ZERO_OPTION       (unsigned char)0x00 /* empty mode flag	*/
#define GSS_SCHEDULER_OPTION  (unsigned char)0x01 /* scheduler mode	*/
#define GSS_DEBUG_OPTION      (unsigned char)0x02 /* debug mode		*/
#define GSS_LEVELIZED_OPTION  (unsigned char)0x04 /* levelized mode	*/

#define GSS_ZERO_DELAY      (unsigned char)0x00	/*      zero delay flag	*/
#define GSS_INERTIAL_DELAY  (unsigned char)0x01	/*  inertial delay flag	*/
#define GSS_TRANSPORT_DELAY (unsigned char)0x02	/* transport delay flag	*/
#define GSS_CUSTOM_DELAY    (unsigned char)0x03	/*    custom delay flag	*/

#define GSS_ZERO_HISTORY                   0x00	/* all events in histor	*/
#define GSS_LASTEVENT_HISTORY              0x01	/* last event in histor	*/
#define GSS_COMPLETE_HISTORY               0x02	/* all events in histor	*/

#define GSS_NOEVENTYET      (unsigned long)0x00	/*      zero delay flag	*/

#define GSS_EMPTY   (unsigned char)0x00	/* empty flag			*/

#define GSS_SIGDEF  (unsigned char)0x01	/* external signal type net	*/
#define GSS_BUSDEF  (unsigned char)0x02	/* external bus    type net	*/
#define GSS_BUXDEF  (unsigned char)0x04	/* internal bus    type net	*/
#define GSS_REGDEF  (unsigned char)0x08	/* internal regist type net	*/
#define GSS_AUXDEF  (unsigned char)0x10	/* internal signal type net	*/
#define GSS_MSGDEF  (unsigned char)0x20	/* internal assert type net	*/

#define GSS_DRVDEF  (unsigned char)0x40	/* driver		    	*/
#define GSS_CNDDEF  (unsigned char)0x80	/* condition    		*/

#define GSS_UNKDEF  (unsigned char)0x00	/* unknown signal type net	*/

#define GSS_TYPMASK (unsigned char)0x3F	/* type mask			*/

#define GSS_SIGTAB     0	/* external signal table field	*/
#define GSS_BUSTAB     1	/* external bus    table field	*/
#define GSS_BUXTAB     2	/* internal bus    table field	*/
#define GSS_REGTAB     3	/* internal regist table field	*/
#define GSS_AUXTAB     4	/* internal signal table field	*/
#define GSS_MSGTAB     5	/* internal assert table field	*/

#define GSS_UNKDFN     8	/* unknown direction	*/
#define GSS_DECDFN     4	/* ??????? direction	*/
#define GSS_REDDFN     2	/* input   direction	*/
#define GSS_WRTDFN     1	/* output  direction	*/

#define GSS_SIGFIELD   0	/* dictionnary field	*/
#define GSS_DIRFIELD   0	/* dictionnary field	*/
#define GSS_MODFIELD   0	/* dictionnary field	*/
#define GSS_CNBFIELD   1	/* dictionnary field	*/
#define GSS_RNBFIELD   2	/* dictionnary field	*/
#define GSS_WNBFIELD   3	/* dictionnary field	*/
#define GSS_PNT0FIELD  5	/* dictionnary field	*/
#define GSS_PNT1FIELD  6	/* dictionnary field	*/

	/* ###------------------------------------------------------### */
	/*    global variables definition                               */
	/* ###------------------------------------------------------### */

extern int                GSS_DELTACYCLE;
extern int                GSS_MAXDELTA;
extern int                GSS_OPTION_MODE;
extern int                GSS_DELAY_MODE;
extern int                GSS_HISTORY_MODE;
extern unsigned long      GSS_CURRENT_TIME;
extern unsigned long      GSS_NEXT_TIME;
extern struct shdul      *GSS_SCHEDULER;
extern struct dct_entry **GSS_DICTIONNARY;
extern unsigned short     GSS_SCG_MODE;

	/* ###------------------------------------------------------### */
	/*    structure definitions					*/
	/* ###------------------------------------------------------### */

struct dct_entry
  {
  struct dct_entry *next;
  struct dct_recrd *data;
  char             *key;
  };

struct dct_recrd
  {
  struct dct_recrd *next;
  char             *key;
  short             fd0_val;
  short             fd1_val;
  short             fd2_val;
  short             fd3_val;
  short             fd4_val;
  int               pnt0_val;
  int               pnt1_val;
  };


typedef struct funpack			/* simul user functions	package	*/
  {
  void *	(*ADDVALUE)();		/* gssfun's name		*/
  void		(*FREEVALUE)();		/* gssfun's name		*/
  int		(*CMPVALUE)();		/* gssfun's name		*/
  int		(*CPYVALUE)();		/* gssfun's name		*/
  int		(*UPDVALUE)();		/* gssfun's name		*/
  int		(*EVALVALUE)();		/* gssfun's name		*/
  int		(*RESOLVALUE)();	/* gssfun's name		*/
  unsigned long	(*EVALDELAY)();		/* gssfun's name		*/
  unsigned long	(*EVALINERT)();		/* gssfun's name		*/
  int		(*COLVALUE)();		/* gssfun's name		*/
  }
funpack;

typedef struct lkdfig			/* linked figure		*/
  {
  struct lkdfig	*NEXT;			/* next figure			*/
  char		*NAME;			/* figure's name		*/
  struct lkdins	*LKDINS;		/* list of instances		*/
  struct simsig	*SIMSIG;		/* list of simple signals	*/
  struct bussig	*BUSSIG;		/* list of bused signals	*/
  }
lkdfig_list;

typedef struct lkdins			/* linked instance		*/
  {
  struct lkdins  *NEXT;			/* next instance		*/
  char           *INSNAME;		/* instance's name		*/
  struct befig   *BEFIG;		/* instance model		*/
  struct wrssig  *WRSSIG;		/* written simple output ports	*/
  struct wrbsig  *WRBSIG;		/* written bused  output ports	*/
  struct wriaux  *WRIAUX;		/* written inter simple signals	*/
  struct wribux  *WRIBUX;		/* written inter bused  signals	*/
  struct wrireg  *WRIREG;		/* written inter registers	*/
  struct wrimsg  *WRIMSG;		/* written inter asserts	*/
  struct redlst  *REDLST;		/* read values table		*/
  struct circuit *CIRCUI;		/* Circuit pointer used for BDDs*/
  }
lkdins_list;

typedef struct simsig			/* extern simple signal		*/
  {
  struct simsig *NEXT;			/* next simple signal		*/
  struct depsig *DEPSIG;		/* list of depending signals	*/
  struct evthis *HISTORY;		/* signal's events history	*/
  void          *CURVAL;		/* signal's current value	*/
  void          *PRJVAL;		/* signal's projected value	*/
  }
simsig_list;

typedef struct bussig			/* extern bused signal		*/
  {
  struct bussig *NEXT;			/* next simple signal		*/
  struct depsig *DEPSIG;		/* list of depending signals 	*/
  struct evthis *HISTORY;		/* signal's events history	*/
  void          *CURVAL;		/* signal's current value	*/
  void          *PRJVAL;		/* signal's projected value	*/
  struct prjbvl *PRJBVL;		/* list of projected bi-values	*/
  char           TYPE;			/* signal's type		*/
  }
bussig_list;

typedef struct wrssig			/* written extern simple signal	*/
  {
  struct wrssig *NEXT;			/* next signal			*/
  struct lkdins	*LKDINS;		/* containing instance		*/
  struct beout 	*BEOUT;			/* formal output port (model)	*/
  struct simsig	*SIMSIG;		/* corresponding extern signal	*/
  char           EVAL;			/* Evaluation's flag		*/
  }
wrssig_list;

typedef struct wrbsig			/* written extern bused signal	*/
  {
  struct wrbsig	*NEXT;			/* next signal			*/
  struct lkdins	*LKDINS;		/* containing instance		*/
  struct bebus  *BEBUS;			/* formal output port (model)	*/
  struct bussig *BUSSIG;		/* corresponding extern signal	*/
  struct prjbvl *PRJBVL;		/* list of projected bi-values	*/
  char           EVAL;			/* Evaluation's flag		*/
  }
wrbsig_list;

typedef struct wriaux                   /* written auxiliary signal 	*/
  {
  struct wriaux *NEXT;			/* next auxiliary signal       	*/
  struct lkdins	*LKDINS;		/* containing instance		*/
  struct depsig	*DEPSIG;		/* list of depending signals	*/
  struct beaux  *BEAUX;			/* formal intern signal (model)	*/
  struct evthis *HISTORY;		/* aux. signal events history	*/
  void          *CURVAL;                /* aux. signal current value	*/
  void          *PRJVAL;		/* aux. signal projected value  */
  char           EVAL;			/* Evaluation's flag		*/
  }
wriaux_list;

typedef struct wrimsg                   /* written assert signal 	*/
  {
  struct wrimsg *NEXT;			/* next assert signal       	*/
  struct lkdins	*LKDINS;		/* containing instance		*/
  struct bemsg  *BEMSG;			/* formal intern assert (model)	*/
  struct evthis *HISTORY;		/* msg. signal events history	*/
  void          *CURVAL;                /* msg. signal current value	*/
  void          *PRJVAL;		/* msg. signal projected value  */
  char           EVAL;			/* Evaluation's flag		*/
  }
wrimsg_list;

typedef struct wrireg			/* written register		*/
  {
  struct wrireg	*NEXT;			/* next register		*/
  struct lkdins	*LKDINS;		/* containing instance		*/
  struct depsig	*DEPSIG;		/* list of depending signals 	*/
  struct bereg  *BEREG;			/* formal register (model)	*/
  struct evthis *HISTORY;		/* register's events history	*/
  void          *CURVAL;		/* register's current value	*/
  void          *PRJVAL;		/* register's projected value	*/
  struct prjbvl *PRJBVL;		/* list of projected bi-values	*/
  char           EVAL;			/* Evaluation's flag		*/
  }
wrireg_list;

typedef struct wribux			/* written internal bused sig.	*/
  {
  struct wribux	*NEXT;			/* next signal			*/
  struct lkdins	*LKDINS;		/* containing instance		*/
  struct depsig	*DEPSIG;		/* list of depending signals 	*/
  struct bebux  *BEBUX;			/* formal bused signal (model)	*/
  struct evthis *HISTORY;		/* bus's events history		*/
  void          *CURVAL;		/* bus's current value		*/
  void          *PRJVAL;		/* signal's projected value	*/
  struct prjbvl *PRJBVL;		/* list of projected bi-values	*/
  char           EVAL;			/* Evaluation's flag		*/
  }
wribux_list;

typedef struct redlst			/* list of read values		*/
  {
  struct redlst *NEXT;			/* next read value (unused)	*/
  void          *VALU;			/* pointer on the value		*/
  struct evthis *HIST;			/* pointer on the history	*/
  }
redlst_list;

typedef struct prjbvl			/* projected bi-value		*/
  {
  struct prjbvl *NEXT;			/* next bi-value (local)	*/
  void          *CNDVAL;		/* driver's condition value	*/
  void          *DRVVAL;		/* drove value			*/
  }
prjbvl_list;

typedef struct gssseq			/* sequence of events		*/
  {
  struct gssseq  	*NEXT    ;	/* next sequence		*/
  char 			*NAME    ;	/* sequence's name		*/
  struct gssevt		*EVENT   ;	/* list of events		*/
  unsigned long		 TIME    ;	/* absolute time (if used)	*/
  char			 ERRFLG  ;	/* error flag			*/
  }
gssseq_list;

typedef struct depsig			/* depending signals		*/
  {
  struct depsig   	*NEXT    ;	/* next depsig			*/
  void			*SIGNAL  ;	/* pointer on the signal	*/
  unsigned char		 TYPE    ;	/* SIG pointer type 		*/
  }
depsig_list;

typedef struct gssevt			/* event			*/
  {
  struct gssevt   	*NEXT    ;	/* next event			*/
  char           	*NAME    ;	/* signal's name		*/
  void          	*VALUE   ;	/* event's value		*/
  void			*SIGNAL  ;	/* pointer on the signal	*/
  unsigned char		 TYPE    ;	/* SIG pointer type 		*/
  }
gssevt_list;

typedef struct evthis			/* event history structure	*/
  {
  struct evthis		*NEXT    ;	/* next event			*/
  unsigned long		 DATE    ;	/* event's date			*/
  int			 DELTA   ;	/* event's deltacycle		*/
  unsigned long		 COUNT   ;	/* event's count		*/
  void          	*VALUE   ;	/* event's value		*/
  }
evthis_list;


	/* ###------------------------------------------------------### */
	/*    functions							*/
	/* ###------------------------------------------------------### */

extern char              *gss_avers   ();
extern struct dct_entry **gss_initab  ();
extern void               gss_addtab  ();
extern int                gss_chktab  ();
extern void               gss_fretab  ();
extern void               gss_warning ();
extern void               gss_toolbug ();
extern int                gss_error   ();

extern void               gss_debug ();

extern struct funpack    *gss_addfunpack ();
extern void               gss_frefunpack ();
extern void              *gss_addvalue   ();
extern int                gss_freevalue  ();
extern int                gss_cmpvalue   ();
extern int                gss_cpyvalue   ();
extern int                gss_updvalue   ();
extern int                gss_resolvalue ();
extern int                gss_evalvalue  ();
extern unsigned long      gss_evaldelay  ();
extern unsigned long      gss_evalinert  ();
extern int                gss_colvalue   ();

extern struct lkdfig     *gss_linker       ();
extern int                gss_restor       ();
extern void               gss_resord       ();
extern struct lofig      *gss_makgst       ();
extern void               gss_frelkdfig    ();

extern void               gss_deltaupdate     ();
extern void               gss_deltaexecute    ();
extern int                gss_deltacheck      ();
extern int                gss_deltaloop       ();
extern int                gss_circuitstable   ();
extern void               gss_circuitinit     ();
extern void               gss_forcesequence   ();
extern void               gss_forceevent      ();
extern void               gss_loadtransaction ();

extern struct gssevt     *gss_addgssevt    ();
extern void               gss_fregssevt    ();
extern struct gssseq     *gss_addgssseq    ();
extern void               gss_fregssseq    ();
extern void               gss_lnkgssseq    ();

extern void              *gss_getcurval    ();
extern void              *gss_getprjval    ();
extern void              *gss_getvalue     ();
extern void               gss_setcurval    ();
extern void               gss_setprjval    ();
extern void               gss_setvalue     ();

extern struct evthis     *gss_gethistory         ();
extern unsigned long      gss_getlasteventdate   ();
extern unsigned long      gss_getlasteventdelta  ();
extern struct evthis     *gss_addevthis          ();
extern void               gss_freevthis          ();
extern void               gss_initevthis         ();


#endif

