

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: gss_sequences.c					*/
/* date		: Jan 28 1998						*/
/* version	: v107							*/
/* authors	: DUNOYER Julien 					*/
/* content	: high-level and low-level gss functions		*/
/*									*/
/* ###--------------------------------------------------------------### */

#include MUT_H
#include GSS_H

/* ###--------------------------------------------------------------### */
/* function	: gss_addgssevt						*/
/* description	: create an EVENT structure at the top of a list	*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct gssevt *gss_addgssevt (lastgssevt, name, type, value)

struct gssevt   *lastgssevt;	/* pointer on the last gssevt structure	*/
char		*name;		/* signal's name			*/
unsigned char  	 type;		/* signal's type			*/
void         	*value;		/* event's value			*/


  {
  struct gssevt *newgssevt;

  newgssevt          = (struct gssevt *) mbkalloc (sizeof(struct gssevt));
  newgssevt->NAME    = name;
  newgssevt->VALUE   = value;
  newgssevt->SIGNAL  = NULL;
  newgssevt->TYPE    = type;
  newgssevt->NEXT    = lastgssevt;

  return (newgssevt);
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_fregssevt						*/
/* description	: release a list of EVENT (gssevt) structures		*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void gss_fregssevt (listgssevt, usr_package)

struct gssevt  *listgssevt;
struct funpack *usr_package;

  {
  struct gssevt  *ptgssevt;

        /* ###------------------------------------------------------### */
        /*    for each object of the list delete the object itself      */
        /* ###------------------------------------------------------### */

  while (listgssevt != NULL)
    {
    ptgssevt   = listgssevt;

    if ((usr_package != NULL) && ((usr_package->FREEVALUE) != NULL))
       {
       (*(usr_package->FREEVALUE)) (ptgssevt->VALUE);
       }

    listgssevt = listgssevt->NEXT;
    mbkfree (ptgssevt);
    }
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_addgssseq						*/
/* description	: create an empty SEQUENCE structure at the top of list	*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct gssseq *gss_addgssseq (lastgssseq,name)

struct gssseq	*lastgssseq;	/* pointer on the last sequence struct	*/
char 		*name;		/* name of the sequence structure 	*/

  {
  struct gssseq *ptgssseq;

  ptgssseq         = (struct gssseq *) mbkalloc (sizeof(struct gssseq));

  ptgssseq->NAME   = (char *) namealloc (name);
  ptgssseq->EVENT  = NULL;
  ptgssseq->TIME   = 0;

  ptgssseq->ERRFLG = 0;

  ptgssseq->NEXT   = lastgssseq;

  return (ptgssseq);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_fregssseq						*/
/* description	: release a list of SEQUENCE (gssseq) structures	*/
/* called func.	: mbkfree, gss_fregssevt				*/
/* ###--------------------------------------------------------------### */

void gss_fregssseq (listgssseq, usr_package)

struct gssseq  *listgssseq;
struct funpack *usr_package;

  {
  struct gssseq *ptgssseq;

        /* ###------------------------------------------------------### */
        /*    for each object of the list delete the object itself      */
        /*    and its attached lists				        */
        /* ###------------------------------------------------------### */

  while (listgssseq != NULL)
    {
    ptgssseq     = listgssseq;
    listgssseq   = listgssseq->NEXT;

    gss_fregssevt (ptgssseq->EVENT, usr_package);

    mbkfree (ptgssseq);
    }
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_lnkgssseq						*/
/* description	: link sequence structure with lkdfig			*/
/* called func.	: vhu_toolbug						*/
/* ###--------------------------------------------------------------### */

void gss_lnkgssseq (pt_seq)

struct gssseq     *pt_seq;

  {
  struct gssevt *pt_evt;
  struct gssevt *prv_evt;
  struct gssevt *tmp_evt;
  unsigned char  type   = GSS_EMPTY;
  int            errflg = 0;

        /* ###------------------------------------------------------### */
        /*   Link sequences structures with the description		*/
        /* ###------------------------------------------------------### */

  pt_evt  = pt_seq->EVENT;
  prv_evt = NULL;

  while (pt_evt != NULL)
    {
    pt_evt->SIGNAL  = (void *) gss_chktab (GSS_DICTIONNARY,pt_evt->NAME,NULL,GSS_PNT0FIELD);
    if (pt_evt->SIGNAL != NULL)
      {
      type = (unsigned char) gss_chktab (GSS_DICTIONNARY,pt_evt->NAME,NULL,GSS_SIGFIELD);
      if (pt_evt->TYPE == GSS_UNKDEF)
        pt_evt->TYPE = type;
      else if ((pt_evt->TYPE & GSS_TYPMASK) != type)
        errflg = gss_error (18, pt_evt->NAME);
      }
    else
      {
      errflg = gss_error (14,pt_evt->NAME);
      }

    if (errflg != 0)
      {
      errflg = 0;

        /* ###------------------------------------------------------### */
        /*   Si l'evenement n'est pas valide : On fait un free de l'evt	*/
        /*   Si c'est la tete de liste : on recalcule la tete de liste	*/
        /*                       Sinon : on ponte la liste		*/
        /* ###------------------------------------------------------### */
      if (prv_evt == NULL)
        {
        pt_seq->EVENT = pt_evt->NEXT;
        tmp_evt       = pt_evt->NEXT;
        mbkfree(pt_evt);
        pt_evt        = tmp_evt;
        }
      else
        {
        tmp_evt       = pt_evt->NEXT;
        prv_evt->NEXT = tmp_evt;
        mbkfree(pt_evt);
        pt_evt        = tmp_evt;
        }
      }
    else
      {
      prv_evt = pt_evt;
      pt_evt  = pt_evt->NEXT;
      }
    }
  }
