
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: gss_xspec.c						*/
/* date		: Jan 13 1998						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* description	: This file contains functions relative to the		*/
/*		  execution of a generic symbolic simulation :		*/
/*									*/
/*		  gss_circuitstable, gss_forceevent  ,gss_forcesequence	*/
/*		  gss_deltaloop , gss_deltaexecute , gss_deltaupdate	*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/time.h>

#include MUT_H
#include "beh.h"
#include LOG_H
#include GSS_H
#include SCG_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */

static chain_list *GSS_EVENT_TABLE[] = 
{
NULL, NULL, NULL, NULL, NULL, NULL
};

static chain_list *GSS_UPDATE_TABLE[] = 
{
NULL, NULL, NULL, NULL, NULL, NULL
};

static chain_list *GSS_RESOL_TABLE[] = 
{
NULL, NULL, NULL, NULL, NULL, NULL
};

static chain_list *GSS_ERROR_TABLE[] = 
{
NULL, NULL, NULL, NULL, NULL, NULL
};


/* ###--------------------------------------------------------------### */
/* function	: gss_deltaresolution					*/
/* description	: resolution of signals computed in the	last evaluation	*/
/*                phase of the current delta cycle.			*/
/* called func.	: gss_muxbit, gss_worbit, gss_regbit, addchain,		*/
/*		  freechain						*/
/* ###--------------------------------------------------------------### */

void gss_deltaresolution (usr_package)

struct funpack *usr_package;

  {
  struct chain  *cur_chain;
  struct chain  *err_chain;

  struct wrireg *ptr_wrireg;
  struct wribux *ptr_wribux;
  struct bussig *ptr_bussig;

  char           already_rcd = 0;
  char		 err_flg     = 0;


	/* ###------------------------------------------------------### */
	/*    Resol extern bused signals's value (also handles errors )	*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_RESOL_TABLE[GSS_BUSTAB];

  while (cur_chain != NULL)
    {
    ptr_bussig = (struct bussig *)cur_chain->DATA;

    err_flg = gss_resolvalue (ptr_bussig->PRJBVL,
                              ptr_bussig->TYPE,
                              ptr_bussig->CURVAL,
                              ptr_bussig->PRJVAL,
                              usr_package);

    if (err_flg != 0) 
      {

 /* find if the error has already been recorded in the error's list */

      err_chain   = GSS_ERROR_TABLE[GSS_BUSTAB];
      already_rcd = 0;

      while (err_chain != NULL)
        {
        if (((struct bussig *)err_chain->DATA) ==  ptr_bussig)
          {
          already_rcd = 1;
          break;
          }
        err_chain = err_chain->NEXT;
        }

      if (already_rcd == 0)
        {
        err_chain = cur_chain;
        cur_chain = cur_chain->NEXT;
        GSS_ERROR_TABLE[GSS_BUSTAB]  =
              addchain( GSS_ERROR_TABLE[GSS_BUSTAB], err_chain->DATA);

        }
      else
        {
        cur_chain = cur_chain->NEXT;
        }
      }
    else
      {
      err_chain = GSS_ERROR_TABLE[GSS_BUSTAB];

      while (err_chain != NULL)
        {
        if (((struct bussig *)err_chain->DATA) ==  ptr_bussig)
          {
          GSS_ERROR_TABLE[GSS_BUSTAB] =
                      delchain(GSS_ERROR_TABLE[GSS_BUSTAB], err_chain);
          break;
          }
        err_chain = err_chain->NEXT;
        }
      cur_chain = cur_chain->NEXT;
      }
    }

  GSS_UPDATE_TABLE[GSS_BUSTAB] = append (GSS_UPDATE_TABLE[GSS_BUSTAB],
                                         GSS_RESOL_TABLE[GSS_BUSTAB]);
  GSS_RESOL_TABLE[GSS_BUSTAB]  = NULL;

	/* ###------------------------------------------------------### */
	/*    Resol intern bused signals's value (also handles errors )	*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_RESOL_TABLE[GSS_BUXTAB] ;

  while (cur_chain != NULL)
    {
    ptr_wribux = (struct wribux *)cur_chain->DATA;

    err_flg = gss_resolvalue (ptr_wribux->PRJBVL,
                              ptr_wribux->BEBUX->TYPE,
                              ptr_wribux->CURVAL,
                              ptr_wribux->PRJVAL,
                              usr_package);

    if (err_flg != 0)
      {

 /* find if the error has already been recorded in the error's list */

      err_chain   = GSS_ERROR_TABLE[GSS_BUXTAB];
      already_rcd = 0;
      while (err_chain != NULL)
        {
        if (((struct wribux *)err_chain->DATA) ==  ptr_wribux)
          {
          already_rcd = 1;
          break;
          }
        err_chain = err_chain->NEXT;
        }

      if (already_rcd == 0)
        {
        err_chain = cur_chain;
        cur_chain = cur_chain->NEXT;
        GSS_ERROR_TABLE[GSS_BUXTAB]  =
              addchain( GSS_ERROR_TABLE[GSS_BUXTAB], err_chain->DATA);

        }
      else
        {
        cur_chain = cur_chain->NEXT;
        }
      }
    else
      {

      err_chain = GSS_ERROR_TABLE[GSS_BUXTAB];
      while (err_chain != NULL)
        {
        if (((struct wribux *)err_chain->DATA) ==  ptr_wribux)
          {
          GSS_ERROR_TABLE[GSS_BUXTAB] =
                      delchain(GSS_ERROR_TABLE[GSS_BUXTAB], err_chain);
          break;
          }
        err_chain = err_chain->NEXT;
        }
      cur_chain = cur_chain->NEXT;
      }
    }

  GSS_UPDATE_TABLE[GSS_BUXTAB] = append (GSS_UPDATE_TABLE[GSS_BUXTAB],
                                         GSS_RESOL_TABLE[GSS_BUXTAB]);
  GSS_RESOL_TABLE[GSS_BUXTAB]  = NULL;

	/* ###------------------------------------------------------### */
	/*    Resol register signals's value (also handles errors )	*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_RESOL_TABLE[GSS_REGTAB];
  
  while (cur_chain != NULL)
    {
    ptr_wrireg = (struct wrireg *)cur_chain->DATA;
    err_flg = gss_resolvalue (ptr_wrireg->PRJBVL,
                              'R',
                              ptr_wrireg->CURVAL,
                              ptr_wrireg->PRJVAL,
                              usr_package);

    if (err_flg != 0)
      {

 /* find if the error has already been recorded in the error's list */

      err_chain = GSS_ERROR_TABLE[GSS_REGTAB];
      already_rcd = 0;

      while (err_chain != NULL)
        {
        if (((struct wrireg *)err_chain->DATA) ==  ptr_wrireg)
          {
          already_rcd = 1;
          break;
          }
        err_chain = err_chain->NEXT;
        }

      if (already_rcd == 0)
        {
        err_chain = cur_chain;
        cur_chain = cur_chain->NEXT;
        GSS_ERROR_TABLE[GSS_REGTAB]  =
              addchain( GSS_ERROR_TABLE[GSS_REGTAB], err_chain->DATA);

        }
      else
        {
        cur_chain = cur_chain->NEXT;
        }
      }
    else
      {
      err_chain = GSS_ERROR_TABLE[GSS_REGTAB];
      while (err_chain != NULL)
        {
        if (((struct wrireg *)err_chain->DATA) ==  ptr_wrireg)
          {
          GSS_ERROR_TABLE[GSS_REGTAB] =
                      delchain(GSS_ERROR_TABLE[GSS_REGTAB], err_chain);
          break;
          }
        err_chain = err_chain->NEXT;
        }
      cur_chain = cur_chain->NEXT;
      }
    }

  GSS_UPDATE_TABLE[GSS_REGTAB] = append (GSS_UPDATE_TABLE[GSS_REGTAB],
                                         GSS_RESOL_TABLE[GSS_REGTAB]);
  GSS_RESOL_TABLE[GSS_REGTAB]  = NULL;
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_deltaevent					*/
/* description	: Used to schedule dependant events			*/
/* called funct	: 							*/
/* ###--------------------------------------------------------------### */

static void gss_deltaevent (ptr_depsig)

struct depsig  *ptr_depsig;

  {
  while (ptr_depsig != NULL)
    {
    switch (ptr_depsig->TYPE)
      {
      case GSS_SIGDEF :
        if (((struct wrssig *)ptr_depsig->SIGNAL)->EVAL == 0)
          {
          GSS_EVENT_TABLE[GSS_SIGTAB] =
              addchain(GSS_EVENT_TABLE[GSS_SIGTAB], ptr_depsig->SIGNAL);
          ((struct wrssig *)ptr_depsig->SIGNAL)->EVAL = 1;
          }
        break;

      case GSS_BUSDEF :
        if (((struct wrbsig *)ptr_depsig->SIGNAL)->EVAL == 0)
          {
          GSS_EVENT_TABLE[GSS_BUSTAB] =
              addchain(GSS_EVENT_TABLE[GSS_BUSTAB], ptr_depsig->SIGNAL);
          ((struct wrbsig *)ptr_depsig->SIGNAL)->EVAL = 1;
          }
        break;

      case GSS_BUXDEF :
        if (((struct wribux *)ptr_depsig->SIGNAL)->EVAL == 0)
          {
          GSS_EVENT_TABLE[GSS_BUXTAB] =
              addchain(GSS_EVENT_TABLE[GSS_BUXTAB], ptr_depsig->SIGNAL);
          ((struct wribux *)ptr_depsig->SIGNAL)->EVAL = 1;
          }
        break;

      case GSS_REGDEF :
        if (((struct wrireg *)ptr_depsig->SIGNAL)->EVAL == 0)
          {
          GSS_EVENT_TABLE[GSS_REGTAB] =
              addchain(GSS_EVENT_TABLE[GSS_REGTAB], ptr_depsig->SIGNAL);
          ((struct wrireg *)ptr_depsig->SIGNAL)->EVAL = 1;
          }
        break;

      case GSS_AUXDEF :
        if (((struct wriaux *)ptr_depsig->SIGNAL)->EVAL == 0)
          {
          GSS_EVENT_TABLE[GSS_AUXTAB] =
              addchain(GSS_EVENT_TABLE[GSS_AUXTAB], ptr_depsig->SIGNAL);
          ((struct wriaux *)ptr_depsig->SIGNAL)->EVAL = 1;
          }
        break;

      case GSS_MSGDEF :
        if (((struct wrimsg *)ptr_depsig->SIGNAL)->EVAL == 0)
          {
          GSS_EVENT_TABLE[GSS_MSGTAB] =
              addchain(GSS_EVENT_TABLE[GSS_MSGTAB], ptr_depsig->SIGNAL);
          ((struct wrimsg *)ptr_depsig->SIGNAL)->EVAL = 1;
          }
        break;

      default :
        gss_error (0,"gss_deltaevent");
        break;

      }

    ptr_depsig = ptr_depsig->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_deltaupdate					*/
/* description	: updates the current value of signals computed in the	*/
/*		  evaluation phase of the current delta cycle.		*/
/* called func.	: gss_cmpvalue, gss_cpyvalue, gss_deltaevent,		*/
/*		  freechain						*/
/* ###--------------------------------------------------------------### */

void gss_deltaupdate (usr_package)

struct funpack *usr_package;

  {
  struct chain  *cur_chain;
  struct chain  *err_chain;

  struct wrireg *ptr_wrireg;
  struct wriaux *ptr_wriaux;
  struct wribux *ptr_wribux;
  struct wrimsg *ptr_wrimsg;
  struct bussig *ptr_bussig;
  struct simsig *ptr_simsig;

  char		 err_flg     = 0;

	/* ###------------------------------------------------------### */
	/*    Update simple signals' value				*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_UPDATE_TABLE[GSS_SIGTAB];
  while (cur_chain != NULL)
    {
    ptr_simsig = (struct simsig *)cur_chain->DATA;

    if (gss_cmpvalue(ptr_simsig->CURVAL, ptr_simsig->PRJVAL, usr_package) != 0)
      {
      gss_updvalue(ptr_simsig->CURVAL, ptr_simsig->PRJVAL, usr_package);

      gss_addevthis (ptr_simsig->HISTORY, GSS_CURRENT_TIME, GSS_DELTACYCLE);

      gss_deltaevent(ptr_simsig->DEPSIG);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (GSS_UPDATE_TABLE[GSS_SIGTAB]);
  GSS_UPDATE_TABLE[GSS_SIGTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Update bused signals's value				*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_UPDATE_TABLE[GSS_BUSTAB];
  while (cur_chain != NULL)
    {
    ptr_bussig = (struct bussig *)cur_chain->DATA;

    if (gss_cmpvalue(ptr_bussig->CURVAL, ptr_bussig->PRJVAL, usr_package) != 0)
      {
      gss_updvalue(ptr_bussig->CURVAL, ptr_bussig->PRJVAL, usr_package);

      gss_addevthis (ptr_bussig->HISTORY, GSS_CURRENT_TIME, GSS_DELTACYCLE);

      gss_deltaevent(ptr_bussig->DEPSIG);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (GSS_UPDATE_TABLE[GSS_BUSTAB]);
  GSS_UPDATE_TABLE[GSS_BUSTAB] = NULL;


	/* ###------------------------------------------------------### */
	/*    Update internal bused signals				*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_UPDATE_TABLE[GSS_BUXTAB] ;

  while (cur_chain != NULL)
    {
    ptr_wribux = (struct wribux *)cur_chain->DATA;

    if (gss_cmpvalue(ptr_wribux->CURVAL, ptr_wribux->PRJVAL, usr_package) != 0)
      {
      gss_updvalue(ptr_wribux->CURVAL, ptr_wribux->PRJVAL, usr_package);

      gss_addevthis (ptr_wribux->HISTORY, GSS_CURRENT_TIME, GSS_DELTACYCLE);

      gss_deltaevent(ptr_wribux->DEPSIG);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (GSS_UPDATE_TABLE[GSS_BUXTAB] );
  GSS_UPDATE_TABLE[GSS_BUXTAB]  = NULL;

	/* ###------------------------------------------------------### */
	/*    Updating registers					*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_UPDATE_TABLE[GSS_REGTAB];
  while (cur_chain != NULL)
    {
    ptr_wrireg = (struct wrireg *)cur_chain->DATA;

    if (gss_cmpvalue(ptr_wrireg->CURVAL, ptr_wrireg->PRJVAL, usr_package) != 0)
      {
      gss_updvalue(ptr_wrireg->CURVAL, ptr_wrireg->PRJVAL, usr_package);

      gss_addevthis (ptr_wrireg->HISTORY, GSS_CURRENT_TIME, GSS_DELTACYCLE);

      gss_deltaevent(ptr_wrireg->DEPSIG);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain(GSS_UPDATE_TABLE[GSS_REGTAB]);
  GSS_UPDATE_TABLE[GSS_REGTAB] = NULL;


	/* ###------------------------------------------------------### */
	/*    Updating simple internal signals				*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_UPDATE_TABLE[GSS_AUXTAB];
  while (cur_chain != NULL)
    {
    ptr_wriaux = (struct wriaux *)cur_chain->DATA;

    if (gss_cmpvalue(ptr_wriaux->CURVAL, ptr_wriaux->PRJVAL, usr_package) != 0)
      {
      gss_updvalue(ptr_wriaux->CURVAL, ptr_wriaux->PRJVAL, usr_package);

      gss_addevthis (ptr_wriaux->HISTORY, GSS_CURRENT_TIME, GSS_DELTACYCLE);

      gss_deltaevent(ptr_wriaux->DEPSIG);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (GSS_UPDATE_TABLE[GSS_AUXTAB]);
  GSS_UPDATE_TABLE[GSS_AUXTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Updating simple assert signals				*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_UPDATE_TABLE[GSS_MSGTAB];
  while (cur_chain != NULL)
    {
    ptr_wrimsg = (struct wrimsg *)cur_chain->DATA;

    if (gss_cmpvalue(ptr_wrimsg->CURVAL, ptr_wrimsg->PRJVAL, usr_package) != 0)
      {
      gss_updvalue(ptr_wrimsg->CURVAL, ptr_wrimsg->PRJVAL, usr_package);

      gss_addevthis (ptr_wrimsg->HISTORY, GSS_CURRENT_TIME, GSS_DELTACYCLE);

      gss_deltaevent(NULL);
      }

    cur_chain = cur_chain->NEXT;
    }
  freechain (GSS_UPDATE_TABLE[GSS_MSGTAB]);
  GSS_UPDATE_TABLE[GSS_MSGTAB] = NULL;
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_forceevent					*/
/* description	: force an event on a signal.				*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void gss_forceevent (pt_event, usr_package)
  
struct gssevt  *pt_event;
struct funpack *usr_package;

  {
  struct chain  *pt_chain;

  switch (pt_event->TYPE)
    {
    case GSS_SIGDEF :
      gss_cpyvalue(((struct simsig *)(pt_event->SIGNAL))->PRJVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_UPDATE_TABLE[GSS_SIGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_UPDATE_TABLE[ GSS_SIGTAB ] = 
           addchain (GSS_UPDATE_TABLE[ GSS_SIGTAB ], pt_event->SIGNAL);
        }
      break;

    case GSS_BUSDEF :
      gss_cpyvalue(((struct bussig *)(pt_event->SIGNAL))->PRJVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_BUSTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_RESOL_TABLE[GSS_BUSTAB] =
                      delchain(GSS_RESOL_TABLE[GSS_BUSTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_BUSTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_UPDATE_TABLE[ GSS_BUSTAB ] = 
           addchain (GSS_UPDATE_TABLE[ GSS_BUSTAB ], pt_event->SIGNAL);
        }
      break;

    case GSS_BUSDEF | GSS_DRVDEF :
      gss_cpyvalue(((struct bussig *)(pt_event->SIGNAL))->PRJBVL->DRVVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_BUSTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_RESOL_TABLE[ GSS_BUSTAB ] = 
           addchain (GSS_RESOL_TABLE[ GSS_BUSTAB ], pt_event->SIGNAL);
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_BUSTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_UPDATE_TABLE[GSS_BUSTAB] =
                      delchain(GSS_UPDATE_TABLE[GSS_BUSTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      break;

    case GSS_BUSDEF | GSS_CNDDEF :
      gss_cpyvalue(((struct bussig *)(pt_event->SIGNAL))->PRJBVL->CNDVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_BUSTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_RESOL_TABLE[ GSS_BUSTAB ] = 
           addchain (GSS_RESOL_TABLE[ GSS_BUSTAB ], pt_event->SIGNAL);
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_BUSTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_UPDATE_TABLE[GSS_BUSTAB] =
                      delchain(GSS_UPDATE_TABLE[GSS_BUSTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      break;

    case GSS_BUXDEF :
      gss_cpyvalue(((struct wribux *)(pt_event->SIGNAL))->PRJVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_BUXTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_RESOL_TABLE[GSS_BUXTAB] =
                      delchain(GSS_RESOL_TABLE[GSS_BUXTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_BUXTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_UPDATE_TABLE[ GSS_BUXTAB ] = 
           addchain (GSS_UPDATE_TABLE[ GSS_BUXTAB ], pt_event->SIGNAL);
        }
      break;

    case GSS_BUXDEF | GSS_DRVDEF :
      gss_cpyvalue(((struct wribux *)(pt_event->SIGNAL))->PRJBVL->DRVVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_BUXTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_RESOL_TABLE[ GSS_BUXTAB ] = 
           addchain (GSS_RESOL_TABLE[ GSS_BUXTAB ], pt_event->SIGNAL);
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_BUXTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_UPDATE_TABLE[GSS_BUXTAB] =
                      delchain(GSS_UPDATE_TABLE[GSS_BUXTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      break;

    case GSS_BUXDEF | GSS_CNDDEF :
      gss_cpyvalue(((struct wribux *)(pt_event->SIGNAL))->PRJBVL->CNDVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_BUXTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_RESOL_TABLE[ GSS_BUXTAB ] = 
           addchain (GSS_RESOL_TABLE[ GSS_BUXTAB ], pt_event->SIGNAL);
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_BUXTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_UPDATE_TABLE[GSS_BUXTAB] =
                      delchain(GSS_UPDATE_TABLE[GSS_BUXTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      break;

    case GSS_REGDEF :
      gss_cpyvalue(((struct wrireg *)(pt_event->SIGNAL))->PRJVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_REGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_RESOL_TABLE[GSS_REGTAB] =
                      delchain(GSS_RESOL_TABLE[GSS_REGTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_REGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_UPDATE_TABLE[ GSS_REGTAB ] = 
           addchain (GSS_UPDATE_TABLE[ GSS_REGTAB ], pt_event->SIGNAL);
        }
      break;

    case GSS_REGDEF | GSS_DRVDEF :
      gss_cpyvalue(((struct wrireg *)(pt_event->SIGNAL))->PRJBVL->DRVVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_REGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_RESOL_TABLE[ GSS_REGTAB ] = 
           addchain (GSS_RESOL_TABLE[ GSS_REGTAB ], pt_event->SIGNAL);
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_REGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_UPDATE_TABLE[GSS_REGTAB] =
                      delchain(GSS_UPDATE_TABLE[GSS_REGTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      break;

    case GSS_REGDEF | GSS_CNDDEF :
      gss_cpyvalue(((struct wrireg *)(pt_event->SIGNAL))->PRJBVL->CNDVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_RESOL_TABLE[GSS_REGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_RESOL_TABLE[ GSS_REGTAB ] = 
           addchain (GSS_RESOL_TABLE[ GSS_REGTAB ], pt_event->SIGNAL);
        }

      pt_chain = GSS_UPDATE_TABLE[GSS_REGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          GSS_UPDATE_TABLE[GSS_REGTAB] =
                      delchain(GSS_UPDATE_TABLE[GSS_REGTAB], pt_chain);
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      break;

    case GSS_AUXDEF :
      gss_cpyvalue(((struct wriaux *)(pt_event->SIGNAL))->PRJVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_UPDATE_TABLE[GSS_AUXTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_UPDATE_TABLE[ GSS_AUXTAB ] = 
           addchain (GSS_UPDATE_TABLE[ GSS_AUXTAB ], pt_event->SIGNAL);
        }
      break;

    case GSS_MSGDEF :
      gss_cpyvalue(((struct wrimsg *)(pt_event->SIGNAL))->PRJVAL,
                   pt_event->VALUE,
                   usr_package);

      pt_chain = GSS_UPDATE_TABLE[GSS_MSGTAB];
      while (pt_chain != NULL)
        {
        if (pt_chain->DATA == pt_event->SIGNAL)
          {
          break;
          }
        pt_chain = pt_chain->NEXT;
        }
      if (pt_chain == NULL)
        {
        GSS_UPDATE_TABLE[ GSS_MSGTAB ] = 
           addchain (GSS_UPDATE_TABLE[ GSS_MSGTAB ], pt_event->SIGNAL);
        }
      break;

    default :
      gss_error (0, "gss_forceevent");
      break;

    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_forcesequence					*/
/* description	: forces an events sequence from a gss sequence		*/
/*		  structure and initialises values in the circuit	*/
/* called func.	: addchain						*/
/* ###--------------------------------------------------------------### */

void gss_forcesequence (pt_sequence, usr_package)

struct gssseq      *pt_sequence;
struct funpack     *usr_package;

  {
  struct gssevt  *pt_event;

  pt_event = pt_sequence->EVENT;

  while (pt_event != NULL)
    {
    gss_forceevent(pt_event, usr_package);

    pt_event = pt_event->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_deltaexecute					*/
/* description	: executes all scheduled events				*/
/*               (execution of a delta cycle)				*/
/* called func.	: gss_evalue, addchain, delchain, freechain		*/
/* ###--------------------------------------------------------------### */

void gss_deltaexecute (usr_package)

struct funpack *usr_package;

  {

  struct chain  *cur_chain;
  struct chain  *err_chain;
  int            errflg      = 0;

  struct lkdins *ptr_lkdins;
  struct beout  *ptr_beout;
  struct bebus  *ptr_bebus;
  struct bebux  *ptr_bebux;
  struct beaux  *ptr_beaux;
  struct bereg  *ptr_bereg;
  struct bemsg  *ptr_bemsg;
  struct wrssig *ptr_wrssig;
  struct wrbsig *ptr_wrbsig;
  struct wrireg *ptr_wrireg;
  struct wriaux *ptr_wriaux;
  struct wrimsg *ptr_wrimsg;
  struct wribux *ptr_wribux;

  struct binode *ptr_binode;
  struct prjbvl *ptr_prjbvl;

  void	        *drvval = NULL;
  void	        *cndval = NULL;

  unsigned long  delay = 0;	/* le delay	*/
  unsigned long  inert = 0;	/* l' inertie	*/

	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRSSIGs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_EVENT_TABLE[GSS_SIGTAB];

    while (cur_chain != NULL)
      {
      ptr_wrssig = (struct wrssig *)cur_chain->DATA;

      drvval = gss_addvalue (usr_package);

/* printf("OK pour evaluer le Wrisig %s\n", ptr_wrssig->BEOUT->NAME); */
      gss_evalvalue (ptr_wrssig->LKDINS,
                     ptr_wrssig->BEOUT->NODE,
                     GSS_SIGDEF,
                     drvval,
                     usr_package);

      delay = gss_evaldelay (ptr_wrssig->LKDINS,
                             ptr_wrssig->BEOUT,
                             GSS_SIGDEF,
                             usr_package);

      inert = gss_evalinert (ptr_wrssig->LKDINS,
                             ptr_wrssig->BEOUT,
                             GSS_SIGDEF,
                             usr_package);

      gss_addtransaction (ptr_wrssig->SIMSIG, ptr_wrssig->SIMSIG,
                          delay, inert, NULL, drvval,
                          (unsigned int) 'S', GSS_SCG_MODE, usr_package);

      ptr_wrssig->EVAL = 0;

      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_EVENT_TABLE[GSS_SIGTAB]);
  GSS_EVENT_TABLE[GSS_SIGTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRBSIGs			*/
	/*								*/
	/*     All drivers of the extern BUS belonging to this instance	*/
	/*     (in this wrbsig) are evaluated, but there is probably	*/
	/*     others prjbvl in the list (before & after) not evaluated	*/
	/*								*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_EVENT_TABLE[GSS_BUSTAB];

    while (cur_chain != NULL)
      {
      ptr_wrbsig = (struct wrbsig *)cur_chain->DATA;

      ptr_prjbvl = ptr_wrbsig->PRJBVL;
      ptr_binode = ptr_wrbsig->BEBUS->BINODE;

      while (ptr_binode != NULL)
        {
        cndval = gss_addvalue (usr_package);

        gss_evalvalue (ptr_wrbsig->LKDINS,
                       ptr_binode->CNDNODE,
                       GSS_BUSDEF | GSS_CNDDEF,
                       cndval,
                       usr_package);

        drvval = gss_addvalue (usr_package);

        gss_evalvalue (ptr_wrbsig->LKDINS,
                       ptr_binode->VALNODE,
                       GSS_BUSDEF | GSS_DRVDEF,
                       drvval,
                       usr_package);

        delay = gss_evaldelay (ptr_wrbsig->LKDINS,
                               ptr_binode,
                               GSS_BUSDEF,
                               usr_package);

        inert = gss_evalinert (ptr_wrbsig->LKDINS,
                               ptr_binode,
                               GSS_BUSDEF,
                               usr_package);

        gss_addtransaction (ptr_prjbvl, ptr_wrbsig->BUSSIG,
                            delay, inert, cndval, drvval,
                            (unsigned int) 'B', GSS_SCG_MODE, usr_package);

        ptr_prjbvl         = ptr_prjbvl->NEXT;
        ptr_binode         = ptr_binode->NEXT;
        }

      ptr_wrbsig->EVAL = 0;

      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_EVENT_TABLE[GSS_BUSTAB]);
  GSS_EVENT_TABLE[GSS_BUSTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRIAUXs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_EVENT_TABLE[GSS_AUXTAB];

    while (cur_chain != NULL)
      {
      drvval = gss_addvalue (usr_package);

      ptr_wriaux = (struct wriaux *)cur_chain->DATA;

/* printf("OK pour evaluer le Wriaux %s\n",ptr_wriaux->BEAUX->NAME); */
      gss_evalvalue (ptr_wriaux->LKDINS,
                     ptr_wriaux->BEAUX->NODE,
                     GSS_AUXDEF,
                     drvval,
                     usr_package);
      delay = gss_evaldelay (ptr_wriaux->LKDINS,
                             ptr_wriaux->BEAUX,
                             GSS_AUXDEF,
                             usr_package);

      inert = gss_evalinert (ptr_wriaux->LKDINS,
                             ptr_wriaux->BEAUX,
                             GSS_AUXDEF,
                             usr_package);

      gss_addtransaction (ptr_wriaux, ptr_wriaux,
                          delay, inert, NULL, drvval,
                          (unsigned int) 'X', GSS_SCG_MODE, usr_package);

      ptr_wriaux->EVAL = 0;

      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_EVENT_TABLE[GSS_AUXTAB]);
  GSS_EVENT_TABLE[GSS_AUXTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Computes projected value of WRIREGs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_EVENT_TABLE[GSS_REGTAB];

    while (cur_chain != NULL)
      {
      ptr_wrireg = (struct wrireg *)cur_chain->DATA;

	/* ###------------------------------------------------------### */
	/*    The first PRJBVL is external & hasn't associated node.	*/
	/* ###------------------------------------------------------### */

      ptr_prjbvl = ptr_wrireg->PRJBVL->NEXT;

      ptr_binode = ptr_wrireg->BEREG->BINODE;

      while (ptr_binode != NULL)
        {
        cndval = gss_addvalue (usr_package);

        gss_evalvalue (ptr_wrireg->LKDINS,
                       ptr_binode->CNDNODE,
                       GSS_REGDEF | GSS_CNDDEF,
                       cndval,
                       usr_package);

        drvval = gss_addvalue (usr_package);

        gss_evalvalue (ptr_wrireg->LKDINS,
                       ptr_binode->VALNODE,
                       GSS_REGDEF | GSS_DRVDEF,
                       drvval,
                       usr_package);

        delay = gss_evaldelay (ptr_wrireg->LKDINS,
                               ptr_binode,
                               GSS_REGDEF,
                               usr_package);

        inert = gss_evalinert (ptr_wrireg->LKDINS,
                               ptr_binode,
                               GSS_REGDEF,
                               usr_package);

        gss_addtransaction (ptr_prjbvl, ptr_wrireg,
                            delay, inert, cndval, drvval,
                            (unsigned int) 'R', GSS_SCG_MODE, usr_package);

        ptr_prjbvl         = ptr_prjbvl->NEXT;
        ptr_binode         = ptr_binode->NEXT;
        }

      ptr_wrireg->EVAL = 0;

      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_EVENT_TABLE[GSS_REGTAB]);
  GSS_EVENT_TABLE[GSS_REGTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Compute projected value of WRIBUXs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_EVENT_TABLE[GSS_BUXTAB];

    while (cur_chain != NULL)
      {
      ptr_wribux = (struct wribux *)cur_chain->DATA;

	/* ###------------------------------------------------------### */
	/*    The first PRJBVL is external & hasn't associated node.	*/
	/* ###------------------------------------------------------### */

      ptr_prjbvl = ptr_wribux->PRJBVL->NEXT;

      ptr_binode = ptr_wribux->BEBUX->BINODE;

      while (ptr_binode != NULL)
        {
        cndval = gss_addvalue (usr_package);

        gss_evalvalue (ptr_wribux->LKDINS,
                       ptr_binode->CNDNODE,
                       GSS_BUXDEF | GSS_CNDDEF,
                       cndval,
                       usr_package);

        drvval = gss_addvalue (usr_package);

        gss_evalvalue (ptr_wribux->LKDINS,
                       ptr_binode->VALNODE,
                       GSS_BUXDEF | GSS_DRVDEF,
                       drvval,
                       usr_package);

        delay = gss_evaldelay (ptr_wribux->LKDINS,
                               ptr_binode,
                               GSS_BUXDEF,
                               usr_package);

        inert = gss_evalinert (ptr_wribux->LKDINS,
                               ptr_binode,
                               GSS_BUXDEF,
                               usr_package);

        gss_addtransaction (ptr_prjbvl, ptr_wribux,
                            delay, inert, cndval, drvval,
                            (unsigned int) 'U', GSS_SCG_MODE, usr_package);

        ptr_prjbvl         = ptr_prjbvl->NEXT;
        ptr_binode         = ptr_binode->NEXT;
        }

      ptr_wribux->EVAL = 0;

      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_EVENT_TABLE[GSS_BUXTAB]);
  GSS_EVENT_TABLE[GSS_BUXTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Compute projected value of WRIMSGs			*/
	/* ###------------------------------------------------------### */

	/*    Messages non ajoutes dans l'echeancier pour le moment.	*/
	/*    Ni delay ni inert ne sont calcules pour les messages. 	*/

  cur_chain = GSS_EVENT_TABLE[GSS_MSGTAB];

    while (cur_chain != NULL)
      {
      ptr_wrimsg = (struct wrimsg *)cur_chain->DATA;

      errflg = gss_evalvalue (ptr_wrimsg->LKDINS,
                              ptr_wrimsg->BEMSG->NODE,
                              GSS_MSGDEF,
                              ptr_wrimsg->PRJVAL,
                              usr_package);

      if (errflg != 0)
        {
        err_chain   = GSS_ERROR_TABLE[GSS_MSGTAB];
        while (err_chain != NULL)
          {
          if (((struct wrimsg *)err_chain->DATA) ==  ptr_wrimsg)
            break;
          else
            err_chain = err_chain->NEXT;
          }
        if (err_chain == NULL)
          GSS_ERROR_TABLE[GSS_MSGTAB] =
            addchain (GSS_ERROR_TABLE[GSS_MSGTAB], ptr_wrimsg);
        }

      else
        {
        err_chain = GSS_ERROR_TABLE[GSS_MSGTAB];
        while (err_chain != NULL)
          {
          if (((struct wrimsg *)err_chain->DATA) ==  ptr_wrimsg)
            {
            GSS_ERROR_TABLE[GSS_MSGTAB] =
              delchain (GSS_ERROR_TABLE[GSS_MSGTAB], err_chain);
            break;
            }
          else
            err_chain = err_chain->NEXT;
          }
        }

      ptr_wrimsg->EVAL = 0;

      cur_chain = cur_chain->NEXT;
      }

  GSS_UPDATE_TABLE[GSS_MSGTAB] = GSS_EVENT_TABLE[GSS_MSGTAB];
  GSS_EVENT_TABLE[GSS_MSGTAB]  = NULL;
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_asserterror					*/
/* description	: report assert violations occured.			*/
/* called func.	: fprintf						*/
/* ###--------------------------------------------------------------### */

static int gss_asserterror (pt_wrimsg)

struct wrimsg  *pt_wrimsg;

  {
  char          *level;
  int            errflg = 0;


  switch (pt_wrimsg->BEMSG->LEVEL)
    {
    case 'W' :
      level = "WARNING";
      break;

    default  :
      level = "ERROR";
      errflg = 1;
      break;
    }

  (void)fprintf (stderr,
                 "%s : `%s` Assert Violation on Cell %s : %s\n",
                 level,
                 pt_wrimsg->BEMSG->LABEL,
                 pt_wrimsg->LKDINS->INSNAME,
                 pt_wrimsg->BEMSG->MESSAGE);
 
  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_deltacheck					*/
/* description	: check errors occured at the end of a delta loop.	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

int gss_deltacheck ()

  {
  int            errflg  = 0;
  struct chain  *cur_chain;

	/* ###------------------------------------------------------### */
	/*    Computes Error records for SIMSIGs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_ERROR_TABLE[GSS_SIGTAB];

    while (cur_chain != NULL)
      {
      errflg += gss_error (0, "gss_deltacheck");
      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_ERROR_TABLE[GSS_SIGTAB]);
  GSS_ERROR_TABLE[GSS_SIGTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Computes Error Records for BUSSIGs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_ERROR_TABLE[GSS_BUSTAB];

    while (cur_chain != NULL)
      {
      errflg += gss_error (16,NULL);
      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_ERROR_TABLE[GSS_BUSTAB]);
  GSS_ERROR_TABLE[GSS_BUSTAB] = NULL;

	/* ###------------------------------------------------------### */
	/*    Computes Error Records for WRIAUXs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_ERROR_TABLE[GSS_AUXTAB];

    while (cur_chain != NULL)
      {
      errflg += gss_error (0, "gss_deltacheck");
      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_ERROR_TABLE[GSS_AUXTAB]);
  GSS_ERROR_TABLE[GSS_AUXTAB]  = NULL;

	/* ###------------------------------------------------------### */
	/*    Computes Error Records for WRIREGs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_ERROR_TABLE[GSS_REGTAB];

    while (cur_chain != NULL)
      {
      errflg += gss_error (17,NULL);
      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_ERROR_TABLE[GSS_REGTAB]);
  GSS_ERROR_TABLE[GSS_REGTAB]  = NULL;

	/* ###------------------------------------------------------### */
	/*    Compute Error Records for WRIBUXs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_ERROR_TABLE[GSS_BUXTAB];

    while (cur_chain != NULL)
      {
      errflg += gss_error (16,NULL);
      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_ERROR_TABLE[GSS_BUXTAB]);
  GSS_ERROR_TABLE[GSS_BUXTAB]  = NULL;

	/* ###------------------------------------------------------### */
	/*    Compute Error Records for WRIMSGs			*/
	/* ###------------------------------------------------------### */

  cur_chain = GSS_ERROR_TABLE[GSS_MSGTAB];

    while (cur_chain != NULL)
      {
      errflg += gss_asserterror((struct wrimsg *) cur_chain->DATA);
      cur_chain = cur_chain->NEXT;
      }

  freechain (GSS_ERROR_TABLE[GSS_MSGTAB]);
  GSS_ERROR_TABLE[GSS_MSGTAB]  = NULL;

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_circuitstable					*/
/* description	: executes all expressions of the circuit before the	*/
/*		  first pattern and until NO_EVENTS.			*/
/* called func.	: gss_evalue, addchain, delchain, freechain		*/
/* ###--------------------------------------------------------------### */

int gss_circuitstable (ptlkdfig, usr_package)

struct lkdfig  *ptlkdfig;
struct funpack *usr_package;

  {
  int    errflg  = 0;
  int    scheduler_empty;
  struct simsig *ptr_simsig;
  struct bussig *ptr_bussig;
  struct wrssig *ptr_wrssig;
  struct wrbsig *ptr_wrbsig;
  struct lkdins *ptr_lkdins;
  struct wriaux *ptr_wriaux;
  struct wrireg *ptr_wrireg;
  struct wrimsg *ptr_wrimsg;
  struct wribux *ptr_wribux;

	/* ###------------------------------------------------------### */
	/*    First : All signals have to be resolved and/or updated.	*/
	/* ###------------------------------------------------------### */

  gss_deltaresolution (usr_package);

  gss_deltaupdate (usr_package);

	/* ###------------------------------------------------------### */
	/*    Then : All drivers have to be evaluated.			*/
	/*           (only for the first delta cycle)			*/
	/* ###------------------------------------------------------### */

  ptr_lkdins  = ptlkdfig->LKDINS;

  while (ptr_lkdins != NULL)
    {
	/* ###------------------------------------------------------### */
	/*            push all WRSSIG on GSS_EVENT_TABLE 		*/
	/* ###------------------------------------------------------### */

    ptr_wrssig  = ptr_lkdins->WRSSIG;
    while (ptr_wrssig != NULL)
      {
      if (ptr_wrssig->EVAL == 0)
        {
        GSS_EVENT_TABLE[GSS_SIGTAB] = addchain
                        (GSS_EVENT_TABLE[GSS_SIGTAB], ptr_wrssig);
        ptr_wrssig->EVAL = 1;
        }
      ptr_wrssig = ptr_wrssig->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all WRBSIG on GSS_EVENT_TABLE 		*/
	/* ###------------------------------------------------------### */

    ptr_wrbsig  = ptr_lkdins->WRBSIG;
    while (ptr_wrbsig != NULL)
      {
      if (ptr_wrbsig->EVAL == 0)
        {
        GSS_EVENT_TABLE[GSS_BUSTAB] = addchain
                        (GSS_EVENT_TABLE[GSS_BUSTAB], ptr_wrbsig);
        ptr_wrbsig->EVAL = 1;
        }
      ptr_wrbsig = ptr_wrbsig->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all WRIAUX on GSS_EVENT_TABLE 		*/
	/* ###------------------------------------------------------### */

    ptr_wriaux  = ptr_lkdins->WRIAUX;
    while (ptr_wriaux != NULL)
      {
      if (ptr_wriaux->EVAL == 0)
        {
        GSS_EVENT_TABLE[GSS_AUXTAB] = addchain
                        (GSS_EVENT_TABLE[GSS_AUXTAB], ptr_wriaux);
        ptr_wriaux->EVAL = 1;
        }
      ptr_wriaux = ptr_wriaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all WRIREG on GSS_EVENT_TABLE 		*/
	/* ###------------------------------------------------------### */

    ptr_wrireg  = ptr_lkdins->WRIREG;
    while (ptr_wrireg != NULL)
      {
      if (ptr_wrireg->EVAL == 0)
        {
        GSS_EVENT_TABLE[GSS_REGTAB] = addchain
                        (GSS_EVENT_TABLE[GSS_REGTAB], ptr_wrireg);
        ptr_wrireg->EVAL = 1;
        }
      ptr_wrireg = ptr_wrireg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all WRIBUX on GSS_EVENT_TABLE 		*/
	/* ###------------------------------------------------------### */

    ptr_wribux  = ptr_lkdins->WRIBUX;
    while (ptr_wribux != NULL)
      {
      if (ptr_wribux->EVAL == 0)
        {
        GSS_EVENT_TABLE[GSS_BUXTAB] = addchain
                        (GSS_EVENT_TABLE[GSS_BUXTAB], ptr_wribux);
        ptr_wribux->EVAL = 1;
        }
      ptr_wribux = ptr_wribux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all WRIMSG on GSS_EVENT_TABLE 		*/
	/* ###------------------------------------------------------### */

    ptr_wrimsg  = ptr_lkdins->WRIMSG;
    while (ptr_wrimsg != NULL)
      {
      if (ptr_wrimsg->EVAL == 0)
        {
        GSS_EVENT_TABLE[GSS_MSGTAB] = addchain
                        (GSS_EVENT_TABLE[GSS_MSGTAB], ptr_wrimsg);
        ptr_wrimsg->EVAL = 1;
        }
      ptr_wrimsg = ptr_wrimsg->NEXT;
      }

    ptr_lkdins = ptr_lkdins->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*           execute the first delta cycle			*/
	/* ###------------------------------------------------------### */

  GSS_DELTACYCLE = 0;

  gss_deltaexecute (usr_package);

	/* ###------------------------------------------------------### */
	/*           execute the following of the first delta loop	*/
	/* ###------------------------------------------------------### */

  GSS_DELTACYCLE = 1;

  scg_SetCurrentTime (GSS_SCHEDULER, GSS_CURRENT_TIME);

  scheduler_empty = scg_GetNextTime (GSS_SCHEDULER, &GSS_NEXT_TIME);

  if ((scheduler_empty != 0) && (GSS_NEXT_TIME == GSS_CURRENT_TIME))
    gss_loadtransaction (usr_package);

  while (GSS_DELTACYCLE < GSS_MAXDELTA)
    {
    gss_deltaresolution (usr_package);

    gss_deltaupdate (usr_package);

    gss_deltaexecute (usr_package);

    scheduler_empty = scg_GetNextTime (GSS_SCHEDULER, &GSS_NEXT_TIME);

    if ((scheduler_empty != 0) && (GSS_NEXT_TIME == GSS_CURRENT_TIME))
      gss_loadtransaction (usr_package);
    else
      break;

    GSS_DELTACYCLE ++;
    }

  errflg = gss_deltacheck ();

  if (GSS_DELTACYCLE >= GSS_MAXDELTA)
    errflg = gss_error (0, "gss_circuitstable");

  return(errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_circuitinit					*/
/* description	: initialise all signals of the circuit.		*/
/* called func.	: gss_evalue, addchain, delchain, freechain		*/
/* ###--------------------------------------------------------------### */

void gss_circuitinit (ptlkdfig, initvalue, usr_package)

struct lkdfig  *ptlkdfig;
void	       *initvalue;
struct funpack *usr_package;

  {
  struct simsig *ptr_simsig;
  struct bussig *ptr_bussig;
  struct lkdins *ptr_lkdins;
  struct wriaux *ptr_wriaux;
  struct wrireg *ptr_wrireg;
  struct wribux *ptr_wribux;
  struct wrimsg *ptr_wrimsg;

	/* ###------------------------------------------------------### */
	/*            push all SIMSIG on GSS_UPDATE_TABLE 		*/
	/*            initialize all SIMSIG 		 		*/
	/* ###------------------------------------------------------### */

  ptr_simsig  = ptlkdfig->SIMSIG;

    while (ptr_simsig != NULL)
      {
      GSS_UPDATE_TABLE[GSS_SIGTAB] = addchain
                      (GSS_UPDATE_TABLE[GSS_SIGTAB], ptr_simsig);

      gss_cpyvalue (ptr_simsig->PRJVAL, initvalue, usr_package);
      gss_initevthis (ptr_simsig->HISTORY);
      ptr_simsig = ptr_simsig->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all BUSSIG on GSS_RESOL_TABLE 		*/
	/*            initialize all BUSSIG 		 		*/
	/* ###------------------------------------------------------### */

  ptr_bussig  = ptlkdfig->BUSSIG;

    while (ptr_bussig != NULL)
      {
      GSS_RESOL_TABLE[GSS_BUSTAB] = addchain
                      (GSS_RESOL_TABLE[GSS_BUSTAB], ptr_bussig);

      gss_cpyvalue (ptr_bussig->PRJVAL, initvalue, usr_package);
      gss_initevthis (ptr_bussig->HISTORY);
      ptr_bussig = ptr_bussig->NEXT;
      }

  ptr_lkdins  = ptlkdfig->LKDINS;

  while (ptr_lkdins != NULL)
    {
	/* ###------------------------------------------------------### */
	/*            push all BUXSIG on GSS_RESOL_TABLE 		*/
	/*            initialize all BUXSIG 		 		*/
	/* ###------------------------------------------------------### */

    ptr_wribux  = ptr_lkdins->WRIBUX;

    while (ptr_wribux != NULL)
      {
      GSS_RESOL_TABLE[GSS_BUXTAB] = addchain
                      (GSS_RESOL_TABLE[GSS_BUXTAB], ptr_wribux);

      gss_cpyvalue (ptr_wribux->PRJVAL, initvalue, usr_package);
      gss_initevthis (ptr_wribux->HISTORY);
      ptr_wribux = ptr_wribux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all REGSIG on GSS_RESOL_TABLE 		*/
	/*            initialize all REGSIG 		 		*/
	/* ###------------------------------------------------------### */

    ptr_wrireg  = ptr_lkdins->WRIREG;

    while (ptr_wrireg != NULL)
      {
      GSS_RESOL_TABLE[GSS_REGTAB] = addchain
                      (GSS_RESOL_TABLE[GSS_REGTAB], ptr_wrireg);

      gss_cpyvalue (ptr_wrireg->PRJVAL, initvalue, usr_package);
      gss_initevthis (ptr_wrireg->HISTORY);
      ptr_wrireg = ptr_wrireg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all AUXSIG on GSS_UPDATE_TABLE 		*/
	/*            initialize all AUXSIG 		 		*/
	/* ###------------------------------------------------------### */

    ptr_wriaux  = ptr_lkdins->WRIAUX;

    while (ptr_wriaux != NULL)
      {
      GSS_UPDATE_TABLE[GSS_AUXTAB] = addchain
                      (GSS_UPDATE_TABLE[GSS_AUXTAB], ptr_wriaux);

      gss_cpyvalue (ptr_wriaux->PRJVAL, initvalue, usr_package);
      gss_initevthis (ptr_wriaux->HISTORY);
      ptr_wriaux = ptr_wriaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*            push all MSGSIG on GSS_UPDATE_TABLE 		*/
	/*            initialize all MSGSIG 		 		*/
	/* ###------------------------------------------------------### */

    ptr_wrimsg  = ptr_lkdins->WRIMSG;

    while (ptr_wrimsg != NULL)
      {
      GSS_UPDATE_TABLE[GSS_MSGTAB] = addchain
                      (GSS_UPDATE_TABLE[GSS_MSGTAB], ptr_wrimsg);

      gss_cpyvalue (ptr_wrimsg->PRJVAL, initvalue, usr_package);
      gss_initevthis (ptr_wrimsg->HISTORY);
      ptr_wrimsg = ptr_wrimsg->NEXT;
      }

    ptr_lkdins = ptr_lkdins->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: gss_deltaloop						*/
/* description	: executes events until NO_EVENTS.			*/
/* called func.	: gss_evalue, gss_deltaupdate, gss_error		*/
/* ###--------------------------------------------------------------### */

int gss_deltaloop (usr_package)

struct funpack *usr_package;

  {
  int              errflg =0;
  int              scheduler_empty;
  unsigned long    next_date;

  scg_SetCurrentTime (GSS_SCHEDULER, GSS_CURRENT_TIME);
  GSS_DELTACYCLE = 0;

  scheduler_empty = scg_GetNextTime (GSS_SCHEDULER, &GSS_NEXT_TIME);

  if ((scheduler_empty != 0) && (GSS_NEXT_TIME == GSS_CURRENT_TIME))
    gss_loadtransaction (usr_package);

  while (GSS_DELTACYCLE < GSS_MAXDELTA)
    {
    gss_deltaresolution (usr_package);

    gss_deltaupdate (usr_package);

    gss_deltaexecute (usr_package);

    scheduler_empty = scg_GetNextTime (GSS_SCHEDULER, &GSS_NEXT_TIME);

    if ((scheduler_empty != 0) && (GSS_NEXT_TIME == GSS_CURRENT_TIME))
      gss_loadtransaction (usr_package);
    else
      break;

    GSS_DELTACYCLE ++;
    }

  errflg = gss_deltacheck ();

  if (GSS_DELTACYCLE >= GSS_MAXDELTA)
    errflg = gss_error (0, "gss_deltaloop");

  return(errflg);
  }


/* ###--------------------------------------------------------------### */
/* function	: gss_loadtransaction					*/
/* description	: load the next set of transactions from the scheduler	*/
/* called func.	: 							*/
/* ###--------------------------------------------------------------### */

void gss_loadtransaction (usr_package)

struct funpack *usr_package;

  {
  struct shtra   *pt_shtra;
  struct shtra  *pt_trans ;
  struct simsig *pt_simsig;
  struct bussig *pt_bussig;
  struct wriaux *pt_wriaux;
  struct wribux *pt_wribux;
  struct wrireg *pt_wrireg;
  struct prjbvl *pt_prjbvl;

  pt_shtra = scg_GetCurrentTransactions (GSS_SCHEDULER);

  pt_trans = pt_shtra;

  while (pt_trans != NULL)
    {
    switch (pt_trans->UFLAG)
      {
      case 'S':
        pt_simsig         = (struct simsig *) pt_trans->IDENT;
        gss_cpyvalue (pt_simsig->PRJVAL, pt_trans->VALU, usr_package);
        gss_freevalue (pt_trans->VALU, usr_package);
        GSS_UPDATE_TABLE[GSS_SIGTAB]  =
              addchain( GSS_UPDATE_TABLE[GSS_SIGTAB], pt_simsig);
        break;

      case 'B':
        pt_prjbvl         = (struct prjbvl *) pt_trans->IDENT;
        pt_bussig         = (struct bussig *) pt_trans->SCDID;
        gss_cpyvalue (pt_prjbvl->CNDVAL, pt_trans->COND, usr_package);
        gss_cpyvalue (pt_prjbvl->DRVVAL, pt_trans->VALU, usr_package);
        gss_freevalue (pt_trans->COND, usr_package);
        gss_freevalue (pt_trans->VALU, usr_package);
        GSS_RESOL_TABLE[GSS_BUSTAB]  =
              addchain( GSS_RESOL_TABLE[GSS_BUSTAB], pt_bussig);
        break;

      case 'X':
        pt_wriaux         = (struct wriaux *) pt_trans->IDENT;
        gss_cpyvalue (pt_wriaux->PRJVAL, pt_trans->VALU, usr_package);
        gss_freevalue (pt_trans->VALU, usr_package);
        GSS_UPDATE_TABLE[GSS_AUXTAB]  =
              addchain( GSS_UPDATE_TABLE[GSS_AUXTAB], pt_wriaux);
        break;

      case 'U':
        pt_prjbvl         = (struct prjbvl *) pt_trans->IDENT;
        pt_wribux         = (struct wribux *) pt_trans->SCDID;
        gss_cpyvalue (pt_prjbvl->CNDVAL, pt_trans->COND, usr_package);
        gss_cpyvalue (pt_prjbvl->DRVVAL, pt_trans->VALU, usr_package);
        gss_freevalue (pt_trans->COND, usr_package);
        gss_freevalue (pt_trans->VALU, usr_package);
        GSS_RESOL_TABLE[GSS_BUXTAB]  =
              addchain( GSS_RESOL_TABLE[GSS_BUXTAB], pt_wribux);
        break;

      case 'R':
        pt_prjbvl         = (struct prjbvl *) pt_trans->IDENT;
        pt_wrireg         = (struct wrireg *) pt_trans->SCDID;
        gss_cpyvalue (pt_prjbvl->CNDVAL, pt_trans->COND, usr_package);
        gss_cpyvalue (pt_prjbvl->DRVVAL, pt_trans->VALU, usr_package);
        gss_freevalue (pt_trans->COND, usr_package);
        gss_freevalue (pt_trans->VALU, usr_package);
        GSS_RESOL_TABLE[GSS_REGTAB]  =
              addchain( GSS_RESOL_TABLE[GSS_REGTAB], pt_wrireg);
        break;
      }
    pt_trans = pt_trans->NEXT;
    }

  scg_FreeTransactions (pt_shtra);
  }
