
/* ###--------------------------------------------------------------### */
/* file		: pat_addpaseq.c					*/
/* date		: Aug 25 1997						*/
/* version	: v107							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level pat function				*/
/* ###--------------------------------------------------------------### */

#include MUT_H
#include PAT_H

/* ###--------------------------------------------------------------### */
/* function	: pat_addpaseq						*/
/* description	: create an empty PASEQ (sequence of patterns)		*/
/*		  structure at the top of a list			*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct paseq *pat_addpaseq (lastpaseq)

struct paseq *lastpaseq;	/* pointer on the last paseq structure	*/

  {
  struct paseq  *ptseq   ;
  unsigned char  err_flg = 0;

  ptseq            = (struct paseq *) mbkalloc (sizeof(struct paseq));

  switch (PAT_TIMEUNIT)
    {
    case PAT_TU__FS:
      PAT_CNV_FS = 1.0    ;
      PAT_CNV_PS = 1.0E3  ;
      PAT_CNV_NS = 1.0E6  ;
      PAT_CNV_US = 1.0E9  ;
      PAT_CNV_MS = 1.0E12 ;
      break;
    case PAT_TU__PS:
      PAT_CNV_FS = 1.0E-3 ;
      PAT_CNV_PS = 1.0    ;
      PAT_CNV_NS = 1.0E3  ;
      PAT_CNV_US = 1.0E6  ;
      PAT_CNV_MS = 1.0E9  ;
      break;
    case PAT_TU__NS:
      PAT_CNV_FS = 1.0E-6 ;
      PAT_CNV_PS = 1.0E-3 ;
      PAT_CNV_NS = 1.0    ;
      PAT_CNV_US = 1.0E3  ;
      PAT_CNV_MS = 1.0E6  ;
      break;
    case PAT_TU__US:
      PAT_CNV_FS = 1.0E-9 ;
      PAT_CNV_PS = 1.0E-6 ;
      PAT_CNV_NS = 1.0E-3 ;
      PAT_CNV_US = 1.0    ;
      PAT_CNV_MS = 1.0E3  ;
      break;
    case PAT_TU__MS:
      PAT_CNV_FS = 1.0E-12;
      PAT_CNV_PS = 1.0E-9 ;
      PAT_CNV_NS = 1.0E-6 ;
      PAT_CNV_US = 1.0E-3 ;
      PAT_CNV_MS = 1.0    ;
      break;
    default:
      err_flg = pat_error (200, NULL, ' ', 0);
    }

  ptseq->NAME      = NULL        ;
  ptseq->PAGRP     = NULL        ;
  ptseq->PAIOL     = NULL        ;
  ptseq->DECCOM    = NULL        ;
  ptseq->OLDPAT    = NULL        ;
  ptseq->OLDCOM    = NULL        ;
  ptseq->CURPAT    = NULL        ;
  ptseq->CURCOM    = NULL        ;
  ptseq->NXTPAT    = NULL        ;
  ptseq->MODEL     = NULL        ;
  ptseq->BUFFER    = NULL        ;
  ptseq->PAINS     = NULL        ;
  ptseq->PATNBR    = 0           ;
  ptseq->IOLNBR    = 0           ;
  ptseq->SUBSEQ    = 0           ;
  ptseq->LINENO    = 1           ;
  ptseq->SAVFLG    = 'N'         ;
  ptseq->ENDFLG    = 'N'         ;
  ptseq->DRVSEQ    = 0           ;
  ptseq->ERRFLG    = err_flg     ;
  ptseq->FLAGS     = 0           ;
  ptseq->TIME_UNIT = PAT_TIMEUNIT;

  ptseq->NEXT      = lastpaseq   ;

  return (ptseq);
  }
