
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: prb.h							*/
/* date		: May 23 1994						*/
/* version	: v109							*/
/* authors	: DUNOYER Julien					*/
/* description	: header for PRB library				*/
/*									*/
/* ###--------------------------------------------------------------### */

#ifndef PRB_PRBDEF
#define PRB_PRBDEF

	/* ###------------------------------------------------------### */
	/*    defines							*/
	/* ###------------------------------------------------------### */

#define DEF_PRB_OUT_SUFFIX   "pro"
#define DEF_PRB_IN_SUFFIX    "prb"

#define DEF_PRB_FLAG_NUL   0 		/* NULL prbseq FLAG		*/
#define DEF_PRB_FLAG_REF   1 		/* REFERENCE prbseq FLAG	*/

#define DEF_PRB_SCAN     'S'
#define DEF_PRB_UNSCAN   'U'

#define PRB_STATIC_MODE       0x01 	/* STATIC     SIMULATION MODE	*/
#define PRB_DYNAMIC_MODE      0x02 	/* DYNAMIC    SIMULATION MODE	*/
#define PRB_MULTICYCLE_MODE   0x04 	/* MULTICYCLE SIMULATION MODE	*/

#define PRB_STEPBYSTEP_SUPERABL   0x00
#define PRB_EXPLODING_SUPERABL    0x01

#define PRB_NOT_EVALUATED 0
#define PRB_EVALUATED     1
#define PRB_END_EVALUATED 6

#define PRB_EVENTDRIVEN_EVALUATION  0x00
#define PRB_LEVELIZED_EVALUATION    0x01

#define DEF_PRB_IN       'I' 		/* IN prbevt MODE		*/
#define DEF_PRB_OUT      'O'
#define DEF_PRB_INOUT    'B'
#define DEF_PRB_TRISTATE 'Z'
#define DEF_PRB_TRANSC   'T'
#define DEF_PRB_PORT     'P'
#define DEF_PRB_SIGNAL   'S'
#define DEF_PRB_ASSERT   'A'
#define DEF_PRB_REGISTER 'R'
#define DEF_PRB_CLOCK    'C'

#define DEF_PRB_UNKNOWN   0

#define DEF_PRB_DEFPROB   0.5 		/* DEFAULT FLOAT prob VALUE	*/
#define DEF_PRB_DEFDENS   0.0 		/* DEFAULT FLOAT dens VALUE	*/
#define DEF_PRB_DEFDISC   0.0 		/* DEFAULT FLOAT disc VALUE	*/

#define DEF_PRB_DEFGATE      0x00 	/* DEFAULT GATE     prbsup FLAG	*/
#define DEF_PRB_SUPERGATE    0x01 	/* SUPERGATE        prbsup FLAG	*/
#define DEF_PRB_DERIVBYTIME  0x02 	/* DERIVATE BY TIME prbsup FLAG	*/
#define DEF_PRB_DERIVBYVAR   0x04 	/* DERIVATE BY VAR  prbsup FLAG	*/
#define DEF_PRB_BYTIMEDDR    0x08 	/* Mode    BYTIME pour la DDR	*/
#define DEF_PRB_BYVARDDR     0x10 	/* Mode    BYVAR  pour la DDR	*/
#define DEF_PRB_BYVOXDDR     0x20 	/* Mode    BYVOX  pour la DDR	*/
#define DEF_PRB_BYVAXDDR     0x40 	/* Mode    BYVAX  pour la DDR	*/
#define DEF_PRB_FINALDDR     0x80 	/* Mode    FINAL  pour la DDR	*/
#define DEF_PRB_MIXEDDDR     0x00 	/* Mode    MIXED  pour la DDR	*/

#define DEF_PRB_NULL_CMPMODE     0 	/* NULL COMPARAISON MODE	*/
#define DEF_PRB_DELTA_CMPMODE    1
#define DEF_PRB_DELTABS_CMPMODE  2
#define DEF_PRB_RMS_CMPMODE      3
#define DEF_PRB_MEAN_CMPMODE     4
#define DEF_PRB_STDDEV_CMPMODE   5
#define DEF_PRB_MINIMUM_CMPMODE  6
#define DEF_PRB_MAXIMUM_CMPMODE  7
#define DEF_PRB_DRELA_CMPMODE    8
#define DEF_PRB_DRELABS_CMPMODE  9

	/* ###------------------------------------------------------### */
	/*    macros definitions					*/
	/* ###------------------------------------------------------### */

#define PRB_SUPERNODE( node )                        \
         ((struct prbsup *)(((bequad_list *)node)->USER))
#define PRB_SUPERFLAG( node )                         \
         (((struct prbsup *)(((bequad_list *)node)->USER))->FLAG)
#define PRB_BDD( node )                         \
         (((bequad_list *)node)->BDD)
#define PRB_ABL( node )                         \
         (((bequad_list *)node)->ABL)
#define PRB_VARLIST( node )                         \
         (((bequad_list *)node)->VARLIST)
#define PRB_DERIVATE( node )                    \
         (((bequad_list *)node)->DERIVATE)
#define PRB_SUPERBDD( node )                         \
         (((struct prbsup *)(((bequad_list *)node)->USER))->BDD)
#define PRB_SUPERABL( node )                         \
         (((struct prbsup *)(((bequad_list *)node)->USER))->ABL)
#define PRB_SUPERVARLIST( node )                         \
         (((struct prbsup *)(((bequad_list *)node)->USER))->VARLIST)
#define PRB_SUPERDERIVATE( node )                    \
         (((struct prbsup *)(((bequad_list *)node)->USER))->DERIVATE)
#define PRB_SUPERDERDEDER( node )                    \
         (((struct prbsup *)(((bequad_list *)node)->USER))->DERDEDER)

	/* ###------------------------------------------------------### */
	/*    global variables						*/
	/* ###------------------------------------------------------### */

extern char          *PRB_OUT_SUFFIX;
extern char          *PRB_IN_SUFFIX;

extern char           PRB_MODE;
extern char           PRB_EVALUATION;
extern char           PRB_SUPERABL_MODE;

extern struct prbval *PRB_INIT_VALUE;
extern struct prbval *PRB_INPUT_VALUE;

extern struct prbseq    *PRB_HEAD_PRBSEQ;

	/* ###------------------------------------------------------### */
	/*    data structures						*/
	/* ###------------------------------------------------------### */

typedef struct prbseq			/* sequence of probabilities	*/
  {
  struct prbseq  	*NEXT    ;	/* next sequence in data base	*/
  char 			*NAME    ;	/* sequence's name		*/
  struct prbini		*PRBINI  ;	/* list of unscannable regs	*/
  struct prbino 	*PRBINO  ;	/* list of input & scannab regs	*/
  struct prbevt		*PRBEVT  ;	/* list of events		*/
  float                  PROBVAL ;	/* default probability value	*/
  float			 DISCVAL ;	/* default disconnect value	*/
  float			 DENSVAL ;	/* default density value	*/
  char			 SCANVAL ;	/* default regaccess type	*/
  char			 FLAG    ;	/* flag (usued for experim.)	*/
  char			 ERRFLG  ;	/* error flag (pars. err. only)	*/
  }
prbseq_list;

	/* ###------------------------------------------------------### */

typedef struct prbini			/* unscannable register		*/
  {
  struct prbini   	*NEXT    ;	/* next register		*/
  char           	*NAME    ;	/* register's name		*/
  }
prbini_list;

	/* ###------------------------------------------------------### */

typedef struct prbino			/* inout			*/
  {
  struct prbino   	*NEXT    ;	/* next inout			*/
  char           	*NAME    ;	/* inout's name			*/
  }
prbino_list;


	/* ###------------------------------------------------------### */

typedef struct prbevt			/* event			*/
  {
  struct prbevt   	*NEXT    ;	/* next event			*/
  char           	*NAME    ;	/* input's name			*/
  struct prbval 	*VALUE   ;	/* input value			*/
  char			 MODE    ;	/* mode (I,O,B,Z,T,P,S,A,R,C)	*/
  char			 TYPE    ;	/* SIG pointer type 		*/
  void			*SIG     ;	/* pointer on the signal	*/
  }
prbevt_list;


	/* ###------------------------------------------------------### */

typedef struct prbval			/* value			*/
  {
  float 		 PROBVAL ;	/* probability value		*/
  float			 DISCVAL ;	/* disconnect value		*/
  float			 DENSVAL ;	/* density value		*/
  float            	 DENSSUM ;	/* density sum (dynamic mode)	*/
  char                   EVALUATED ;    /* flag                         */
  };


	/* ###------------------------------------------------------### */

typedef struct prbwav			/* waveform			*/
  {
  struct prbwav		*NEXT    ;	/* next transaction		*/
  float			 DENSVAL ;	/* density value		*/
  unsigned long		 DATE    ;	/* transaction date		*/
  };


	/* ###------------------------------------------------------### */

typedef struct prbsup			/* supergate expression struct.	*/
  {
  struct chain		*ABL          ;	/* Super Boolean Expression	*/
  struct node		*BDD          ;	/* Super BDD			*/
  struct chain		*VARLIST      ;	/* Super Variable List		*/
  struct beder		*DERIVATE     ;	/* Super list of derivate BDDs	*/
  struct beder		*DERDEDER     ;	/* Super list of dynamic deriv.	*/
  char			 FLAG    ;	/* flag (usued for identif.)	*/
  };

struct prb_entry
  {
  struct prb_entry *next;
  struct prb_recrd *data;
  char             *key;
  };

struct prb_recrd
  {
  struct prb_recrd *next;
  char             *key;
  float             value;
  unsigned long     date;
  };


	/* ###------------------------------------------------------### */
	/*    functions							*/
	/* ###------------------------------------------------------### */

		/* ###--------------------------------------### */
		/*    prb_dict					*/
		/* ###--------------------------------------### */

		/* ###--------------------------------------### */
		/*    prb_util					*/
		/* ###--------------------------------------### */

extern char              *prb_avers   ();
extern void               prb_warning ();
extern void               prb_toolbug ();
extern void               prb_message ();
extern int                prb_error   ();

		/* ###--------------------------------------### */
		/*    prb_supernodes				*/
		/* ###--------------------------------------### */

extern void               prb_makprbsup     ();
extern struct prbsup     *prb_addprbsup     ();
extern void               prb_maksupabl     ();
extern struct chain      *prb_superabl      ();
extern void               prb_maksupvarlist ();
extern void               prb_maksupbdd     ();

		/* ###--------------------------------------### */
		/*    prb_package				*/
		/* ###--------------------------------------### */

extern void              *prb_addvalue ();
extern void               prb_freevalue ();
extern int                prb_cmpvalue ();
extern int                prb_cpyvalue ();
extern int                prb_updvalue ();
extern int                prb_evalvalue ();
extern int                prb_resolvalue ();

extern int                prb_dyncmpvalue ();
extern int                prb_dyncpyvalue ();
extern int                prb_dynupdvalue ();
extern int                prb_dynevalvalue ();
extern int                prb_dynresolvalue ();
extern unsigned long      prb_dynevaldelay ();
extern unsigned long      prb_dynevalinert ();
extern int                prb_dyncolvalue ();

		/* ###--------------------------------------### */
		/*    prb_debug					*/
		/* ###--------------------------------------### */

extern void               prb_debug ();

		/* ###--------------------------------------### */
		/*    prb_compare				*/
		/* ###--------------------------------------### */

extern struct prbseq     *prb_seqcompare ();
extern struct prbseq     *prb_seqchncompare ();

		/* ###--------------------------------------### */
		/*    prb_getprbseq				*/
		/* ###--------------------------------------### */

extern struct prbseq     *prb_randprbseq ();

		/* ###--------------------------------------### */
		/*    prb_getprbseq				*/
		/* ###--------------------------------------### */

extern struct prbseq     *prb_getprbseq ();

		/* ###--------------------------------------### */
		/*    prb_parser				*/
		/* ###--------------------------------------### */

extern struct prbseq     *prb_parser ();

		/* ###--------------------------------------### */
		/*    prb_driver				*/
		/* ###--------------------------------------### */

extern int                prb_driver ();

		/* ###--------------------------------------### */
		/*    prb_sequences				*/
		/* ###--------------------------------------### */

extern struct gssseq     *prb_loadsequence ();
extern void               prb_storesequence ();
extern void               prb_writesequence ();
extern struct prbseq     *prb_savesequence ();

		/* ###--------------------------------------### */
		/*    prb high level functions			*/
		/* ###--------------------------------------### */

extern struct dct_entry **prb_simulator ();

		/* ###--------------------------------------### */
		/*    prb low level functions			*/
		/* ###--------------------------------------### */

extern struct prbini     *prb_addprbini   ();
extern struct prbino     *prb_addprbino   ();
extern struct prbevt     *prb_addprbevt   ();
extern struct prbseq     *prb_addprbseq   ();

extern void               prb_freprbini   ();
extern void               prb_freprbevt   ();
extern void               prb_freprbino   ();
extern void               prb_freprbseq   ();

#endif
