
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: prb_dict.c						*/
/* date		: Jun  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* description	: This file contains the set of probability dedicated	*/
/*                dictionnary functions 			 	*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/time.h>

#include MUT_H
#include LOG_H
#include "beh.h"
#include "bhl.h"
#include GSS_H
#include PRB_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */

#define PRB_ALODFN      600
#define PRB_HSZDFN    16001

static struct prb_entry  *PRB_DCEHED;   /* free prb_entry's head        */
static struct prb_recrd  *PRB_DCRHED;   /* free prb_recrd's head        */

/* ###--------------------------------------------------------------### */
/* function	: prb_addent						*/
/* description	: add a new entry in a dictionary			*/
/* called func.	: mbkaloc						*/
/* ###--------------------------------------------------------------### */

static struct prb_entry *prb_addent (head, key)

struct prb_entry *head;
char             *key;

  {
  struct prb_entry *entry;
  int               i;

  if (PRB_DCEHED == NULL)
    {
    PRB_DCEHED = (struct prb_entry *)
                 mbkalloc (sizeof(struct prb_entry) * PRB_ALODFN);

    entry = PRB_DCEHED;
    for (i=1 ; i<PRB_ALODFN ; i++)
      {
      entry->next = entry + 1;
      entry++;
      }
    entry->next = NULL;
    }

  entry       = PRB_DCEHED;
  PRB_DCEHED  = PRB_DCEHED->next;

  entry->next = head;
  entry->data = NULL;
  entry->key  = key;

  return (entry);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_addrcd						*/
/* description	: add a new record in a dictionary			*/
/* called func.	: mbkaloc						*/
/* ###--------------------------------------------------------------### */

struct prb_recrd *prb_addrcd (head , key)

struct prb_recrd *head;
char             *key;

  {
  struct prb_recrd *recrd;
  int               i;

  if (PRB_DCRHED == NULL)
    {
    PRB_DCRHED = (struct prb_recrd *)
                 mbkalloc (sizeof(struct prb_recrd) * PRB_ALODFN);

    recrd = PRB_DCRHED;
    for (i=1 ; i<PRB_ALODFN ; i++)
      {
      recrd->next = recrd + 1;
      recrd++;
      }
    recrd->next = NULL;
    }

  recrd           = PRB_DCRHED;
  PRB_DCRHED      = PRB_DCRHED->next;

  recrd->next     = head;
  recrd->key      = key;
  recrd->value    = 0.0;
  recrd->date     = 0;

  return (recrd);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_initab						*/
/* description	: create a new dictionary				*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct prb_entry **prb_initab ()

  {
  struct prb_entry **head;
  int                i;

  head = (struct prb_entry **)
         mbkalloc (sizeof(struct prb_entry *) * PRB_HSZDFN);

  for (i=0 ; i<PRB_HSZDFN ; i++)
    head[i] = NULL;

  return (head);
  }


/* ###--------------------------------------------------------------### */
/* function	: prb_addtab						*/
/* description	: add a new information into a dictionary. If the	*/
/*		  key or the context doesn't exist create a new record	*/
/* called func.	: prb_addrcd, prb_addent				*/
/* ###--------------------------------------------------------------### */

void prb_addtab (head,key_str,ctx_str,value,date)

struct prb_entry **head;
char              *key_str;
char              *ctx_str;
float              value;
int                date;
  {
  int               found = 0;
  int               index;
  struct prb_entry *entry_pnt;
  struct prb_recrd *recrd_pnt;

  index     = (int) key_str % PRB_HSZDFN;
  entry_pnt = head[index];

  while (entry_pnt != NULL)
    {
    if (entry_pnt->key == key_str)
      {
      found = 1;
      break;
      }
    entry_pnt = entry_pnt->next;
    }

  if (found == 0)
    {
    head[index] = prb_addent (head[index],key_str); 
    entry_pnt = head[index]; 
    }

  found = 0;
  recrd_pnt = entry_pnt->data;
  while (recrd_pnt != NULL)
    {
    if (recrd_pnt->key == ctx_str)
      {
      found = 1;
      break;
      }
    recrd_pnt = recrd_pnt->next;
    }

  if (found == 0)
    {
    entry_pnt->data = prb_addrcd (entry_pnt->data,ctx_str); 
    recrd_pnt       = entry_pnt->data ;
    }

  recrd_pnt->date  = date;
  recrd_pnt->value = value;
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_chktab						*/
/* description	: extract an information from a dictionary. Return	*/
/*		  0 if the information has not been found.		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

struct prb_recrd *prb_chktab (head,key_str,ctx_str)

struct prb_entry **head;
char              *key_str;
char              *ctx_str;

  {
  int               found = 0;
  struct prb_entry *entry_pnt = NULL;
  struct prb_recrd *recrd_pnt = NULL;

  entry_pnt = head [(int)key_str % PRB_HSZDFN];

  while (entry_pnt != NULL)
    {
    if (entry_pnt->key == key_str)
      {
      found = 1;
      break;
      }
    entry_pnt = entry_pnt->next;
    }

  if (found == 1)
    {
    found = 0;
    recrd_pnt = entry_pnt->data;
    while (recrd_pnt != NULL)
      {
      if (recrd_pnt->key == ctx_str)
        {
        found = 1;
        break;
        }
      recrd_pnt = recrd_pnt->next;
      }
    }
  return (recrd_pnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_fretab						*/
/* description	: remove entirely a dictionary				*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void prb_fretab (pt_hash)

struct prb_entry **pt_hash;
  {
  struct prb_entry *pt_entry;
  struct prb_entry *pt_nxtentry;
  struct prb_recrd *pt_record;
  int               i;

  if (pt_hash != NULL)
    {
    for (i=0 ; i<PRB_HSZDFN ; i++)
      {
      if ((pt_entry = pt_hash[i]) != NULL)
        {
        while (pt_entry != NULL)
          {
          pt_record = pt_entry->data;

          while (pt_record->next != NULL)
            pt_record = pt_record->next;

          pt_record->next = PRB_DCRHED;
          PRB_DCRHED      = pt_entry->data;

          pt_nxtentry     = pt_entry->next;
          pt_entry->next  = PRB_DCEHED;
          PRB_DCEHED      = pt_entry;
          pt_entry        = pt_nxtentry;
          }
        }
      }
    mbkfree (pt_hash);
    }
  }

