
/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: prb_maksupbdd.c					*/
/* date		: Jan 20 1998						*/
/* version	: v107							*/
/* authors	: DUNOYER Julien					*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include "beh.h"
#include "bhl.h"
#include PRB_H

/* La limitation du nombre de noeuds BDD crees est desactivee */
/* ainsi que la fonction prb_abl2bdd                          */

static unsigned int BDD_CREATED_NODES = 0;
static unsigned int BDD_MAX_NODES = 2000;

/* ###--------------------------------------------------------------### */
/* function	: prb_abl2bdd						*/
/* description	: transform recursively an abl into a bdd 		*/
/* called func.	: searchInputCct, createNodeTermBdd, addListBdd,	*/
/*		  prb_abl2bdd   , applyBdd         , freechain ,	*/
/*		  prb_error     ,					*/
/* ###--------------------------------------------------------------### */

static pNode prb_abl2bdd (pt_befig, expr)

struct befig *pt_befig;		/* pointer on  behavioural figure	*/
struct chain *expr    ;		/* pointer abl expression		*/

  {
  short         oper        ;
  struct node  *pt_bdd      = zero;
  struct chain *lstGbd      = NULL;
  pCircuit      ptr_circuit = pt_befig->CIRCUI;
  int           err_flg     = 0;
  int           index       ;
  char         *atom        ;
  char          name [256]  ;
  unsigned int  initial_total_nodes = 0;
  unsigned int  initial_created_nodes = 0;


if (BDD_CREATED_NODES >= BDD_MAX_NODES)
  {
  return (pt_bdd);
  }

initial_created_nodes = BDD_CREATED_NODES;
initial_total_nodes   = BDD_CREATED_NODES;

	/* ###------------------------------------------------------### */
	/*    check that the expression exists				*/
	/* ###------------------------------------------------------### */

  if (expr == NULL)
    err_flg = beh_error (2, NULL);
  else
    {

    if (ATOM(expr))
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is a terminal (atom)			*/
	/* ###------------------------------------------------------### */

      atom = (char *) VALUE_ATOM (expr);

	/* ###------------------------------------------------------### */
	/*    if the terminal is a primary signal create a single node	*/
	/* with its index and return					*/
	/* ###------------------------------------------------------### */

      if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
        {
        pt_bdd = createNodeTermBdd (index);
        BDD_CREATED_NODES ++;
        }
      else
        {
	/* ###------------------------------------------------------### */
	/* the terminal should be a const. '0', '1' or 'd' (don't care)	*/
	/* ###------------------------------------------------------### */

        if ((!strcmp(atom, "'0'")) || (!strcmp(atom, "'d'")))
          pt_bdd = zero;
        else
          {
          if (!strcmp(atom, "'1'"))
            pt_bdd = one;
          else
            {
            err_flg = beh_error (3, atom);
            }
          }
        }
      }
    else
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is not a terminal (is a complex expr.)	*/
	/* get the operator, create the bdd of each operand then,	*/
	/* combine operands applying the operator			*/
	/* ###------------------------------------------------------### */

      lstGbd = NULL;

      if ((oper = OPER(expr)) != STABLE)
       {
       while ((expr = CDR(expr)) != NULL)
        {
        lstGbd = addListBdd (lstGbd, prb_abl2bdd (pt_befig, CAR(expr)));
        }

       pt_bdd = applyBdd (oper, lstGbd);
       freechain (lstGbd);
       }
      else
       {
	expr = CDR(expr);
	if (!ATOM(expr))
	  err_flg = beh_error (4, NULL);
        else
          {
	  atom = (char *)VALUE_ATOM (CAR(expr));

          if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
            {
            lstGbd = addListBdd (lstGbd, createNodeTermBdd (index));
	    sprintf (name, "%s'delayed", atom);
	    atom = namealloc (name);
            if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
              {
              lstGbd = addListBdd (lstGbd, createNodeTermBdd (index));
              pt_bdd = applyBdd (NXOR, lstGbd);
              }
            else
	      err_flg = beh_error (3, atom);
            freechain (lstGbd);
            }
          else
	    err_flg = beh_error (3, atom);
          }
        }
      }
    }

  if (err_flg != 0)
    pt_befig->ERRFLG = 1;

  return (pt_bdd);
  }


/* ###--------------------------------------------------------------### */
/* function	: prb_msupbddtot					*/
/* description	: make bdds for all primary signals: simple and bussed	*/
/*		  outputs, simple and bussed internal signals and	*/
/*		  registers						*/
/* called func.	: prb_abl2bdd, addchain , delchain,			*/
/* ###--------------------------------------------------------------### */

static void prb_msupbddtot (ptr_befig)

struct befig *ptr_befig;		/* pointer on current BEFIG	*/

  {
  struct beaux  *ptr_beaux;
  struct beaux  *ptr_bedly;
  struct bemsg  *ptr_bemsg;
  struct beout  *ptr_beout;
  struct bebus  *ptr_bebus;
  struct bebux  *ptr_bebux;
  struct bereg  *ptr_bereg;
  struct binode *ptr_binode;

	/* ###------------------------------------------------------### */
	/*    make a bdd for each simple internal signal		*/
	/* ###------------------------------------------------------### */

  ptr_beaux = ptr_befig->BEAUX;

  while (ptr_beaux != NULL)
    {
    BDD_CREATED_NODES = 0;

    if (PRB_SUPERBDD(ptr_beaux->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(ptr_beaux->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERBDD(ptr_beaux->NODE) =
            beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_beaux->NODE));

        if (BDD_CREATED_NODES >= BDD_MAX_NODES)
          {
          printf("BDD_MAX_NODES depasse avec %s\n", ptr_beaux->NAME);
          PRB_SUPERFLAG(ptr_beaux->NODE) = DEF_PRB_DEFGATE;
          PRB_SUPERBDD(ptr_beaux->NODE) = PRB_BDD(ptr_beaux->NODE);
          }
        }
      else
        {
        PRB_SUPERBDD(ptr_beaux->NODE) = PRB_BDD(ptr_beaux->NODE);
        }
      }
    ptr_beaux = ptr_beaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each delayed internal signal		*/
	/* ###------------------------------------------------------### */

  ptr_bedly = ptr_befig->BEDLY;
  while (ptr_bedly != NULL)
    {
    BDD_CREATED_NODES = 0;

    if (PRB_SUPERBDD(ptr_bedly->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(ptr_bedly->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERBDD(ptr_bedly->NODE) =
            beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_bedly->NODE));

        if (BDD_CREATED_NODES >= BDD_MAX_NODES)
          {
          PRB_SUPERFLAG(ptr_bedly->NODE) = DEF_PRB_DEFGATE;
          PRB_SUPERBDD(ptr_bedly->NODE) = PRB_BDD(ptr_bedly->NODE);
          }
        }
      else
        {
        PRB_SUPERBDD(ptr_bedly->NODE) = PRB_BDD(ptr_bedly->NODE);
        }
      }

    ptr_bedly = ptr_bedly->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each assertion				*/
	/* ###------------------------------------------------------### */

  ptr_bemsg = ptr_befig->BEMSG;
  while (ptr_bemsg != NULL)
    {
    BDD_CREATED_NODES = 0;

    if (PRB_SUPERBDD(ptr_bemsg->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(ptr_bemsg->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERBDD(ptr_bemsg->NODE) =
            beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_bemsg->NODE));

        if (BDD_CREATED_NODES >= BDD_MAX_NODES)
          {
          PRB_SUPERFLAG(ptr_bemsg->NODE) = DEF_PRB_DEFGATE;
          PRB_SUPERBDD(ptr_bemsg->NODE) = PRB_BDD(ptr_bemsg->NODE);
          }
        }
      else
        {
        PRB_SUPERBDD(ptr_bemsg->NODE) = PRB_BDD(ptr_bemsg->NODE);
        }
      }

    ptr_bemsg = ptr_bemsg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each simple output port			*/
	/* ###------------------------------------------------------### */

  ptr_beout = ptr_befig->BEOUT;
  while (ptr_beout != NULL)
    {
    BDD_CREATED_NODES = 0;

    if (PRB_SUPERBDD(ptr_beout->NODE) == NULL)
      {
      if (PRB_SUPERFLAG(ptr_beout->NODE) == DEF_PRB_SUPERGATE)
        {
        PRB_SUPERBDD(ptr_beout->NODE) =
            beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_beout->NODE));

        if (BDD_CREATED_NODES >= BDD_MAX_NODES)
          {
          printf("BDD_MAX_NODES depasse avec %s\n", ptr_beout->NAME);
          PRB_SUPERFLAG(ptr_beout->NODE) = DEF_PRB_DEFGATE;
          PRB_SUPERBDD(ptr_beout->NODE) = PRB_BDD(ptr_beout->NODE);
          }
        }
      else
        {
        PRB_SUPERBDD(ptr_beout->NODE) = PRB_BDD(ptr_beout->NODE);
        }
      }

    ptr_beout = ptr_beout->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each bussed output port			*/
	/* ###------------------------------------------------------### */

  ptr_bebus = ptr_befig->BEBUS;
  while (ptr_bebus != NULL)
    {
    ptr_binode = ptr_bebus->BINODE;

    while (ptr_binode != NULL)
      {
      BDD_CREATED_NODES = 0;

      if (PRB_SUPERBDD(ptr_binode->CNDNODE) == NULL)
        {
        if (PRB_SUPERFLAG(ptr_binode->CNDNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERBDD(ptr_binode->CNDNODE) =
              beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_binode->CNDNODE));

          if (BDD_CREATED_NODES >= BDD_MAX_NODES)
            {
            PRB_SUPERFLAG(ptr_binode->CNDNODE) = DEF_PRB_DEFGATE;
            PRB_SUPERBDD(ptr_binode->CNDNODE) = PRB_BDD(ptr_binode->CNDNODE);
            }
          }
        else
          {
          PRB_SUPERBDD(ptr_binode->CNDNODE) = PRB_BDD(ptr_binode->CNDNODE);
          }
        }

      BDD_CREATED_NODES = 0;

      if (PRB_SUPERBDD(ptr_binode->VALNODE) == NULL)
        {
        if (PRB_SUPERFLAG(ptr_binode->VALNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERBDD(ptr_binode->VALNODE) =
              beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_binode->VALNODE));

          if (BDD_CREATED_NODES >= BDD_MAX_NODES)
            {
            PRB_SUPERFLAG(ptr_binode->VALNODE) = DEF_PRB_DEFGATE;
            PRB_SUPERBDD(ptr_binode->VALNODE) = PRB_BDD(ptr_binode->VALNODE);
            }
          }
        else
          {
          PRB_SUPERBDD(ptr_binode->VALNODE) = PRB_BDD(ptr_binode->VALNODE);
          }
        }

      ptr_binode = ptr_binode->NEXT;
      }

    ptr_bebus = ptr_bebus->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each bussed internal signal		*/
	/* ###------------------------------------------------------### */

  ptr_bebux = ptr_befig->BEBUX;
  while (ptr_bebux != NULL)
    {
    ptr_binode = ptr_bebux->BINODE;

    while (ptr_binode != NULL)
      {
      BDD_CREATED_NODES = 0;

      if (PRB_SUPERBDD(ptr_binode->CNDNODE) == NULL)
        {
        if (PRB_SUPERFLAG(ptr_binode->CNDNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERBDD(ptr_binode->CNDNODE) =
              beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_binode->CNDNODE));

          if (BDD_CREATED_NODES >= BDD_MAX_NODES)
            {
            PRB_SUPERFLAG(ptr_binode->CNDNODE) = DEF_PRB_DEFGATE;
            PRB_SUPERBDD(ptr_binode->CNDNODE) = PRB_BDD(ptr_binode->CNDNODE);
            }
          }
        else
          {
          PRB_SUPERBDD(ptr_binode->CNDNODE) = PRB_BDD(ptr_binode->CNDNODE);
          }
        }

      BDD_CREATED_NODES = 0;

      if (PRB_SUPERBDD(ptr_binode->VALNODE) == NULL)
        {
        if (PRB_SUPERFLAG(ptr_binode->VALNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERBDD(ptr_binode->VALNODE) =
              beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_binode->VALNODE));

          if (BDD_CREATED_NODES >= BDD_MAX_NODES)
            {
            PRB_SUPERFLAG(ptr_binode->VALNODE) = DEF_PRB_DEFGATE;
            PRB_SUPERBDD(ptr_binode->VALNODE) = PRB_BDD(ptr_binode->VALNODE);
            }
          }
        else
          {
          PRB_SUPERBDD(ptr_binode->VALNODE) = PRB_BDD(ptr_binode->VALNODE);
          }
        }

      ptr_binode = ptr_binode->NEXT;
      }
    ptr_bebux = ptr_bebux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each internal register			*/
	/* ###------------------------------------------------------### */

  ptr_bereg = ptr_befig->BEREG;
  while (ptr_bereg != NULL)
    {
    ptr_binode = ptr_bereg->BINODE;

    while (ptr_binode != NULL)
      {
      BDD_CREATED_NODES = 0;

      if (PRB_SUPERBDD(ptr_binode->CNDNODE) == NULL)
        {
        if (PRB_SUPERFLAG(ptr_binode->CNDNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERBDD(ptr_binode->CNDNODE) =
              beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_binode->CNDNODE));

          if (BDD_CREATED_NODES >= BDD_MAX_NODES)
            {
            PRB_SUPERFLAG(ptr_binode->CNDNODE) = DEF_PRB_DEFGATE;
            PRB_SUPERBDD(ptr_binode->CNDNODE) = PRB_BDD(ptr_binode->CNDNODE);
            }
          }
        else
          {
          PRB_SUPERBDD(ptr_binode->CNDNODE) = PRB_BDD(ptr_binode->CNDNODE);
          }
        }

      BDD_CREATED_NODES = 0;

      if (PRB_SUPERBDD(ptr_binode->VALNODE) == NULL)
        {
        if (PRB_SUPERFLAG(ptr_binode->VALNODE) == DEF_PRB_SUPERGATE)
          {
          PRB_SUPERBDD(ptr_binode->VALNODE) =
              beh_abl2bdd (ptr_befig, PRB_SUPERABL(ptr_binode->VALNODE));

          if (BDD_CREATED_NODES >= BDD_MAX_NODES)
            {
            PRB_SUPERFLAG(ptr_binode->VALNODE) = DEF_PRB_DEFGATE;
            PRB_SUPERBDD(ptr_binode->VALNODE) = PRB_BDD(ptr_binode->VALNODE);
            }
          }
        else
          {
          PRB_SUPERBDD(ptr_binode->VALNODE) = PRB_BDD(ptr_binode->VALNODE);
          }
        }

      ptr_binode = ptr_binode->NEXT;
      }
    ptr_bereg = ptr_bereg->NEXT;
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: prb_maksupbdd						*/
/* description	: make supbdd for each signal of  a befig structure 	*/
/* called func.	: initializeCct, addInputCct, prb_mbddtot		*/
/* ###--------------------------------------------------------------### */

void prb_maksupbdd (ptr_befig)

struct befig *ptr_befig;

  {
  berin_list *ptr_rin;
  beout_list *ptr_out;
  bebus_list *ptr_bus;
  int         count_rin = 0;
  int         count_out = 0;
  int         err_flg   = 0;


  if ((!(ptr_befig->TYPE & BEH_NODE_BDD ))
  ||  (!(ptr_befig->TYPE & BEH_NODE_QUAD))
  ||  (!(ptr_befig->TYPE & BEH_NODE_USER)))
    {
    prb_toolbug (4, "prb_maksupbdd", ' ');
    }


        /* ###------------------------------------------------------### */
        /*    Change ABLs to BDDs					*/
        /* ###------------------------------------------------------### */

  prb_msupbddtot (ptr_befig);


  if (err_flg != 0)
    {
    ptr_befig->ERRFLG = 1;
    }
  }
