/* ###--------------------------------------------------------------### */
/* file		: prb_maksupderiv.c					*/
/* date		: Jan 24 1998						*/
/* version	: v111							*/
/* authors	: Julien DUNOYER					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H
#include LOG_H
#include "beh.h"
#include "bhl.h"
#include PRB_H


/* ###--------------------------------------------------------------### */
/* function     : prb_derivSig                                          */
/* description  : Interface between prb world and beh world             */
/* called func. : beh_derivSig                                          */
/* ###--------------------------------------------------------------### */

static struct beder *prb_derivSig (pt_bdd, pt_varlist, time, mode)

pNode          pt_bdd      ;
struct chain  *pt_varlist  ;
unsigned int   time        ;
unsigned char  mode        ;

  {
  unsigned char  beh_mode = 0x00 ;

  if (mode & DEF_PRB_DERIVBYVAR)
    {
    beh_mode |= BEH_NODE_DERIVBYVAR;
    }

  if (mode & DEF_PRB_DERIVBYTIME)
    {
    beh_mode |= BEH_NODE_DERIVBYTIME;
    }

  return (beh_derivSig (pt_bdd, pt_varlist, time, beh_mode));
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_maksupderiv					*/
/* description	: build the list of supderivate for all signals		*/
/*                the mode is used to make a choice between		*/
/*                supderivate by time or by all variables.		*/
/* called func.	: prb_derivSig						*/
/* ###--------------------------------------------------------------### */

void prb_maksupderiv (pt_befig, mode)

struct befig    *pt_befig;
unsigned char    mode;

  {
  struct beaux  *pt_beaux;
  struct beout  *pt_beout;
  struct bebus  *pt_bebus;
  struct bebux  *pt_bebux;
  struct bereg  *pt_bereg;
  struct bemsg  *pt_bemsg;
  struct binode *pt_binod;
  struct chain  *list_der;


	/* ###------------------------------------------------------### */
	/*   check that the BEFIG is ready to support the derivation	*/
	/* ###------------------------------------------------------### */

  if ((pt_befig->TYPE & BEH_NODE_QUAD)
  &&  (pt_befig->TYPE & BEH_NODE_USER))
    {

	/* ###------------------------------------------------------### */
	/*   simple output ports					*/
	/* ###------------------------------------------------------### */

    pt_beout = pt_befig->BEOUT;
    while (pt_beout != NULL)
      {
      PRB_SUPERDERIVATE (pt_beout->NODE) =
             prb_derivSig (PRB_SUPERBDD (pt_beout->NODE),
                PRB_SUPERVARLIST (pt_beout->NODE), pt_beout->TIME, mode);

      PRB_SUPERFLAG(pt_beout->NODE) |= mode;

      pt_beout = pt_beout->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   bussed output ports					*/
	/* ###------------------------------------------------------### */

    pt_bebus = pt_befig->BEBUS;
    while (pt_bebus != NULL)
      {
      pt_binod = pt_bebus->BINODE;
      while (pt_binod != NULL)
         {
         PRB_SUPERDERIVATE (pt_binod->CNDNODE) =
              prb_derivSig (PRB_SUPERBDD (pt_binod->CNDNODE),
                 PRB_SUPERVARLIST (pt_binod->CNDNODE), pt_binod->TIME, mode);

         PRB_SUPERFLAG(pt_binod->CNDNODE) |= mode;

         PRB_SUPERDERIVATE (pt_binod->VALNODE) =
             prb_derivSig (PRB_SUPERBDD (pt_binod->VALNODE),
                 PRB_SUPERVARLIST (pt_binod->VALNODE), pt_binod->TIME, mode);

         PRB_SUPERFLAG(pt_binod->VALNODE) |= mode;

         pt_binod = pt_binod->NEXT;
         }

      pt_bebus = pt_bebus->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   bussed internal signals					*/
	/* ###------------------------------------------------------### */

    pt_bebux = pt_befig->BEBUX;
    while (pt_bebux != NULL)
      {
      pt_binod = pt_bebux->BINODE;
      while (pt_binod != NULL)
         {
         PRB_SUPERDERIVATE (pt_binod->CNDNODE) =
              prb_derivSig (PRB_SUPERBDD (pt_binod->CNDNODE),
                 PRB_SUPERVARLIST (pt_binod->CNDNODE), pt_binod->TIME, mode);

         PRB_SUPERFLAG(pt_binod->CNDNODE) |= mode;

         PRB_SUPERDERIVATE (pt_binod->VALNODE) =
             prb_derivSig (PRB_SUPERBDD (pt_binod->VALNODE),
                 PRB_SUPERVARLIST (pt_binod->VALNODE), pt_binod->TIME, mode);

         PRB_SUPERFLAG(pt_binod->VALNODE) |= mode;

         pt_binod = pt_binod->NEXT;
         }
      pt_bebux = pt_bebux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   internal registers						*/
	/* ###------------------------------------------------------### */

    pt_bereg = pt_befig->BEREG;
    while (pt_bereg != NULL)
      {
      pt_binod = pt_bereg->BINODE;
      while (pt_binod != NULL)
         {
         PRB_SUPERDERIVATE (pt_binod->CNDNODE) =
              prb_derivSig (PRB_SUPERBDD (pt_binod->CNDNODE),
                 PRB_SUPERVARLIST (pt_binod->CNDNODE), pt_binod->TIME, mode);

         PRB_SUPERFLAG(pt_binod->CNDNODE) |= mode;

         PRB_SUPERDERIVATE (pt_binod->VALNODE) =
             prb_derivSig (PRB_SUPERBDD (pt_binod->VALNODE),
                 PRB_SUPERVARLIST (pt_binod->VALNODE), pt_binod->TIME, mode);

         PRB_SUPERFLAG(pt_binod->VALNODE) |= mode;

         pt_binod = pt_binod->NEXT;
         }
      pt_bereg = pt_bereg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   internal simple signals					*/
	/* ###------------------------------------------------------### */

    pt_beaux = pt_befig->BEAUX;
    while (pt_beaux != NULL)
      {
      PRB_SUPERDERIVATE (pt_beaux->NODE) =
             prb_derivSig (PRB_SUPERBDD (pt_beaux->NODE),
                PRB_SUPERVARLIST (pt_beaux->NODE), pt_beaux->TIME, mode);

      PRB_SUPERFLAG(pt_beaux->NODE) |= mode;

      pt_beaux = pt_beaux->NEXT;
      }


	/* ###------------------------------------------------------### */
	/*   delayed internal signals					*/
	/* ###------------------------------------------------------### */

    pt_beaux = pt_befig->BEDLY;
    while (pt_beaux != NULL)
      {
      PRB_SUPERDERIVATE (pt_beaux->NODE) =
             prb_derivSig (PRB_SUPERBDD (pt_beaux->NODE),
                PRB_SUPERVARLIST (pt_beaux->NODE), pt_beaux->TIME, mode);

      PRB_SUPERFLAG(pt_beaux->NODE) |= mode;

      pt_beaux = pt_beaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   assertions							*/
	/* ###------------------------------------------------------### */

    pt_bemsg = pt_befig->BEMSG;
    while (pt_bemsg != NULL)
      {
      PRB_SUPERDERIVATE (pt_bemsg->NODE) =
             prb_derivSig (PRB_SUPERBDD (pt_bemsg->NODE),
                PRB_SUPERVARLIST (pt_bemsg->NODE), 0, mode);

      PRB_SUPERFLAG(pt_bemsg->NODE) |= mode;

      pt_bemsg = pt_bemsg->NEXT;
      }
    }

  else
    {
    prb_error (0, NULL);
    }
  }
