
/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: prb_maksupgates.c					*/
/* date		: May 13 1995						*/
/* version	: v000							*/
/* authors	: DUNOYER Julien					*/
/* content	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include MUT_H
#include MLO_H
#include "beh.h"
#include "bhl.h"
#include LOG_H
#include COX_H
#include GSS_H
#include PRB_H

/* ###--------------------------------------------------------------### */
/* function     : prb_maksupgates  					*/
/* description  : Make a superabl, superbdd, and a supervarlist for all	*/
/*                signals of a befig					*/
/* called func  : prb_superabl,cox_simulator				*/
/* ###--------------------------------------------------------------### */

void prb_maksupgates (ptr_lofig, ptr_befig)

struct lofig *ptr_lofig;
struct befig *ptr_befig;
  {
  struct dct_entry  **spg_dico;
  struct dct_entry  **cox_dico;

  struct bemsg       *ptr_bemsg ;
  struct beaux       *ptr_beaux ;
  struct beaux       *ptr_bedly;
  struct beout       *ptr_beout ;
  struct bereg       *ptr_bereg;
  struct binode      *ptr_binode;
  struct biabl       *ptr_biabl;
  char               *name;

  struct rusage       rusage1, rusage2;
  double              delta_utime = 0.0;

	/* ###------------------------------------------------------### */
	/*    prb_maksupabl use the USER fields of NODEs.		*/
	/*    NODES may be of QUAD type.				*/
	/* ###------------------------------------------------------### */

  if ( (ptr_befig->TYPE & BEH_NODE_USER)
  || (!(ptr_befig->TYPE & BEH_NODE_QUAD)))
    {
    prb_error(0, NULL);
    }

  else
    {
    COX_DICTIONNARY = gss_initab();

	/* ###------------------------------------------------------### */
	/*    Record each BASIC ABL in the COX_DICTIONNARY.		*/
	/* ###------------------------------------------------------### */

    ptr_beaux = ptr_befig->BEAUX;
    while (ptr_beaux != NULL)
      {
      name = ptr_beaux->NAME;

      gss_addtab (COX_DICTIONNARY, name, COX_BASICABL_DEF,
                  GSS_PNT0FIELD, ptr_beaux->ABL);

      PRB_SUPERABL(ptr_beaux->NODE) = ptr_beaux->ABL ;
      PRB_SUPERBDD(ptr_beaux->NODE) = PRB_BDD(ptr_beaux->NODE) ;
      PRB_SUPERVARLIST(ptr_beaux->NODE) = PRB_VARLIST(ptr_beaux->NODE) ;

      ptr_beaux       = ptr_beaux->NEXT;
      }

    ptr_beaux = ptr_befig->BEDLY;
    while (ptr_beaux != NULL)
      {
      name = ptr_beaux->NAME;

      gss_addtab (COX_DICTIONNARY, name, COX_BASICABL_DEF,
                  GSS_PNT0FIELD, ptr_beaux->ABL);

      PRB_SUPERABL(ptr_beaux->NODE) = ptr_beaux->ABL ;
      PRB_SUPERBDD(ptr_beaux->NODE) = PRB_BDD(ptr_beaux->NODE) ;
      PRB_SUPERVARLIST(ptr_beaux->NODE) = PRB_VARLIST(ptr_beaux->NODE) ;

      ptr_beaux       = ptr_beaux->NEXT;
      }

    ptr_beout = ptr_befig->BEOUT;
    while (ptr_beout != NULL)
      {
      name = ptr_beout->NAME;

      gss_addtab(COX_DICTIONNARY,name,COX_BASICABL_DEF,
                 GSS_PNT0FIELD,ptr_beout->ABL);

      PRB_SUPERABL(ptr_beout->NODE) = ptr_beout->ABL ;
      PRB_SUPERBDD(ptr_beout->NODE) = PRB_BDD(ptr_beout->NODE) ;
      PRB_SUPERVARLIST(ptr_beout->NODE) = PRB_VARLIST(ptr_beout->NODE) ;
      ptr_beout       = ptr_beout->NEXT;
      }

    ptr_bereg =  ptr_befig->BEREG;
    while (ptr_bereg != NULL)
      {
      ptr_binode = ptr_bereg->BINODE;
      ptr_biabl  = ptr_bereg->BIABL ;
      while ((ptr_binode != NULL) && (ptr_biabl != NULL))
        {
        PRB_SUPERABL(ptr_binode->CNDNODE) = ptr_biabl->CNDABL ;
        PRB_SUPERBDD(ptr_binode->CNDNODE) = PRB_BDD(ptr_binode->CNDNODE) ;
        PRB_SUPERVARLIST(ptr_binode->CNDNODE) = PRB_VARLIST(ptr_binode->CNDNODE) ;
        PRB_SUPERABL(ptr_binode->VALNODE) = ptr_biabl->VALABL ;
        PRB_SUPERBDD(ptr_binode->VALNODE) = PRB_BDD(ptr_binode->VALNODE) ;
        PRB_SUPERVARLIST(ptr_binode->VALNODE) = PRB_VARLIST(ptr_binode->VALNODE) ;
        
        ptr_binode = ptr_binode->NEXT;
        ptr_biabl  = ptr_biabl->NEXT;
        }
      ptr_bereg =  ptr_bereg->NEXT;
      }
	/* ###------------------------------------------------------### */
	/*    Complete the Cox simulation => SuperAbl,			*/
	/*                                   SuperBdd,			*/
	/*                                   SuperVarlist.		*/
	/* ###------------------------------------------------------### */

    GSS_MAXDELTA = 100000;

    getrusage(RUSAGE_SELF,&rusage1);

    cox_simulator(ptr_lofig, ptr_befig);

    getrusage(RUSAGE_SELF,&rusage2);

    delta_utime  =
           (rusage2.ru_utime.tv_sec*1000.0+rusage2.ru_utime.tv_usec/1000.0)
         - (rusage1.ru_utime.tv_sec*1000.0+rusage1.ru_utime.tv_usec/1000.0);

    fprintf(stdout, "TEMPS DE COX SIMULATION\n");
    fprintf(stdout, "  +   DE COX EXPANSION  : %.2f\n",delta_utime);

    ptr_befig->TYPE |= BEH_NODE_USER;
    }
  }
