
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: prb_package.c						*/
/* date		: Jun  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER Julien					*/
/* description	: This file contains the set of probability dedicated	*/
/*                functions to link with generic symbolic simulator 	*/
/*		  prb_addvalue , prb_freevalue, prb_cpyvalue,		*/
/*		  prb_cmpvalue , prb_evalvalue, prb_resolvalue,		*/
/*									*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/time.h>

#include MUT_H
#include LOG_H
#include "beh.h"
#include "bhl.h"
#include GSS_H
#include PRB_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */

static struct prb_entry **local_dic;

  static unsigned long count_hit  =0;
  static unsigned long count_miss =0;

/* ###--------------------------------------------------------------### */
/* function	: prb_addvalue						*/
/* description	: addition function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void *prb_addvalue ()

  {
  struct prbval *value;

  value = (struct prbval *) mbkalloc (sizeof(struct prbval));
  value->PROBVAL = DEF_PRB_DEFPROB;
  value->DISCVAL = DEF_PRB_DEFDISC;
  value->DENSVAL = DEF_PRB_DEFDENS;
  value->DENSSUM = DEF_PRB_DEFDENS;
  value->EVALUATED = PRB_NOT_EVALUATED;

  return((void*)value);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_freevalue						*/
/* description	: liberation function for values			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void prb_freevalue (value)

  float *value;
  {
  if (value != NULL)
     {
     mbkfree (value);
     }
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_cmpvalue						*/
/* description	: comparaison function for values			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_cmpvalue (value1,value2)

  struct prbval    *value1;
  struct prbval    *value2;

  {
/*-- une solution pour gerer le mode levelized avec init des registres --*/

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    if ( value1->EVALUATED != PRB_NOT_EVALUATED )
      {
      return(0);
      }
    else
      {
      if ( value2->EVALUATED == PRB_NOT_EVALUATED )
        {
        return(0);
        }
      else
        {
        return(1);
        }
      }
    }

  if ((value1->PROBVAL  != value2->PROBVAL ) ||
      (value1->DENSVAL  != value2->DENSVAL )    )
    {
    return(1);
    }
  else
    {
    return(0);
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_cpyvalue						*/
/* description	: copy function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_cpyvalue (value1,value2)

  struct prbval  *value1;
  struct prbval  *value2;

  {
  int errflg = 0;

  value1->PROBVAL = value2->PROBVAL;
  value1->DENSVAL = value2->DENSVAL;

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    value1->EVALUATED = value2->EVALUATED;
    }

  return(errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_updvalue						*/
/* description	: copy function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_updvalue (value1,value2)

  struct prbval  *value1;
  struct prbval  *value2;

  {
  int errflg = 0;

  value1->PROBVAL = value2->PROBVAL;
  value1->DENSVAL = value2->DENSVAL;

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    value1->EVALUATED = value2->EVALUATED;
    }

  return(errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_resolvalue					*/
/* description	: resolution function for typed signals			*/
/*		  return flag:0 ok, 1 error				*/
/*		  return default value 0				*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

int prb_resolvalue (projected_bivalue,
                    resolution_function_type,
                    current_value,
                    result_value)

  struct prjbvl *projected_bivalue;
  char		 resolution_function_type;
  struct prbval *current_value;
  struct prbval *result_value;

  {
  struct prjbvl *tmp_bivalue;
  int  errflg = 0;
  int  nbdriver = 0;

/*   ces 3 lignes doivent etre virees le jour ou on */
/*   veut traiter bien les choses                   */
/*---
    result_value->EVALUATED = PRB_EVALUATED;
    result_value->DENSVAL   = current_value->DENSVAL;
    result_value->PROBVAL   = current_value->PROBVAL;
    return(errflg);
-----*/

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    for ( tmp_bivalue = projected_bivalue->NEXT;
          tmp_bivalue != NULL ;
          tmp_bivalue = tmp_bivalue->NEXT)
      {
      if ( (((struct prbval *)(tmp_bivalue->CNDVAL))->EVALUATED
                                                     == PRB_NOT_EVALUATED) ||
           (((struct prbval *)(tmp_bivalue->DRVVAL))->EVALUATED
                                                     == PRB_NOT_EVALUATED) )
        {
        return(errflg);
        }
      }
    result_value->EVALUATED = PRB_EVALUATED;
    }

  switch (resolution_function_type)
    {
    case 'R' :
      result_value->PROBVAL = 0.0;
      result_value->DENSVAL = 0.0;
      nbdriver = 0;
      projected_bivalue = projected_bivalue->NEXT;
      while (projected_bivalue != NULL)
        {
        result_value->PROBVAL +=
        ((struct prbval *)(projected_bivalue->DRVVAL))->PROBVAL;

        result_value->DENSVAL +=
        ((struct prbval *)(projected_bivalue->DRVVAL))->DENSVAL;

        nbdriver += 1;
        projected_bivalue = projected_bivalue->NEXT;
        }

      result_value->PROBVAL = result_value->PROBVAL / nbdriver;
      result_value->DENSVAL = result_value->DENSVAL / nbdriver;

      if ( fabs(result_value->PROBVAL - current_value->PROBVAL) < 0.01)
        {
        result_value->PROBVAL = current_value->PROBVAL;
        }

      if ( fabs(result_value->DENSVAL - current_value->DENSVAL) < 0.01)
        {
        result_value->DENSVAL = current_value->DENSVAL;
        }
      break;

    case 'W' :
      result_value->PROBVAL = 0.0;
      result_value->DENSVAL = 0.0;
      projected_bivalue = projected_bivalue->NEXT;
      while (projected_bivalue != NULL)
        {
        result_value->PROBVAL +=
                ((struct prbval *)(projected_bivalue->CNDVAL))->PROBVAL
       * (1.0 - ((struct prbval *)(projected_bivalue->DRVVAL))->PROBVAL);

        result_value->DENSVAL +=
              ((struct prbval *)(projected_bivalue->CNDVAL))->PROBVAL
            * ((struct prbval *)(projected_bivalue->DRVVAL))->DENSVAL;

        projected_bivalue = projected_bivalue->NEXT;
        }

      result_value->PROBVAL = 1.0 - result_value->PROBVAL;
      break;

    case 'M' :
      result_value->PROBVAL = 0.0;
      result_value->DENSVAL = 0.0;
      projected_bivalue = projected_bivalue->NEXT;
      while (projected_bivalue != NULL)
        {
        result_value->PROBVAL +=
               ((struct prbval *)(projected_bivalue->CNDVAL))->PROBVAL
       * (1.0- ((struct prbval *)(projected_bivalue->DRVVAL))->PROBVAL);

        result_value->DENSVAL +=
              ((struct prbval *)(projected_bivalue->CNDVAL))->PROBVAL
            * ((struct prbval *)(projected_bivalue->DRVVAL))->DENSVAL;

        projected_bivalue = projected_bivalue->NEXT;
        }

      result_value->PROBVAL = 1.0 - result_value->PROBVAL;
      break;

    default :
      prb_toolbug (2,"prb_resolvalue",resolution_function_type);
      break;
    }
  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_bddevalvalue					*/
/* description	: computes a BDD's value depending on its node's value	*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

static float prb_bddevalvalue (instance , bdd_expression, date)

struct lkdins *instance;
struct node   *bdd_expression;
unsigned long  date;
  {
  short              indexbdd;
  float              value;
  float              lowvalue = 0.0;
  float              highvalue = 0.0;
  struct prb_recrd  *record = NULL;

  if (bdd_expression == zero) 
    {
    return (0.0);
    }
  else if (bdd_expression == one) 
    {
    return (1.0);
    }

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    record = prb_chktab (local_dic, bdd_expression, instance);

    if ((record != NULL) && (record->date == date)) 
      {
      count_hit ++;
      return (record->value);
      }

    count_miss ++;
    }

  indexbdd = bdd_expression->index ;

  value = ((struct prbval *) gss_getvalue(instance, indexbdd))->PROBVAL;

  if ( value != 1.0)
    {
    lowvalue = prb_bddevalvalue(instance,
                                bdd_expression->low,
                                date);

    lowvalue = (1.0 - value) * lowvalue;
    }

  if ( value != 0.0)
    {
    highvalue = prb_bddevalvalue(instance,
                                 bdd_expression->high,
                                 date);

    highvalue = value * highvalue;
    }

  value = lowvalue + highvalue;

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    prb_addtab(local_dic, bdd_expression, instance, value, date);
    }

  return(value);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_dbtevalvalue					*/
/* description	: computes a DBT's value depending on its node's value	*/
/*            	:          a DBT's value is a Derivate By Time   value	*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

static float prb_dbtevalvalue (instance , bdd_expression, date)

struct lkdins *instance;
struct node   *bdd_expression;
unsigned long  date;
  {
  short              indexbdd;
  unsigned long      datevalue;
  float              value         = 0.0 ;
  float              probvalue     = 0.0 ;
  float              densvalue     = 0.0 ;
  float              lowlowvalue   = 0.0 ;
  float              lowhighvalue  = 0.0 ;
  float              highhighvalue = 0.0 ;
  float              highlowvalue  = 0.0 ;
  struct prb_recrd  *record = NULL;

  if (bdd_expression == zero) 
    {
    return (0.0);
    }

  if (bdd_expression == one) 
    {
    return (1.0);
    }

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    record = prb_chktab (local_dic, bdd_expression, instance);

    if ((record != NULL) && (record->date == date)) 
      {
      count_hit ++;
      return (record->value);
      }
    else
      {
      count_miss ++;
      }
    }

  indexbdd  = bdd_expression->index ;

  probvalue = ((struct prbval *) gss_getvalue(instance,indexbdd))->PROBVAL;

  if ( gss_getlasteventdate(instance,indexbdd,&datevalue) != GSS_NOEVENTYET)
    {
    if ( datevalue == date )
      {
      densvalue = ((struct prbval *) gss_getvalue(instance,indexbdd))->DENSVAL;
      }
    else
      densvalue = 0.0;
    }
  else
    densvalue = 0.0;

  if ((bdd_expression->low->index >> 1) != (indexbdd >> 1))
    {
    lowlowvalue   = prb_dbtevalvalue(instance,
                                     bdd_expression->low,
                                     date);

    lowhighvalue  = lowlowvalue;
    }
  else
    {
    lowlowvalue   = prb_dbtevalvalue(instance,
                                     bdd_expression->low->low,
                                     date);

    if ( densvalue != 0.0 )
      {
      lowhighvalue  = prb_dbtevalvalue(instance,
                                       bdd_expression->low->high,
                                       date);
      }
    else
      {
      lowhighvalue  = 0.0;
      }
    }

  if ((bdd_expression->high->index >> 1) != (indexbdd >> 1))
    {
    highlowvalue  = prb_dbtevalvalue(instance,
                                     bdd_expression->high,
                                     date);

    highhighvalue = highlowvalue;
    }
  else
    {
    if ( densvalue != 0.0 )
      {
      highlowvalue  = prb_dbtevalvalue(instance,
                                       bdd_expression->high->low,
                                       date);
      }
    else
      {
      highlowvalue  = 0.0;
      }

    highhighvalue = prb_dbtevalvalue(instance,
                                     bdd_expression->high->high,
                                     date);
    }

  value = ( 0.5 * densvalue   * ( highlowvalue  + lowhighvalue ) )
        - ( 0.5 * densvalue   * ( highhighvalue + lowlowvalue  ) )
        + (       probvalue   * ( highhighvalue                ) )
        + ( ( 1 - probvalue ) * ( lowlowvalue                  ) ) ;

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    prb_addtab(local_dic, bdd_expression, instance, value, date);
    }

  return(value);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_evalvalue						*/
/* description	: computes a node's value depending on variable's value	*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

int prb_evalvalue (instance , expression, target_type, value)

struct lkdins *instance;
void          *expression;
unsigned char  target_type;
struct prbval *value;
  {
  int           errflg    = 0   ;
  float         probvalue = 0.0 ;
  float         densvalue = 0.0 ;
  struct chain *var_list  = NULL;

  if (local_dic == NULL)
    local_dic = prb_initab();

  var_list = PRB_SUPERVARLIST (expression);

  if ( PRB_EVALUATION & PRB_LEVELIZED_EVALUATION )
    {
    while ( var_list != NULL )
      {
      if (((struct prbval *)gss_getvalue(instance, var_list->DATA))->EVALUATED
                      == PRB_NOT_EVALUATED)
        {
        return(errflg);
        }

      var_list = var_list->NEXT;
      }

    value->EVALUATED = PRB_EVALUATED;
    }

  errflg+=prb_evalproba  (instance, expression, target_type, &(probvalue));
  value->PROBVAL  = probvalue;

  if (! ( PRB_MODE & PRB_DYNAMIC_MODE ))
    {
    errflg+=prb_evaldensity(instance, expression, target_type, &(densvalue));
    value->DENSVAL  = densvalue;
    }

  fprintf (stdout,"%9d hit, %9d mis", count_hit, count_miss);
  fprintf (stdout,"\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
  fflush (stdout);

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_evalproba						*/
/* description	: computes a BDD's proba depending on its node's value	*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

static int prb_evalproba (instance , expression, target_type, value)

struct lkdins *instance;
void          *expression;
unsigned char  target_type;
float         *value;
  {
  int           errflg = 0;

  switch ( target_type )
    {
    case GSS_SIGDEF :
    case GSS_AUXDEF :
    case GSS_BUSDEF | GSS_DRVDEF :
    case GSS_BUXDEF | GSS_DRVDEF :
    case GSS_REGDEF | GSS_DRVDEF :
    case GSS_BUSDEF | GSS_CNDDEF :
    case GSS_BUXDEF | GSS_CNDDEF :
    case GSS_REGDEF | GSS_CNDDEF :
    case GSS_MSGDEF :
      *value = prb_bddevalvalue(instance,
                                PRB_SUPERBDD (expression),
                                GSS_CURRENT_TIME) ;
      break;

    default :
      prb_toolbug (1,"prb_evalvalue",target_type);
      break;
    }

  if ((target_type == GSS_MSGDEF ) && ( *value != 1.0 ))
    errflg = 1;

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_evaldensity					*/
/* description	: computes a density					*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

static int prb_evaldensity (instance , expression, target_type, value)

struct lkdins *instance;
void          *expression;
unsigned char  target_type;
float         *value;
  {
  struct beder *derivate = NULL;
  int           errflg   = 0;
  float         tmpval   = 0.0;


  switch ( target_type )
    {
    case GSS_SIGDEF :
    case GSS_AUXDEF :
    case GSS_BUSDEF | GSS_DRVDEF :
    case GSS_BUXDEF | GSS_DRVDEF :
    case GSS_REGDEF | GSS_DRVDEF :
    case GSS_BUSDEF | GSS_CNDDEF :
    case GSS_BUXDEF | GSS_CNDDEF :
    case GSS_REGDEF | GSS_CNDDEF :
    case GSS_MSGDEF :
      derivate = PRB_SUPERDERIVATE (expression);
      if (PRB_SUPERFLAG (expression) & DEF_PRB_DERIVBYTIME)
        {
        *value = prb_dbtevalvalue(instance,
                                  derivate->BDD,
                                  GSS_CURRENT_TIME) ;
        }

      else if (PRB_SUPERFLAG (expression) & DEF_PRB_DERIVBYVAR)
        {
        if ((derivate != NULL) && (derivate->INDEX < 2))
          derivate = derivate->NEXT;

        *value = 0.0;
        while (derivate != NULL)
          {
          tmpval = prb_bddevalvalue(instance,
                                    derivate->BDD,
                                    GSS_CURRENT_TIME) ;

          *value += tmpval *
          ((struct prbval *) gss_getvalue(instance, derivate->INDEX))->DENSVAL;
          derivate = derivate->NEXT;
          }
        }

      else
        {
        tmpval = prb_bddevalvalue(instance,
                                  PRB_SUPERBDD(expression),
                                  GSS_CURRENT_TIME) ;
        *value = 2.0 * tmpval * ( 1.0 - tmpval );
        }
      break;


    default :
      prb_toolbug (1,"prb_evalvalue",target_type);
      break;
    }

  if ((target_type == GSS_MSGDEF ) && ( *value != 0.0 ))
    errflg = 1;

  return (errflg);
  }


/* ###--------------------------------------------------------------### */
/* function	: prb_dyncmpvalue					*/
/* description	: comparaison function for values			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_dyncmpvalue (value1,value2)

  struct prbval    *value1;
  struct prbval    *value2;

  {
  if ( value2->DENSVAL != 0.0)
    return(1);
  else 
    return (0);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_dyncpyvalue					*/
/* description	: copy function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_dyncpyvalue (value1,value2)

  struct prbval  *value1;
  struct prbval  *value2;

  {
  int errflg = 0;

  value1->DENSVAL  = value2->DENSVAL;
  value1->DENSSUM  = value2->DENSSUM;

  return(errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_dynupdvalue					*/
/* description	: copy function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_dynupdvalue (value1,value2)

  struct prbval  *value1;
  struct prbval  *value2;

  {
  int errflg = 0;

  value1->DENSVAL   = value2->DENSVAL;
  value1->DENSSUM  += value2->DENSVAL;

/* printf("UPDATE DYNAMIQUE : %d ps\n", GSS_CURRENT_TIME); */
/* printf("UPDATE DYNAMIQUE : densval = %-6.4f\n", value1->DENSVAL); */
/* printf("UPDATE DYNAMIQUE : denssum = %-6.4f\n", value1->DENSSUM); */

  return(errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_dynresolvalue					*/
/* description	: resolution function for typed signals			*/
/*		  return flag:0 ok, 1 error				*/
/*		  return default value 0				*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

int prb_dynresolvalue (projected_bivalue,
                    resolution_function_type,
                    current_value,
                    result_value)

  struct prjbvl *projected_bivalue;
  char		 resolution_function_type;
  struct prbval *current_value;
  struct prbval *result_value;

  {
  int  errflg = 0;

  switch (resolution_function_type)
    {
    case 'R' :
/* printf("RESOL DYNAMIQUE REGISTRE : %d ps\n", GSS_CURRENT_TIME); */
      result_value->DENSVAL = 0.0;
      break;

    case 'W' :
      result_value->DENSVAL = 0.0;
      projected_bivalue = projected_bivalue->NEXT;
      while (projected_bivalue != NULL)
        {
        result_value->DENSVAL +=
              ((struct prbval *)(projected_bivalue->CNDVAL))->PROBVAL
            * ((struct prbval *)(projected_bivalue->DRVVAL))->DENSVAL;
        projected_bivalue = projected_bivalue->NEXT;
        }

      break;

    case 'M' :
      result_value->DENSVAL = 0.0;
      projected_bivalue = projected_bivalue->NEXT;
      while (projected_bivalue != NULL)
        {
        result_value->DENSVAL +=
              ((struct prbval *)(projected_bivalue->CNDVAL))->PROBVAL
            * ((struct prbval *)(projected_bivalue->DRVVAL))->DENSVAL;
        projected_bivalue = projected_bivalue->NEXT;
        }

      break;

    default :
      prb_toolbug (2,"prb_dynresolvalue",resolution_function_type);
      break;
    }

  return (errflg);
  }


/* ###--------------------------------------------------------------### */
/* function	: prb_dyndbtevalvalue					*/
/* description	: computes a DBT's value depending on its node's value	*/
/*            	:          a DBT's value is a Derivate By Time   value	*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

static float prb_dyndbtevalvalue (instance , bdd_expression, date)

struct lkdins *instance;
struct node   *bdd_expression;
unsigned long  date;
  {
  short              indexbdd;
  unsigned long      datevalue;
  float              value         = 0.0 ;
  float              probvalue     = 0.0 ;
  float              densvalue     = 0.0 ;
  float              lowlowvalue   = 0.0 ;
  float              lowhighvalue  = 0.0 ;
  float              highhighvalue = 0.0 ;
  float              highlowvalue  = 0.0 ;
  struct prb_recrd  *record = NULL;

  if (bdd_expression == zero) 
    {
/* printf("dyndbtevalvalue noeud zero\n"); */
    return (0.0);
    }

  if (bdd_expression == one) 
    {
/* printf("dyndbtevalvalue noeud un\n"); */
    return (1.0);
    }

  if (date != 0)
    {
    record = prb_chktab (local_dic, bdd_expression, instance);

    if ((record != NULL) && (record->date == date)) 
      {
      count_hit ++;
      return (record->value);
      }
    else
      {
      count_miss ++;
      }
    }

  indexbdd  = bdd_expression->index ;
/* printf("dyndbtevalvalue indexbdd %d\n", indexbdd); */

  probvalue = ((struct prbval *) gss_getvalue(instance,indexbdd))->PROBVAL;

  if ( gss_getlasteventdate(instance,indexbdd,&datevalue) != GSS_NOEVENTYET)
    {
    if ( datevalue == date )
      {
      densvalue = ((struct prbval *) gss_getvalue(instance,indexbdd))->DENSVAL;
      }
    else
      densvalue = 0.0;
    }
  else
    {
    densvalue = 0.0;
    }
/* printf("dyndbtevalvalue proba[%d] = %f\n", indexbdd, probvalue); */
/* printf("dyndbtevalvalue densi[%d] = %f\n", indexbdd, densvalue); */

  if ((densvalue > 1.0) || (densvalue < 0.0))
    {
    printf("WARNING : dyndbtevalvalue dens[%d]=%f : invalid value\n", indexbdd, densvalue);
    }

  if ((bdd_expression->low->index >> 1) != (indexbdd >> 1))
    {
    lowlowvalue   = prb_dyndbtevalvalue(instance,
                                     bdd_expression->low,
                                     date);

    lowhighvalue  = lowlowvalue;
    }
  else
    {
    lowlowvalue   = prb_dyndbtevalvalue(instance,
                                     bdd_expression->low->low,
                                     date);

    if ( densvalue != 0.0 )
      {
      lowhighvalue  = prb_dyndbtevalvalue(instance,
                                       bdd_expression->low->high,
                                       date);
      }
    else
      {
      lowhighvalue  = 0.0;
      }
    }

  if ((bdd_expression->high->index >> 1) != (indexbdd >> 1))
    {
    highlowvalue  = prb_dyndbtevalvalue(instance,
                                     bdd_expression->high,
                                     date);

    highhighvalue = highlowvalue;
    }
  else
    {
    if ( densvalue != 0.0 )
      {
      highlowvalue  = prb_dyndbtevalvalue(instance,
                                       bdd_expression->high->low,
                                       date);
      }
    else
      {
      highlowvalue  = 0.0;
      }

    highhighvalue = prb_dyndbtevalvalue(instance,
                                     bdd_expression->high->high,
                                     date);
    }

  value = ( 0.5 * densvalue   * ( highlowvalue  + lowhighvalue ) )
        - ( 0.5 * densvalue   * ( highhighvalue + lowlowvalue  ) )
        + (       probvalue   * ( highhighvalue                ) )
        + ( ( 1 - probvalue ) * ( lowlowvalue                  ) ) ;

  prb_addtab(local_dic, bdd_expression, instance, value, date);

/* printf("dyndbtevalvalue retourne %f\n", value); */
  return(value);
  }


/* ###--------------------------------------------------------------### */
/* function	: prb_dynevalvalue					*/
/* description	: computes a BDD's value depending on its node's value	*/
/* called func.	: prb_toolbug						*/
/* ###--------------------------------------------------------------### */

int prb_dynevalvalue (instance , expression, target_type, value)

struct lkdins *instance;
void          *expression;
unsigned char  target_type;
struct prbval *value;
  {
  struct beder *derivate = NULL;
  int           errflg   = 0;
  float         tmpval   = 0.0;
  float         densval  = 0.0;
  unsigned long date     = 0;

/* printf("debut evaluation dynamique\n"); */

  switch ( target_type )
    {
    case GSS_SIGDEF :
    case GSS_AUXDEF :
    case GSS_BUSDEF | GSS_DRVDEF :
    case GSS_BUXDEF | GSS_DRVDEF :
    case GSS_REGDEF | GSS_DRVDEF :
    case GSS_BUSDEF | GSS_CNDDEF :
    case GSS_BUXDEF | GSS_CNDDEF :
    case GSS_REGDEF | GSS_CNDDEF :
    case GSS_MSGDEF :

        derivate = PRB_SUPERDERDEDER (expression);

        if (   ( (PRB_SUPERFLAG (expression) & DEF_PRB_BYTIMEDDR)
              && (PRB_SUPERFLAG (expression) & DEF_PRB_DERIVBYTIME) )
        ||     ( (PRB_SUPERFLAG (expression) & DEF_PRB_BYVARDDR)
              && (PRB_SUPERFLAG (expression) & DEF_PRB_DERIVBYTIME) ) )
          {
          while ((derivate != NULL) && (derivate->INDEX >1))
            {
            derivate = derivate->NEXT;
            }

          value->DENSVAL = prb_dyndbtevalvalue(instance,
                                               derivate->BDD,
                                               GSS_CURRENT_TIME) ;
          }

        else if (   ( (PRB_SUPERFLAG (expression) & DEF_PRB_BYTIMEDDR)
                   && (PRB_SUPERFLAG (expression) & DEF_PRB_DERIVBYVAR) )
             ||     ( (PRB_SUPERFLAG (expression) & DEF_PRB_BYVARDDR)
                   && (PRB_SUPERFLAG (expression) & DEF_PRB_DERIVBYVAR) ) )
          {
          value->DENSVAL = 0.0;
          while (derivate != NULL)
            {
            if (derivate->INDEX > 1)
              {
              if ( gss_getlasteventdate(instance,derivate->INDEX,&date)
                                       != GSS_NOEVENTYET )
                {
                if (date == GSS_CURRENT_TIME )
                  {
                  densval = ((struct prbval *)
                            gss_getvalue(instance, derivate->INDEX))->DENSVAL;
                  }
                else
                  {
                  densval = 0.0;
                  }
                }
              else
                {
                densval = 0.0;
                }

              if ( densval != 0.0 )
                {
                tmpval = prb_bddevalvalue(instance,
                                          derivate->BDD,
                                          GSS_CURRENT_TIME) ;
                }

              value->DENSVAL  += tmpval * densval;
              }
            derivate = derivate->NEXT;
            }
          }

        else if (PRB_SUPERFLAG (expression) & DEF_PRB_BYVOXDDR)
          {
          value->DENSVAL = 0.0;
          while (derivate != NULL)
            {
            if (derivate->INDEX > 1)
              {
              if ( gss_getlasteventdate(instance,derivate->INDEX,&date)
                                       != GSS_NOEVENTYET )
                {
                if (date == GSS_CURRENT_TIME )
                  {
                  densval = ((struct prbval *)
                            gss_getvalue(instance, derivate->INDEX))->DENSVAL;
                  }
                else
                  {
                  densval = 0.0;
                  }
                }
              else
                {
                densval = 0.0;
                }

              if ( densval != 0.0 )
                {
                tmpval = prb_bddevalvalue(instance,
                                          derivate->BDD,
                                          GSS_CURRENT_TIME) ;
                }

              value->DENSVAL  += tmpval * densval;
              }
            derivate = derivate->NEXT;
            }
          }

        else if (PRB_SUPERFLAG (expression) & DEF_PRB_BYVAXDDR)
          {
          value->DENSVAL = 0.0;
          while (derivate != NULL)
            {
            if (derivate->INDEX > 1)
              {
              if ( gss_getlasteventdate(instance,derivate->INDEX,&date)
                                       != GSS_NOEVENTYET )
                {
                if (date == GSS_CURRENT_TIME )
                  {
                  densval = ((struct prbval *)
                            gss_getvalue(instance, derivate->INDEX))->DENSVAL;
                  }
                else
                  {
                  densval = 0.0;
                  }
                }
              else
                {
                densval = 0.0;
                }

              if ( densval != 0.0 )
                {
                tmpval = prb_bddevalvalue(instance,
                                          derivate->BDD,
                                          GSS_CURRENT_TIME) ;
                }

              value->DENSVAL  += tmpval * densval;
              }
            derivate = derivate->NEXT;
            }
          }

        else if (PRB_SUPERFLAG (expression) & DEF_PRB_FINALDDR)
          {
          value->DENSVAL = 0.0;
          while (derivate != NULL)
            {
            if (derivate->INDEX > 1)
              {
/* printf("evaluation dynamique : c'est une BYVAR\n"); */
              if ( gss_getlasteventdate(instance,derivate->INDEX,&date)
                                       != GSS_NOEVENTYET )
                {
                if (date == GSS_CURRENT_TIME )
                  {
                  densval = ((struct prbval *)
                            gss_getvalue(instance, derivate->INDEX))->DENSVAL;
                  }
                else
                  {
                  densval = 0.0;
                  }
                }
              else
                {
                densval = 0.0;
                }

              if ( densval != 0.0 )
                {
                tmpval = prb_bddevalvalue(instance,
                                          derivate->BDD,
                                          GSS_CURRENT_TIME) ;
                }

              value->DENSVAL  += tmpval * densval;
/* printf("evaluation dynamique derive partielle : %f\n", value->DENSVAL); */
              }
            else
              {
/* printf("evaluation dynamique : c'est une BYTIME\n"); */
              value->DENSVAL += prb_dyndbtevalvalue(instance,
                                                    derivate->BDD,
                                                    GSS_CURRENT_TIME) ;
/* printf("evaluation dynamique fn de transition : %f\n", value->DENSVAL); */
              }
            derivate = derivate->NEXT;
            }
          }

        else
          {
          prb_toolbug (1,"prb_dynevalvalue",target_type);
          }
        break;

    default :
      prb_toolbug (1,"prb_dynevalvalue",target_type);
      break;
    }

  if ((target_type == GSS_MSGDEF ) && ( value->DENSVAL != 0.0 ))
    errflg = 1;

  return (errflg);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_dynevaldelay					*/
/* description	: dynamic eval delay          				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

unsigned long prb_dynevaldelay (instance , target, target_type)

struct lkdins *instance;
void          *target;
unsigned char  target_type;
  {
  unsigned long   result = 0;

  switch (target_type)
    {
    case GSS_SIGDEF :
      result = ((struct beout *)target)->TIME;
      break;

    case GSS_BUSDEF :
      result = ((struct binode *)target)->TIME;
      break;

    case GSS_BUXDEF :
      result = ((struct binode *)target)->TIME;
      break;

    case GSS_REGDEF :
      result = ((struct binode *)target)->TIME;
      break;

    case GSS_AUXDEF :
      result = ((struct beaux *)target)->TIME;
      break;

    case GSS_MSGDEF :
      result = 0;
      break;

    default :
      prb_toolbug (1,"prb_dynevaldelay",target_type);
      break;
    }

  return(result);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_dynevalinert					*/
/* description	: dynamic eval inert          				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

unsigned long prb_dynevalinert (instance , target, target_type)

struct lkdins *instance;
void          *target;
unsigned char  target_type;
  {
  unsigned long   result = 0;

  return(result);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_dyncolvalue					*/
/* description	: collapse function for values				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_dyncolvalue (value1,value2)

  struct prbval  *value1;
  struct prbval  *value2;

  {
  int errflg = 0;

/* printf("appel a la fonction collector : %f + %f = ", value1->DENSVAL, value2->DENSVAL); */

  if ((value1 != NULL) && (value2 != NULL))
    {
    value1->DENSVAL  += value2->DENSVAL;
    value1->DENSSUM  += value2->DENSSUM;
    }
  else
    {
    prb_toolbug (5,"prb_dyncolvalue",0);
    }

  return(errflg);
  }
