
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: prb_sequences.c					*/
/* date		: Jun  1 1994						*/
/* version	: v1.00							*/
/* author	: DUNOYER julien					*/
/* description	: This file contains functions relative to the 		*/
/*		  interface with sequences :				*/
/*		  prb_loadsequence  , prb_writesequence			*/
/*		  prb_storesequence , prb_statistics			*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include <sys/time.h>

#include MUT_H
#include LOG_H
#include MLO_H
#include "beh.h"
#include "bhl.h"
#include PRB_H
#include GSS_H

/* ###--------------------------------------------------------------### */
/*                       defines                                        */
/*                       static variables declaration                   */
/*                       externals                                      */
/* ###--------------------------------------------------------------### */

#define DEF_VALUE_CTX   1

/* ###--------------------------------------------------------------### */
/* function	: prb_savesequence					*/
/* description	: save the result of the simulation cycle		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

struct prbseq  *prb_savesequence (dict, seqname, ptlofig, head_befig)

struct dct_entry **dict;
char              *seqname;
struct lofig      *ptlofig;
struct befig      *head_befig;
  {
  struct prbseq     *pt_prbseq = NULL;
  struct prbevt     *pt_prbevt = NULL;
  struct prbval     *value ;
  struct bemsg      *ptbemsg ;
  struct beaux      *ptbeaux ;
  struct bebux      *ptbebux ;
  struct bereg      *ptbereg ;
  struct befig      *ptbefig ;
  struct loins      *ptloins ;
  struct losig      *ptlosig ;
  struct locon      *ptlocon ;
  char              *name ;
  char              *insname ;
  float              densval ;

  pt_prbseq = prb_addprbseq(seqname);

  if (ptlofig != NULL)
    {
    ptlocon  = ptlofig->LOCON;
    while (ptlocon != NULL)
      {
      ptlosig = ptlocon->SIG;

      name = getsigname(ptlosig);

      value=(struct prbval *)gss_chktab(dict,name,DEF_VALUE_CTX,GSS_PNT0FIELD);

      if (PRB_MODE & PRB_DYNAMIC_MODE)
        densval = value->DENSSUM;
      else
        densval = value->DENSVAL;

      pt_prbevt = prb_addprbevt( (struct prbevt *) pt_prbevt,
                                          (char *) name,
                                          (float ) value->DISCVAL,
                                          (float ) densval       ,
                                          (float ) value->PROBVAL,
                                          (char  ) DEF_PRB_UNKNOWN,
                                          (char  ) DEF_PRB_UNKNOWN);

      switch (ptlocon->DIRECTION )
        {
        case 'I' :
          pt_prbevt->MODE = DEF_PRB_IN;
          break;
        case 'O' :
          pt_prbevt->MODE = DEF_PRB_OUT;
          break;
        case 'B' :
          pt_prbevt->MODE = DEF_PRB_INOUT;
          break;
        case 'Z' :
          pt_prbevt->MODE = DEF_PRB_TRISTATE;
          break;
        case 'T' :
          pt_prbevt->MODE = DEF_PRB_TRANSC;
          break;
        default  :
          pt_prbevt->MODE = DEF_PRB_PORT;
          break;
        }

      ptlocon  = ptlocon->NEXT;
      }

     pt_prbseq->PRBEVT = pt_prbevt;
     pt_prbevt = NULL;

    ptlosig  = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      if (ptlosig->TYPE == 'I')
        {
        name = getsigname (ptlosig);

        value=(struct prbval *)gss_chktab(dict,name,DEF_VALUE_CTX,GSS_PNT0FIELD);

        if (PRB_MODE & PRB_DYNAMIC_MODE)
          densval = value->DENSSUM;
        else
          densval = value->DENSVAL;

        pt_prbevt = prb_addprbevt( (struct prbevt *) pt_prbevt,
                                            (char *) name,
                                            (float ) value->DISCVAL,
                                            (float ) densval       ,
                                            (float ) value->PROBVAL,
                                            (char  ) DEF_PRB_SIGNAL,
                                            (char  ) DEF_PRB_UNKNOWN);
        }
      ptlosig  = ptlosig->NEXT;
      }

    ptloins = ptlofig->LOINS;
    while ( ptloins != NULL)
      {
      insname = ptloins->INSNAME;

      ptbefig = head_befig;
      while (ptbefig != NULL)
        {
        if (ptbefig->NAME == ptloins->FIGNAME)
          {
          break;
          }
        ptbefig = ptbefig->NEXT;
        }

      if (ptbefig == NULL)
        {
        prb_toolbug (4, NULL, 0);
        }

/*---------------------
      ptbemsg  = ptbefig->BEMSG;
      while (ptbemsg != NULL)
        {
        name = concatname (insname, ptbemsg->LABEL);

        value=(struct prbval *)gss_chktab(dict,name,DEF_VALUE_CTX,GSS_PNT0FIELD);

        if (PRB_MODE & PRB_DYNAMIC_MODE)
          densval = value->DENSSUM;
        else
          densval = value->DENSVAL;

        pt_prbevt = prb_addprbevt( (struct prbevt *) pt_prbevt,
                                            (char *) name,
                                            (float ) value->DISCVAL,
                                            (float ) densval       ,
                                            (float ) value->PROBVAL,
                                            (char  ) DEF_PRB_ASSERT,
                                            (char  ) DEF_PRB_UNKNOWN);

        ptbemsg  = ptbemsg->NEXT;
        }
----------------------*/

/*---------------------
      ptbeaux  = ptbefig->BEDLY;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        value=(struct prbval *)gss_chktab(dict,name,DEF_VALUE_CTX,GSS_PNT0FIELD);

        if (PRB_MODE & PRB_DYNAMIC_MODE)
          densval = value->DENSSUM;
        else
          densval = value->DENSVAL;

        pt_prbevt = prb_addprbevt( (struct prbevt *) pt_prbevt,
                                            (char *) name,
                                            (float ) value->DISCVAL,
                                            (float ) densval       ,
                                            (float ) value->PROBVAL,
                                            (char  ) DEF_PRB_SIGNAL,
                                            (char  ) DEF_PRB_UNKNOWN);

        ptbeaux  = ptbeaux->NEXT;
        }
----------------------*/
 
      ptbeaux  = ptbefig->BEAUX;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        value=(struct prbval *)gss_chktab(dict,name,DEF_VALUE_CTX,GSS_PNT0FIELD);

        if (PRB_MODE & PRB_DYNAMIC_MODE)
          densval = value->DENSSUM;
        else
          densval = value->DENSVAL;

        pt_prbevt = prb_addprbevt( (struct prbevt *) pt_prbevt,
                                            (char *) name,
                                            (float ) value->DISCVAL,
                                            (float ) densval       ,
                                            (float ) value->PROBVAL,
                                            (char  ) DEF_PRB_SIGNAL,
                                            (char  ) DEF_PRB_UNKNOWN);

        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbebux  = ptbefig->BEBUX;
      while (ptbebux != NULL)
        {
        name = concatname (insname, ptbebux->NAME);

        value=(struct prbval *)gss_chktab(dict,name,DEF_VALUE_CTX,GSS_PNT0FIELD);

        if (PRB_MODE & PRB_DYNAMIC_MODE)
          densval = value->DENSSUM;
        else
          densval = value->DENSVAL;

        pt_prbevt = prb_addprbevt( (struct prbevt *) pt_prbevt,
                                            (char *) name,
                                            (float ) value->DISCVAL,
                                            (float ) densval       ,
                                            (float ) value->PROBVAL,
                                            (char  ) DEF_PRB_TRISTATE,
                                            (char  ) DEF_PRB_UNKNOWN);

        ptbebux  = ptbebux->NEXT;
        }
 
      ptbereg  = ptbefig->BEREG;
      while (ptbereg != NULL)
        {
        name = concatname (insname, ptbereg->NAME);

        value=(struct prbval *)gss_chktab(dict,name,DEF_VALUE_CTX,GSS_PNT0FIELD);

        if (PRB_MODE & PRB_DYNAMIC_MODE)
          densval = value->DENSSUM;
        else
          densval = value->DENSVAL;

        pt_prbevt = prb_addprbevt( (struct prbevt *) pt_prbevt,
                                            (char *) name,
                                            (float ) value->DISCVAL,
                                            (float ) densval       ,
                                            (float ) value->PROBVAL,
                                            (char  ) DEF_PRB_REGISTER,
                                            (char  ) DEF_PRB_UNKNOWN);

        ptbereg  = ptbereg->NEXT;
        }
 
      ptloins = ptloins->NEXT;
      }
    }

  pt_prbevt = (struct prbevt *) reverse((struct chain *)pt_prbevt);

  pt_prbseq->PRBEVT = (struct prbevt *)
          append((struct chain *)pt_prbseq->PRBEVT, (struct chain *)pt_prbevt);

  return (pt_prbseq);
  }


/* ###--------------------------------------------------------------### */
/* function	: prb_storesequence					*/
/* description	: store the result of the simulation cycle in the dict	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void prb_storesequence (dict, ptlofig, head_befig)

struct dct_entry **dict;
struct lofig      *ptlofig ;
struct befig      *head_befig ;
  {
  struct prbval     *value ;
  struct bemsg      *ptbemsg ;
  struct beaux      *ptbeaux ;
  struct bebux      *ptbebux ;
  struct bereg      *ptbereg ;
  struct befig      *ptbefig ;
  struct loins      *ptloins ;
  struct losig      *ptlosig ;
  char              *name ;
  char              *insname ;


  if (ptlofig != NULL)
    {

    ptlosig  = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      name = getsigname(ptlosig);

      value = prb_addvalue();

      prb_cpyvalue (value, gss_getcurval(name));
      prb_dyncpyvalue (value, gss_getcurval(name));

      gss_addtab(dict, name, DEF_VALUE_CTX, GSS_PNT0FIELD, value);

      ptlosig  = ptlosig->NEXT;
      }

    ptloins = ptlofig->LOINS;
    while ( ptloins != NULL)
      {
      insname = ptloins->INSNAME;

      ptbefig = head_befig;
      while (ptbefig != NULL)
        {
        if (ptbefig->NAME == ptloins->FIGNAME)
          {
          break;
          }
        ptbefig = ptbefig->NEXT;
        }

      if (ptbefig == NULL)
        {
        prb_toolbug (4, NULL, 0);
        }

      ptbemsg  = ptbefig->BEMSG;
      while (ptbemsg != NULL)
        {
        name = concatname (insname, ptbemsg->LABEL);

        value = prb_addvalue();

        prb_cpyvalue (value, gss_getcurval(name));
        prb_dyncpyvalue (value, gss_getcurval(name));

        gss_addtab(dict, name, DEF_VALUE_CTX, GSS_PNT0FIELD, value);

        ptbemsg  = ptbemsg->NEXT;
        }

      ptbeaux  = ptbefig->BEDLY;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        value = prb_addvalue();

        prb_cpyvalue (value, gss_getcurval(name));
        prb_dyncpyvalue (value, gss_getcurval(name));

        gss_addtab(dict, name, DEF_VALUE_CTX, GSS_PNT0FIELD, value);

        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbeaux  = ptbefig->BEAUX;
      while (ptbeaux != NULL)
        {
        name = concatname (insname, ptbeaux->NAME);

        value = prb_addvalue();

        prb_cpyvalue (value, gss_getcurval(name));
        prb_dyncpyvalue (value, gss_getcurval(name));

        gss_addtab(dict, name, DEF_VALUE_CTX, GSS_PNT0FIELD, value);

        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbebux  = ptbefig->BEBUX;
      while (ptbebux != NULL)
        {
        name = concatname (insname, ptbebux->NAME);

        value = prb_addvalue();

        prb_cpyvalue (value, gss_getcurval(name));
        prb_dyncpyvalue (value, gss_getcurval(name));

        gss_addtab(dict, name, DEF_VALUE_CTX, GSS_PNT0FIELD, value);

        ptbebux  = ptbebux->NEXT;
        }
 
      ptbereg  = ptbefig->BEREG;
      while (ptbereg != NULL)
        {
        name = concatname (insname, ptbereg->NAME);

        value = prb_addvalue();

        prb_cpyvalue (value, gss_getcurval(name));
        prb_dyncpyvalue (value, gss_getcurval(name));

        gss_addtab(dict, name, DEF_VALUE_CTX, GSS_PNT0FIELD, value);

        ptbereg  = ptbereg->NEXT;
        }
 
      ptloins = ptloins->NEXT;
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_writesequence					*/
/* description	: write the result of the simulation cycle		*/
/* called func.	: prb_error						*/
/* ###--------------------------------------------------------------### */

void prb_writesequence (dict, ptlofig, head_befig, filename)

struct dct_entry **dict;
struct lofig      *ptlofig ;
struct befig      *head_befig ;
char              *filename;

{
FILE              *fp;
time_t             clock;
char              *name;
struct bemsg      *ptbemsg ;
struct beaux      *ptbeaux ;
struct bebux      *ptbebux ;
struct bereg      *ptbereg ;
struct befig      *ptbefig ;
struct loins      *ptloins ;
struct losig      *ptlosig ;
struct locon      *ptlocon ;
struct prbval     *value   ;
char              *insname ;
char               printname[256];

fp = mbkfopen (filename, PRB_OUT_SUFFIX, WRITE_TEXT);

if (fp == NULL)
  prb_error (2, filename, 0.0, ' ', 0);

else
  {
  time (&clock);

  if ((ptlofig != NULL) && (head_befig != NULL))
    {

  (void)fprintf (fp,"#\t\tPrb file generated from %s\n", ptlofig->NAME);
  (void)fprintf (fp,"#\t\tdate : %s \n \n", ctime(&clock));

        /* ###------------------------------------------------------### */
        /*   For each signal write the values				*/
        /* ###------------------------------------------------------### */

    ptlocon  = ptlofig->LOCON;
    while (ptlocon != NULL)
      {
      ptlosig = ptlocon->SIG;

      name = getsigname(ptlosig);

      if (vectorindex(name) == -1)
        sprintf (printname, "%s", vectorradical(name));
      else
        sprintf (printname, "%s(%d)", vectorradical(name), vectorindex(name));

      switch (ptlocon->DIRECTION )
        {
        case 'I' :
          (void)fprintf(fp, "in       %-15s : p('1') = ", printname);
          break;
        case 'O' :
          (void)fprintf(fp, "out      %-15s : p('1') = ", printname);
          break;
        case 'B' :
          (void)fprintf(fp, "inout    %-15s : p('1') = ", printname);
          break;
        case 'Z' :
          (void)fprintf(fp, "tristate %-15s : p('1') = ", printname);
          break;
        case 'T' :
          (void)fprintf(fp, "transc   %-15s : p('1') = ", printname);
          break;
        default  :
          (void)fprintf(fp, "port     %-15s : p('1') = ", printname);
          break;
        }

      value = (struct prbval *)gss_chktab(dict, name, DEF_VALUE_CTX , GSS_PNT0FIELD);
      (void)fprintf(fp,"%f ;\n", value->PROBVAL);

      ptlocon  = ptlocon->NEXT;
      }

    ptlosig  = ptlofig->LOSIG;
    while (ptlosig != NULL)
      {
      if (ptlosig->TYPE == 'I')
        {
        name = getsigname (ptlosig);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s(%d)", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "signal   %-15s : p('1') = ", printname);

        value = (struct prbval *)gss_chktab(dict, name, DEF_VALUE_CTX , GSS_PNT0FIELD);
        (void)fprintf(fp,"%f ;\n", value->PROBVAL);
        }

      ptlosig  = ptlosig->NEXT;
      }

    ptloins = ptlofig->LOINS;
    while ( ptloins != NULL)
      {
      insname = ptloins->INSNAME;

      ptbefig = head_befig;
      while (ptbefig->NAME != ptloins->FIGNAME)
        {
        ptbefig = ptbefig->NEXT;
        }
/*---------------------
      ptbemsg  = ptbefig->BEMSG;
      while (ptbemsg != NULL)
        {
        name = concatname(insname, ptbemsg->LABEL);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s(%d)", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "assert %-15s : p('1') = ", printname);

        value = (struct prbval *)gss_chktab(dict, name, DEF_VALUE_CTX , GSS_PNT0FIELD);
        (void)fprintf(fp,"%f ;\n", value->PROBVAL);

        ptbemsg  = ptbemsg->NEXT;
        }
----------------------*/
 
/*---------------------
      ptbeaux  = ptbefig->BEDLY;
      while (ptbeaux != NULL)
        {
        name = concatname(insname, ptbeaux->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s(%d)", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "signal  %-15s : p('1') = ", printname);

        value = (struct prbval *)gss_chktab(dict, name, DEF_VALUE_CTX , GSS_PNT0FIELD);
        (void)fprintf(fp,"%f ;\n", value->PROBVAL);

        ptbeaux  = ptbeaux->NEXT;
        }
----------------------*/
 
      ptbeaux  = ptbefig->BEAUX;
      while (ptbeaux != NULL)
        {
        name = concatname(insname, ptbeaux->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s(%d)", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "signal   %-15s : p('1') = ", printname);

        value = (struct prbval *)gss_chktab(dict, name, DEF_VALUE_CTX , GSS_PNT0FIELD);
        (void)fprintf(fp,"%f ;\n", value->PROBVAL);

        ptbeaux  = ptbeaux->NEXT;
        }
 
      ptbebux  = ptbefig->BEBUX;
      while (ptbebux != NULL)
        {
        name = concatname(insname, ptbebux->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s(%d)", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "tristate  %-15s : p('1') = ", printname);

        value = (struct prbval *)gss_chktab(dict, name, DEF_VALUE_CTX , GSS_PNT0FIELD);
        (void)fprintf(fp,"%f ;\n", value->PROBVAL);

        ptbebux  = ptbebux->NEXT;
        }
 
      ptbereg  = ptbefig->BEREG;
      while (ptbereg != NULL)
        {
        name = concatname(insname, ptbereg->NAME);

        if (vectorindex(name) == -1)
          sprintf (printname, "%s", vectorradical(name));
        else
          sprintf (printname, "%s(%d)", vectorradical(name), vectorindex(name));

        (void)fprintf(fp, "register %-15s : p('1') = ", printname);

        value = (struct prbval *)gss_chktab(dict, name, DEF_VALUE_CTX , GSS_PNT0FIELD);
        (void)fprintf(fp,"%f ;\n", value->PROBVAL);

        ptbereg  = ptbereg->NEXT;
        }
      ptloins = ptloins->NEXT;
      }
    }
  }
}



/* ###--------------------------------------------------------------### */
/* function	: prb_loadsequence					*/
/* description	: get a probability events sequence			*/
/*		  add it in a new gss sequence list			*/
/* called func.	: prb_getprbseq, gss_addgssseq, gss_addgssevt		*/
/* ###--------------------------------------------------------------### */

struct gssseq *prb_loadsequence (seqname)

char  *seqname  ;		/* name of the prb sequence to be get	*/

  {
  struct gssseq *pt_gssseq  = NULL;
  struct prbseq *pt_prbseq  = NULL;
  struct prbevt *pt_prbevt  = NULL;
  struct prbval *value      = NULL;

  pt_prbseq = prb_getprbseq (seqname);

  PRB_INPUT_VALUE = (struct prbval *)prb_addvalue ();

  PRB_INPUT_VALUE->PROBVAL  = pt_prbseq->PROBVAL;
  PRB_INPUT_VALUE->DISCVAL  = pt_prbseq->DISCVAL;
  PRB_INPUT_VALUE->DENSVAL  = pt_prbseq->DENSVAL;
  PRB_INPUT_VALUE->DENSSUM  = (float) DEF_PRB_DEFDENS;

  PRB_INIT_VALUE = (struct prbval *)prb_addvalue ();

  PRB_INIT_VALUE->PROBVAL  = (float) DEF_PRB_DEFPROB;
  PRB_INIT_VALUE->DISCVAL  = (float) DEF_PRB_DEFDISC;
  PRB_INIT_VALUE->DENSVAL  = (float) DEF_PRB_DEFDENS;
  PRB_INIT_VALUE->DENSSUM  = (float) DEF_PRB_DEFDENS;

  pt_prbevt = pt_prbseq->PRBEVT;

  pt_gssseq = gss_addgssseq(NULL, NULL);

  while (pt_prbevt)
    {
    value = prb_addvalue();
    value->PROBVAL = pt_prbevt->VALUE->PROBVAL;
    value->DENSVAL = pt_prbevt->VALUE->DENSVAL;
    value->DISCVAL = pt_prbevt->VALUE->DISCVAL;
    value->EVALUATED = PRB_EVALUATED;

    pt_gssseq->EVENT = gss_addgssevt (pt_gssseq->EVENT,
                                      pt_prbevt->NAME,
                                      GSS_UNKDEF,
                                      value);
    pt_prbevt = pt_prbevt->NEXT;
    }

  return (pt_gssseq);
  }

