
/* ###--------------------------------------------------------------### */
/*									*/
/* file		: util.c						*/
/* date		: Feb 13 1993						*/
/* authors	: Julien DUNOYER					*/
/* version	: v1.0							*/
/* description	: This file contains some utility functions :		*/
/*		  prb_error, prb_toolbug, prb_message,			*/
/*                prb_warning, prb_avers				*/
/* ###--------------------------------------------------------------### */

#include <stdio.h>
#include MUT_H

/* ###--------------------------------------------------------------### */
/* function	: prb_avers						*/
/* description	: return the current version				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

char *prb_avers ()
  {
  return ("v1.00");
  }


/* ###--------------------------------------------------------------### */
/* function	: prb_warning						*/
/* description	: print a warning message corresponding to the argument	*/
/*		  code							*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void prb_warning (code,str1)

int   code;
char *str1;

  {
  fprintf (stderr,"Warning %d : ",code);
  switch (code)
    {
    case 1:
      fprintf (stderr,"cannot find register `%s`\n",str1);
      break;
    case 2:
      fprintf (stderr,"consistency checks will be disabled\n");
      break;
    case 3:
      fprintf (stderr,"signal `%s` never used\n",str1);
      break;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_toolbug						*/
/* description	: print a fatal error message corresponding to the	*/
/*		  argument code and exit				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void prb_toolbug (code,str1,nbr1)

int   code;
char *str1;
int   nbr1;

  {
  (void) fprintf (stderr,"Fatal Prb error %d executing `%s`: ", code,str1);
  switch (code)
    {
    case 1  :
      (void) fprintf (stderr,"Invalid Target Type \n\n");
      break;
    case 2  :
      (void) fprintf (stderr,"Invalid Resolution Type \n\n");
      break;
    case 3  :
      (void) fprintf (stderr,"Cant find corresponding signal : %s\n\n", str1);
      break;
    case 4  :
      (void) fprintf (stderr,"Cant construct SuperBdds when Bdds already exists\n\n", str1);
      break;

    default:
      fprintf (stderr,"Unknown Prb Fatal Error\n\n");
      break;
    }
  exit (1);
  }

/* ###--------------------------------------------------------------### */
/* function	: prb_message						*/
/* description	: print a message corresponding to the argument code	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void prb_message (code, str1, nmb1)

int   code;
char *str1;
int   nmb1;

  {
  switch (code)
    {
    case 0:
  (void) fprintf (stdout,"\nPROBA %s\n\n", prb_avers());
      break;
    case 1:
  (void) fprintf (stdout,"  ##-- USAGE --## \n\n");
  (void) fprintf (stdout,"=> proba [options] vhdl_file prb_file res_file\n");
  (void) fprintf (stdout,"=> option -h             : display this help\n");
  (void) fprintf (stdout,"=> option -t             : run with trace (debug)\n");
  (void) fprintf (stdout,"=> option -c             : compilation phase only\n");
  (void) fprintf (stdout,"=> option -b             : behav. description\n");
  (void) fprintf (stdout,"=> option -sav save_file : save after treatment\n");
  (void) fprintf (stdout,"=> option -cox           : with full     correl. extraction\n");
  (void) fprintf (stdout,"=> option -cox n         : with level n  correl. extraction\n");
  (void) fprintf (stdout,"=> option -pvx           : with complete prim. inputs extraction\n");
  (void) fprintf (stdout,"=> option -pvx n         : with level n  prim. inputs extraction\n");
  (void) fprintf (stdout,"\n  ##-- !! Don't add suffix to file's name !! --##\n\n");
      break;
    case 2:
      (void) fprintf (stdout,"Initializing ...\n\n");
      break;
    case 3:
      (void) fprintf (stdout,"Compiling `%s` (Behaviour) ...\n\n",str1);
      break;
    case 4:
      (void) fprintf (stdout,"Compiling `%s` (Structural) ...\n\n",str1);
      break;
    case 5:
      (void) fprintf (stdout,"Searching `%s` (Behaviour) ...\n\n",str1);
      break;
    case 6:
      (void) fprintf (stdout,"Searching `%s` (Probabilities)  ...\n\n",str1);
      break;
    case 8:
      (void) fprintf (stdout,"Compiling `%s` (Probabilities) ...\n\n",str1);
      break;
    case 9:
      (void) fprintf (stdout,"Linking ...\n\n");
      break;
    case 10:
      (void) fprintf (stdout,"###----- processing pattern %d -----###\n",nmb1);
      break;
    case 11:
      (void) fprintf (stdout,"Making Dependencies ...\n\n");
      break;
    case 12:
      (void) fprintf (stdout,"\n");
      (void) fprintf (stdout,"Generating the file '%s' ...\n\n",str1);
      break;
    case 13:
      (void) fprintf (stdout,"of the figure '%s'.\n\n",str1);
      break;
    case 14:
     (void) fprintf (stdout,"File '%s' has been generated.\n\n",str1);
      break;
    case 15:
      (void) fprintf (stdout,"Making Nodes ...\n\n");
      break;
    case 16:
      (void) fprintf (stdout,"Restoring ...\n\n");
      break;
    case 17:
      (void) fprintf (stdout,"Creating the result file ...\n\n");
      break;
    case 18:
      (void) fprintf (stdout,"Flattening the root figure ...\n\n");
      break;
    case 19:
      (void) fprintf (stdout,"Making Ghost Structural Level ...\n\n");
      break;
    case 20:
      (void) fprintf (stdout,"Making Super Abl  ...\n\n");
      break;
    case 21:
      (void) fprintf (stdout,"Making Super Bdd  ...\n\n");
      break;
    case 22:
      (void) fprintf (stdout,"Making Simple Bdd  ...\n\n");
      break;

    default:
      (void) fprintf (stderr,"Unknown Prb Message\n\n",code);
      break;

    }
  }
/* ###--------------------------------------------------------------### */
/* function	: prb_error						*/
/* description	: print an error message depending on the argument code	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int prb_error (code,str1,fdat,cdat,idat)

int           code;
char         *str1;
float         fdat ;
char          cdat ;
int           idat ;

  {
  if (code < 100)
    (void)fprintf (stderr,"Prb : Error %d : ", code);
  else
    (void)fprintf (stderr,"Prb : Parse Error %d line %d : ", code, idat);

  switch (code)
    {
    case 0 :
      (void) fprintf (stderr,"\n        cannot continue further more.\n");
      (void) fprintf (stderr,"\n                Have a nice day...\n");
      break;
    case 1  :
      (void) fprintf (stderr,"Invalid terminal Value\n\n");
      break;
    case 2  :
      (void) fprintf (stderr,"Cannot open result File %s\n\n", str1);
      break;
    case 3  :
      (void) fprintf (stderr,"Cannot open source File %s\n\n", str1);
      break;

    case 104  :
      (void) fprintf (stderr,"Unknown prb Parser Error : %s\n", str1);
      (void) fprintf (stderr,"Compilation Aborted... Using Default Values\n\n");
      break;
    case 114:
      fprintf (stderr,"too many hierarchical names\n\n");
      break;
    case 201:
      fprintf (stderr,"Invalid float value : %-6.4f\n\n", fdat);
      break;
    case 202:
      fprintf (stderr,"Invalid name : %s\n\n", str1);
      break;
    case 203:
      fprintf (stderr,"Invalid name : %s\n\n", str1);
      break;
 

    default :
      (void) fprintf (stderr,"Unknown Prb Error\n\n");
      break;
    }

  return (1);
  }
