/*
 * This file is part of the Alliance CAD System
 * Copyright (C) Laboratoire LIP6 - Dpartement ASIM
 * Universite Pierre et Marie Curie
 *
 * Home page          : http://www-asim.lip6.fr/alliance/
 * E-mail support     : mailto:alliance-support@asim.lip6.fr
 *
 * This progam is  free software; you can redistribute it  and/or modify it
 * under the  terms of the GNU  General Public License as  published by the
 * Free Software Foundation;  either version 2 of the License,  or (at your
 * option) any later version.
 *
 * Alliance VLSI  CAD System  is distributed  in the hope  that it  will be
 * useful, but WITHOUT  ANY WARRANTY; without even the  implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details.
 *
 * You should have received a copy  of the GNU General Public License along
 * with the GNU C Library; see the  file COPYING. If not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Tool    : Logic Synthesis - Standard Cell Mapping
 * Date    : 1991,92
 * Author  : Pascale Allegre
 * Modified by Czo <Olivier.Sirol@lip6.fr> 1996,97
 */




#ident "$Id: mapping_cout.c,v 1.2 2000/11/07 16:15:21 czo Exp $"

#include <stdio.h>
#include <stdlib.h>

#include MUT_H
#include LOG_H
#include BEH_H
#include MLO_H

#include "lax_param.h"
#include "sl_local.h"
#include "mapping_local.h"


/*---------------------------------------------------------------------------
	delais	:	calcul de delai sur une expression mappee
-----------------------------------------------------------------------------
	return	:	son delai
---------------------------------------------------------------------------*/

long
delais (expr)
     chain_list *expr;
{
  sig *s;
  struct virtuel *v;

  if (ATOM (expr))
    {
      if ((v = (struct virtuel *) (testObj (VALUE_ATOM (expr), 'v'))) != NULL)
	return v->delais;
      if ((s = (sig *) testObj (VALUE_ATOM (expr), 's')) != NULL)
	return s->delais;
      if ((VALUE_ATOM (expr) == nameUn) ||
	  (VALUE_ATOM (expr) == nameZero))
	return 1;
      if (SL_TRACE)
	printf ("delais : object undefined = %s\n", VALUE_ATOM (expr));
      return 0;
    }
  else
    {
      cellList *cell = operToCell (expr);
      num_list *del;
      long max_del = 0;

      if (!cell)
	{
/*
   if (SL_TRACE) printf("delay : %s%d doesn't exist\n",
   searchCharOper(OPER(expr)),ARITE(expr));
 */
	  return -2;
	}
      del = cell->delais;
      while ((expr = CDR (expr)))
	{
	  long delCour = delais (CAR (expr));

	  if ((delCour + del->DATA) > max_del)
	    max_del = delCour + del->DATA;
	  del = del->NEXT;
	}
      return max_del;
    }
}

/*---------------------------------------------------------------------------
	delRC	:	calcul de delai sur une expression mappee
-----------------------------------------------------------------------------
	return	:	son delai
---------------------------------------------------------------------------*/
long
delRC (virt, expr)
     virtuelList *virt;
     chain_list *expr;
{
  sig *signal;
  struct virtuel *virtuel;

  if (virt->del_rc != -1)
    return virt->del_rc;
  if (ATOM (expr))
    {
      if ((signal = (sig *) testObj (VALUE_ATOM (expr), 's')))
	return signal->delais + signal->capa;
      if ((virtuel = (virtuelList *) testObj (VALUE_ATOM (expr), 'v')))
	{
	  if (virtuel->del_rc != -1)
	    return virtuel->del_rc;
	  return delRC (virtuel, CAR (virtuel->solutions));
	}
      return 0;
    }
  else
    {
      tabNumList *delai = virt->op_log->delais_tech;
      num_list *del;
      ptype_list *resist = virt->op_log->resistance_tech;
      char *nameDef = virt->cellule;
      long maxDel = 0;
      chain_list *args;

      while ((char *) resist->DATA != nameDef)
	{
	  resist = CDR (resist);
	  delai = CDR (delai);
	}
      del = delai->ptNum;
      for (args = CDR (expr); args; args = CDR (args))
	{
	  long d = delRC (virt, CAR (args)) + del->DATA;

	  del = CDR (del);
	  if (maxDel < d)
	    maxDel = d;
	}
      virt->del_rc = maxDel + (resist->TYPE * virt->capa) / 1000;
      return virt->del_rc;
    }
}

long
entrance (virtFan, name)
     virtuelList *virtFan;
     char *name;
{
  chain_list *sol = CAR (virtFan->solutions);
  chain_list *args;
  tabNumList *capalist = virtFan->op_log->capa_tech;
  num_list *capasNum;

  while (virtFan->cellule != capalist->name)
    capalist = CDR (capalist);
  capasNum = capalist->ptNum;
  for (args = CDR (sol); args; args = CDR (args))
    {
      if (VALUE_ATOM (CAR (args)) == name)
	return capasNum->DATA;
      capasNum = capasNum->NEXT;
    }
  return (long) 0;
}
/*---------------------------------------------------------------------------
	capa	:	calcul de capa
-----------------------------------------------------------------------------
	return	:	sa capa
---------------------------------------------------------------------------*/

void
capa (name, virt)
     char *name;
     virtuelList *virt;
{
  long totCapa = 0;
  chain_list *f;

  for (f = virt->fanout; f; f = CDR (f))
    {
      virtuelList *virtFan = (virtuelList *) testObj (f->DATA, 'v');
      sig *signal = (sig *) testObj (f->DATA, 's');

      if (name != (char *) f->DATA)
	if (virtFan && virtFan->op_log)		/* cas atomique... */
	  totCapa += entrance (virtFan, name);

      if (signal && (signal->type == 5))
	{
	  cellList *cell = ((cellList *) testObj (nameNot1, 'c'));

	  if (cell)
	    {
	      tabNumList *capalist = cell->capa_tech;

	      while (cell->nom != capalist->name)
		capalist = capalist->NEXT;
	      totCapa += capalist->ptNum->DATA;
	      signal->capa = capalist->ptNum->DATA;
	    }
	  else
	    printf ("capa : any not\n");
	}
      if (signal && (signal->name == name))
	signal->capa = totCapa;
    }
  if (!totCapa)
    {
      chain_list *i;

      for (f = virt->fanout; f; f = CDR (f))
	{
	  sig *s = (sig *) testObj ((char *) (f->DATA), 's');

	  if (s)
	    {
	      for (i = s->util; i; i = CDR (i))
		if (testObj ((char *) i->DATA, 'v'))
		  if (s->type == 3)
		    totCapa += entrance ((virtuelList *) testObj ((char *) i->DATA, 'v'),
					 s->name);
		  else
		    totCapa += entrance ((virtuelList *) testObj ((char *) i->DATA, 'v'),
					 name);
	      s->capa = totCapa;
	    }
	}
    }
  virt->capa = totCapa;
  if (!totCapa)			/* corrige le 21/12/92 on enleve meme les signaux */
    {
      if (virt->fanout)
	{
	  freechain (virt->fanout);
	  virt->fanout = NULL;
	}

      if (!testObj (name, 's'))
	{
	    /* printf ("sup %s\n", name); */
	  /* AFAC */
	  /*	  suppVirtuel(name,NULL); */

	}

    }
}

/*---------------------------------------------------------------------------
	surface	:	calcul de surface sur une expression mappee
-----------------------------------------------------------------------------
	return	:	son delai
---------------------------------------------------------------------------*/

long
surface (expr)
     chain_list *expr;
{
  sig *s;
  struct virtuel *v;

  if (ATOM (expr))
    {
      if ((v = (struct virtuel *) testObj (VALUE_ATOM (expr), 'v')) != NULL)
	if (v->cellule)
	  return 0;
	else
	  return v->surface_cone;
      if ((s = (sig *) testObj (VALUE_ATOM (expr), 's')) != NULL)
	return 0;
      if ((VALUE_ATOM (expr) == nameUn) ||
	  (VALUE_ATOM (expr) == nameZero))
	{
	  cellList *cell;

	  if (VALUE_ATOM (expr) == nameUn)
	    cell = (cellList *) testObj (nameHigh, 'c');
	  else
	    cell = (cellList *) testObj (nameLow, 'c');
	  if (cell)
	    return cell->surface;
	  return 0;
	}
      if (SL_TRACE)
	printf ("undefined object : %s\n", VALUE_ATOM (expr));
      return 0;
    }
  else
    {
      cellList *cell = operToCell (expr);
      long max_surf = 0;

      if (!cell)
	{
/*
   if (SL_TRACE) printf("area : %s%d doesn't exist\n",
   searchCharOper(OPER(expr)),ARITE(expr));
 */
	  return -2;
	}
      /* ajoute  le 18/09/92 */
      if (profExpr (expr) == 1 && testObj (exprToCharRed (expr), 'a'))
	return 0;
      max_surf = cell->surface;
      while ((expr = CDR (expr)))
	max_surf += surface (CAR (expr));
      return max_surf;
    }
}

long
surface_cone (cell)
     virtuelList *cell;
{
  if (cell->cellule)
    return 0;
  return surface (CAR (cell->solutions));
}

long
Mapcout (expr)
     chain_list *expr;
{
  long val;

  switch (mappSys->coutSig)
    {
    case 0:
      return surface (expr);
    case 1:
      val = (mappSys->coef * 2) * surface (expr);
      if (val >= 0)
	val += delais (expr);
      break;
    case 2:
      val = mappSys->coef * surface (expr);
      if (val >= 0)
	val += delais (expr);
      break;
    case 3:
      val = (mappSys->coef / 2) * surface (expr);
      if (val >= 0)
	val += delais (expr);
      break;
    case 4:
      return delais (expr);
      break;
    }
  if (val < 0)
    return -2;
  return val;
}

void
genDelta (signal)
     sig *signal;
{
  long del;
  cellList *cell = (cellList *) testObj (nameNot1, 'c');

  if (cell)
    {
      del = cell->delais->DATA;
      mappSys->coef = cell->delais->DATA / cell->surface;
    }
  else
    {
      del = 400;
      mappSys->coef = 200;
    }
  switch (signal->cout)
    {
    case 0:
      mappSys->delta = del * 13;
      break;
    case 1:
      mappSys->delta = del * 9;
      break;
    case 2:
      mappSys->delta = del * 7;
      break;
    case 3:
      mappSys->delta = del * 5;
      break;
    case 4:
      mappSys->delta = del * 3;
      break;
/*
   case 0 : mappSys->delta = del * 13;break;
   case 1 : mappSys->delta = del * 9;break;
   case 2 : mappSys->delta = del * 8;break;
   case 3 : mappSys->delta = del * 7;break;
   case 4 : mappSys->delta = del * 5;break;
 */
    }
}
