# include MUT_H
# include LOG_H
# include "beh.h"
# include "bhl.h"
# include "bvl.h"
# include "beh_debug.h"

/* ###--------------------------------------------------------------### */
/* file		: beh_chkbefig.c					*/
/* date		: Oct 30 1995						*/
/* version	: v109							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_chkbefig						*/
/* description	: check the consistency of a behavioural description	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

int beh_chkbefig (pt_befig, mode)

struct befig *pt_befig;
unsigned int  mode    ;

  {
  struct beout *pt_beout ;
  struct beaux *pt_beaux ;
  struct bebus *pt_bebus ;
  struct bebux *pt_bebux ;
  struct bereg *pt_bereg ;
  int           err_flg  = 0;

	/* ###------------------------------------------------------### */
	/*   check that the description is not empty			*/
	/* ###------------------------------------------------------### */

  if (pt_befig == NULL)
    err_flg = 1;

  else
    { 
	/* ###------------------------------------------------------### */
	/*   check that each output has at least one driver. Outputs	*/
	/* are simple output ports, bussed output ports, simple		*/
	/* internal signals, delayed internal signals, bussed internal	*/
	/* signals, and internal registers.				*/
	/* ###------------------------------------------------------### */

    if ((mode & BEH_CHK_DRIVERS) != 0)
      {
	/* ###------------------------------------------------------### */
	/*   simple output ports					*/
	/* ###------------------------------------------------------### */

      pt_beout = pt_befig->BEOUT;
      while (pt_beout != NULL)
        {
        if (pt_beout->ABL == NULL)
          err_flg += beh_error (40, pt_beout->NAME);
        pt_beout = pt_beout->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*   simple internal signals					*/
	/* ###------------------------------------------------------### */

      pt_beaux = pt_befig->BEAUX;
      while (pt_beaux != NULL)
        {
        if (pt_beaux->ABL == NULL)
          err_flg += beh_error (40, pt_beaux->NAME);
        pt_beaux = pt_beaux->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*   delayed internal signals					*/
	/* ###------------------------------------------------------### */

      pt_beaux = pt_befig->BEDLY;
      while (pt_beaux != NULL)
        {
        if (pt_beaux->ABL == NULL)
          err_flg += beh_error (40, pt_beaux->NAME);
        pt_beaux = pt_beaux->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*   bussed output ports					*/
	/* ###------------------------------------------------------### */

      pt_bebus = pt_befig->BEBUS;
      while (pt_bebus != NULL)
        {
        if (pt_bebus->BIABL == NULL)
          err_flg += beh_error (40, pt_bebus->NAME);
        pt_bebus = pt_bebus->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*   bussed internal signals					*/
	/* ###------------------------------------------------------### */

      pt_bebux = pt_befig->BEBUX;
      while (pt_bebux != NULL)
        {
        if (pt_bebux->BIABL == NULL)
          err_flg += beh_error (40, pt_bebux->NAME);
        pt_bebux = pt_bebux->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*   internal registers						*/
	/* ###------------------------------------------------------### */

      pt_bereg = pt_befig->BEREG;
      while (pt_bereg != NULL)
        {
        if (pt_bereg->BIABL == NULL)
          err_flg += beh_error (40, pt_bereg->NAME);
        pt_bereg = pt_bereg->NEXT;
        }

      }

	/* ###------------------------------------------------------### */
	/*   check that the descriptiona has an empty architecture.	*/
	/* that means, no internal signal is declared and output ports	*/
	/* have no dreivers.						*/
	/* ###------------------------------------------------------### */

    if ((mode & BEH_CHK_EMPTY) != 0)
      {
	/* ###------------------------------------------------------### */
	/*   check that simple output ports have no drivers		*/
	/* ###------------------------------------------------------### */

      pt_beout = pt_befig->BEOUT;
      while (pt_beout != NULL)
        {
        if (pt_beout->ABL != NULL)
          break;
        pt_beout = pt_beout->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*   check that bussed output ports have no drivers		*/
	/* ###------------------------------------------------------### */

      pt_bebus = pt_befig->BEBUS;
      while (pt_bebus != NULL)
        {
        if (pt_bebus->BIABL != NULL)
          break;
        pt_bebus = pt_bebus->NEXT;
        }

	/* ###------------------------------------------------------### */
	/*   check that the architecture is empty			*/
	/* ###------------------------------------------------------### */

      if ((pt_befig->BEAUX != NULL) || (pt_befig->BEBUX != NULL) ||
          (pt_befig->BEDLY != NULL) || (pt_befig->BEREG != NULL) ||
          (pt_befig->BEMSG != NULL) || (pt_beout        != NULL) ||
          (pt_bebus        != NULL) )
        err_flg += beh_error (41, pt_befig->NAME);

      }

    pt_befig->ERRFLG = err_flg;
    }

  return (err_flg);
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_debug.c						*/
/* date		: Nov  6 1995						*/
/* version	: v109							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: high level function					*/
/* ###--------------------------------------------------------------### */


static char           *buffer      = NULL;
static unsigned int    buff_size   = 0;
static struct circuit *circuit_pnt = NULL;

/* ###--------------------------------------------------------------### */
/* function	: beh_debug						*/
/* description	: dispaly any mad structure				*/
/* called func.	: goforward, translate, dispaly, splitline, pop, push	*/
/* ###--------------------------------------------------------------### */

void beh_debug (head_pnt, type)

void *head_pnt;				/* structure's pointer		*/
char *type;				/* structure's type		*/

  {
  static void          disp_immd ();
  static void          getsize   ();
  static void          display   ();
  static void          push      ();
  static void          pop       ();
  static struct chain *goforward ();
  static int           splitline ();
  static int           translate ();

  char          line   [128];		/* buffer to read a cmd line	*/
  char          heap   [128];		/* buffer to split the cmd line	*/

  char         *words  [ 10];		/* number of words on a line	*/
  int           nmbrs  [ 10];		/* words translated into number	*/
  char          flags  [ 10];		/* set if words is a number	*/
  int           indxs  [ 10];		/* index of words		*/

  struct stack  jtab   [ 10];		/* list of memorized addresses	*/
  int           idx, i;
  int           dispflg = 0;
  int           code;
  unsigned int  size;
  char         *pntr   = NULL;
  long          pshtype;
  int           wrdcnt = 1;

  struct stack  stk [STKSIZ_DFN];
  int           stkpnt = -1;

  union value   pnt [MAXCMD_DFN];
  long          typ [MAXCMD_DFN];
  unsigned int  siz [MAXCMD_DFN];

  static char  *str [] = {
                          "_back"    , "_exit"    , "_jump"    , "_save"    ,
                          "_stop"    , "_top"     , "_up"      , "_display" ,

                          "character", "short"    , "integer"  , "long"     ,
                          "void"     , "string"   ,
                          "chain"    , "ptype"    , "befig"    , "bereg"    ,
                          "bemsg"    , "berin"    , "beout"    , "bebus"    ,
                          "beaux"    , "bedly"    , "bebux"    , "bepor"    ,
                          "begen"    , "biabl"    , "binode"   , "bequad"   ,
                          "beder"    ,
                          "next"     , "name"     , "user"     , "errflg"   ,
                          "type"     , "level"    , "label"    , "message"  ,
                          "abl"      , "node"     , "direction", "cndabl"   ,
                          "valabl"   , "cndnode"  , "valnode"  , "out_ref"  ,
                          "bus_ref"  , "aux_ref"  , "reg_ref"  , "bux_ref"  ,
                          "msg_ref"  , "dly_ref"  , "value"    , "data"     ,
                          "time"     , "bdd"      , "gex"      , "varlist"  ,
                          "derivate" , "index"    , "flag"     , "time_unit"
                          };

	/* ###------------------------------------------------------### */
	/*    initialisation :						*/
	/*    - allocate a buffer for read words			*/
	/*    - break the argument that identifies the structure (type)	*/
	/*      into words						*/
	/*    - search that words among recognized strings		*/
	/* ###------------------------------------------------------### */

  words [0] = heap;
  getsize (siz);

  wrdcnt        = splitline (words, type);
  idx           = translate (words, wrdcnt, str, nmbrs, flags, indxs);

  typ [idx]     = POINTER_DFN | s_DFN | idx;
  pnt [idx].dat = head_pnt;

	/* ###------------------------------------------------------### */
	/*    process the command line until the _exit command		*/
	/* ###------------------------------------------------------### */

  while ((idx & TYPE_DFN) != _exit_DFN)
    {
	/* ###------------------------------------------------------### */
	/*    if the first word of the line has not been recognized,	*/
	/* print an error message. Otherwise, proccess the command line	*/
	/* (generally it is a request for displaying a specific field).	*/
	/*								*/
	/* At this point :						*/
	/*    - pnt [] contains all available pointers			*/
	/*    - typ [] contains the type of availabale pointers		*/
	/* ###------------------------------------------------------### */

    if (idx == _error_DFN)
      fprintf (stderr ,"unrecognized command\n");
    else
      {

	/* ###------------------------------------------------------### */
	/*    define the type of the structure that may be pushed on	*/
	/* the stack :							*/
	/*    - if the pointer is a VOID pointer the exact type must be	*/
	/*      defined on the command line (last word of the command)	*/
	/* ###------------------------------------------------------### */

      pshtype = POINTER_DFN | s_DFN | void_DFN;
      if ((typ [idx] & TYPE_DFN) == void_DFN)
        {
        if ((wrdcnt >= 2) && (indxs [wrdcnt - 1] != _error_DFN))
          pshtype = POINTER_DFN | s_DFN | indxs [wrdcnt - 1];
        else
          fprintf (stderr ,"structure has unknown type (void *)\n");
        }
      else
       pshtype = typ [idx];

	/* ###------------------------------------------------------### */
	/*    depending on the kind of the first word of the command	*/
	/* activate actions :						*/
	/*    - COMMAND (_top, _up, ...)				*/
	/*    - POINTER							*/
	/*        - for NEXT go forward until the Nth element of the	*/
	/*          list. Then, push it on the stack and dispaly it	*/
	/*        - for others push and display				*/
	/*    - ARRAY							*/
	/*        - push and display the Nth element of the array	*/
	/* ###------------------------------------------------------### */

      switch (typ [idx] & KIND_DFN)
        {

        case COMMAND_DFN :
          switch (typ [idx] & TYPE_DFN)
            {
	/* ###------------------------------------------------------### */
	/*    _top COMMAND : reset the stack pointer, call read_field	*/
	/* to read the structure on the top of stack			*/
	/* ###------------------------------------------------------### */

            case _top_DFN :
              stkpnt  = 0;
              dispflg = 1;
              break;

	/* ###------------------------------------------------------### */
	/*    _stop COMMAND : set the stop mark for the structure on	*/
	/* the top of stack						*/
	/* ###------------------------------------------------------### */

            case _stop_DFN :
              stk [stkpnt].mark = 1;
              fprintf (stdout, "beh_debug : stop mark set\n");
              break;

	/* ###------------------------------------------------------### */
	/*    _back COMMAND : pop structures from the stack until a	*/
	/* marked structure is found. Call read_field to read the	*/
	/* structure on the top of stack				*/
	/* ###------------------------------------------------------### */

            case _back_DFN :
              while ((stkpnt != 0) && (stk[--stkpnt].mark != 1));
              dispflg = 1;
              break;

	/* ###------------------------------------------------------### */
	/*    _jump COMMAND : push on the stack the structure saved in	*/
	/* the jump table. Call read_field to read the structure on the	*/
	/* top of stack							*/
	/* ###------------------------------------------------------### */

            case _jump_DFN :
              if ((wrdcnt == 2) && (flags [1] == 1) && (nmbrs [1] < 10))
                {
                push (stk, &stkpnt, jtab[nmbrs[1]].data, jtab[nmbrs[1]].type);
                dispflg = 1;
                }
              else
                fprintf (stderr ,"unrecognized command\n");
              break;

	/* ###------------------------------------------------------### */
	/*    _up COMMAND : pop structure from the stack. Call		*/
	/* read_field to read the structure on the top of stack		*/
	/* ###------------------------------------------------------### */

            case _up_DFN :
              if (wrdcnt == 1)
                {
                pop (&stkpnt, 1);
                dispflg = 1;
                }
              else
                {
                if ((wrdcnt == 2) && (flags [1] == 1))
                  {
                  pop (&stkpnt, nmbrs [1] + 1);
                  dispflg = 1;
                  }
                else
                  fprintf (stderr ,"unrecognized command\n");
                }
              break;

	/* ###------------------------------------------------------### */
	/*    _save COMMAND : save the structure on the top of stack in	*/
	/* the jump table						*/
	/* ###------------------------------------------------------### */

            case _save_DFN :
              if ((wrdcnt == 2) && (flags [1] == 1) && (nmbrs [1] < 10))
                jtab [nmbrs[1]] = stk [stkpnt];
              else
                fprintf (stderr ,"unrecognized command\n");
              break;

	/* ###------------------------------------------------------### */
	/*    _display COMMAND : display a specified field as an	*/
	/* immediate							*/
	/* ###------------------------------------------------------### */

            case _display_DFN :
              if ((wrdcnt == 3) && (indxs [1] != _error_DFN))
                disp_immd (str, pnt [indxs [1]], indxs [2]);
              else
                fprintf (stderr ,"unrecognized command\n");
              break;

            }
          break;

        case POINTER_DFN :
          if (idx == next_DFN)
            {
            if ((wrdcnt == 1) || ((wrdcnt == 2) && (flags [1] == 0)))
              {
              push (stk, &stkpnt, pnt [idx].dat, pshtype);
              dispflg = 1;
              }
            else
              {
              if (((wrdcnt == 3) || (wrdcnt == 2)) && (flags [1] == 1))
                {
                pnt [idx].dat = (void *) goforward (pnt [idx].dat, nmbrs [1]);
                push (stk, &stkpnt, pnt [idx].dat, pshtype);
                dispflg = 1;
                }
              else
                fprintf (stderr ,"unrecognized command\n");
              }
            }
          else
            {
            push (stk, &stkpnt, pnt [idx].dat, pshtype);
            dispflg = 1;
            }
          break;

        case ARRAY_DFN :
          if ((wrdcnt > 1) && (flags [1] == 1))
            {
            size = siz [(typ [idx] & TYPE_DFN)];
            pntr = (void *)
                   (((unsigned int) pnt [idx].dat) + (size * nmbrs [1]));
            push (stk, &stkpnt, pntr, pshtype);
            dispflg = 1;
            }
          else
            fprintf (stderr ,"unrecognized command\n");
          break;

        case ARYOFPNT_DFN :
          if ((wrdcnt > 1) && (flags [1] == 1))
            {
            size = sizeof (void *);
            pntr = * (void **)
                     (((unsigned int) pnt [idx].dat) + (size * nmbrs [1]));
            push (stk, &stkpnt, pntr, pshtype);
            dispflg = 1;
            }
          else
            fprintf (stderr ,"unrecognized command\n");
          break;

        }

      if (dispflg == 1)
        {
        dispflg = 0;
        display (stk [stkpnt], pnt, typ, str);
        }

      }

    printf ("\n\nCOMMAND > ");

    gets (line);
    if (strcmp (line ,"."))
      {
      wrdcnt = splitline (words, line);
      idx    = translate (words, wrdcnt, str, nmbrs, flags, indxs);
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: display						*/
/* description	: display the fields of a structure			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static void display (curpnt, pnt, typ, str)

struct stack   curpnt;			/* current struct. for display	*/
union value   *pnt;			/* child structures or data	*/
long          *typ;			/* child struct. or data type	*/
char         **str;			/* recognized strings		*/

  {
  extern char         *exprToCharInt ();
  extern struct chain *bddToAbl ();

  int                     i;
  struct chain           *ptr_abl;

  struct chain           *chain_pnt;
  struct ptype           *ptype_pnt;

  struct befig           *befig_pnt;
  struct bereg           *bereg_pnt;
  struct bemsg           *bemsg_pnt;
  struct berin           *berin_pnt;
  struct beout           *beout_pnt;
  struct bebus           *bebus_pnt;
  struct beaux           *beaux_pnt;
  struct beaux           *bedly_pnt;
  struct bebux           *bebux_pnt;
  struct bepor           *bepor_pnt;
  struct begen           *begen_pnt;
  struct biabl           *biabl_pnt;
  struct binode          *binode_pnt;
  struct bequad          *bequad_pnt;
  struct beder           *beder_pnt;

  char                  **string_pnt;
  char                   *character_pnt;
  int                    *int_pnt;

	/* ###------------------------------------------------------### */
	/*    initialization						*/
	/* ###------------------------------------------------------### */

  if (buffer == NULL)
    {
    buff_size = 1024;
    buffer    = (char *) mbkalloc (buff_size);
    }

  for (i=0 ; i<MAXCMD_DFN ; i++)
    {
    typ [i]     = _error_DFN;
    pnt [i].dat = NULL;
    }

  typ [_exit_DFN]     = COMMAND_DFN | _exit_DFN;
  typ [_up_DFN]       = COMMAND_DFN | _up_DFN;
  typ [_top_DFN]      = COMMAND_DFN | _top_DFN;
  typ [_stop_DFN]     = COMMAND_DFN | _stop_DFN;
  typ [_back_DFN]     = COMMAND_DFN | _back_DFN;
  typ [_jump_DFN]     = COMMAND_DFN | _jump_DFN;
  typ [_save_DFN]     = COMMAND_DFN | _save_DFN;
  typ [_display_DFN]  = COMMAND_DFN | _display_DFN;

  switch (curpnt.type & TYPE_DFN)
    {
	/* ###------------------------------------------------------### */
	/*    depending on the type of the current structure :		*/
	/*								*/
	/*       - define the type of each field :			*/
	/*          - the field contains an immediate or a pointer	*/
	/*          - what format must be used to print it		*/
	/*          - kind of data it contains				*/
	/*								*/
	/*       - register the data contained in the field		*/
	/* ###------------------------------------------------------### */

	/* ###------------------------------------------------------### */
	/*    befig							*/
	/* ###------------------------------------------------------### */

    case befig_DFN :

      befig_pnt = (struct befig *) curpnt.data;

      typ [next_DFN]      = POINTER_DFN   | s_DFN | befig_DFN;
      typ [name_DFN]      = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [bereg_DFN]     = POINTER_DFN   | s_DFN | bereg_DFN;
      typ [bemsg_DFN]     = POINTER_DFN   | s_DFN | bemsg_DFN;
      typ [berin_DFN]     = POINTER_DFN   | s_DFN | berin_DFN;
      typ [beout_DFN]     = POINTER_DFN   | s_DFN | beout_DFN;
      typ [bebus_DFN]     = POINTER_DFN   | s_DFN | bebus_DFN;
      typ [beaux_DFN]     = POINTER_DFN   | s_DFN | beaux_DFN;
      typ [bedly_DFN]     = POINTER_DFN   | s_DFN | beaux_DFN;
      typ [bebux_DFN]     = POINTER_DFN   | s_DFN | bebux_DFN;
      typ [bepor_DFN]     = POINTER_DFN   | s_DFN | bepor_DFN;
      typ [begen_DFN]     = POINTER_DFN   | s_DFN | begen_DFN;
      typ [user_DFN]      = POINTER_DFN   | s_DFN | ptype_DFN;
      typ [errflg_DFN]    = IMMEDIATE_DFN | d_DFN | character_DFN;
      typ [type_DFN]      = IMMEDIATE_DFN | x_DFN | character_DFN;
      typ [flag_DFN]      = IMMEDIATE_DFN | x_DFN | character_DFN;
      typ [time_unit_DFN] = IMMEDIATE_DFN | x_DFN | character_DFN;

      pnt [next_DFN].dat      = befig_pnt->NEXT     ;
      pnt [name_DFN].dat      = befig_pnt->NAME     ;
      pnt [bereg_DFN].dat     = befig_pnt->BEREG    ;
      pnt [bemsg_DFN].dat     = befig_pnt->BEMSG    ;
      pnt [berin_DFN].dat     = befig_pnt->BERIN    ;
      pnt [beout_DFN].dat     = befig_pnt->BEOUT    ;
      pnt [bebus_DFN].dat     = befig_pnt->BEBUS    ;
      pnt [beaux_DFN].dat     = befig_pnt->BEAUX    ;
      pnt [bedly_DFN].dat     = befig_pnt->BEDLY    ;
      pnt [bebux_DFN].dat     = befig_pnt->BEBUX    ;
      pnt [bepor_DFN].dat     = befig_pnt->BEPOR    ;
      pnt [begen_DFN].dat     = befig_pnt->BEGEN    ;
      pnt [user_DFN].dat      = befig_pnt->USER     ;
      pnt [errflg_DFN].imd    = befig_pnt->ERRFLG   ;
      pnt [type_DFN].imd      = befig_pnt->TYPE     ;
      pnt [flag_DFN].imd      = befig_pnt->FLAG     ;
      pnt [time_unit_DFN].imd = befig_pnt->TIME_UNIT;

      circuit_pnt = befig_pnt->CIRCUI;

      break;

	/* ###------------------------------------------------------### */
	/*    bereg							*/
	/* ###------------------------------------------------------### */

    case bereg_DFN :

      bereg_pnt = (struct bereg *) curpnt.data;

      typ [next_DFN]   = POINTER_DFN   | s_DFN | bereg_DFN;
      typ [name_DFN]   = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [biabl_DFN]  = POINTER_DFN   | s_DFN | biabl_DFN;
      typ [binode_DFN] = POINTER_DFN   | s_DFN | binode_DFN;

      pnt [next_DFN].dat   = bereg_pnt->NEXT;
      pnt [name_DFN].dat   = bereg_pnt->NAME;
      pnt [biabl_DFN].dat  = bereg_pnt->BIABL;
      pnt [binode_DFN].dat = bereg_pnt->BINODE;

      break;

	/* ###------------------------------------------------------### */
	/*    bemsg							*/
	/* ###------------------------------------------------------### */

    case bemsg_DFN :

      bemsg_pnt = (struct bemsg *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN   | bemsg_DFN;
      typ [level_DFN]   = IMMEDIATE_DFN | c_DFN   | character_DFN;
      typ [label_DFN]   = IMMEDIATE_DFN | s_DFN   | string_DFN;
      typ [message_DFN] = IMMEDIATE_DFN | s_DFN   | string_DFN;
      typ [abl_DFN]     = IMMEDIATE_DFN | ABL_DFN | abl_DFN;
      typ [node_DFN]    = POINTER_DFN   | s_DFN   | void_DFN;

      pnt [next_DFN].dat    = bemsg_pnt->NEXT;
      pnt [level_DFN].imd   = bemsg_pnt->LEVEL;
      pnt [label_DFN].dat   = bemsg_pnt->LABEL;
      pnt [message_DFN].dat = bemsg_pnt->MESSAGE;
      pnt [abl_DFN].dat     = bemsg_pnt->ABL;
      pnt [node_DFN].dat    = bemsg_pnt->NODE;

      break;

	/* ###------------------------------------------------------### */
	/*    beout							*/
	/* ###------------------------------------------------------### */

    case beout_DFN :

      beout_pnt = (struct beout *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN   | beout_DFN;
      typ [name_DFN]    = IMMEDIATE_DFN | s_DFN   | string_DFN;
      typ [time_DFN]    = IMMEDIATE_DFN | l_DFN   | long_DFN;
      typ [abl_DFN]     = IMMEDIATE_DFN | ABL_DFN | abl_DFN;
      typ [node_DFN]    = POINTER_DFN   | s_DFN   | void_DFN;

      pnt [next_DFN].dat    = beout_pnt->NEXT;
      pnt [name_DFN].dat    = beout_pnt->NAME;
      pnt [time_DFN].imd    = beout_pnt->TIME;
      pnt [abl_DFN].dat     = beout_pnt->ABL;
      pnt [node_DFN].dat    = beout_pnt->NODE;

      break;

	/* ###------------------------------------------------------### */
	/*    bebus							*/
	/* ###------------------------------------------------------### */

    case bebus_DFN :

      bebus_pnt = (struct bebus *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN | bebus_DFN;
      typ [name_DFN]    = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [biabl_DFN]   = POINTER_DFN   | s_DFN | biabl_DFN;
      typ [binode_DFN]  = POINTER_DFN   | s_DFN | binode_DFN;
      typ [type_DFN]    = IMMEDIATE_DFN | c_DFN | character_DFN;

      pnt [next_DFN].dat    = bebus_pnt->NEXT;
      pnt [name_DFN].dat    = bebus_pnt->NAME;
      pnt [biabl_DFN].dat   = bebus_pnt->BIABL;
      pnt [binode_DFN].dat  = bebus_pnt->BINODE;
      pnt [type_DFN].imd    = bebus_pnt->TYPE;

      break;

	/* ###------------------------------------------------------### */
	/*    beaux							*/
	/* ###------------------------------------------------------### */

    case beaux_DFN :

      beaux_pnt = (struct beaux *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN   | beaux_DFN;
      typ [name_DFN]    = IMMEDIATE_DFN | s_DFN   | string_DFN;
      typ [time_DFN]    = IMMEDIATE_DFN | l_DFN   | long_DFN;
      typ [abl_DFN]     = IMMEDIATE_DFN | ABL_DFN | abl_DFN;
      typ [node_DFN]    = POINTER_DFN   | s_DFN   | void_DFN;

      pnt [next_DFN].dat    = beaux_pnt->NEXT;
      pnt [name_DFN].dat    = beaux_pnt->NAME;
      pnt [time_DFN].imd    = beaux_pnt->TIME;
      pnt [abl_DFN].dat     = beaux_pnt->ABL;
      pnt [node_DFN].dat    = beaux_pnt->NODE;

      break;

	/* ###------------------------------------------------------### */
	/*    bebux							*/
	/* ###------------------------------------------------------### */

    case bebux_DFN :

      bebux_pnt = (struct bebux *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN | bebux_DFN;
      typ [name_DFN]    = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [biabl_DFN]   = POINTER_DFN   | s_DFN | biabl_DFN;
      typ [binode_DFN]  = POINTER_DFN   | s_DFN | binode_DFN;
      typ [type_DFN]    = IMMEDIATE_DFN | c_DFN | character_DFN;

      pnt [next_DFN].dat    = bebux_pnt->NEXT;
      pnt [name_DFN].dat    = bebux_pnt->NAME;
      pnt [biabl_DFN].dat   = bebux_pnt->BIABL;
      pnt [binode_DFN].dat  = bebux_pnt->BINODE;
      pnt [type_DFN].imd    = bebux_pnt->TYPE;

      break;

	/* ###------------------------------------------------------### */
	/*    bepor							*/
	/* ###------------------------------------------------------### */

    case bepor_DFN :

      bepor_pnt = (struct bepor *) curpnt.data;

      typ [next_DFN]      = POINTER_DFN   | s_DFN | bepor_DFN;
      typ [name_DFN]      = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [direction_DFN] = IMMEDIATE_DFN | c_DFN | character_DFN;
      typ [type_DFN]      = IMMEDIATE_DFN | c_DFN | character_DFN;

      pnt [next_DFN].dat      = bepor_pnt->NEXT;
      pnt [name_DFN].dat      = bepor_pnt->NAME;
      pnt [direction_DFN].imd = bepor_pnt->DIRECTION;
      pnt [type_DFN].imd      = bepor_pnt->TYPE;

      break;

	/* ###------------------------------------------------------### */
	/*    biabl							*/
	/* ###------------------------------------------------------### */

    case biabl_DFN :

      biabl_pnt = (struct biabl *) curpnt.data;

      typ [next_DFN]      = POINTER_DFN   | s_DFN   | biabl_DFN;
      typ [label_DFN]     = IMMEDIATE_DFN | s_DFN   | string_DFN;
      typ [time_DFN]      = IMMEDIATE_DFN | l_DFN   | long_DFN;
      typ [cndabl_DFN]    = IMMEDIATE_DFN | ABL_DFN | abl_DFN;
      typ [valabl_DFN]    = IMMEDIATE_DFN | ABL_DFN | abl_DFN;
      typ [flag_DFN]      = IMMEDIATE_DFN | x_DFN   | character_DFN;
      typ [user_DFN]      = POINTER_DFN   | s_DFN   | ptype_DFN;

      pnt [next_DFN].dat   = biabl_pnt->NEXT;
      pnt [label_DFN].dat  = biabl_pnt->LABEL;
      pnt [time_DFN].imd   = biabl_pnt->TIME;
      pnt [cndabl_DFN].dat = biabl_pnt->CNDABL;
      pnt [valabl_DFN].dat = biabl_pnt->VALABL;
      pnt [flag_DFN].imd   = biabl_pnt->FLAG;
      pnt [user_DFN].dat   = biabl_pnt->USER;

      break;

	/* ###------------------------------------------------------### */
	/*    binode							*/
	/* ###------------------------------------------------------### */

    case binode_DFN :

      binode_pnt = (struct binode *) curpnt.data;

      typ [next_DFN]      = POINTER_DFN   | s_DFN   | binode_DFN;
      typ [time_DFN]      = IMMEDIATE_DFN | l_DFN   | long_DFN;
      typ [cndnode_DFN]   = POINTER_DFN   | s_DFN   | void_DFN;
      typ [valnode_DFN]   = POINTER_DFN   | s_DFN   | void_DFN;

      pnt [next_DFN].dat      = binode_pnt->NEXT;
      pnt [time_DFN].imd      = binode_pnt->TIME;
      pnt [cndnode_DFN].dat   = binode_pnt->CNDNODE;
      pnt [valnode_DFN].dat   = binode_pnt->VALNODE;

      break;

	/* ###------------------------------------------------------### */
	/*    berin							*/
	/* ###------------------------------------------------------### */

    case berin_DFN :

      berin_pnt = (struct berin *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN | berin_DFN;
      typ [name_DFN]    = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [out_ref_DFN] = POINTER_DFN   | s_DFN | chain_DFN;
      typ [bus_ref_DFN] = POINTER_DFN   | s_DFN | chain_DFN;
      typ [aux_ref_DFN] = POINTER_DFN   | s_DFN | chain_DFN;
      typ [reg_ref_DFN] = POINTER_DFN   | s_DFN | chain_DFN;
      typ [bux_ref_DFN] = POINTER_DFN   | s_DFN | chain_DFN;
      typ [msg_ref_DFN] = POINTER_DFN   | s_DFN | chain_DFN;
      typ [dly_ref_DFN] = POINTER_DFN   | s_DFN | chain_DFN;

      pnt [next_DFN].dat    = berin_pnt->NEXT;
      pnt [name_DFN].dat    = berin_pnt->NAME;
      pnt [out_ref_DFN].dat = berin_pnt->OUT_REF;
      pnt [bus_ref_DFN].dat = berin_pnt->BUS_REF;
      pnt [aux_ref_DFN].dat = berin_pnt->AUX_REF;
      pnt [reg_ref_DFN].dat = berin_pnt->REG_REF;
      pnt [bux_ref_DFN].dat = berin_pnt->BUX_REF;
      pnt [msg_ref_DFN].dat = berin_pnt->MSG_REF;
      pnt [dly_ref_DFN].dat = berin_pnt->DLY_REF;

      break;

	/* ###------------------------------------------------------### */
	/*    begen							*/
	/* ###------------------------------------------------------### */

    case begen_DFN :

      begen_pnt = (struct begen *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN | begen_DFN;
      typ [name_DFN]    = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [type_DFN]    = IMMEDIATE_DFN | s_DFN | string_DFN;
      typ [value_DFN]   = POINTER_DFN   | s_DFN | void_DFN;

      pnt [next_DFN].dat    = begen_pnt->NEXT;
      pnt [name_DFN].dat    = begen_pnt->NAME;
      pnt [type_DFN].dat    = begen_pnt->TYPE;
      pnt [value_DFN].dat   = begen_pnt->VALUE;

      break;

	/* ###------------------------------------------------------### */
	/*    bequad							*/
	/* ###------------------------------------------------------### */

    case bequad_DFN :

      bequad_pnt = (struct bequad *) curpnt.data;

      typ [bdd_DFN]      = IMMEDIATE_DFN | BDD_DFN | bdd_DFN;
      typ [gex_DFN]      = IMMEDIATE_DFN | GEX_DFN | gex_DFN;
      typ [varlist_DFN]  = POINTER_DFN   | s_DFN   | chain_DFN;
      typ [derivate_DFN] = POINTER_DFN   | s_DFN   | beder_DFN;
      typ [user_DFN]     = POINTER_DFN   | s_DFN   | void_DFN;

      pnt [bdd_DFN].dat      = bequad_pnt->BDD     ;
      pnt [gex_DFN].dat      = bequad_pnt->GEX     ;
      pnt [varlist_DFN].dat  = bequad_pnt->VARLIST ;
      pnt [derivate_DFN].dat = bequad_pnt->DERIVATE;
      pnt [user_DFN].dat     = bequad_pnt->USER    ;

      break;

	/* ###------------------------------------------------------### */
	/*    beder							*/
	/* ###------------------------------------------------------### */

    case beder_DFN :

      beder_pnt = (struct beder *) curpnt.data;

      typ [next_DFN]    = POINTER_DFN   | s_DFN   | beder_DFN;
      typ [bdd_DFN]     = IMMEDIATE_DFN | BDD_DFN | bdd_DFN  ;
      typ [index_DFN]   = IMMEDIATE_DFN | u_DFN   | short_DFN;
      typ [time_DFN]    = IMMEDIATE_DFN | l_DFN   | long_DFN;

      pnt [next_DFN].dat    = beder_pnt->NEXT ;
      pnt [bdd_DFN].dat     = beder_pnt->BDD  ;
      pnt [index_DFN].imd   = beder_pnt->INDEX;
      pnt [time_DFN].imd    = beder_pnt->TIME ;

      break;

	/* ###------------------------------------------------------### */
	/*    ptype							*/
	/* ###------------------------------------------------------### */

    case ptype_DFN :

      ptype_pnt = (struct ptype *) curpnt.data;

      typ [next_DFN] = POINTER_DFN   | s_DFN | ptype_DFN;
      typ [data_DFN] = POINTER_DFN   | s_DFN | void_DFN;
      typ [type_DFN] = IMMEDIATE_DFN | l_DFN | long_DFN;

      pnt [next_DFN].dat = ptype_pnt->NEXT;
      pnt [data_DFN].dat = ptype_pnt->DATA;
      pnt [type_DFN].imd = ptype_pnt->TYPE;

      break;

	/* ###------------------------------------------------------### */
	/*    chain							*/
	/* ###------------------------------------------------------### */

    case chain_DFN :

      chain_pnt = (struct chain *) curpnt.data;

      typ [next_DFN] = POINTER_DFN | s_DFN | chain_DFN;
      typ [data_DFN] = POINTER_DFN | s_DFN | void_DFN;

      pnt [next_DFN].dat = chain_pnt->NEXT;
      pnt [data_DFN].dat = chain_pnt->DATA;

      break;

	/* ###------------------------------------------------------### */
	/*    string							*/
	/* ###------------------------------------------------------### */

    case string_DFN :

      string_pnt = (char **) curpnt.data;

      typ [string_DFN]     = IMMEDIATE_DFN | s_DFN | string_DFN;
      pnt [string_DFN].dat = *string_pnt;

      break;

	/* ###------------------------------------------------------### */
	/*    integer							*/
	/* ###------------------------------------------------------### */

    case integer_DFN :

      int_pnt = (int *) curpnt.data;

      typ [integer_DFN]     = IMMEDIATE_DFN | d_DFN | integer_DFN;
      pnt [integer_DFN].imd = *int_pnt;

      break;

	/* ###------------------------------------------------------### */
	/*    character							*/
	/* ###------------------------------------------------------### */

    case character_DFN :

      character_pnt = (char *) curpnt.data;

      typ [character_DFN]     = IMMEDIATE_DFN | x_DFN | character_DFN;
      pnt [character_DFN].imd = *character_pnt;

      break;

    }

	/* ###------------------------------------------------------### */
	/*    scan the list of recongnized strings to print those	*/
	/* values that are not empty (different form _error_DFN) :	*/
	/*								*/
	/*   - if it contains an immediate, print it following the	*/
	/*     required format						*/
	/*								*/
	/*   - if it contains a pointer, just print the pointer		*/
	/*     indication, ->, and "availbale"				*/
	/*								*/
	/*   - if it contains an array, just print the array		*/
	/*     indication, =>, and "availbale"				*/
	/* ###------------------------------------------------------### */

  for (i=0 ; i<MAXCMD_DFN ; i++)
    {
    if ((typ [i] & KIND_DFN) == IMMEDIATE_DFN)
      {
      printf ("   %-15s : ", str [i]);
      switch (typ [i] & FORMAT_DFN)
        {
        case s_DFN :
          if (pnt[i].dat != NULL)
            printf ("%s", pnt[i].dat);
          printf ("\n");
          break;

        case u_DFN :
          printf ("%u\n", pnt[i].imd);
          break;

        case c_DFN :
          printf ("'%c'\n", pnt[i].imd);
          break;

        case d_DFN :
          printf ("%d\n", pnt[i].imd);
          break;

        case l_DFN :
          printf ("%ld\n", pnt[i].imd);
          break;

        case x_DFN :
          printf ("0x%x\n", pnt[i].imd);
          break;

        case ABL_DFN :
          if (pnt[i].dat != NULL)
            {
            buffer [0] = '\0';
            buffer = exprToCharInt (pnt[i].dat, 1, buffer, &buff_size);
            printf ("%s", buffer);
            }
          printf ("\n");
          break;

        case BDD_DFN :
          if ((pnt[i].dat != NULL) && (circuit_pnt != NULL))
            {
            buffer [0] = '\0';
            ptr_abl = bddToAbl (pnt[i].dat, circuit_pnt->pNameI);
            buffer  = exprToCharInt (ptr_abl, 1, buffer, &buff_size);
            printf ("%s", buffer);
            }
          printf ("\n");
          break;

        case GEX_DFN :
          if (pnt[i].dat != NULL)
            printf ("Gex displayer not yet available\n");
          break;
        }
      }

    if ((typ [i] & KIND_DFN) == POINTER_DFN)
      {
      printf ("-> %-15s : ", str [i]);
      if (pnt[i].dat != NULL)
        printf ("available");
      printf ("\n");
      }

    if ( ((typ [i] & KIND_DFN) == ARRAY_DFN   ) ||
         ((typ [i] & KIND_DFN) == ARYOFPNT_DFN) )
      {
      printf ("=> %-15s : ", str [i]);
      if (pnt[i].dat != NULL)
        printf ("available");
      printf ("\n");
      }
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: splitline						*/
/* description	: read a line (the space must have been reserved by the	*/
/*		  caller - *words) from the standard input and split it	*/
/*		  into seperate words. Return the number of words read.	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static int splitline (words, line)

char **words;
char  *line ;
  {

  char *heap   = *words;
  int   new    =  1;
  int   wrdcnt =  0;
  int   i      =  0;
  int   j      =  0;

	/* ###------------------------------------------------------### */
	/*    copy the line read from input into the words until a	*/
	/* space has been found. Then, increment the word count and	*/
	/* begin filling the next word					*/
	/* ###------------------------------------------------------### */

  while ((heap[j] = line[i++]) != '\0')
    {
    if ((heap[j] != ' ') && (heap[j] != '\t'))
      {
      new = 0;
      j++;
      }
    else
      {
      if (new != 1)
        {
        heap  [j++]      = '\0';
        words [++wrdcnt] = heap + j;
        new = 1;
        }
      }
    }

  if (new != 1)
    wrdcnt++;

  return (wrdcnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: push							*/
/* description	: push a structure on the stack				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static void push (stk, stkpnt, data, type)

struct stack *stk;
int          *stkpnt;
void         *data;
short         type;

  {
  int stkidx = *stkpnt;

  if (stkidx == STKSIZ_DFN)
    fprintf (stdout, "beh_debug :\tSTACK OVERFLOW !!\tSTACK OVERFLOW !!\n");
  else
    {
    if (data != NULL)
      {
      stkidx++;
      stk [stkidx].data = data;
      stk [stkidx].type = type;
      stk [stkidx].mark = 0;
      *stkpnt           = stkidx;
      }
    else
      fprintf (stderr ,"cannot push NULL pointer on the stack\n");
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: pop							*/
/* description	: pop N structures from the stack			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static void pop (stkpnt, count)

int          *stkpnt;
int           count;

  {
  int stkidx = *stkpnt;

  if (stkidx <= count)
    stkidx = 0;
  else
    stkidx -= count;

  *stkpnt = stkidx;
  }

/* ###--------------------------------------------------------------### */
/* function	: goforward						*/
/* description	: go forward through the NEXT field to find the Nth	*/
/*		  structure in the list and return it			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static struct chain *goforward (pnt, count)

struct chain *pnt;
int           count;

  {
  while ((count > 0) && (pnt != NULL) && (pnt->NEXT != NULL))
    {
    pnt = pnt->NEXT;
    count--;
    }
  return (pnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: translate						*/
/* description	: find the index of each word read from the command	*/
/*		  line. For each word check if it is a number. Return	*/
/*		  the index of the first word.				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static int translate (words, wrdcnt, strgs, nmbrs, flags, indxs)

char **words;				/* list of words		*/
int    wrdcnt;				/* number of words		*/
char **strgs;				/* list of recognized strings	*/
int   *nmbrs;				/* words translated to numbers	*/
char  *flags;				/* set if word is a number	*/
int   *indxs;				/* words' index in strgs table	*/

  {
  int i , j;

	/* ###------------------------------------------------------### */
	/*    initialize flags and indxs				*/
	/* ###------------------------------------------------------### */

  for (i=0 ; i<10 ; i++)
    {
    flags [i] = 0;
    indxs [i] = _error_DFN;
    }

	/* ###------------------------------------------------------### */
	/*    for each word :						*/
	/*      - check if it is a number				*/
	/*      - try to locate the word among the recognized strings.	*/
	/*        If found register it's index				*/
	/* ###------------------------------------------------------### */

  for (i=0 ; i<wrdcnt ; i++)
    {
    flags [i] = sscanf (words [i], "%u", &nmbrs [i]);
    for (j=0; j<MAXCMD_DFN ; j++)
      {
      if (!strcmp (strgs [j], words [i]))
        {
        indxs [i] = j;
        break;
        }
      }
    }
  return (indxs [0]);
  }

static void getsize (siz)

unsigned int siz [];

  {
  int i;
 
  for (i=0 ; i<MAXCMD_DFN ; i++)
    siz [i] = 0;


  siz [character_DFN] = sizeof (char);
  siz [short_DFN]     = sizeof (short);
  siz [integer_DFN]   = sizeof (int);
  siz [long_DFN]      = sizeof (long);
  siz [string_DFN]    = sizeof (char *);

  siz [chain_DFN]     = sizeof (struct chain);
  siz [ptype_DFN]     = sizeof (struct ptype);

  siz [befig_DFN]     = sizeof (struct befig);
  siz [bereg_DFN]     = sizeof (struct bereg);
  siz [bemsg_DFN]     = sizeof (struct bemsg);
  siz [berin_DFN]     = sizeof (struct berin);
  siz [beout_DFN]     = sizeof (struct beout);
  siz [bebus_DFN]     = sizeof (struct bebus);
  siz [beaux_DFN]     = sizeof (struct beaux);
  siz [bedly_DFN]     = sizeof (struct beaux);
  siz [bebux_DFN]     = sizeof (struct bebux);
  siz [bepor_DFN]     = sizeof (struct bepor);
  siz [begen_DFN]     = sizeof (struct begen);
  siz [biabl_DFN]     = sizeof (struct biabl);
  siz [binode_DFN]    = sizeof (struct binode);
  siz [bequad_DFN]    = sizeof (struct bequad);
  siz [beder_DFN]     = sizeof (struct beder);
  }

/* ###--------------------------------------------------------------### */
/* function	: disp_immd						*/
/* description	: display a specific field as an immediate		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

static void disp_immd (str, pnt, type)

char        **str;
union value pnt  ;
int         type ;

  {
  extern char  *exprToCharInt ();
  char         *lcl_str         ;
  struct chain *ptr_abl         ;

  printf ("   %-15s: ", str [type]);
  switch (type)
    {
    case string_DFN :
      if (((char *) pnt.dat) != NULL)
        printf ("%s", (char *) pnt.dat);
      printf ("\n");
      break;

    case character_DFN :
      printf ("%c\n", pnt.imd);
      break;

    case short_DFN     :
    case integer_DFN   :
    case long_DFN      :
      printf ("0x%x\n", pnt.imd);
      break;

    case abl_DFN :
      if (pnt.dat != NULL)
        {
        buffer [0] = '\0';
        buffer = exprToCharInt (pnt.dat, 1, buffer, &buff_size);
        printf ("%s", buffer);
        }
      printf ("\n");
      break;

    case bdd_DFN :
      if ((pnt.dat != NULL) && (circuit_pnt != NULL))
        {
        buffer [0] = '\0';
        ptr_abl = bddToAbl (pnt.dat, circuit_pnt->pNameI);
        buffer  = exprToCharInt (ptr_abl, 1, buffer, &buff_size);
        printf ("%s", buffer);
        }
      printf ("\n");
      break;

    case gex_DFN :
      if (pnt.dat != NULL)
        printf ("Gex displayer not yet available\n");
      break;

    default :
      fprintf (stderr ,"unrecognized command\n");
    }
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_depend.c						*/
/* date		: Aug 27 1997						*/
/* version	: v110							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


	/* ###------------------------------------------------------### */
	/*    A hash table containing inputs name and their pointers	*/
	/* ###------------------------------------------------------### */

static ht *rin_list = NULL;

/* ###--------------------------------------------------------------### */
/* function	: beh_expdepend						*/
/* description	: find dependencies for an expression			*/
/* called func.	: beh_namelist, gethtitem				*/
/* ###--------------------------------------------------------------### */

static struct chain *beh_expdepend (pt_fig, pt_exp)

struct befig *pt_fig;
struct chain *pt_exp;

  {
  struct chain *res_chain = NULL;
  struct chain *del_chain = NULL;
  struct chain *tmp_chain = NULL;
  struct chain *mem_chain = NULL;
  int           value           ;

  if (pt_fig->ERRFLG == 0)
    {
	/* ###------------------------------------------------------### */
	/*   if there is no error in the current figure, find the list	*/
	/* of terminals for the expression				*/
	/* ###------------------------------------------------------### */

    res_chain = beh_namelist (pt_exp);

	/* ###------------------------------------------------------### */
	/*   scan the list of terminals to replace terminal names by	*/
	/* the pointer on the corresponding primary input (berin).	*/
	/*   check the unicity of terminals ...				*/
	/* ###------------------------------------------------------### */

    tmp_chain = res_chain;

    while (tmp_chain != NULL)
      {
      del_chain = tmp_chain->NEXT;
      while (del_chain != NULL)
        {
        mem_chain = del_chain->NEXT;

        if (del_chain->DATA == tmp_chain->DATA)
          tmp_chain = delchain (tmp_chain, del_chain);

        del_chain = mem_chain;
        }

      value = gethtitem (rin_list, tmp_chain->DATA);

      if (value == EMPTYHT)
        pt_fig->ERRFLG = beh_error (3, tmp_chain->DATA);
      else
        tmp_chain->DATA = (void *) value;

      tmp_chain = tmp_chain->NEXT;
      }
    }

  return (res_chain);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_depend						*/
/* description	: find dependencies between signals for a behavioural	*/
/*		  description.						*/
/* called func.	: beh_expdepend, addht    , addhtitem, addchain,	*/
/*		  freechain    , append   , delhtitem 			*/
/* ###--------------------------------------------------------------### */

void beh_depend (pt_fig)

struct befig *pt_fig;

  {
  struct chain *res_chain = NULL;
  struct beaux *pt_aux    ;
  struct beaux *pt_dly    ;
  struct bebux *pt_bux    ;
  struct bereg *pt_reg    ;
  struct bemsg *pt_msg    ;
  struct berin *pt_rin    ;
  struct beout *pt_out    ;
  struct bebus *pt_bus    ;
  struct biabl *pt_biabl  ;
  unsigned int  count     = 0;

	/* ###------------------------------------------------------### */
	/*   if there is no error in the current figure ...		*/
	/* ###------------------------------------------------------### */

  if ((pt_fig != NULL) && (pt_fig->ERRFLG == 0))
    {
	/* ###------------------------------------------------------### */
	/*   initialize a hash table with inputs			*/
	/* ###------------------------------------------------------### */

    pt_rin = pt_fig->BERIN;
    while (pt_rin != NULL)
      {
      count++;
      pt_rin = pt_rin->NEXT;
      }

    pt_rin = pt_fig->BERIN;

    if (pt_rin != NULL)
      rin_list = addht (count);

    while (pt_rin != NULL)
      {
      addhtitem (rin_list, pt_rin->NAME, pt_rin);
      pt_rin = pt_rin->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process simple internal signals.				*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    pt_aux = pt_fig->BEAUX;
    while (pt_aux != NULL)
      {
      res_chain = beh_expdepend (pt_fig, pt_aux->ABL);
      while (res_chain != NULL)
        {
        pt_rin = (struct berin *) res_chain->DATA;

        if ((pt_rin->AUX_REF == NULL) || (pt_rin->AUX_REF->DATA != pt_aux))
          pt_rin->AUX_REF = addchain (pt_rin->AUX_REF, pt_aux);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_aux = pt_aux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process delayed internal signals				*/
	/*     check that the signal does not already belong to the	*/
	/*     dependency list before adding it to the list		*/
	/* ###------------------------------------------------------### */

    pt_dly = pt_fig->BEDLY;
    while (pt_dly != NULL)
      {
      res_chain = beh_expdepend (pt_fig, pt_dly->ABL);
      while (res_chain != NULL)
        {
        pt_rin = (struct berin *) res_chain->DATA;

        if ((pt_rin->DLY_REF == NULL) || (pt_rin->DLY_REF->DATA != pt_dly))
          pt_rin->DLY_REF = addchain (pt_rin->DLY_REF, pt_dly);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_dly = pt_dly->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process simple ouputs					*/
	/*     check that the signal does not already belong to the	*/
	/*     dependency list before adding it to the list		*/
	/* ###------------------------------------------------------### */

    pt_out = pt_fig->BEOUT;
    while (pt_out != NULL)
      {
      res_chain = beh_expdepend (pt_fig, pt_out->ABL);
      while (res_chain != NULL)
        {
        pt_rin = (struct berin *) res_chain->DATA;

        if ((pt_rin->OUT_REF == NULL) || (pt_rin->OUT_REF->DATA != pt_out))
          pt_rin->OUT_REF = addchain (pt_rin->OUT_REF, pt_out);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_out = pt_out->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process bussed ouputs.					*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_bus    = pt_fig->BEBUS;
    while (pt_bus != NULL)
      {
      pt_biabl = pt_bus->BIABL;
      while (pt_biabl != NULL)
        {
        res_chain = append (res_chain, beh_expdepend (pt_fig,pt_biabl->VALABL));
        res_chain = append (res_chain, beh_expdepend (pt_fig,pt_biabl->CNDABL));
        pt_biabl  = pt_biabl->NEXT;
        }

      while (res_chain != NULL)
        {
        pt_rin = (struct berin *) res_chain->DATA;

        if ((pt_rin->BUS_REF == NULL) || (pt_rin->BUS_REF->DATA != pt_bus))
          pt_rin->BUS_REF = addchain (pt_rin->BUS_REF, pt_bus);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_bus = pt_bus->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process bussed internal signals.				*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_bux    = pt_fig->BEBUX;
    while (pt_bux != NULL)
      {
      pt_biabl = pt_bux->BIABL;
      while (pt_biabl != NULL)
        {
        res_chain = append (res_chain, beh_expdepend (pt_fig,pt_biabl->VALABL));
        res_chain = append (res_chain, beh_expdepend (pt_fig,pt_biabl->CNDABL));
        pt_biabl  = pt_biabl->NEXT;
        }

      while (res_chain != NULL)
        {
        pt_rin = (struct berin *) res_chain->DATA;

        if ((pt_rin->BUX_REF == NULL) || (pt_rin->BUX_REF->DATA != pt_bux))
          pt_rin->BUX_REF = addchain (pt_rin->BUX_REF, pt_bux);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_bux = pt_bux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process internal registers.				*/
	/*   check that the signal does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_reg    = pt_fig->BEREG;
    while (pt_reg != NULL)
      {
      pt_biabl = pt_reg->BIABL;
      while (pt_biabl != NULL)
        {
        res_chain = append (res_chain, beh_expdepend (pt_fig,pt_biabl->VALABL));
        res_chain = append (res_chain, beh_expdepend (pt_fig,pt_biabl->CNDABL));
        pt_biabl  = pt_biabl->NEXT;
        }

      while (res_chain != NULL)
        {
        pt_rin = (struct berin *) res_chain->DATA;

        if ((pt_rin->REG_REF == NULL) || (pt_rin->REG_REF->DATA != pt_reg))
          pt_rin->REG_REF = addchain (pt_rin->REG_REF, pt_reg);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_reg = pt_reg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   process assert statements.					*/
	/*   check that the statement does not already belong to the	*/
	/* dependency list before adding it to the list			*/
	/* ###------------------------------------------------------### */

    res_chain = NULL;
    pt_msg    = pt_fig->BEMSG;
    while (pt_msg != NULL)
      {
      res_chain = beh_expdepend (pt_fig, pt_msg->ABL);
      while (res_chain != NULL)
        {
        pt_rin = (struct berin *) res_chain->DATA;

        if ((pt_rin->MSG_REF == NULL) || (pt_rin->MSG_REF->DATA != pt_msg))
          pt_rin->MSG_REF = addchain (pt_rin->MSG_REF, pt_msg);

        res_chain = res_chain->NEXT;
        }
      freechain (res_chain);
      pt_msg = pt_msg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   release hash tables and lists that they contain		*/
	/* ###------------------------------------------------------### */

    if (pt_fig->BERIN != NULL)
      delht(rin_list);

    }
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_freabl.c						*/
/* date		: Sep  9 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_freabl						*/
/* description	: free all ABLs in BEFIG structure			*/
/* called func.	: freeExpr						*/
/* ###--------------------------------------------------------------### */

void beh_freabl (ptr_befig)

struct befig *ptr_befig;		/* pointer on current BEFIG	*/
  {
  struct beaux *ptr_beaux;
  struct bemsg *ptr_bemsg;
  struct beout *ptr_beout;
  struct bebus *ptr_bebus;
  struct biabl *ptr_biabl;
  struct bebux *ptr_bebux;
  struct bereg *ptr_bereg;

	/* ###------------------------------------------------------### */
	/*    check that the unit exists				*/
	/* ###------------------------------------------------------### */

  if (ptr_befig != NULL)
    {
	/* ###------------------------------------------------------### */
	/*    release expression in simple internal signals' list	*/
	/* ###------------------------------------------------------### */

    ptr_beaux = ptr_befig->BEAUX;
    while (ptr_beaux != NULL)
      {
      if (ptr_beaux->ABL != NULL)
        {
        freeExpr (ptr_beaux->ABL);
        ptr_beaux->ABL = NULL;
        }
      ptr_beaux = ptr_beaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    release expression in delayed signals' list		*/
	/* ###------------------------------------------------------### */

    ptr_beaux = ptr_befig->BEDLY;
    while (ptr_beaux != NULL)
      {
      if (ptr_beaux->ABL != NULL)
        {
        freeExpr (ptr_beaux->ABL);
        ptr_beaux->ABL = NULL;
        }
      ptr_beaux = ptr_beaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    release expression in assertions' list			*/
	/* ###------------------------------------------------------### */

    ptr_bemsg = ptr_befig->BEMSG;
    while (ptr_bemsg != NULL)
      {
      if (ptr_bemsg->ABL != NULL)
        {
        freeExpr (ptr_bemsg->ABL);
        ptr_bemsg->ABL = NULL;
        }
      ptr_bemsg = ptr_bemsg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    release expression in simple output ports' list		*/
	/* ###------------------------------------------------------### */

    ptr_beout = ptr_befig->BEOUT;
    while (ptr_beout != NULL)
      {
      if (ptr_beout->ABL != NULL) 
        {
        freeExpr (ptr_beout->ABL);
        ptr_beout->ABL = NULL;
        }
      ptr_beout = ptr_beout->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    release expression in bussed output ports' list		*/
	/* ###------------------------------------------------------### */

    ptr_bebus = ptr_befig->BEBUS;
    while (ptr_bebus != NULL)
      {
      ptr_biabl = ptr_bebus->BIABL;
      while (ptr_biabl != NULL)
        {
        if (ptr_biabl->CNDABL != NULL)
          {
          freeExpr(ptr_biabl->CNDABL);
          ptr_biabl->CNDABL = NULL;
          }
        if (ptr_biabl->VALABL != NULL)
          {
          freeExpr(ptr_biabl->VALABL);
          ptr_biabl->VALABL = NULL;
          }
        ptr_biabl = ptr_biabl->NEXT;
        }
      ptr_bebus = ptr_bebus->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    release expression in bussed internal signals' list	*/
	/* ###------------------------------------------------------### */

    ptr_bebux = ptr_befig->BEBUX;
    while (ptr_bebux != NULL)
      {
      ptr_biabl  = ptr_bebux->BIABL;
      while (ptr_biabl != NULL)
        {
        if (ptr_biabl->CNDABL != NULL)
          {
          freeExpr (ptr_biabl->CNDABL);
          ptr_biabl->CNDABL = NULL;
          }
        if (ptr_biabl->VALABL != NULL)
          {
          freeExpr   (ptr_biabl->VALABL);
          ptr_biabl->VALABL = NULL;
          }

        ptr_biabl  = ptr_biabl->NEXT;
        }
      ptr_bebux = ptr_bebux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    release expression in internal registers' list		*/
	/* ###------------------------------------------------------### */

    ptr_bereg = ptr_befig->BEREG;
    while (ptr_bereg != NULL)
      {
      ptr_biabl = ptr_bereg->BIABL;
      while (ptr_biabl != NULL)
        {
        if ( ptr_biabl->CNDABL != NULL )
          {
          freeExpr(ptr_biabl->CNDABL);
          ptr_biabl->CNDABL = NULL;
          }
        if ( ptr_biabl->VALABL != NULL )
          {
          freeExpr(ptr_biabl->VALABL);
          ptr_biabl->VALABL = NULL;
          }
        ptr_biabl = ptr_biabl->NEXT;
        }
      ptr_bereg = ptr_bereg->NEXT;
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_indexbdd.c					*/
/* date		: Aug 28 1997						*/
/* version	: v110							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


#define CIRCUI_SEC_FACTOR  2
#define CIRCUI_EXP_FACTOR  2

/* ###--------------------------------------------------------------### */
/* function	: beh_indexbdd						*/
/* description	: initilize the index table for bdd			*/
/* called func.	: initializeCct, addInputCct				*/
/* ###--------------------------------------------------------------### */

void beh_indexbdd (pt_fig)

struct befig *pt_fig;

  {
  struct berin *pt_rin  = NULL;
  struct beout *pt_out  = NULL;
  struct bebus *pt_bus  = NULL;
  struct beaux *pt_aux  = NULL;
  struct bebux *pt_bux  = NULL;
  struct bereg *pt_reg  = NULL;
  struct bemsg *pt_msg  = NULL;
  struct biabl *pt_bia  = NULL;
  unsigned int  cnt_in  = 0   ;
  unsigned int  cnt_out = 0   ;
  char          tmpname[256]  ;

  if ((pt_fig != NULL) && (pt_fig->CIRCUI == NULL))
    {
      /* ###--------------------------------------------------------### */
      /*    count inputs						*/
      /* ###--------------------------------------------------------### */

    pt_rin = pt_fig->BERIN;
    while (pt_rin != NULL)
      {
      cnt_in ++;
      pt_rin = pt_rin->NEXT;
      }

      /* ###--------------------------------------------------------### */
      /*    count outputs						*/
      /* ###--------------------------------------------------------### */

    pt_out = pt_fig->BEOUT;
    while (pt_out != NULL)
      {
      cnt_out ++;
      pt_out = pt_out->NEXT;
      }

    pt_bus = pt_fig->BEBUS;
    while (pt_bus != NULL)
      {
      pt_bia = pt_bus->BIABL;
      while (pt_bia != NULL)
        {
        cnt_out += 2;
        pt_bia   = pt_bia->NEXT;
        }
      pt_bus = pt_bus->NEXT;
      }

    pt_aux = pt_fig->BEAUX;
    while (pt_aux != NULL)
      {
      cnt_out ++;
      pt_aux = pt_aux->NEXT;
      }

    pt_aux = pt_fig->BEDLY;
    while (pt_aux != NULL)
      {
      cnt_out ++;
      pt_aux = pt_aux->NEXT;
      }

    pt_bux = pt_fig->BEBUX;
    while (pt_bux != NULL)
      {
      pt_bia = pt_bux->BIABL;
      while (pt_bia != NULL)
        {
        cnt_out += 2;
        pt_bia   = pt_bia->NEXT;
        }
      pt_bux = pt_bux->NEXT;
      }

    pt_reg = pt_fig->BEREG;
    while (pt_reg != NULL)
      {
      pt_bia = pt_reg->BIABL;
      while (pt_bia != NULL)
        {
        cnt_out += 2;
        pt_bia   = pt_bia->NEXT;
        }
      pt_reg = pt_reg->NEXT;
      }

    pt_msg = pt_fig->BEMSG;
    while (pt_msg != NULL)
      {
      cnt_out ++;
      pt_msg = pt_msg->NEXT;
      }

      /* ###--------------------------------------------------------### */
      /*    initialization						*/
      /* ###--------------------------------------------------------### */

    if (BEH_CIRCUI & BEH_CIRCUI_EXPANDED)
      cnt_in = CIRCUI_SEC_FACTOR * CIRCUI_EXP_FACTOR * cnt_in;
    else
      cnt_in = CIRCUI_SEC_FACTOR * cnt_in;

    pt_fig->CIRCUI = initializeCct (pt_fig->NAME, cnt_in, cnt_out);

      /* ###--------------------------------------------------------### */
      /*    define an index for each primary signal			*/
      /* ###--------------------------------------------------------### */

    pt_rin = pt_fig->BERIN;
    while (pt_rin != NULL)
      {
      addInputCct (pt_fig->CIRCUI, pt_rin->NAME);

      if (BEH_CIRCUI & BEH_CIRCUI_EXPANDED)
        {
        sprintf (tmpname, "%s__GNT",pt_rin->NAME);
        addInputCct (pt_fig->CIRCUI, namealloc(tmpname));
        }
      pt_rin = pt_rin->NEXT;
      }
    }
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_makbdd.c						*/
/* date		: Sep 20 1994						*/
/* version	: v106							*/
/* authors	: VUONG H.N., L.A. TABUSSE, P. BAZARGAN, J. DUNOYER	*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */
/* modifs	: DUNOYER Julien, VUILLEMIN Laurent			*/
/*		- Refonte du pointeur NODE pour avoir BDD, GEX,		*/
/*		  VARIABLE-LISTS, DERIVATE.				*/
/* 		- Changement de l'interface de beh_makbdd,		*/
/* 		  Plus de trace_flg, plus de aux_flg, mais un node_flg.	*/
/* 		- la fonction beh_abl2bdd est rendue publique.		*/
/* 		- ajout des appels a addOutputCct (package BDD luc).	*/
/* 		  pour les BEAUX et les BEDLY (utilise par PRB).	*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_abl2bdd						*/
/* description	: transform recursively an abl into a bdd 		*/
/* called func.	: searchInputCct, createNodeTermBdd, addListBdd,	*/
/*		  beh_abl2bdd   , applyBdd         , freechain ,	*/
/*		  beh_error     ,					*/
/* ###--------------------------------------------------------------### */

pNode beh_abl2bdd (pt_befig, expr)

struct befig *pt_befig;		/* pointer on  behavioural figure	*/
struct chain *expr    ;		/* pointer abl expression		*/

  {
  short         oper        ;
  struct node  *pt_bdd      = zero;
  struct chain *lstGbd      = NULL;
  pCircuit      ptr_circuit = pt_befig->CIRCUI;
  int           err_flg     = 0;
  int           index       ;
  char         *atom        ;
  char          name [256]  ;

	/* ###------------------------------------------------------### */
	/*    check that the expression exists				*/
	/* ###------------------------------------------------------### */

  if (expr == NULL)
    err_flg = beh_error (2, NULL);
  else
    {

    if (ATOM(expr))
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is a terminal (atom)			*/
	/* ###------------------------------------------------------### */

      atom = (char *) VALUE_ATOM (expr);

	/* ###------------------------------------------------------### */
	/*    if the terminal is a primary signal create a single node	*/
	/* with its index and return					*/
	/* ###------------------------------------------------------### */

      if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
        pt_bdd = createNodeTermBdd (index);
      else
        {
	/* ###------------------------------------------------------### */
	/* the terminal should be a const. '0', '1' or 'd' (don't care)	*/
	/* ###------------------------------------------------------### */

        if ((!strcmp(atom, "'0'")) || (!strcmp(atom, "'d'")))
          pt_bdd = zero;
        else
          {
          if (!strcmp(atom, "'1'"))
            pt_bdd = one;
          else
            {
            err_flg = beh_error (3, atom);
            }
          }
        }
      }
    else
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is not a terminal (is a complex expr.)	*/
	/* get the operator, create the bdd of each operand then,	*/
	/* combine operands applying the operator			*/
	/* ###------------------------------------------------------### */

      lstGbd = NULL;

      if ((oper = OPER(expr)) != STABLE)
       {
       while ((expr = CDR(expr)) != NULL)
        lstGbd = addListBdd (lstGbd, beh_abl2bdd (pt_befig, CAR(expr)));

       pt_bdd = applyBdd (oper, lstGbd);
       freechain (lstGbd);
       }
      else
       {
	expr = CDR(expr);
	if (!ATOM(expr))
	  err_flg = beh_error (4, NULL);
        else
          {
	  atom = (char *)VALUE_ATOM (CAR(expr));

          if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
            {
            lstGbd = addListBdd (lstGbd, createNodeTermBdd (index));
	    sprintf (name, "%s'delayed", atom);
	    atom = namealloc (name);
            if ((index = searchInputCct (ptr_circuit, atom)) != VIDE)
              {
              lstGbd = addListBdd (lstGbd, createNodeTermBdd (index));
              pt_bdd = applyBdd (NXOR, lstGbd);
              }
            else
	      err_flg = beh_error (3, atom);
            freechain (lstGbd);
            }
          else
	    err_flg = beh_error (3, atom);
          }
        }
      }
    }

  if (err_flg != 0)
    pt_befig->ERRFLG = 1;

  return (pt_bdd);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_mbddtot						*/
/* description	: make bdds for all primary signals: simple and bussed	*/
/*		  outputs, simple and bussed internal signals and	*/
/*		  registers						*/
/* called func.	: beh_abl2bdd, addchain , delchain			*/
/* ###--------------------------------------------------------------### */

static void beh_mbddtot (pt_befig)

struct befig *pt_befig;		/* pointer on current BEFIG	*/

  {
  struct beaux  *pt_beaux;
  struct bemsg  *pt_bemsg;
  struct beout  *pt_beout;
  struct bebus  *pt_bebus;
  struct bebux  *pt_bebux;
  struct bereg  *pt_bereg;
  struct biabl  *pt_biabl;
  struct binode *pt_binode;
  struct node   *pt_cndbdd;
  struct node   *pt_valbdd;

	/* ###------------------------------------------------------### */
	/*    make a bdd for each simple internal signal		*/
	/* ###------------------------------------------------------### */

  pt_beaux = pt_befig->BEAUX;

  while (pt_beaux != NULL)
    {
    pt_valbdd = beh_abl2bdd (pt_befig, pt_beaux->ABL);
    addOutputCct (pt_befig->CIRCUI, pt_beaux->NAME, pt_valbdd);

    if (pt_befig->TYPE & BEH_NODE_QUAD)
      ((struct bequad *)(pt_beaux->NODE))->BDD = pt_valbdd;
    else
      pt_beaux->NODE = pt_valbdd;

    pt_beaux = pt_beaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each delayed internal signal		*/
	/* ###------------------------------------------------------### */

  pt_beaux = pt_befig->BEDLY;
  while (pt_beaux != NULL)
    {
    pt_valbdd = beh_abl2bdd (pt_befig, pt_beaux->ABL);
    addOutputCct (pt_befig->CIRCUI, pt_beaux->NAME, pt_valbdd);

    if (pt_befig->TYPE & BEH_NODE_QUAD)
      ((struct bequad *)(pt_beaux->NODE))->BDD = pt_valbdd;
    else
      pt_beaux->NODE = pt_valbdd;
    pt_beaux = pt_beaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each assertion				*/
	/* ###------------------------------------------------------### */

  pt_bemsg = pt_befig->BEMSG;
  while (pt_bemsg != NULL)
    {
    pt_valbdd = beh_abl2bdd (pt_befig, pt_bemsg->ABL);
    addOutputCct (pt_befig->CIRCUI, pt_bemsg->LABEL, pt_valbdd);

    if (pt_befig->TYPE & BEH_NODE_QUAD)
      ((bequad_list *)(pt_bemsg->NODE))->BDD = pt_valbdd;
    else
      pt_bemsg->NODE = pt_valbdd;

    pt_bemsg = pt_bemsg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each simple output port			*/
	/* ###------------------------------------------------------### */

  pt_beout = pt_befig->BEOUT;
  while (pt_beout != NULL)
    {
    pt_valbdd = beh_abl2bdd (pt_befig, pt_beout->ABL);
    addOutputCct (pt_befig->CIRCUI, pt_beout->NAME, pt_valbdd);

    if (pt_befig->TYPE & BEH_NODE_QUAD)
      ((bequad_list *)(pt_beout->NODE))->BDD = pt_valbdd;
    else
      pt_beout->NODE = pt_valbdd;

    pt_beout = pt_beout->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each bussed output port			*/
	/* ###------------------------------------------------------### */

  pt_bebus = pt_befig->BEBUS;
  while (pt_bebus != NULL)
    {
    pt_biabl  = pt_bebus->BIABL;
    pt_binode = pt_bebus->BINODE;

    while (pt_biabl != NULL)
      {
      pt_cndbdd = beh_abl2bdd (pt_befig, pt_biabl->CNDABL);
      pt_valbdd = beh_abl2bdd (pt_befig, pt_biabl->VALABL);
      addOutputCct (pt_befig->CIRCUI, pt_bebus->NAME, pt_cndbdd);
      addOutputCct (pt_befig->CIRCUI, pt_bebus->NAME, pt_valbdd);

      if (pt_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(pt_binode->CNDNODE))->BDD = pt_cndbdd;
        ((bequad_list *)(pt_binode->VALNODE))->BDD = pt_valbdd;
        }
      else
        {
        pt_binode->CNDNODE = pt_cndbdd;
        pt_binode->VALNODE = pt_valbdd;
        }

      pt_biabl  = pt_biabl->NEXT;
      pt_binode = pt_binode->NEXT;
      }

    pt_bebus = pt_bebus->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each bussed internal signal		*/
	/* ###------------------------------------------------------### */

  pt_bebux = pt_befig->BEBUX;
  while (pt_bebux != NULL)
    {
    pt_biabl  = pt_bebux->BIABL;
    pt_binode = pt_bebux->BINODE;

    while (pt_biabl != NULL)
      {
      pt_cndbdd = beh_abl2bdd (pt_befig, pt_biabl->CNDABL);
      pt_valbdd = beh_abl2bdd (pt_befig, pt_biabl->VALABL);
      addOutputCct (pt_befig->CIRCUI, pt_bebux->NAME, pt_cndbdd);
      addOutputCct (pt_befig->CIRCUI, pt_bebux->NAME, pt_valbdd);

      if (pt_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(pt_binode->CNDNODE))->BDD = pt_cndbdd;
        ((bequad_list *)(pt_binode->VALNODE))->BDD = pt_valbdd;
        }
      else
        {
        pt_binode->CNDNODE = pt_cndbdd;
        pt_binode->VALNODE = pt_valbdd;
        }

      pt_biabl  = pt_biabl->NEXT ;
      pt_binode = pt_binode->NEXT;
      }
    pt_bebux = pt_bebux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a bdd for each internal register			*/
	/* ###------------------------------------------------------### */

  pt_bereg = pt_befig->BEREG;
  while (pt_bereg != NULL)
    {
    pt_biabl  = pt_bereg->BIABL;
    pt_binode = pt_bereg->BINODE;

    while (pt_biabl != NULL)
      {
      pt_cndbdd = beh_abl2bdd (pt_befig, pt_biabl->CNDABL);
      pt_valbdd = beh_abl2bdd (pt_befig, pt_biabl->VALABL);
      addOutputCct (pt_befig->CIRCUI, pt_bereg->NAME, pt_cndbdd);
      addOutputCct (pt_befig->CIRCUI, pt_bereg->NAME, pt_valbdd);

      if (pt_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(pt_binode->CNDNODE))->BDD = pt_cndbdd;
        ((bequad_list *)(pt_binode->VALNODE))->BDD = pt_valbdd;
        }
      else
        {
        pt_binode->CNDNODE = pt_cndbdd;
        pt_binode->VALNODE = pt_valbdd;
        }

      pt_biabl  = pt_biabl->NEXT;
      pt_binode = pt_binode->NEXT;
      }
    pt_bereg = pt_bereg->NEXT;
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: beh_makbdd						*/
/* description	: translate all abls into bdds				*/
/* called func.	: beh_indexbdd, beh_mbddtot				*/
/* ###--------------------------------------------------------------### */

void beh_makbdd (pt_befig)

struct befig *pt_befig;

  {
  int err_flg = 0;

  if (!(pt_befig->TYPE & BEH_NODE_BDD));
    {
    beh_indexbdd (pt_befig);

        /* ###------------------------------------------------------### */
        /*    change ABLs to BDDs					*/
        /* ###------------------------------------------------------### */

    beh_mbddtot (pt_befig);
    pt_befig->TYPE |= BEH_NODE_BDD;

    if (err_flg != 0)
      pt_befig->ERRFLG = 1;
    }
  }
/* ###--------------------------------------------------------------### */
/* file		: beh_makderiv.c					*/
/* date		: Jan 24 1998						*/
/* version	: v111							*/
/* authors	: Laurent VUILLEMIN, Julien DUNOYER			*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


#define BEH_TIME_IDX 0	/* Index utilise pour la derivee dS/dt	*/
#define BEH_NEXT_IDX 1	/* Delta Index utilise pour la derivee dS/dt	*/

/* ###--------------------------------------------------------------### */
/* function     : beh_nextTimeBdd					*/
/* description  : copy a BDD with shifted (+BEH_NEXT_IDX) var index	*/
/* called func. : initVertexBdd						*/
/* ###--------------------------------------------------------------### */

static pNode beh_nextTimeBdd (pt_bdd)

pNode pt_bdd   ;

  {
  pNode varbdd;

  if (pt_bdd == zero)
    {
    varbdd = zero;
    }
  else if (pt_bdd == one)
    {
    varbdd = one;
    }
  else
    {
    varbdd = initVertexBdd ( pt_bdd->index + BEH_NEXT_IDX,
                             beh_nextTimeBdd (pt_bdd->high),
                             beh_nextTimeBdd (pt_bdd->low) );
    }

  return (varbdd);
  }

/* ###--------------------------------------------------------------### */
/* function     : beh_derivSigByTime					*/
/* description  : derivate a BDD by the time				*/
/* called func. : beh_nextTimeBdd, applyBinBdd				*/
/* ###--------------------------------------------------------------### */

static pNode beh_derivSigByTime (pt_bdd)

pNode pt_bdd   ;

  {
  pNode varbdd;

  varbdd   = beh_nextTimeBdd (pt_bdd);

  return (applyBinBdd (XOR, pt_bdd, varbdd));
  }

/* ###--------------------------------------------------------------### */
/* function     : beh_derivSigByVar					*/
/* description  : derivate a BDD by a variable				*/
/* called func. : constraintBdd, applyBinBdd				*/
/* ###--------------------------------------------------------------### */

static pNode beh_derivSigByVar (pt_bdd, index_var)

pNode pt_bdd   ;
short index_var;

  {
  pNode high  ;
  pNode low   ;
  pNode varbdd;

  varbdd = createNodeTermBdd (index_var);

  high   = constraintBdd (pt_bdd, varbdd);
  low    = constraintBdd (pt_bdd, notBdd(varbdd));

  return (applyBinBdd (XOR,high,low));
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_derivSig						*/
/* description	: derivate a BDD by all variables and by time [option]	*/
/* called func.	: beh_derivSigByVar, beh_derivSigByTime			*/
/* ###--------------------------------------------------------------### */

struct beder *beh_derivSig (pt_bdd, pt_varlist, time, mode)

pNode          pt_bdd      ;
struct chain  *pt_varlist  ;
unsigned int   time        ;
unsigned char  mode        ;

  {
  pNode         pt_node     ;
  struct beder *list_derive = NULL;

  if (mode & BEH_NODE_DERIVBYVAR)
    {
    while (pt_varlist != NULL)
      {
      pt_node     = beh_derivSigByVar (pt_bdd, pt_varlist->DATA);
      list_derive = beh_addbeder (list_derive, pt_node, pt_varlist->DATA, time);

      pt_varlist   = pt_varlist->NEXT;
      }
    }

  if (mode & BEH_NODE_DERIVBYTIME)
    {
    pt_node     = beh_derivSigByTime (pt_bdd);
    list_derive = beh_addbeder (list_derive, pt_node, BEH_TIME_IDX, time);
    }

  return (list_derive);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_makderiv						*/
/* description	: build the list of derivate for all signals		*/
/*                the mode is used to make a choice between		*/
/*                derivate by time or by all variables.			*/
/* called func.	: beh_derivSig						*/
/* ###--------------------------------------------------------------### */

void beh_makderiv (pt_befig, mode)

struct befig    *pt_befig;
unsigned char    mode;

  {
  struct beaux  *pt_beaux;
  struct beout  *pt_beout;
  struct bebus  *pt_bebus;
  struct bebux  *pt_bebux;
  struct bereg  *pt_bereg;
  struct bemsg  *pt_bemsg;
  struct binode *pt_binod;
  struct chain  *list_der;


	/* ###------------------------------------------------------### */
	/*   check that the BEFIG is ready to support the derivation	*/
	/* ###------------------------------------------------------### */

  if (pt_befig->TYPE & BEH_NODE_QUAD)
    {
	/* ###------------------------------------------------------### */
	/*   Use BEH_NODE_DERIVATE as a mask.				*/
	/* ###------------------------------------------------------### */

    mode = mode & BEH_NODE_DERIVATE;

	/* ###------------------------------------------------------### */
	/*   simple output ports					*/
	/* ###------------------------------------------------------### */

    pt_beout = pt_befig->BEOUT;
    while (pt_beout != NULL)
      {
      ((struct bequad *) (pt_beout->NODE))->DERIVATE =
        beh_derivSig (((struct bequad *) (pt_beout->NODE))->BDD,
                      ((struct bequad *) (pt_beout->NODE))->VARLIST,
                      pt_beout->TIME   , mode);

      pt_beout = pt_beout->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   bussed output ports					*/
	/* ###------------------------------------------------------### */

    pt_bebus = pt_befig->BEBUS;
    while (pt_bebus != NULL)
      {
      pt_binod = pt_bebus->BINODE;
      while (pt_binod != NULL)
         {
         ((struct bequad *) (pt_binod->CNDNODE))->DERIVATE =
           beh_derivSig (((struct bequad *) (pt_binod->CNDNODE))->BDD,
                      ((struct bequad *) (pt_binod->CNDNODE))->VARLIST,
                      pt_binod->TIME   , mode);

         ((struct bequad *) (pt_binod->VALNODE))->DERIVATE =
           beh_derivSig (((struct bequad *) (pt_binod->VALNODE))->BDD,
                      ((struct bequad *) (pt_binod->VALNODE))->VARLIST,
                      pt_binod->TIME   , mode);

         pt_binod = pt_binod->NEXT;
         }

      pt_bebus = pt_bebus->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   bussed internal signals					*/
	/* ###------------------------------------------------------### */

    pt_bebux = pt_befig->BEBUX;
    while (pt_bebux != NULL)
      {
      pt_binod = pt_bebux->BINODE;
      while (pt_binod != NULL)
         {
         ((struct bequad *) (pt_binod->CNDNODE))->DERIVATE =
           beh_derivSig (((struct bequad *) (pt_binod->CNDNODE))->BDD,
                      ((struct bequad *) (pt_binod->CNDNODE))->VARLIST,
                      pt_binod->TIME   , mode);

         ((struct bequad *) (pt_binod->VALNODE))->DERIVATE =
           beh_derivSig (((struct bequad *) (pt_binod->VALNODE))->BDD,
                      ((struct bequad *) (pt_binod->VALNODE))->VARLIST,
                      pt_binod->TIME   , mode);

         pt_binod = pt_binod->NEXT;
         }
      pt_bebux = pt_bebux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   internal registers						*/
	/* ###------------------------------------------------------### */

    pt_bereg = pt_befig->BEREG;
    while (pt_bereg != NULL)
      {
      pt_binod = pt_bereg->BINODE;
      while (pt_binod != NULL)
         {
         ((struct bequad *) (pt_binod->CNDNODE))->DERIVATE =
           beh_derivSig (((struct bequad *) (pt_binod->CNDNODE))->BDD,
                      ((struct bequad *) (pt_binod->CNDNODE))->VARLIST,
                      pt_binod->TIME   , mode);

         ((struct bequad *) (pt_binod->VALNODE))->DERIVATE =
           beh_derivSig (((struct bequad *) (pt_binod->VALNODE))->BDD,
                      ((struct bequad *) (pt_binod->VALNODE))->VARLIST,
                      pt_binod->TIME   , mode);

         pt_binod = pt_binod->NEXT;
         }
      pt_bereg = pt_bereg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   internal simple signals					*/
	/* ###------------------------------------------------------### */

    pt_beaux = pt_befig->BEAUX;
    while (pt_beaux != NULL)
      {
      ((struct bequad *) (pt_beaux->NODE))->DERIVATE =
        beh_derivSig (((struct bequad *) ( pt_beaux->NODE))->BDD,
                      ((struct bequad *) (pt_beaux->NODE))->VARLIST,
                      pt_beaux->TIME   , mode);

      pt_beaux = pt_beaux->NEXT;
      }


	/* ###------------------------------------------------------### */
	/*   delayed internal signals					*/
	/* ###------------------------------------------------------### */

    pt_beaux = pt_befig->BEDLY;
    while (pt_beaux != NULL)
      {
      ((struct bequad *) (pt_beaux->NODE))->DERIVATE =
        beh_derivSig (((struct bequad *) (pt_beaux->NODE))->BDD,
                      ((struct bequad *) (pt_beaux->NODE))->VARLIST,
                      pt_beaux->TIME   , mode);

      pt_beaux = pt_beaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*   assertions							*/
	/* ###------------------------------------------------------### */

    pt_bemsg = pt_befig->BEMSG;
    while (pt_bemsg != NULL)
      {
      ((struct bequad *) (pt_bemsg->NODE))->DERIVATE =
        beh_derivSig (((struct bequad *) (pt_bemsg->NODE))->BDD,
                      ((struct bequad *) (pt_bemsg->NODE))->VARLIST,
                      0   , mode);

      pt_bemsg = pt_bemsg->NEXT;
      }

    pt_befig->TYPE |= mode ;
    }

  else
    {
    beh_error (6, NULL);
    }
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_makgex.c						*/
/* date		: Mar  9 1994						*/
/* version	: v109							*/
/* authors	: Xavier Picat						*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


#define BEH_GEXBLK	256

static struct begex *BEH_GEXHED = NULL;
static ht           *HashTable  = NULL;

/* ###--------------------------------------------------------------### */
/* function	: beh_addgexex						*/
/* description	: create a GEX expression				*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct begex *beh_addgexex (oper, type)

unsigned int   oper;
unsigned short type;

  {
  struct begex *ptgex;
  int           i;

  if (BEH_GEXHED == NULL)
    {
	/* ###------------------------------------------------------### */
	/*    if there is no more structure allocate a new block	*/
	/* (use OPERAND field to chain structures in a list)		*/
	/* ###------------------------------------------------------### */

    BEH_GEXHED = mbkalloc (sizeof (struct begex) * BEH_GEXBLK);

    ptgex = BEH_GEXHED;
    for (i=1 ; i<BEH_GEXBLK ; i++)
      {
      ptgex->OPERAND = (struct chain *) (ptgex + 1);
      ptgex++;
      }
    ptgex->OPERAND = NULL;
    }

  ptgex      = BEH_GEXHED;
  BEH_GEXHED = (struct begex *) BEH_GEXHED->OPERAND;

  ptgex->TERM    = oper;
  ptgex->TYPE    = type;
  ptgex->OPERAND = NULL;

  return (ptgex);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_addterm						*/
/* description	: create a GEX for a terminal. For each terminal there	*/
/*		  is a unique GEX					*/
/* called func.	: beh_addgexex, mbkalloc, addchain			*/
/* ###--------------------------------------------------------------### */

static struct begex *beh_addterm (oper)

unsigned int oper;

  {
  static struct chain  *term_tab = NULL;
  static unsigned int   bank_nbr = 0;
  struct chain         *tmp_chn  = NULL;
  struct begex         *locl_tab;
  unsigned int          bank;
  struct begex         *resgex;
  int                   i;
  int                   j;

  bank = oper / 64;
  if (bank >= bank_nbr)
    {
    for (i=bank_nbr ; i<=bank ; i++)
      {
      locl_tab = (struct begex *) mbkalloc (64 * sizeof (struct begex));

      for (j=0 ; j<64 ; j++)
        {
        (locl_tab [j]).TERM    = j + (i * 64);
        (locl_tab [j]).TYPE    = 0;
        (locl_tab [j]).OPERAND = NULL;
        }

      if (term_tab == NULL)
        term_tab = addchain (NULL, locl_tab);
      else
        {
        tmp_chn = term_tab;
        while (tmp_chn->NEXT != NULL)
          tmp_chn = tmp_chn->NEXT;
        tmp_chn->NEXT = addchain (NULL, locl_tab);
        }
      }

    bank_nbr = bank + 1;
    }

  tmp_chn = term_tab;
  for (i=0 ; i<bank ; i++)
    tmp_chn = tmp_chn->NEXT;

  locl_tab = (struct begex *) tmp_chn->DATA;

  resgex = & (locl_tab [(unsigned int) (oper % 64)]);

  return (resgex);
  }

/* ###--------------------------------------------------------------### */
/* function	: mad_fregex						*/
/* description	: release a GEX expression (excluding terminals)	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void *mad_fregex (ptgex, mode)

struct begex *ptgex;
char          mode ;

  {
  struct begex *tmpgex;
  struct chain *tmpchn;

  if (ptgex != NULL)
    {
    if ((tmpchn = ptgex->OPERAND) != NULL)
      {
      if (mode == 'a')
        {
        while (tmpchn != NULL)
          {
          mad_fregex ((struct begex *) tmpchn->DATA, 'a');
          tmpchn = tmpchn->NEXT;
          }
        freechain (ptgex->OPERAND);
        }

      ptgex->OPERAND = (struct chain *) BEH_GEXHED;
      BEH_GEXHED     = ptgex;
      }
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: flatgex						*/
/* description	: flatten the top level of a complex expression		*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct begex *flatgex (gexpnt)

struct begex *gexpnt;

  {
  int           top_operator;
  int           operator;
  struct chain *tmp_chn;
  struct chain *cur_oper;
  struct chain *oper_list = NULL;
  struct begex *operand;
  struct begex *resgex   = gexpnt;
  char          flatflag = 0;

  if ((gexpnt != NULL) && (gexpnt->OPERAND != NULL))
    {
    cur_oper = gexpnt->OPERAND;

    while (cur_oper != NULL)
      {
      flatflag     = 0;
      top_operator = gexpnt->TERM;
      operand      = (struct begex *) cur_oper->DATA;

      if (operand->OPERAND != NULL)
        {
        operator = operand->TERM;
        switch (operator)
          {
          case OR :
            if (top_operator == NOT)
              {
              gexpnt->TERM = NOR;
              flatflag = 1;
              }
            else
              {
              if ((top_operator == OR) || (top_operator == NOR))
                flatflag = 1;
              }
            break;

          case AND :
            if (top_operator == NOT)
              {
              gexpnt->TERM = NAND;
              flatflag = 1;
              }
            else
              {
              if ((top_operator == AND) || (top_operator == NAND))
                flatflag = 1;
              }
            break;

          case XOR :
            if (top_operator == NOT)
              {
              gexpnt->TERM = NXOR;
              flatflag = 1;
              }
            else
              {
              if ((top_operator == XOR) || (top_operator == NXOR))
                flatflag = 1;
              }
            break;

          case NXOR :
            if (top_operator == XOR)
              {
              gexpnt->TERM = NXOR;
              flatflag     = 1;
              }
            else
              {
              if ((top_operator == NXOR) || (top_operator == NOT))
                {
                gexpnt->TERM = XOR;
                flatflag     = 1;
                }
              }
            break;
          }
        }

      if (flatflag == 1)
        {
        tmp_chn = operand->OPERAND;
        while (tmp_chn->NEXT != NULL)
          tmp_chn = tmp_chn->NEXT;

        tmp_chn->NEXT = oper_list;
        oper_list     = operand->OPERAND;
        mad_fregex (operand, 'p');
        }
      else
        oper_list = addchain (oper_list, operand);

      cur_oper = cur_oper->NEXT;
      }

    freechain (gexpnt->OPERAND);

    gexpnt->OPERAND = oper_list;
    resgex          = gexpnt;
    }

  return (resgex);
  }

/* ###--------------------------------------------------------------### */
/* function	: rmvconst						*/
/* description	: remove constantes that appear at the top level of a	*/
/*		  espression						*/
/* called func.	: addchain, freechain, mad_fregex			*/
/* ###--------------------------------------------------------------### */

static struct begex *rmvconst (gexpnt)

struct begex *gexpnt;

  {
  struct begex *resgex;
  struct begex *operand;
  struct begex *gex_zero;
  struct begex *gex_one ;
  struct chain *new_oper = NULL;
  struct chain *cur_oper;
  struct chain *tmp_chn = NULL;
  int           operator;
  char          freflag = 0;

  resgex   = gexpnt;
  gex_zero = beh_addterm (0);
  gex_one  = beh_addterm (1);

  if (gexpnt != NULL)
    {
    if (gexpnt->OPERAND != NULL)
      {
      cur_oper  = gexpnt->OPERAND;

      while (cur_oper != NULL)
        {
        operator = gexpnt->TERM;
        operand  = (struct begex *) cur_oper->DATA;
        freflag  = 0;

        if (operand == gex_zero)
          {
          switch (operator)
            {
            case AND :
              freflag = 2; resgex = gex_zero; break;
            case NAND :
            case NOT  :
              freflag = 2; resgex = gex_one ; break;
            case OR   :
            case XOR  :
            case NOR  :
            case NXOR :
              freflag = 1; break;
              break;
            }
          }

        if (operand == gex_one)
          {
          switch (operator)
            {
            case OR :
              freflag = 2; resgex = gex_one ; break;
            case NOR :
            case NOT :
              freflag = 2; resgex = gex_zero; break;
            case NAND :
            case AND  :
              freflag = 1; break;
              break;
            case XOR  :
              gexpnt->TERM = NXOR; freflag = 1; break;
              break;
            case NXOR :
              gexpnt->TERM = XOR ; freflag = 1; break;
              break;
            }
          }

        if (freflag == 2)
          {
          mad_fregex (gexpnt);
          freechain (new_oper);
          gexpnt = NULL;
          new_oper = NULL;
          break;
          }
        else
          {
          if (freflag == 0)
            new_oper = addchain (new_oper, cur_oper->DATA);
          cur_oper = cur_oper->NEXT;
          }
        }
      if (gexpnt != NULL)
        {
        operator = gexpnt->TERM;
        if (new_oper == NULL)
          {
          switch (operator)
            {
            case AND  :
            case NOR  :
            case NXOR :
              resgex = gex_one ;
              break;
            case OR   :
            case XOR  :
            case NAND :
              resgex = gex_zero;
              break;
            }
          mad_fregex (gexpnt);
          gexpnt = NULL;
          }
        else
          {
          if (new_oper->NEXT == NULL)
            {
            switch (operator)
              {
              case AND  :
              case OR   :
              case XOR  :
                resgex = (struct begex *) new_oper->DATA;
                freechain (new_oper);
                gexpnt = NULL;
                new_oper = NULL;
                break;
              case NOR  :
              case NXOR :
              case NAND :
              case NOT  :
                tmp_chn = gexpnt->OPERAND;
                gexpnt->OPERAND = new_oper;
                gexpnt->TERM    = NOT;
                freechain (tmp_chn);
                tmp_chn = NULL;
                resgex  = gexpnt;
                break;
              }
            }
          }
        }
      }
    }

  return (resgex);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_abl2gex						*/
/* description	: transform recursively abl in gex 			*/
/* called func.	: beh_toolbug, beh_abl2gex, gethtitem beh_addgexex,	*/
/*		  namealloc  , addchain					*/
/* ###--------------------------------------------------------------### */

static struct begex *beh_abl2gex (ptr_befig, expr)

struct befig *ptr_befig;
struct chain *expr;

  {
struct chain *my_operand;

  char          delayed [128];
  static char  *str_zero   = NULL;
  static char  *str_one    = NULL;
  struct begex *resgex       ;
  struct begex *gex_elt      ;
  struct chain *oper_list    ;
  struct chain *operand      ;
  char         *term         ;
  long          term_idx     ;
  long          operator     ;

  if (expr == NULL)
    beh_toolbug (103, "beh_abl2gex", NULL, 0);

  if (str_zero == NULL)
    {
    str_zero = namealloc ("'0'");
    str_one  = namealloc ("'1'");
    }

  if (expr->NEXT != NULL)
    {
	/* ###------------------------------------------------------### */
	/*    if the expression is not a terminal ...			*/
	/* ###------------------------------------------------------### */

    operator = (long) ((struct chain *)expr->DATA)->DATA;
    switch (operator)
      {
      case STABLE:

	/* ###------------------------------------------------------### */
	/*    translate signal'stable into a logical expression.	*/
	/* notice :							*/
	/*    signal'stable = not (signal xor signal'delayed)		*/
	/* ###------------------------------------------------------### */

        resgex = beh_addgexex (NXOR, 0);

        operand = (struct chain *) expr->NEXT->DATA;

        if (operand->NEXT != NULL)
          beh_toolbug (100, "beh_abl2gex", NULL, 0);
        else
          {
          term     = (char *) operand->DATA;
          term_idx = gethtitem (HashTable, term);

          if (term_idx == EMPTYHT)
            beh_toolbug (101, "beh_abl2gex", term, 0);
          else
            {
            gex_elt = beh_addterm (term_idx);

            resgex->OPERAND = addchain (NULL, gex_elt);

            sprintf (delayed, "%s'delayed", term);
            term     = namealloc (delayed);
            term_idx = gethtitem (HashTable, term);

            if (term_idx == EMPTYHT)
              beh_toolbug (101, "beh_abl2gex", term, 0);
            else
              {
              gex_elt = beh_addterm (term_idx);
              resgex->OPERAND->NEXT = addchain (NULL, gex_elt);
              }
            }
          }
        break;

      case OR:
      case AND:
      case XOR:
      case NOR:
      case NAND:
      case NXOR:
      case NOT:

	/* ###------------------------------------------------------### */
	/*    for other operator make the top level, then, translate	*/
	/* each operand							*/
	/* ###------------------------------------------------------### */

        resgex    = beh_addgexex (operator, 0);
        oper_list = expr->NEXT;
        while (oper_list != NULL)
          {
          gex_elt         = beh_abl2gex (ptr_befig, oper_list->DATA);
          resgex->OPERAND = addchain (resgex->OPERAND, gex_elt);
          oper_list    = oper_list->NEXT;
          }
        break;

      default:
        beh_toolbug (102, "beh_abl2gex", NULL, operator);
      }

    resgex = rmvconst (resgex);
    resgex = flatgex  (resgex);
    }

  else
    {
	/* ###------------------------------------------------------### */
	/*    if the expression is a terminal create a terminal gex	*/
	/* with the index correspnding to the terminal			*/
	/* ###------------------------------------------------------### */

    term     = (char *) expr->DATA;
    term_idx = gethtitem (HashTable, term);
    if (term_idx == EMPTYHT)
      {
      if (expr->DATA == str_one)
        resgex = beh_addterm (1);
      else
        {
        if (expr->DATA == str_zero)
          resgex = beh_addterm (0);
        else
          beh_toolbug (101, "beh_abl2gex", term, 0);
        }
      }
    else
      resgex = beh_addterm (term_idx);

    }

  return (resgex);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_makgex						*/
/* description	: make gex for each signal of a befig structure 	*/
/* called func.	: beh_abl2gex, addht, addhtitem, delht			*/
/* ###--------------------------------------------------------------### */

void beh_makgex (pt_befig)

struct befig *pt_befig;			/* befig containing expressions	*/

  {
  struct beaux  *pt_beaux ;
  struct bemsg  *pt_bemsg ;
  struct beout  *pt_beout ;
  struct bebus  *pt_bebus ;
  struct bebux  *pt_bebux ;
  struct bereg  *pt_bereg ;
  struct biabl  *pt_biabl ;
  struct binode *pt_binode;
  struct berin  *pt_rin   ;
  unsigned int   count    ;

  if (!(pt_befig->TYPE & BEH_NODE_GEX));
    {
	/* ###------------------------------------------------------### */
	/*    initialize hash table					*/
	/* ###------------------------------------------------------### */

    count  = 1;
    pt_rin = pt_befig->BERIN;

    while (pt_rin != NULL)
      {
      count ++;
      pt_rin = pt_rin->NEXT;
      }

    HashTable = addht (count);

    count  = 2;
    pt_rin = pt_befig->BERIN;

    while (pt_rin != NULL)
      {
      addhtitem (HashTable, pt_rin->NAME, count);
      count ++;
      pt_rin = pt_rin->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    make a gex for each simple internal signal		*/
	/* ###------------------------------------------------------### */

    pt_beaux = pt_befig->BEAUX;
    while (pt_beaux != NULL)
      {
      if (pt_befig->TYPE & BEH_NODE_QUAD)
        ((bequad_list *)(pt_beaux->NODE))->GEX =
		      beh_abl2gex (pt_befig, pt_beaux->ABL);
      else
        pt_beaux->NODE= (pNode)beh_abl2gex (pt_befig, pt_beaux->ABL);

      pt_beaux = pt_beaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    make a gex for each delayed internal signal		*/
	/* ###------------------------------------------------------### */

    pt_beaux = pt_befig->BEDLY;
    while (pt_beaux != NULL)
      {
      if (pt_befig->TYPE & BEH_NODE_QUAD)
        ((bequad_list *)(pt_beaux->NODE))->GEX =
		      beh_abl2gex (pt_befig, pt_beaux->ABL);
      else 
        pt_beaux->NODE= (pNode)beh_abl2gex(pt_befig,pt_beaux->ABL);

      pt_beaux = pt_beaux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    make a gex for each assertion				*/
	/* ###------------------------------------------------------### */

    pt_bemsg = pt_befig->BEMSG;
    while (pt_bemsg != NULL)
      {
      if (pt_befig->TYPE & BEH_NODE_QUAD)
        ((bequad_list *)(pt_bemsg->NODE))->GEX =
		      beh_abl2gex (pt_befig, pt_bemsg->ABL);
      else
        pt_bemsg->NODE=(pNode)beh_abl2gex(pt_befig,pt_bemsg->ABL);
      pt_bemsg       = pt_bemsg->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    make a gex for each simple output				*/
	/* ###------------------------------------------------------### */

    pt_beout = pt_befig->BEOUT;
    while (pt_beout != NULL)
      {
      if (pt_befig->TYPE & BEH_NODE_QUAD)
         ((bequad_list *)(pt_beout->NODE))->GEX =
		      beh_abl2gex (pt_befig, pt_beout->ABL);
      else 
        pt_beout->NODE=(pNode) beh_abl2gex (pt_befig,pt_beout->ABL);

      pt_beout = pt_beout->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    make a gex for each bussed output				*/
	/* ###------------------------------------------------------### */

    pt_bebus = pt_befig->BEBUS;
    while (pt_bebus != NULL)
      {
      pt_biabl  = pt_bebus->BIABL;
      pt_binode = pt_bebus->BINODE;

      while (pt_biabl != NULL)
        {
        if (pt_befig->TYPE & BEH_NODE_QUAD)
	  {
          ((bequad_list *)(pt_binode->CNDNODE))->GEX =
		        beh_abl2gex (pt_befig, pt_biabl->CNDABL);
          ((bequad_list *)(pt_binode->VALNODE))->GEX =
		        beh_abl2gex (pt_befig, pt_biabl->VALABL);
          }
        else 
	  {
	  pt_binode->CNDNODE=(pNode)beh_abl2gex(pt_befig,pt_biabl->CNDABL);
	  pt_binode->VALNODE=(pNode)beh_abl2gex(pt_befig,pt_biabl->VALABL);
	  }

        pt_biabl  = pt_biabl->NEXT;
        pt_binode = pt_binode->NEXT;
        }
      pt_bebus = pt_bebus->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    make a gex for each bussed internal signal		*/
	/* ###------------------------------------------------------### */

    pt_bebux = pt_befig->BEBUX;
    while (pt_bebux != NULL)
      {
      pt_biabl  = pt_bebux->BIABL;
      pt_binode = pt_bebux->BINODE;

      while (pt_biabl != NULL)
        {
        if (pt_befig->TYPE & BEH_NODE_QUAD)
	  {
          ((bequad_list *)(pt_binode->CNDNODE))->GEX  =
		        beh_abl2gex (pt_befig, pt_biabl->CNDABL);
          ((bequad_list *)(pt_binode->VALNODE))->GEX   =
		        beh_abl2gex (pt_befig, pt_biabl->VALABL);
          }
        else 
	  {
	  pt_binode->CNDNODE=(pNode)beh_abl2gex(pt_befig,pt_biabl->CNDABL);
	  pt_binode->VALNODE=(pNode)beh_abl2gex(pt_befig,pt_biabl->VALABL);
	  }

        pt_biabl  = pt_biabl->NEXT;
        pt_binode = pt_binode->NEXT;
        }
      pt_bebux = pt_bebux->NEXT;
      }

	/* ###------------------------------------------------------### */
	/*    make a gex for each internal register			*/
	/* ###------------------------------------------------------### */

    pt_bereg = pt_befig->BEREG;
    while (pt_bereg != NULL)
      {
      pt_biabl  = pt_bereg->BIABL;
      pt_binode = pt_bereg->BINODE;

      while (pt_biabl != NULL)
        {
        if (pt_befig->TYPE & BEH_NODE_QUAD)
	  {
          ((bequad_list *)(pt_binode->CNDNODE))->GEX  =
		        beh_abl2gex (pt_befig, pt_biabl->CNDABL);
          ((bequad_list *)(pt_binode->VALNODE))->GEX   =
		        beh_abl2gex (pt_befig, pt_biabl->VALABL);
          }
        else 
	  {
	  pt_binode->CNDNODE=(pNode)beh_abl2gex(pt_befig,pt_biabl->CNDABL);
	  pt_binode->VALNODE=(pNode)beh_abl2gex(pt_befig,pt_biabl->VALABL);
	  }


        pt_biabl  = pt_biabl->NEXT;
        pt_binode = pt_binode->NEXT;
        }
      pt_bereg = pt_bereg->NEXT;
      }

    delht (HashTable);
    pt_befig->TYPE |= BEH_NODE_GEX;
    }
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_maknode.c						*/
/* date		: Jan  12 1998						*/
/* version	: v111							*/
/* authors	: Laurent VUILLEMIN, Julien DUNOYER			*/
/* description	: high level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function     : beh_maknode						*/
/* description  : create node for each signal register or bus		*/ 
/* called func. : beh_makquad beh_makbdd beh_makderiv beh_makgex	*/
/* ###--------------------------------------------------------------### */

void beh_maknode (pt_befig, node_flag, trace)

struct befig *pt_befig ;
char          node_flag;
char          trace    ;

  {
  unsigned int count = 0;

	/* ###------------------------------------------------------### */
	/*   setting missing flags :					*/
	/*     - deriv. by var  need quads, Bdd and list of variables	*/
	/*     - deriv. by time need quads and Bdd			*/
	/* ###------------------------------------------------------### */

  if ((node_flag & BEH_NODE_DERIVBYVAR) != 0)
    {
    node_flag |= BEH_NODE_BDD | BEH_NODE_VARLIST;
    count++;
    }

  if ((node_flag & BEH_NODE_DERIVBYTIME) != 0)
    {
    node_flag |= BEH_NODE_BDD;
    count++;
    }
	/* ###------------------------------------------------------### */
	/*   setting missing flags :					*/
	/*     - if two flags are set, quads are needed			*/
	/* ###------------------------------------------------------### */

  if ((node_flag & BEH_NODE_VARLIST) != 0)
    count++;

  if ((node_flag & BEH_NODE_BDD)     != 0)
    count++;

  if ((node_flag & BEH_NODE_GEX)     != 0)
    count++;

  if ((node_flag & BEH_NODE_USER)    != 0)
    count++;

  if ( count > 1 )
    {
    node_flag |= BEH_NODE_QUAD ;
    }

	/* ###------------------------------------------------------### */
	/*    call the appropriate function two build the requested	*/
	/* structures							*/
	/* ###------------------------------------------------------### */

	/* ###------------------------------------------------------### */
	/*   build quads to support further structures			*/
	/* ###------------------------------------------------------### */

  if ((node_flag & BEH_NODE_QUAD) != 0)
    {
    if (trace != 0)
      beh_message (14, "QUAD");
    beh_makquad (pt_befig);
    }

	/* ###------------------------------------------------------### */
	/*   build general expressions					*/
	/* ###------------------------------------------------------### */

  if ((node_flag & BEH_NODE_GEX) != 0)
    {
    if (trace != 0)
      beh_message (14, "GEX");
    beh_makgex (pt_befig);
    }

	/* ###------------------------------------------------------### */
	/*   build binary decision diagrams				*/
	/* ###------------------------------------------------------### */

  if ((node_flag & BEH_NODE_BDD) != 0)
    {
    if (trace != 0)
      beh_message (14, "BDD");
    beh_makbdd (pt_befig);
    }

	/* ###------------------------------------------------------### */
	/*   build list of variables					*/
	/* ###------------------------------------------------------### */

  if ((node_flag & BEH_NODE_VARLIST) != 0)
    {
    if (trace != 0)
      beh_message (14, "list of variables");
    beh_makvarlist (pt_befig);
    }

	/* ###------------------------------------------------------### */
	/*   build derivatives of binary expressions (list of Bdds)	*/
	/* ###------------------------------------------------------### */

  if ((node_flag & BEH_NODE_DERIVATE) != 0)
    {
    if (trace != 0)
      beh_message (14, "derivatives");
    beh_makderiv (pt_befig);
    }

  }

/* ###--------------------------------------------------------------### */
/* file		: beh_makquad.c						*/
/* date		: Mar  23 1995						*/
/* version	: v109							*/
/* authors	: Laurent VUILLEMIN					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function     : beh_makquad						*/
/* description  : add a BEQUAD structure for each expression		*/
/* called func. : beh_addbequad	 					*/
/* ###--------------------------------------------------------------### */

void beh_makquad (pt_befig)

struct befig *pt_befig;

  {
  struct beaux  *pt_beaux;
  struct berin  *pt_berin;
  struct beout  *pt_beout;
  struct bebus  *pt_bebus;
  struct bebux  *pt_bebux;
  struct bereg  *pt_bereg;
  struct bemsg  *pt_bemsg;
  struct binode *pt_binod;
  struct biabl  *pt_biabl;

	/* ###------------------------------------------------------### */
	/*   simple output ports					*/
	/* ###------------------------------------------------------### */

  pt_beout = pt_befig->BEOUT;
  while (pt_beout != NULL)
     {
     pt_beout->NODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);
     pt_beout       = pt_beout->NEXT;
     }

	/* ###------------------------------------------------------### */
	/*   bussed ouput ports						*/
	/* ###------------------------------------------------------### */

  pt_bebus = pt_befig->BEBUS;
  while (pt_bebus != NULL)
     {
     pt_binod = pt_bebus->BINODE;
     while (pt_binod != NULL)
       {
       pt_binod->CNDNODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);
       pt_binod->VALNODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);

       pt_binod          = pt_binod->NEXT;
       }
     pt_bebus = pt_bebus->NEXT;
     }

	/* ###------------------------------------------------------### */
	/*   internal bussed signals					*/
	/* ###------------------------------------------------------### */

  pt_bebux = pt_befig->BEBUX;
  while (pt_bebux != NULL)
     {
     pt_binod = pt_bebux->BINODE;
     while (pt_binod != NULL)
       {
       pt_binod->CNDNODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);
       pt_binod->VALNODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);

       pt_binod          = pt_binod->NEXT;
       }
     pt_bebux = pt_bebux->NEXT;
     }

	/* ###------------------------------------------------------### */
	/*   internal registers						*/
	/* ###------------------------------------------------------### */

  pt_bereg = pt_befig->BEREG;
  while (pt_bereg != NULL)
     {
     pt_binod = pt_bereg->BINODE;
     while (pt_binod != NULL)
       {
       pt_binod->CNDNODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);
       pt_binod->VALNODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);

       pt_binod          = pt_binod->NEXT;
       }
     pt_bereg = pt_bereg->NEXT;
     }

	/* ###------------------------------------------------------### */
	/*   internal simple signals					*/
	/* ###------------------------------------------------------### */
    
  pt_beaux = pt_befig->BEAUX;
  while (pt_beaux != NULL)
     {
     pt_beaux->NODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);
     pt_beaux       = pt_beaux->NEXT;
     }

	/* ###------------------------------------------------------### */
	/*   internal delayed signals					*/
	/* ###------------------------------------------------------### */

  pt_beaux = pt_befig->BEDLY;
  while (pt_beaux != NULL)
     {
     pt_beaux->NODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);
     pt_beaux       = pt_beaux->NEXT;
     }

	/* ###------------------------------------------------------### */
	/*   assertions							*/
	/* ###------------------------------------------------------### */

  pt_bemsg = pt_befig->BEMSG;
  while (pt_bemsg != NULL)
     {
     pt_bemsg->NODE = (pNode) beh_addbequad (NULL,NULL,NULL,NULL);
     pt_bemsg       = pt_bemsg->NEXT;
     }

  pt_befig->TYPE = pt_befig->TYPE | BEH_NODE_QUAD;
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_makvarlist.c					*/
/* date		: Aug 29 1997						*/
/* version	: v110							*/
/* authors	: DUNOYER Julien.					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_abl2varlist					*/
/* description	: for an expression create the list of terminals. Each	*/
/*		  terminal may be present only once in the list.	*/
/*		  Constante values are not present in the list. Each	*/
/*		  terminal is represented by its BDD index		*/
/* called func.	: searchInputCct, beh_abl2varlist , beh_error, 		*/
/* ###--------------------------------------------------------------### */

struct chain *beh_abl2varlist (pt_befig, expr)

struct befig *pt_befig;		/* pointer on  behavioural figure	*/
struct chain *expr    ;		/* pointer abl expression		*/

  {
  short           oper         ;
  struct chain   *set_pnt      = NULL;
  struct circuit *pt_circuit   = pt_befig->CIRCUI;
  int             err_flg      = 0;
  int             index        ;
  char           *atom         ;

  static char     buffer [128] ;

	/* ###------------------------------------------------------### */
	/*    check that the expression exists				*/
	/* ###------------------------------------------------------### */

  if (expr == NULL)
    err_flg = beh_error (2, NULL);
  else
    {
    if (expr->NEXT == NULL)
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is a terminal (atom)			*/
	/* ###------------------------------------------------------### */

      atom = (char *) expr->DATA;

	/* ###------------------------------------------------------### */
	/*    if the terminal is a primary signal create an element	*/
	/* with its index and return					*/
	/* ###------------------------------------------------------### */

      if ((index = searchInputCct (pt_circuit, atom)) != VIDE)
        set_pnt = cst_AddSetElement (set_pnt, index);
      }
    else
      {
	/* ###------------------------------------------------------### */
	/*    if the expression is not a terminal (is a complex expr.)	*/
	/* ###------------------------------------------------------### */

      if ((oper = OPER(expr)) != STABLE)
       {
       while ((expr = CDR(expr)) != NULL)
         {
         set_pnt =cst_CreateUnionSet
                      (set_pnt, beh_abl2varlist (pt_befig, CAR(expr)));
         }
       }
      else
       {
	expr = CDR(expr);
	if (!ATOM(expr))
	  err_flg = beh_error (4, NULL);
        else
          {
	  atom = (char *)VALUE_ATOM (CAR(expr));

          if ((index = searchInputCct (pt_circuit, atom)) != VIDE)
            {
            set_pnt = cst_AddSetElement (set_pnt, index);

	    sprintf (buffer, "%s'delayed", atom);
	    atom = namealloc (buffer);

            if ((index = searchInputCct (pt_circuit, atom)) != VIDE)
              {
              set_pnt = cst_AddSetElement (set_pnt, index);
              }
            else
	      err_flg = beh_error (3, atom);
            }
          else
	    err_flg = beh_error (3, atom);
          }
        }
      }
    }

  if (err_flg != 0)
    pt_befig->ERRFLG = 1;

  return (set_pnt);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_mvlsttot						*/
/* description	: make list of variables for all primary signals	*/
/*		  simple and bussed outputs,				*/
/*		  simple and bussed internal signals, registers.	*/
/* called func.	: beh_abl2varlist					*/
/* ###--------------------------------------------------------------### */

static void beh_mvlsttot (pt_befig)

struct befig *pt_befig;		/* pointer on current BEFIG	*/

  {
  struct beaux  *pt_beaux;
  struct beaux  *pt_bedly;
  struct bemsg  *pt_bemsg;
  struct beout  *pt_beout;
  struct bebus  *pt_bebus;
  struct bebux  *pt_bebux;
  struct bereg  *pt_bereg;
  struct biabl  *pt_biabl;
  struct binode *pt_binode;

	/* ###------------------------------------------------------### */
	/*    make a var list for each simple internal signal		*/
	/* ###------------------------------------------------------### */

  pt_beaux = pt_befig->BEAUX;

  while (pt_beaux != NULL)
    {
    if (pt_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(pt_beaux->NODE))->VARLIST = 
                        beh_abl2varlist (pt_befig, pt_beaux->ABL);
      }
    else
      {
      pt_beaux->NODE = (struct node *)beh_abl2varlist (pt_befig, pt_beaux->ABL);
      }
    pt_beaux = pt_beaux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each delayed internal signal		*/
	/* ###------------------------------------------------------### */

  pt_bedly = pt_befig->BEDLY;
  while (pt_bedly != NULL)
    {
    if (pt_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(pt_bedly->NODE))->VARLIST = 
                        beh_abl2varlist (pt_befig, pt_bedly->ABL);
      }
    else
      {
      pt_bedly->NODE = (struct node *)beh_abl2varlist (pt_befig, pt_bedly->ABL);
      }
    pt_bedly = pt_bedly->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each assertion			*/
	/* ###------------------------------------------------------### */

  pt_bemsg = pt_befig->BEMSG;
  while (pt_bemsg != NULL)
    {
    if (pt_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(pt_bemsg->NODE))->VARLIST = 
                        beh_abl2varlist (pt_befig, pt_bemsg->ABL);
      }
    else
      {
      pt_bemsg->NODE = (struct node *)beh_abl2varlist (pt_befig, pt_bemsg->ABL);
      }

    pt_bemsg = pt_bemsg->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each simple output port		*/
	/* ###------------------------------------------------------### */

  pt_beout = pt_befig->BEOUT;
  while (pt_beout != NULL)
    {
    if (pt_befig->TYPE & BEH_NODE_QUAD)
      {
      ((bequad_list *)(pt_beout->NODE))->VARLIST = 
                        beh_abl2varlist (pt_befig, pt_beout->ABL);
      }
    else
      {
      pt_beout->NODE = (struct node *)beh_abl2varlist (pt_befig, pt_beout->ABL);
      }

    pt_beout = pt_beout->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each bussed output port		*/
	/* ###------------------------------------------------------### */

  pt_bebus = pt_befig->BEBUS;
  while (pt_bebus != NULL)
    {
    pt_biabl  = pt_bebus->BIABL;
    pt_binode = pt_bebus->BINODE;

    while (pt_biabl != NULL)
      {
      if (pt_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(pt_binode->CNDNODE))->VARLIST = 
                              beh_abl2varlist (pt_befig, pt_biabl->CNDABL);
        ((bequad_list *)(pt_binode->VALNODE))->VARLIST = 
                              beh_abl2varlist (pt_befig, pt_biabl->VALABL);
        }
      else
        {
        pt_binode->CNDNODE = (struct node *)beh_abl2varlist (pt_befig, pt_biabl->CNDABL);
        pt_binode->VALNODE = (struct node *)beh_abl2varlist (pt_befig, pt_biabl->VALABL);
        }

      pt_biabl = pt_biabl->NEXT;
      pt_binode = pt_binode->NEXT;
      }

    pt_bebus = pt_bebus->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each bussed internal signal		*/
	/* ###------------------------------------------------------### */

  pt_bebux = pt_befig->BEBUX;
  while (pt_bebux != NULL)
    {
    pt_biabl  = pt_bebux->BIABL;
    pt_binode = pt_bebux->BINODE;

    while (pt_biabl != NULL)
      {
      if (pt_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(pt_binode->CNDNODE))->VARLIST = 
                              beh_abl2varlist (pt_befig, pt_biabl->CNDABL);
        ((bequad_list *)(pt_binode->VALNODE))->VARLIST = 
                              beh_abl2varlist (pt_befig, pt_biabl->VALABL);
        }
      else
        {
        pt_binode->CNDNODE = (struct node *)beh_abl2varlist (pt_befig, pt_biabl->CNDABL);
        pt_binode->VALNODE = (struct node *)beh_abl2varlist (pt_befig, pt_biabl->VALABL);
        }

      pt_biabl  = pt_biabl->NEXT;
      pt_binode = pt_binode->NEXT;
      }
    pt_bebux = pt_bebux->NEXT;
    }

	/* ###------------------------------------------------------### */
	/*    make a var list for each internal register		*/
	/* ###------------------------------------------------------### */

  pt_bereg = pt_befig->BEREG;
  while (pt_bereg != NULL)
    {
    pt_biabl  = pt_bereg->BIABL;
    pt_binode = pt_bereg->BINODE;

    while (pt_biabl != NULL)
      {
      if (pt_befig->TYPE & BEH_NODE_QUAD)
        {
        ((bequad_list *)(pt_binode->CNDNODE))->VARLIST = 
                              beh_abl2varlist (pt_befig, pt_biabl->CNDABL);
        ((bequad_list *)(pt_binode->VALNODE))->VARLIST = 
                              beh_abl2varlist (pt_befig, pt_biabl->VALABL);
        }
      else
        {
        pt_binode->CNDNODE = (struct node *)beh_abl2varlist (pt_befig, pt_biabl->CNDABL);
        pt_binode->VALNODE = (struct node *)beh_abl2varlist (pt_befig, pt_biabl->VALABL);
        }

      pt_biabl  = pt_biabl->NEXT;
      pt_binode = pt_binode->NEXT;
      }
    pt_bereg = pt_bereg->NEXT;
    }
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_makvarlist					*/
/* description	: make var list for each signal of a befig structure 	*/
/* called func.	: initializeCct, addInputCct, beh_mvlsttot		*/
/* ###--------------------------------------------------------------### */

void beh_makvarlist (pt_befig)

struct befig *pt_befig;

  {
  int         err_flg   = 0;

  if (!(pt_befig->TYPE & BEH_NODE_VARLIST));
    {
    beh_indexbdd (pt_befig);

        /* ###------------------------------------------------------### */
        /*    get the list of variables for all expressions		*/
        /* ###------------------------------------------------------### */

    beh_mvlsttot (pt_befig);

    pt_befig->TYPE |= BEH_NODE_VARLIST;

    if (err_flg != 0)
      pt_befig->ERRFLG = 1;
    }
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_namelist.c					*/
/* date		: Mar 10 1997						*/
/* version	: v110							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* description	: expression level function				*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_namelist						*/
/* description	: find the list of terminals' name for an expression.	*/
/*		  The value returned by the function is the list of	*/
/*		  terminals. Each terminal is represented by its name	*/
/*		  in the list. A given terminal may appear several	*/
/*		  times in the list. Constante values are not added	*/
/*		  to the list.						*/
/* called func.	: beh_namelist, namealloc, addchain, append		*/
/* ###--------------------------------------------------------------### */

struct chain *beh_namelist (pt_exp)

struct chain *pt_exp;

  {
  struct chain *res_chain    = NULL;
  struct chain *tmp_chain    = NULL;
  struct chain *pt_opr       ;
  char         *name         ;

  static char  *str_z        = NULL;
  static char  *str_o        = NULL;
  static char  *str_d        = NULL;
  static char   buffer [128] ;

	/* ###------------------------------------------------------### */
	/*   initialize 3 variable with constantes (first call only)	*/
	/* ###------------------------------------------------------### */

  if (str_z == NULL)
    {
    str_z = namealloc ("'0'");
    str_o = namealloc ("'1'");
    str_d = namealloc ("'d'");
    }

  if (pt_exp != NULL)
    {
	/* ###------------------------------------------------------### */
	/*   check that the expression is not empty			*/
	/* ###------------------------------------------------------### */

    if (pt_exp->NEXT != NULL)
      {
	/* ###------------------------------------------------------### */
	/*   If the expression is not a terminal :			*/
	/*								*/
	/*     - if the operator is STABLE, add the varaiable's name	*/
	/*       to the list. Add also the "delayed" variable.		*/
	/*     - otherwies, call recursively the function on each	*/
	/*       operand, and append the returned list to the previous	*/
	/*       variable list.						*/
	/* ###------------------------------------------------------### */

      if (((int) ((struct chain *) pt_exp->DATA)->DATA) == STABLE)
        {
        name      = ((struct chain *) pt_exp->NEXT->DATA)->DATA;
        res_chain = addchain (NULL, name);
        sprintf   (buffer, "%s'delayed", name);
        name      = namealloc (buffer);
        res_chain = addchain (res_chain, name);
        }
      else
        {
        pt_opr = pt_exp->NEXT;
        while (pt_opr != NULL)
          {
          tmp_chain = beh_namelist (pt_opr->DATA);
          res_chain = append (res_chain, tmp_chain);
          pt_opr    = pt_opr->NEXT;
          }
        }
      }
    else
      {
	/* ###------------------------------------------------------### */
	/*   if the expression is a terminal but not a constante, add	*/
	/* the variable's name to the list of variables (do not check	*/
	/* if the variable has already been added to the list).		*/
	/* ###------------------------------------------------------### */

      if ((pt_exp->DATA != str_z) && (pt_exp->DATA != str_o) &&
          (pt_exp->DATA != str_d)    )
        {
        res_chain = addchain (NULL, pt_exp->DATA);
        }
      }
    }

  return (res_chain);
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_setdelay.c					*/
/* date		: Jan 20 1998						*/
/* version	: v3.0							*/
/* author	: Julien DUNOYER					*/
/* description	: high level function					*/
/* ###--------------------------------------------------------------### */


#define BEH_DELAY_FIELD             7   /* delay dico field             */

/* ###--------------------------------------------------------------### */
/* function	: beh_backdelay						*/
/* ###--------------------------------------------------------------### */

beh_backdelay ( delay_dict, delay_filename )

struct beden **delay_dict;
char          *delay_filename;

  {
  FILE	*delay_file;
  char	 line[256];
  char   signame[256];
  char	 vector_radical[256];
  int	 index;
  int	 mean_delay;
  int	 min_delay;
  int	 max_delay;
  int	 nbr_delay;

  delay_file = fopen( delay_filename, "r" );

  if ( delay_file == NULL )
    {
    return(1);
    }
 
   
  while ( fgets(line, 256, delay_file) != NULL)
    {
    if ( sscanf( line, "%s : %d : %d : %d : %d", signame,
                      &mean_delay, &min_delay, &max_delay, &nbr_delay ) == 5 )
      {
      beh_addtab (delay_dict, namealloc(signame), NULL, BEH_DELAY_FIELD, max_delay);
      }
    else if ( sscanf( line, "%s %d : %d : %d : %d : %d", vector_radical, &index,
                      &mean_delay, &min_delay, &max_delay, &nbr_delay ) == 6 )
      {
      sprintf(signame, "%s %d", vector_radical, index);
/*       printf("mise au dico de %s avec %d\n", signame, max_delay); */
      beh_addtab (delay_dict, namealloc(signame), NULL, BEH_DELAY_FIELD, max_delay);
      }
    else
      {
      sscanf( line, "%s", signame );
      printf("WARNING : invalid line in file %s : \n", delay_file);
      printf("        : %s\n", line);
      }
    }
  fclose(delay_file);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_setdelay						*/
/* ###--------------------------------------------------------------### */

void beh_setdelay (pt_befig, max_delay, delay_mode)

struct befig  *pt_befig;
unsigned int   max_delay;
unsigned char  delay_mode;

  {
  struct beout         *pt_beout ;
  struct beaux         *pt_beaux ;
  struct bebus         *pt_bebus ;
  struct bebux         *pt_bebux ;
  struct biabl         *pt_biabl ;
  static unsigned int   time     = 0;
  struct beden        **delay_dict;
  char                  filename [200];
 
  if ( delay_mode == BEH_DELAY_RANDOM ) 
    {
    pt_beout = pt_befig->BEOUT;
    while (pt_beout != NULL)
      {
      if (pt_beout->TIME == 0)
        {
        time           += 5;
        pt_beout->TIME  = time % max_delay + 3;
        }
      pt_beout = pt_beout->NEXT;
      }

    pt_beaux = pt_befig->BEAUX;
    while (pt_beaux != NULL)
      {
      if (pt_beaux->TIME == 0)
        {
        time           += 5;
        pt_beaux->TIME  = time % max_delay + 3;
        }
      pt_beaux = pt_beaux->NEXT;
      }

/*----------
    pt_bebus = pt_befig->BEBUS;
    while (pt_bebus != NULL)
      {
      pt_biabl = pt_bebus->BIABL;
      while (pt_biabl != NULL)
        {
        time           += 5;
        pt_biabl->TIME  = time % max_delay;
        pt_biabl        = pt_biabl->NEXT;
        }
      pt_bebus = pt_bebus->NEXT;
      }

    pt_bebux = pt_befig->BEBUX;
    while (pt_bebux != NULL)
      {
      pt_biabl = pt_bebux->BIABL;
      while (pt_biabl != NULL)
        {
        time           += 5;
        pt_biabl->TIME  = time % max_delay;
        pt_biabl        = pt_biabl->NEXT;
        }
      pt_bebux = pt_bebux->NEXT;
      }
----------*/
    }
 
  if ( delay_mode == BEH_DELAY_FIXED ) 
    {
    pt_beout = pt_befig->BEOUT;
    while (pt_beout != NULL)
      {
      if (pt_beout->TIME == 0)
        {
        pt_beout->TIME  = max_delay ;
        }
      pt_beout = pt_beout->NEXT;
      }

    pt_beaux = pt_befig->BEAUX;
    while (pt_beaux != NULL)
      {
      if (pt_beaux->TIME == 0)
        {
        pt_beaux->TIME  = max_delay ;
        }
      pt_beaux = pt_beaux->NEXT;
      }

/*----------
    pt_bebus = pt_befig->BEBUS;
    while (pt_bebus != NULL)
      {
      pt_biabl = pt_bebus->BIABL;
      while (pt_biabl != NULL)
        {
        pt_biabl->TIME  = max_delay;
        pt_biabl        = pt_biabl->NEXT;
        }
      pt_bebus = pt_bebus->NEXT;
      }

    pt_bebux = pt_befig->BEBUX;
    while (pt_bebux != NULL)
      {
      pt_biabl = pt_bebux->BIABL;
      while (pt_biabl != NULL)
        {
        pt_biabl->TIME  = max_delay;
        pt_biabl        = pt_biabl->NEXT;
        }
      pt_bebux = pt_bebux->NEXT;
      }
----------*/
    }
 
  if ( delay_mode == BEH_DELAY_BACKANOTATED ) 
    {

    delay_dict = beh_initab ();
    sprintf (filename, "%s.dly", pt_befig->NAME);
    beh_backdelay (delay_dict, filename);

    pt_beout = pt_befig->BEOUT;
    while (pt_beout != NULL)
      {
      if (pt_beout->TIME == 0)
        {
        pt_beout->TIME = beh_chktab (delay_dict, pt_beout->NAME,
                                     NULL, BEH_DELAY_FIELD) ;
        }
      pt_beout = pt_beout->NEXT;
      }

    pt_beaux = pt_befig->BEAUX;
    while (pt_beaux != NULL)
      {
      if (pt_beaux->TIME == 0)
        {
        pt_beaux->TIME  = beh_chktab (delay_dict, pt_beaux->NAME,
                                      NULL, BEH_DELAY_FIELD) ;
        }
      pt_beaux = pt_beaux->NEXT;
      }

/*----------
    pt_bebus = pt_befig->BEBUS;
    while (pt_bebus != NULL)
      {
      pt_biabl = pt_bebus->BIABL;
      while (pt_biabl != NULL)
        {
        pt_biabl->TIME  = beh_chktab (delay_dict, pt_bebus->NAME,
                                      NULL, BEH_DELAY_FIELD) ;
        pt_biabl        = pt_biabl->NEXT;
        }
      pt_bebus = pt_bebus->NEXT;
      }

    pt_bebux = pt_befig->BEBUX;
    while (pt_bebux != NULL)
      {
      pt_biabl = pt_bebux->BIABL;
      while (pt_biabl != NULL)
        {
        pt_biabl->TIME  = beh_chktab (delay_dict, pt_bebux->NAME,
                                      NULL, BEH_DELAY_FIELD) ;
        pt_biabl        = pt_biabl->NEXT;
        }
      pt_bebux = pt_bebux->NEXT;
      }
----------*/
 
    beh_fretab (delay_dict);
    }
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_unamlist.c					*/
/* date		: Mar 10 1997						*/
/* version	: v110							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* description	: expression level function				*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_unamlist						*/
/* description	: find the list of terminals' name for an expression.	*/
/*		  The value returned by the function is the list of	*/
/*		  terminals. Each terminal is represented by its name	*/
/*		  in the list. A given terminal may only appear once	*/
/*		  in the list. Constante values are not added		*/
/*		  to the list.						*/
/* called func.	: beh_namelist, delchain				*/
/* ###--------------------------------------------------------------### */

struct chain *beh_unamlist (pt_exp)

struct chain *pt_exp;

  {
  struct chain *res_chain;
  struct chain *tmp_chain;
  struct chain *del_chain;
  struct chain *prv_chain;

	/* ###------------------------------------------------------### */
	/*   find the list of terminals					*/
	/* ###------------------------------------------------------### */

  res_chain = beh_namelist (pt_exp);

	/* ###------------------------------------------------------### */
	/*   delete multiple refernces to the same terminal		*/
	/* ###------------------------------------------------------### */

  tmp_chain = res_chain;

  while (tmp_chain != NULL)
    {
    prv_chain = tmp_chain      ;
    del_chain = tmp_chain->NEXT;

    while (del_chain != NULL)
      {
      if (del_chain->DATA == tmp_chain->DATA)
        {
        del_chain       = delchain (del_chain, del_chain);
        prv_chain->NEXT = del_chain;
        }
      else
        {
        prv_chain = del_chain      ;
        del_chain = del_chain->NEXT;
        }
      }

    tmp_chain = tmp_chain->NEXT;
    }

  return (res_chain);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbeaux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbeaux						*/
/* description	: create a BEAUX structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct beaux *beh_addbeaux (lastbeaux, name, abl_expr, bdd_expr)

struct beaux *lastbeaux;	/* pointer on the last beaux structure	*/
char         *name;		/* signal's name			*/
struct chain *abl_expr;		/* signal's expression (ABL)		*/
struct node  *bdd_expr;		/* signal's expression (BDD)		*/

  {
  struct beaux *ptaux;

  name        = namealloc (name);

  ptaux       = (struct beaux *) mbkalloc (sizeof(struct beaux));
  ptaux->NAME = name;
  ptaux->TIME = 0;
  ptaux->ABL  = abl_expr;
  ptaux->NODE = bdd_expr;
  ptaux->NEXT = lastbeaux;

  return (ptaux);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbebus.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbebus						*/
/* description	: create a BEBUS structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct bebus *beh_addbebus (lastbebus, name, biabl, binode, type)

struct bebus  *lastbebus;	/* pointer on the last bebus structure	*/
char          *name;		/* port's name				*/
struct biabl  *biabl;		/* port's drivers (ABL)			*/
struct binode *binode;		/* port's drivers (BDD)			*/
char           type;		/* port type mark (M or W)		*/

  {
  struct bebus *ptbus;

  name          = namealloc (name);

  ptbus         = (struct bebus *) mbkalloc (sizeof(struct bebus));
  ptbus->NAME   = name;
  ptbus->BIABL  = biabl;
  ptbus->BINODE = binode;
  ptbus->TYPE   = type;
  ptbus->NEXT   = lastbebus;

  return (ptbus);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbebux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbebux						*/
/* description	: create a BEBUX structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct bebux *beh_addbebux (lastbebux, name, biabl, binode, type)

struct bebux  *lastbebux;	/* pointer on the last bebux structure	*/
char          *name;		/* signal's name			*/
struct biabl  *biabl;		/* signal's expression (ABL)		*/
struct binode *binode;		/* signal's expression (BDD)		*/
char           type;		/* signal's type mark (M or W)		*/

  {
  struct bebux *ptbux;

  name          = namealloc (name);

  ptbux         = (struct bebux *) mbkalloc (sizeof(struct bebux));
  ptbux->NAME   = name;
  ptbux->BIABL  = biabl;
  ptbux->BINODE = binode;
  ptbux->TYPE   = type;
  ptbux->NEXT   = lastbebux;

  return (ptbux);
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_addbeder.c					*/
/* date		: Mar  17 1995						*/
/* version	: v109							*/
/* authors	: Laurent VUILLEMIN, Julien DUNOYER			*/
/* content	: low-level function					*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbeder						*/
/* description	: create a BEDER structure at the top of the list	*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct beder *beh_addbeder (last_beder, bdd_expr, index, delay)

struct beder *last_beder;
struct node  *bdd_expr  ;
short         index     ;
unsigned int  delay     ;

  {
  struct beder *ptder;

  ptder        = (struct beder *) mbkalloc (sizeof(struct beder));

  ptder->BDD   = bdd_expr  ;
  ptder->INDEX = index     ;
  ptder->TIME  = delay     ;
  ptder->NEXT  = last_beder;

  return (ptder);
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_addbefig.c					*/
/* date		: Mar 10 1997						*/
/* version	: v110							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbefig						*/
/* description	: create an empty BEFIG structure at the top of the list*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct befig *beh_addbefig (lastbefig, name)

struct befig *lastbefig;	/* pointer on the last befig structure	*/
char         *name     ;	/* figure's name			*/

  {
  struct befig  *ptfig   ;
  unsigned char  err_flg = 0;

  ptfig = (struct befig *) mbkalloc (sizeof(struct befig));

  switch (BEH_TIMEUNIT)
    {
    case BEH_TU__FS:
      BEH_CNV_FS = 1.0    ;
      BEH_CNV_PS = 1.0E3  ;
      BEH_CNV_NS = 1.0E6  ;
      BEH_CNV_US = 1.0E9  ;
      BEH_CNV_MS = 1.0E12 ;
      break;
    case BEH_TU__PS:
      BEH_CNV_FS = 1.0E-3 ;
      BEH_CNV_PS = 1.0    ;
      BEH_CNV_NS = 1.0E3  ;
      BEH_CNV_US = 1.0E6  ;
      BEH_CNV_MS = 1.0E9  ;
      break;
    case BEH_TU__NS:
      BEH_CNV_FS = 1.0E-6 ;
      BEH_CNV_PS = 1.0E-3 ;
      BEH_CNV_NS = 1.0    ;
      BEH_CNV_US = 1.0E3  ;
      BEH_CNV_MS = 1.0E6  ;
      break;
    case BEH_TU__US:
      BEH_CNV_FS = 1.0E-9 ;
      BEH_CNV_PS = 1.0E-6 ;
      BEH_CNV_NS = 1.0E-3 ;
      BEH_CNV_US = 1.0    ;
      BEH_CNV_MS = 1.0E3  ;
      break;
    case BEH_TU__MS:
      BEH_CNV_FS = 1.0E-12;
      BEH_CNV_PS = 1.0E-9 ;
      BEH_CNV_NS = 1.0E-6 ;
      BEH_CNV_US = 1.0E-3 ;
      BEH_CNV_MS = 1.0    ;
      break;
    default:
      err_flg = beh_error (200, NULL, ' ', 0);
    }

  ptfig->NAME      = namealloc (name);
  ptfig->BEREG     = NULL            ;
  ptfig->BEMSG     = NULL            ;
  ptfig->BERIN     = NULL            ;
  ptfig->BEOUT     = NULL            ;
  ptfig->BEBUS     = NULL            ;
  ptfig->BEAUX     = NULL            ;
  ptfig->BEDLY     = NULL            ;
  ptfig->BEBUX     = NULL            ;
  ptfig->BEPOR     = NULL            ;
  ptfig->CIRCUI    = NULL            ;
  ptfig->USER      = NULL            ;
  ptfig->BEGEN     = NULL            ;
  ptfig->ERRFLG    = err_flg         ;
  ptfig->TYPE      = 0               ;
  ptfig->FLAG      = 0               ;
  ptfig->TIME_UNIT = BEH_TIMEUNIT    ;
  ptfig->NEXT      = lastbefig       ;

  return (ptfig);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbegen.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbegen						*/
/* description	: create a BEGEN structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct begen *beh_addbegen (lastbegen, name, type, value)

struct begen *lastbegen;	/* pointer on the last begen structure	*/
char         *name;		/* generic's name			*/
char         *type;		/* generic's type mark			*/
void         *value;		/* generic's value			*/

  {

  struct begen *ptgen;

  name         = namealloc (name);
  type         = namealloc (type);
  
  ptgen        = (struct begen *) mbkalloc (sizeof(struct begen));
  ptgen->NAME  = name;
  ptgen->TYPE  = type;
  ptgen->VALUE = value;
  ptgen->NEXT  = lastbegen;

  return (ptgen);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbemsg.c					*/
/* date		: Nov  6 1995						*/
/* version	: v109							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbemsg						*/
/* description	: create BEMSG structure at the top the list		*/
/* called func.	: mbkalloc, namealloc					*/
/* ###--------------------------------------------------------------### */

struct bemsg *beh_addbemsg (lastbemsg,label,level,message,abl_expr,bdd_expr)

struct bemsg *lastbemsg;	/* pointer on the last bemsg structure	*/
char         *label;		/* assert's label			*/
char          level;		/* severity level (E or W)		*/
char         *message;		/* reported message			*/
struct chain *abl_expr;		/* assert's condition (ABL)		*/
struct node  *bdd_expr;		/* assert's condition (BDD)		*/

  {

  struct bemsg *ptmsg;

  label          = namealloc (label);

  ptmsg          = (struct bemsg *) mbkalloc (sizeof(struct bemsg));
  ptmsg->LABEL   = label;
  ptmsg->LEVEL   = level;
  ptmsg->MESSAGE = message;
  ptmsg->ABL     = abl_expr;
  ptmsg->NODE    = bdd_expr;
  ptmsg->NEXT    = lastbemsg;

  return (ptmsg);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbeout.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbeout						*/
/* description	: create a BEOUT structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct beout *beh_addbeout (lastbeout, name, abl_expr, bdd_expr)

struct beout *lastbeout;	/* pointer on the last beout structure	*/
char         *name;		/* port's name				*/
struct chain *abl_expr;		/* port's expression (ABL)		*/
struct node  *bdd_expr;		/* port's expression (BDD)		*/

  {
  struct beout *ptout;

  name         = namealloc (name);

  ptout        = (struct beout *) mbkalloc (sizeof(struct beout));
  ptout->NAME  = name;
  ptout->ABL   = abl_expr;
  ptout->TIME  = 0;
  ptout->NODE  = bdd_expr;
  ptout->NEXT  = lastbeout;

  return (ptout);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbepor.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbepor						*/
/* description	: create a BEPOR structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct bepor *beh_addbepor (lastbepor, name, dir, type)

struct bepor *lastbepor;	/* pointer on the last bepor structure	*/
char         *name;		/* signal's name			*/
char          dir;     		/* signal's mode (I, O, B, Z, or T)	*/
char          type;     	/* signal'type mark (M or W)		*/

  {
  struct bepor *ptbepor;

  name               = namealloc (name);

  ptbepor            = (struct bepor *) mbkalloc (sizeof(struct bepor));
  ptbepor->NAME      = name;
  ptbepor->DIRECTION = dir;
  ptbepor->TYPE      = type;
  ptbepor->NEXT      = lastbepor;

  return (ptbepor);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbequad.c					*/
/* date		: Mar  17 1995						*/
/* version	: v000							*/
/* authors	: Laurent VUILLEMIN, Julien DUNOYER			*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbequad						*/
/* description	: create a BEQUAD structure at the top of the list	*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct bequad *beh_addbequad ( bdd_expr, gex_expr, var_list, der_list)

struct node *bdd_expr;
struct begex *gex_expr;
struct chain *var_list;
struct beder *der_list;

  {
  struct bequad *ptquad;


  ptquad           = (struct bequad *) mbkalloc (sizeof(struct bequad));
  ptquad->BDD      = bdd_expr;
  ptquad->GEX      = gex_expr;
  ptquad->VARLIST  = var_list;
  ptquad->DERIVATE = der_list;
  ptquad->USER     = NULL;
  return (ptquad);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbereg.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbereg						*/
/* description	: create a BEREG structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct bereg *beh_addbereg (lastbereg, name, biabl, binode)

struct bereg  *lastbereg;	/* pointer on the last bereg structure	*/
char          *name;		/* register's name			*/
struct biabl  *biabl;		/* register's drivers (ABL)		*/
struct binode *binode;		/* register's drivers (BDD)		*/

  {
  struct bereg *ptreg;

  name          = namealloc (name);

  ptreg         = (struct bereg *) mbkalloc (sizeof(struct bereg));
  ptreg->NAME   = name;
  ptreg->BIABL  = biabl;
  ptreg->BINODE = binode;
  ptreg->NEXT   = lastbereg;

  return (ptreg);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addberin.c					*/
/* date		: Sep 20 1994						*/
/* version	: v107							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addberin						*/
/* description	: create a BERIN structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct berin *beh_addberin (lastberin, name)

struct berin *lastberin;	/* pointer on the last berin structure	*/
char         *name;		/* signal's name			*/

  {
  struct berin *ptrin;

  name         = namealloc (name);

  ptrin        = (struct berin *) mbkalloc (sizeof(struct berin));
  ptrin->NAME  = name;
  ptrin->NEXT  = lastberin;

  ptrin->REG_REF = NULL;
  ptrin->MSG_REF = NULL;
  ptrin->AUX_REF = NULL;
  ptrin->BUX_REF = NULL;
  ptrin->BUS_REF = NULL;
  ptrin->OUT_REF = NULL;
  ptrin->DLY_REF = NULL;

  return (ptrin);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbiabl.c					*/
/* date		: Nov  6 1995						*/
/* version	: v109							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbiabl						*/
/* description	: create a BIABL structure at the top the list		*/
/* called func.	: namealloc, mbkalloc					*/
/* ###--------------------------------------------------------------### */

struct biabl *beh_addbiabl (lastbiabl, label, condition, value)

char         *label;		/* block's label			*/
struct biabl *lastbiabl;	/* pointer on the last biabl structure	*/
struct chain *condition;	/* guard expression (ABL)		*/
struct chain *value;		/* value expression (ABL)		*/

  {
  struct biabl  *ptbiabl;

  ptbiabl         = (struct biabl *) mbkalloc (sizeof(struct biabl));
  label           = namealloc (label);

  ptbiabl->LABEL  = label;
  ptbiabl->CNDABL = condition;
  ptbiabl->VALABL = value;
  ptbiabl->TIME   = 0;
  ptbiabl->FLAG   = 0;
  ptbiabl->USER   = NULL;
  ptbiabl->NEXT   = lastbiabl;

  return (ptbiabl);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_addbinod.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_addbinode						*/
/* description	: create a BINODE structure at the top the list		*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct binode *beh_addbinode (lastbinode, condition, value)

struct binode *lastbinode;	/* pointer on the last binode structure	*/
struct node   *condition;	/* guard expression (BDD)		*/
struct node   *value;		/* value expression (BDD)		*/

  {
  struct binode *ptbinode;

  ptbinode          = (struct binode *) mbkalloc (sizeof(struct binode));
  ptbinode->TIME    = 0;
  ptbinode->CNDNODE = condition;
  ptbinode->VALNODE = value;
  ptbinode->NEXT    = lastbinode;

  return (ptbinode);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbeaux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbeaux						*/
/* description	: delete a BEAUX structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, freeExpr					*/
/* ###--------------------------------------------------------------### */

struct beaux *beh_delbeaux (listbeaux, ptbeaux, mode)

struct beaux *listbeaux;	/* list of beaux containing the object	*/
struct beaux *ptbeaux;		/* pointer of the BEAUX to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct beaux  headaux;
  struct beaux *ptlastaux;

  if ((listbeaux != NULL) && (ptbeaux != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headaux.NEXT = listbeaux;
    ptlastaux    = &headaux;
    while ((ptlastaux != NULL) && (ptlastaux->NEXT != ptbeaux))
      ptlastaux = ptlastaux->NEXT;

    if (ptlastaux != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbeaux->ABL != NULL)
          (void) fprintf (stderr,"BEH_warning : beaux `%s` not empty\n",
                          ptbeaux->NAME);
        }
      else
        freeExpr (ptbeaux->ABL);

      ptlastaux->NEXT = ptbeaux->NEXT;
      mbkfree (ptbeaux);
      }

    listbeaux = headaux.NEXT;
    }

  return(listbeaux);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbebus.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbebus						*/
/* description	: delete a BEBUS structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, beh_frebiabl, beh_frebinode			*/
/* ###--------------------------------------------------------------### */

struct bebus *beh_delbebus (listbebus, ptbebus, mode)

struct bebus *listbebus;	/* list of bebus containing the object	*/
struct bebus *ptbebus;		/* pointer of the BEBUS to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bebus  headbus;
  struct bebus *ptlastbus;

  if ((listbebus != NULL) && (ptbebus != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headbus.NEXT = listbebus;
    ptlastbus    = &headbus;
    while ((ptlastbus != NULL) && (ptlastbus->NEXT != ptbebus))
      ptlastbus = ptlastbus->NEXT;

    if (ptlastbus != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbebus->BIABL != NULL) || (ptbebus->BINODE != NULL))
          (void) fprintf (stderr,"BEH_warning : bebus `%s` not empty\n",
                          ptbebus->NAME);
        }
      else
        {
        beh_frebiabl  (ptbebus->BIABL);
        beh_frebinode (ptbebus->BINODE);
        }

      ptlastbus->NEXT = ptbebus->NEXT;
      mbkfree (ptbebus);
      }

    listbebus = headbus.NEXT;
    }

  return(listbebus);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbebux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbebux						*/
/* description	: delete a BEBUX structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, beh_frebiabl, beh_frebinode			*/
/* ###--------------------------------------------------------------### */

struct bebux *beh_delbebux (listbebux, ptbebux, mode)

struct bebux *listbebux;	/* list of bebux containing the object	*/
struct bebux *ptbebux;		/* pointer of the BEBUX to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bebux  headbux;
  struct bebux *ptlastbux;

  if ((listbebux != NULL) && (ptbebux != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headbux.NEXT = listbebux;
    ptlastbux    = &headbux;
    while ((ptlastbux != NULL) && (ptlastbux->NEXT != ptbebux))
      ptlastbux = ptlastbux->NEXT;

    if (ptlastbux != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbebux->BIABL != NULL) || (ptbebux->BINODE != NULL))
          (void) fprintf (stderr,"BEH_warning : bebux `%s` not empty\n",
                          ptbebux->NAME);
        }
      else
        {
        beh_frebiabl  (ptbebux->BIABL);
        beh_frebinode (ptbebux->BINODE);
        }

      ptlastbux->NEXT = ptbebux->NEXT;
      mbkfree (ptbebux);
      }

    listbebux = headbux.NEXT;
    }

  return(listbebux);
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_delbeder.c					*/
/* date		: Oct 25 1995						*/
/* version	: v109							*/
/* authors	: Pirouz BAZARGAN SABET, Julien DUNOYER			*/
/* content	: low-level function					*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbeder						*/
/* description	: delete a BEDER structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct beder *beh_delbeder (listbeder, ptbeder, mode)

struct beder *listbeder;	/* list of beder containing the object	*/
struct beder *ptbeder  ;	/* pointer of the beder to be deleted	*/
char          mode     ;	/* recursive delete or not (Y or N)	*/

  {
  struct beder  headder  ;
  struct beder *ptlastder;

  if ((listbeder != NULL) && (ptbeder != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headder.NEXT = listbeder;
    ptlastder    = &headder;

    while ((ptlastder != NULL) && (ptlastder->NEXT != ptbeder))
      ptlastder = ptlastder->NEXT;

    if (ptlastder != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbeder->BDD != NULL) )
          (void) fprintf (stderr,"BEH_warning : beder not empty\n");
        }

      ptlastder->NEXT = ptbeder->NEXT;
      mbkfree (ptbeder);
      }

    listbeder = headder.NEXT;
    }

  return(listbeder);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbefig.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbefig						*/
/* description	: delete a BEFIG structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N.	*/
/* called func.	: mbkfree     , beh_frebereg, beh_frebemsg,		*/
/*		  beh_freberin, beh_frebeout, beh_frebebus,		*/
/*		  beh_frebeaux, beh_frebebux, beh_frebepor,		*/
/*		  beh_frebegen, getptype				*/
/* ###--------------------------------------------------------------### */

struct befig *beh_delbefig (listbefig, ptbefig, mode)

struct befig *listbefig;	/* list of befig containing the object	*/
struct befig *ptbefig;		/* pointer of the BEFIG to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct befig  headfig;
  struct befig *ptlastfig;
  struct ptype *ptptype;

  if ((listbefig != NULL) && (ptbefig != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headfig.NEXT = listbefig;
    ptlastfig    = &headfig;
    while ((ptlastfig != NULL) && (ptlastfig->NEXT != ptbefig))
      ptlastfig = ptlastfig->NEXT;

    if (ptlastfig != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbefig->BEREG != NULL) || (ptbefig->BEMSG != NULL) ||
            (ptbefig->BERIN != NULL) || (ptbefig->BEOUT != NULL) ||
            (ptbefig->BEBUS != NULL) || (ptbefig->BEAUX != NULL) ||
            (ptbefig->BEBUX != NULL) || (ptbefig->BEDLY != NULL) ||
            (ptbefig->BEPOR != NULL) || (ptbefig->USER  != NULL))

          (void) fprintf (stderr,"BEH_warning : befig `%s` not empty\n",
                          ptbefig->NAME);
        }
      else
        {
        beh_frebereg (ptbefig->BEREG);
        beh_frebemsg (ptbefig->BEMSG);
        beh_freberin (ptbefig->BERIN);
        beh_frebeout (ptbefig->BEOUT);
        beh_frebebus (ptbefig->BEBUS);
        beh_frebeaux (ptbefig->BEAUX);
        beh_frebeaux (ptbefig->BEDLY);
        beh_frebebux (ptbefig->BEBUX);
        beh_frebepor (ptbefig->BEPOR);
        if ((ptptype = getptype (ptbefig->USER,BEH_GENERIC)) != NULL)
          beh_frebegen (ptptype->DATA);
        }

      ptlastfig->NEXT = ptbefig->NEXT;
      mbkfree (ptbefig);
      }

    listbefig = headfig.NEXT;
    }

  return(listbefig);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbegen.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbegen						*/
/* description	: delete a BEGEN structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct begen *beh_delbegen (listbegen, ptbegen, mode)

struct begen *listbegen;	/* list of begen containing the object	*/
struct begen *ptbegen;		/* pointer of the BEGEN to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct begen  headgen;
  struct begen *ptlastgen;

  if ((listbegen != NULL) && (ptbegen != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headgen.NEXT = listbegen;
    ptlastgen    = &headgen;
    while ((ptlastgen != NULL) && (ptlastgen->NEXT != ptbegen))
      ptlastgen = ptlastgen->NEXT;

    if (ptlastgen != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbegen->VALUE != NULL)
          (void) fprintf (stderr,"BEH_warning : begen `%s` not empty\n",
                          ptbegen->NAME);
        }
      else
        mbkfree (ptbegen->VALUE);

      ptlastgen->NEXT = ptbegen->NEXT;
      mbkfree (ptbegen);
      }

    listbegen = headgen.NEXT;
    }

  return(listbegen);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbemsg.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbemsg						*/
/* description	: delete a BEMSG structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, freeExpr					*/
/* ###--------------------------------------------------------------### */

struct bemsg *beh_delbemsg (listbemsg, ptbemsg, mode)

struct bemsg *listbemsg;	/* list of bemsg containing the object	*/
struct bemsg *ptbemsg;		/* pointer of the BEMSG to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bemsg  headmsg;
  struct bemsg *ptlastmsg;

  if ((listbemsg != NULL) && (ptbemsg != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headmsg.NEXT = listbemsg;
    ptlastmsg    = &headmsg;
    while ((ptlastmsg != NULL) && (ptlastmsg->NEXT != ptbemsg))
      ptlastmsg = ptlastmsg->NEXT;

    if (ptlastmsg != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbemsg->ABL != NULL)
          (void) fprintf (stderr,"BEH_warning : bemsg `%s` not empty\n",
                          ptbemsg->LABEL);
        }
      else
        freeExpr (ptbemsg->ABL);

      ptlastmsg->NEXT = ptbemsg->NEXT;
      mbkfree (ptbemsg);
      }

    listbemsg = headmsg.NEXT;
    }

  return(listbemsg);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbeout.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbeout						*/
/* description	: delete a BEOUT structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, freeExpr					*/
/* ###--------------------------------------------------------------### */

struct beout *beh_delbeout (listbeout, ptbeout, mode)

struct beout *listbeout;	/* list of beout containing the object	*/
struct beout *ptbeout;		/* pointer of the BEOUT to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct beout  headout;
  struct beout *ptlastout;

  if ((listbeout != NULL) && (ptbeout != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headout.NEXT = listbeout;
    ptlastout    = &headout;
    while ((ptlastout != NULL) && (ptlastout->NEXT != ptbeout))
      ptlastout = ptlastout->NEXT;

    if (ptlastout != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbeout->ABL != NULL)
          (void) fprintf (stderr,"BEH_warning : beout `%s` not empty\n",
                          ptbeout->NAME);
        }
      else
        freeExpr (ptbeout->ABL);

      ptlastout->NEXT = ptbeout->NEXT;
      mbkfree (ptbeout);
      }

    listbeout = headout.NEXT;
    }

  return(listbeout);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbepor.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbepor						*/
/* description	: delete a BEPOR structure and return the pointer of	*/
/*		  the next object.					*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct bepor *beh_delbepor (listbepor, ptbepor)

struct bepor *listbepor;	/* list of bepor containing the object	*/
struct bepor *ptbepor;		/* pointer of the BEPOR to be deleted	*/

  {
  struct bepor  headpor;
  struct bepor *ptlastpor;

  if ((listbepor != NULL) && (ptbepor != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headpor.NEXT = listbepor;
    ptlastpor    = &headpor;
    while ((ptlastpor != NULL) && (ptlastpor->NEXT != ptbepor))
      ptlastpor = ptlastpor->NEXT;

    if (ptlastpor != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/* ###------------------------------------------------------### */

      ptlastpor->NEXT = ptbepor->NEXT;
      mbkfree (ptbepor);
      }

    listbepor = headpor.NEXT;
    }

  return(listbepor);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbequad.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Laurent VUILLEMIN, Julien DUNOYER 			*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbequad						*/
/* description	: delete a BEQUAD structure. A warning is printed out	*/
/*		  if the object to be deleted is not empty when the	*/
/*		  mode is N						*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_delbequad (listbequad, ptbequad, mode)

struct bequad *ptbequad;	/* pointer of the BEQUAD to be deleted	*/
char           mode;		/* recursive delete or not (Y or N)	*/

  {

  if (ptbequad != NULL)
    {
	/* ###------------------------------------------------------### */
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

    if (mode == 'N')
      {
      if ((ptbequad->BDD     != NULL) || (ptbequad->GEX      != NULL) ||
	  (ptbequad->VARLIST != NULL) || (ptbequad->DERIVATE != NULL))
        (void) fprintf (stderr, "BEH_warning : bequad not empty\n");
      }
    else
      {
      beh_frebeder (ptbequad->DERIVATE);
      freechain    (ptbequad->VARLIST );
      }

    mbkfree (ptbequad);
    }
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbereg.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbereg						*/
/* description	: delete a BEREG structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, beh_frebiabl, beh_frebinode			*/
/* ###--------------------------------------------------------------### */

struct bereg *beh_delbereg (listbereg, ptbereg, mode)

struct bereg *listbereg;	/* list of bereg containing the object	*/
struct bereg *ptbereg;		/* pointer of the BEREG to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bereg  headreg;
  struct bereg *ptlastreg;

  if ((listbereg != NULL) && (ptbereg != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headreg.NEXT = listbereg;
    ptlastreg    = &headreg;
    while ((ptlastreg != NULL) && (ptlastreg->NEXT != ptbereg))
      ptlastreg = ptlastreg->NEXT;

    if (ptlastreg != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbereg->BIABL != NULL) || (ptbereg->BINODE != NULL))
          (void) fprintf (stderr,"BEH_warning : bereg `%s` not empty\n",
                          ptbereg->NAME);
        }
      else
        {
        beh_frebiabl  (ptbereg->BIABL);
        beh_frebinode (ptbereg->BINODE);
        }

      ptlastreg->NEXT = ptbereg->NEXT;
      mbkfree (ptbereg);
      }

    listbereg = headreg.NEXT;
    }

  return(listbereg);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delberin.c					*/
/* date		: Sep 20 1994						*/
/* version	: v107							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delberin						*/
/* description	: delete a BERIN structure and return the pointer of	*/
/*		  the next object.					*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct berin *beh_delberin (listberin, ptberin)

struct berin *listberin;	/* list of berin containing the object	*/
struct berin *ptberin;		/* pointer of the BERIN to be deleted	*/

  {
  struct berin  headrin;
  struct berin *ptlastrin;

  if ((listberin != NULL) && (ptberin != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headrin.NEXT = listberin;
    ptlastrin    = &headrin;
    while ((ptlastrin != NULL) && (ptlastrin->NEXT != ptberin))
      ptlastrin = ptlastrin->NEXT;

    if (ptlastrin != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/* ###------------------------------------------------------### */

      ptlastrin->NEXT = ptberin->NEXT;
      mbkfree (ptberin);
      }

    listberin = headrin.NEXT;
    }

  return(listberin);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbiabl.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbiabl						*/
/* description	: delete a BIABL structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, freeExpr					*/
/* ###--------------------------------------------------------------### */

struct biabl *beh_delbiabl (listbiabl, ptbiabl, mode)

struct biabl *listbiabl;	/* list of biabl containing the object	*/
struct biabl *ptbiabl;		/* pointer of the BIABL to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct biabl  headabl;
  struct biabl *ptlastabl;

  if ((listbiabl != NULL) && (ptbiabl != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headabl.NEXT = listbiabl;
    ptlastabl    = &headabl;
    while ((ptlastabl != NULL) && (ptlastabl->NEXT != ptbiabl))
      ptlastabl = ptlastabl->NEXT;

    if (ptlastabl != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbiabl->VALABL != NULL) || (ptbiabl->CNDABL != NULL))
          (void) fprintf (stderr,"BEH_warning : biabl not empty\n");
        }
      else
        {
        freeExpr (ptbiabl->VALABL);
        freeExpr (ptbiabl->CNDABL);
        freeptype(ptbiabl->USER);
        }

      ptlastabl->NEXT = ptbiabl->NEXT;
      mbkfree (ptbiabl);
      }

    listbiabl = headabl.NEXT;
    }

  return(listbiabl);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_delbinod.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_delbinode						*/
/* description	: delete a BINODE structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct binode *beh_delbinode (listbinode, ptbinode, mode)

struct binode *listbinode;	/* list of binode containing the object	*/
struct binode *ptbinode;	/* pointer of the BINODE to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct binode  headnode;
  struct binode *ptlastnode;

  if ((listbinode != NULL) && (ptbinode != NULL))
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headnode.NEXT = listbinode;
    ptlastnode    = &headnode;
    while ((ptlastnode != NULL) && (ptlastnode->NEXT != ptbinode))
      ptlastnode = ptlastnode->NEXT;

    if (ptlastnode != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbinode->VALNODE != NULL) || (ptbinode->CNDNODE != NULL))
          (void) fprintf (stderr,"BEH_warning : binode not empty\n");
        }

      ptlastnode->NEXT = ptbinode->NEXT;
      mbkfree (ptbinode);
      }

    listbinode = headnode.NEXT;
    }

  return(listbinode);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_dict.c						*/
/* date		: Jun 15 1994						*/
/* version	: v107							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


#define BEH_ALODFN    64
#define BEH_HSZDFN   307

static struct beden *BEH_DCEHED = NULL;		/* free entries list	*/
static struct bedrd *BEH_DCRHED = NULL;		/* free records list	*/

/* ###--------------------------------------------------------------### */
/* function	: beh_addent						*/
/* description	: add a new entry point to the dictionary		*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct beden *beh_addent (head, key)

struct beden *head;
char         *key;

  {
  struct beden *entry;
  int           i    ;

  if (BEH_DCEHED == NULL)
    {
    BEH_DCEHED = (struct beden *)
                 mbkalloc (sizeof (struct beden) * BEH_ALODFN);

    entry = BEH_DCEHED;
    for (i=1 ; i<BEH_ALODFN ; i++)
      {
      entry->NEXT = entry + 1;
      entry++;
      }
    entry->NEXT = NULL;
    }

  entry       = BEH_DCEHED;
  BEH_DCEHED  = BEH_DCEHED->NEXT;

  entry->NEXT = head;
  entry->DATA = NULL;
  entry->KEY  = key;

  return (entry);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_addrcd						*/
/* description	: add a new record to the dictionary			*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

static struct bedrd *beh_addrcd (head, key)

struct bedrd *head;
char         *key;

  {
  struct bedrd *recrd;
  int           i    ;

  if (BEH_DCRHED == NULL)
    {
    BEH_DCRHED = (struct bedrd *)
                 mbkalloc (sizeof (struct bedrd) * BEH_ALODFN);

    recrd = BEH_DCRHED;
    for (i=1 ; i<BEH_ALODFN ; i++)
      {
      recrd->NEXT = recrd + 1;
      recrd++;
      }
    recrd->NEXT = NULL;
    }

  recrd           = BEH_DCRHED;
  BEH_DCRHED      = BEH_DCRHED->NEXT;

  recrd->NEXT     = head;
  recrd->FD0_VAL  = 0;
  recrd->FD1_VAL  = 0;
  recrd->FD2_VAL  = 0;
  recrd->FD3_VAL  = 0;
  recrd->FD4_VAL  = 0;
  recrd->FD5_VAL  = 0;
  recrd->FD6_VAL  = 0;
  recrd->PNT_VAL  = 0;
  recrd->KEY      = key;

  return (recrd);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_initab						*/
/* description	: create a new dictionary				*/
/* called func.	: mbkalloc						*/
/* ###--------------------------------------------------------------### */

struct beden **beh_initab ()

  {
  struct beden **head;
  int                i;

  head = (struct beden **)
         mbkalloc (sizeof(struct beden *) * BEH_HSZDFN);

  for (i=0 ; i<BEH_HSZDFN ; i++)
    head[i] = NULL;

  return (head);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_addtab						*/
/* description	: save a data in a dictionary				*/
/* called func.	: beh_addent, beh_addrcd				*/
/* ###--------------------------------------------------------------### */

void beh_addtab (head, key_str, ctx_str, field, valu)

struct beden **head;
char          *key_str;
char          *ctx_str;
int            field;
int            valu;

  {
  int           found     = 0;
  unsigned int  key       ;
  unsigned int  index     ;
  struct beden *entry_pnt ;
  struct bedrd *recrd_pnt ;

  key       = ((unsigned int) key_str) + ((unsigned int) ctx_str);

  index     = key % BEH_HSZDFN;
  entry_pnt = head [index];

  while (entry_pnt != NULL)
    {
    if (entry_pnt->KEY == key_str)
      {
      found = 1;
      break;
      }
    entry_pnt = entry_pnt->NEXT;
    }

  if (found == 0)
    {
    head [index] = beh_addent (head [index], key_str); 
    entry_pnt    = head [index]; 
    }

  found     = 0;
  recrd_pnt = entry_pnt->DATA;
  while (recrd_pnt != NULL)
    {
    if (recrd_pnt->KEY == ctx_str)
      {
      found = 1;
      break;
      }
    recrd_pnt = recrd_pnt->NEXT;
    }

  if (found == 0)
    {
    entry_pnt->DATA = beh_addrcd (entry_pnt->DATA, ctx_str); 
    recrd_pnt       = entry_pnt->DATA ;
    }

  switch (field)
    {
    case 0 :
      recrd_pnt->FD0_VAL = valu; break;
    case 1 :
      recrd_pnt->FD1_VAL = valu; break;
    case 2 :
      recrd_pnt->FD2_VAL = valu; break;
    case 3 :
      recrd_pnt->FD3_VAL = valu; break;
    case 4 :
      recrd_pnt->FD4_VAL = valu; break;
    case 5 :
      recrd_pnt->FD5_VAL = valu; break;
    case 6 :
      recrd_pnt->FD6_VAL = valu; break;
    case 7 :
      recrd_pnt->PNT_VAL = valu; break;
    }

  }

/* ###--------------------------------------------------------------### */
/* function	: beh_chktab						*/
/* description	: extract a data from a dictionary			*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

long beh_chktab (head, key_str, ctx_str, field)

struct beden **head   ;
char          *key_str;
char          *ctx_str;
int            field  ;

  {
  int           found    = 0;
  long          valu     = 0;
  unsigned int  key      ;
  unsigned int  index    ;
  struct beden *entry_pnt;
  struct bedrd *recrd_pnt;

  key       = ((unsigned int) key_str) + ((unsigned int) ctx_str);

  index     = key % BEH_HSZDFN;
  entry_pnt = head [index];

  while (entry_pnt != NULL)
    {
    if (entry_pnt->KEY == key_str)
      {
      found = 1;
      break;
      }
    entry_pnt = entry_pnt->NEXT;
    }

  if (found == 1)
    {
    found     = 0;
    recrd_pnt = entry_pnt->DATA;
    while (recrd_pnt != NULL)
      {
      if (recrd_pnt->KEY == ctx_str)
        {
        found = 1;
        break;
        }
      recrd_pnt = recrd_pnt->NEXT;
      }
    if (found == 1)
      {
      switch (field)
        {
        case 0 :
          valu = recrd_pnt->FD0_VAL; break;
        case 1 :
          valu = recrd_pnt->FD1_VAL; break;
        case 2 :
          valu = recrd_pnt->FD2_VAL; break;
        case 3 :
          valu = recrd_pnt->FD3_VAL; break;
        case 4 :
          valu = recrd_pnt->FD4_VAL; break;
        case 5 :
          valu = recrd_pnt->FD5_VAL; break;
        case 6 :
          valu = recrd_pnt->FD6_VAL; break;
        case 7 :
          valu = recrd_pnt->PNT_VAL; break;
        }
      }
    }

  return (valu);
  }

/* ###--------------------------------------------------------------### */
/* function	: beh_fretab						*/
/* description	: release a dictionary					*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_fretab (pt_hash)

struct beden **pt_hash;

  {
  struct beden *pt_entry   ;
  struct beden *pt_nxtentry;
  struct bedrd *pt_record  ;
  int           i          ;

  if (pt_hash != NULL)
    {
    for (i=0 ; i<BEH_HSZDFN ; i++)
      {
      if ((pt_entry = pt_hash [i]) != NULL)
        {
        while (pt_entry != NULL)
          {
          pt_record = pt_entry->DATA;

          while (pt_record->NEXT != NULL)
            pt_record = pt_record->NEXT;

          pt_record->NEXT = BEH_DCRHED;
          BEH_DCRHED      = pt_entry->DATA;

          pt_nxtentry     = pt_entry->NEXT;
          pt_entry->NEXT  = BEH_DCEHED;
          BEH_DCEHED      = pt_entry;
          pt_entry        = pt_nxtentry;
          }
        }
      }
    mbkfree (pt_hash);
    }
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_error.c						*/
/* date		: Oct 30 1995						*/
/* version	: v109							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_error						*/
/* description	: print an errorr message				*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

unsigned int beh_error (code, str1)

int   code;
char *str1;

  {
  fprintf (stderr, "BEH : Error %d: ", code);

  switch (code)
    {
    case 1:
      fprintf (stderr, "combinatory loop: `%s`\n", str1);
      break;
    case 2:
      fprintf (stderr, "cannot make bdd of empty expression\n");
      break;
    case 3:
      fprintf (stderr, "cannot find terminal `%s`\n", str1);
      break;
    case 4:
      fprintf (stderr, "illegal use of STABLE attribute\n");
      break;
    case 5:
      fprintf (stderr, "cannot simplify internal signals\n");
      break;
    case 6:
      fprintf (stderr, "cannot make derivatives of expressions\n");
      break;

    case 40:
      fprintf (stderr, "signal `%s` never assigned\n", str1);
      break;
    case 41:
      fprintf (stderr, "`%s` has not an empty architecture\n", str1);
      break;
    case 68:
      fprintf (stderr, "port `%s` has unknwon type\n", str1);
      break;
    case 69:
      fprintf (stderr, "port `%s` has unknwon mode\n", str1);
      break;
    case 70:
      fprintf (stderr, "unknown time unit\n");
      break;

    case 100:
      fprintf (stderr, "cannot find `%s`\n", str1);
      break;
    case 107:
      fprintf (stderr, "cannot open result file\n");
      break;

    default:
      fprintf (stderr, "syntax error\n");
      break;
    }

  return (1);
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebeaux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebeaux						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebeaux (listbeaux)

struct beaux *listbeaux;	/* list of beaux to be deleted		*/

  {
  struct beaux *ptbeaux;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbeaux != NULL)
    {
    freeExpr (listbeaux->ABL);

    ptbeaux   = listbeaux;
    listbeaux = listbeaux->NEXT;
    mbkfree (ptbeaux);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebebus.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebebus						*/
/* description	: delete a list of BEBUS structures and all objects	*/
/*		  pointed by any os BEBUSs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebebus (listbebus)

struct bebus *listbebus;	/* list of bebus to be deleted		*/

  {
  struct bebus *ptbebus;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbebus != NULL)
    {
    beh_frebiabl  (listbebus->BIABL);
    beh_frebinode (listbebus->BINODE);

    ptbebus   = listbebus;
    listbebus = listbebus->NEXT;
    mbkfree (ptbebus);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebebux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebebux						*/
/* description	: delete a list of BEBUX structures and all objects	*/
/*		  pointed by any os BEBUXs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebebux (listbebux)

struct bebux *listbebux;	/* list of bebux to be deleted		*/

  {
  struct bebux *ptbebux;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbebux != NULL)
    {
    beh_frebiabl  (listbebux->BIABL);
    beh_frebinode (listbebux->BINODE);

    ptbebux   = listbebux;
    listbebux = listbebux->NEXT;
    mbkfree (ptbebux);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebeder.c					*/
/* date		: Mar  22 1995						*/
/* version	: v100							*/
/* authors	: Laurent VUILLEMIN, Julien DUNOYER			*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebeder						*/
/* description	: delete a list of BEDER structures and all objects	*/
/*		  pointed by any os BEDERs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebeder (listbeder)

struct beder *listbeder;	/* list of beder to be deleted		*/

  {
  struct beder *ptbeder;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbeder != NULL)
    {
    ptbeder   = listbeder;
    listbeder = listbeder->NEXT;
    mbkfree (ptbeder);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebefig.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebefig						*/
/* description	: delete a list of BEFIG structures and all objects	*/
/*		  pointed by any os BEFIGs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebefig (listbefig)

struct befig *listbefig;	/* list of befig to be deleted		*/

  {
  struct befig *ptbefig;
  struct ptype *ptptype;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbefig != NULL)
    {
    beh_frebereg (listbefig->BEREG);
    beh_frebemsg (listbefig->BEMSG);
    beh_freberin (listbefig->BERIN);
    beh_frebeout (listbefig->BEOUT);
    beh_frebebus (listbefig->BEBUS);
    beh_frebeaux (listbefig->BEAUX);
    beh_frebeaux (listbefig->BEDLY);
    beh_frebebux (listbefig->BEBUX);
    beh_frebepor (listbefig->BEPOR);
    if ((ptptype = getptype (listbefig->USER,BEH_GENERIC)) != NULL)
      beh_frebegen (ptptype->DATA);

    ptbefig   = listbefig;
    listbefig = listbefig->NEXT;
    mbkfree (ptbefig);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebegen.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebegen						*/
/* description	: delete a list of BEGEN structures and all objects	*/
/*		  pointed by any os BEGENs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebegen (listbegen)

struct begen *listbegen;	/* list of begen to be deleted		*/

  {
  struct begen *ptbegen;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbegen != NULL)
    {
    mbkfree (listbegen->VALUE);

    ptbegen   = listbegen;
    listbegen = listbegen->NEXT;
    mbkfree (ptbegen);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebemsg.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebemsg						*/
/* description	: delete a list of BEMSG structures and all objects	*/
/*		  pointed by any os BEMSGs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebemsg (listbemsg)

struct bemsg *listbemsg;	/* list of bemsg to be deleted		*/

  {
  struct bemsg *ptbemsg;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbemsg != NULL)
    {
    freeExpr (listbemsg->ABL);

    ptbemsg   = listbemsg;
    listbemsg = listbemsg->NEXT;
    mbkfree (ptbemsg);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebeout.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebeout						*/
/* description	: delete a list of BEOUT structures and all objects	*/
/*		  pointed by any os BEOUTs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebeout (listbeout)

struct beout *listbeout;	/* list of beout to be deleted		*/

  {
  struct beout *ptbeout;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbeout != NULL)
    {
    freeExpr (listbeout->ABL);

    ptbeout   = listbeout;
    listbeout = listbeout->NEXT;
    mbkfree (ptbeout);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebepor.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebepor						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebepor (listbepor)

struct bepor *listbepor;	/* list of bepor to be deleted		*/

  {
  struct bepor *ptbepor;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, delete the object itself	*/
	/* ###------------------------------------------------------### */

  while (listbepor != NULL)
    {
    ptbepor   = listbepor;
    listbepor = listbepor->NEXT;
    mbkfree (ptbepor);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebereg.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebereg						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebereg (listbereg)

struct bereg *listbereg;	/* list of bereg to be deleted		*/

  {
  struct bereg *ptbereg;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbereg != NULL)
    {
    beh_frebiabl  (listbereg->BIABL);
    beh_frebinode (listbereg->BINODE);

    ptbereg   = listbereg;
    listbereg = listbereg->NEXT;
    mbkfree (ptbereg);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_freberin.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_freberin						*/
/* description	: delete a list of BEAUX structures and all objects	*/
/*		  pointed by any os BEAUXs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_freberin (listberin)

struct berin *listberin;	/* list of berin to be deleted		*/

  {
  struct berin *ptberin;

	/* ###------------------------------------------------------### */
	/*    for each object of the list delete the object itself	*/
	/* ###------------------------------------------------------### */

  while (listberin != NULL)
    {
    ptberin   = listberin;
    listberin = listberin->NEXT;
    mbkfree (ptberin);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebiabl.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebiabl						*/
/* description	: delete a list of BIABL structures and all objects	*/
/*		  pointed by any os BIABLs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebiabl (listbiabl)

struct biabl *listbiabl;	/* list of biabl to be deleted		*/

  {
  struct biabl *ptbiabl;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbiabl != NULL)
    {
    freeExpr (listbiabl->VALABL);
    freeExpr (listbiabl->CNDABL);
    freeptype(listbiabl->USER);

    ptbiabl   = listbiabl;
    listbiabl = listbiabl->NEXT;
    mbkfree (ptbiabl);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_frebinod.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_frebinode						*/
/* description	: delete a list of BINODE structures and all objects	*/
/*		  pointed by any os BINODEs in the list			*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

void beh_frebinode (listbinode)

struct binode *listbinode;	/* list of binode to be deleted		*/

  {
  struct binode *ptbinode;

	/* ###------------------------------------------------------### */
	/*    for each object of the list, first delete pointed objects	*/
	/* then, delete the object itself				*/
	/* ###------------------------------------------------------### */

  while (listbinode != NULL)
    {
    ptbinode   = listbinode;
    listbinode = listbinode->NEXT;
    mbkfree (ptbinode);
    }

  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_getgenva.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_getgenval						*/
/* description	: Search an find a BEGEN structure (known from its name)*/
/*		  then return its value field. If the BEGEN is not found*/
/*		  a NULL pointer is returned				*/
/* called func.	:							*/
/* ###--------------------------------------------------------------### */

void *beh_getgenval (listbegen, begenname)

struct begen *listbegen;	/* head of BEGEN list			*/
char         *begenname;	/* name of the structure 		*/

  {
  struct begen *ptbegen;
  void         *value = NULL;

	/* ###------------------------------------------------------### */
	/*    searching the object					*/
	/* ###------------------------------------------------------### */

  ptbegen = listbegen;
  while ((ptbegen != NULL) && (ptbegen->NAME != begenname))
    ptbegen = ptbegen->NEXT;

	/* ###------------------------------------------------------### */
	/*    if found return its value field				*/
	/* ###------------------------------------------------------### */

  if (ptbegen != NULL)
    value = ptbegen->VALUE;

  return (value);
  }

/* ###--------------------------------------------------------------### */
/* file		: beh_globals.c						*/
/* date		: Aug 27 1997						*/
/* version	: v110							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/* ###--------------------------------------------------------------### */


unsigned int   BEH_LINNUM     = 1         ;

unsigned char  BEH_TIMEUNIT   = BEH_TU__PS;

float          BEH_CNV_FS     = 1.0E-3    ;
float          BEH_CNV_PS     = 1.0       ;
float          BEH_CNV_NS     = 1.0E3     ;
float          BEH_CNV_US     = 1.0E6     ;
float          BEH_CNV_MS     = 1.0E9     ;

unsigned char  BEH_CIRCUI     = BEH_CIRCUI_EXPANDED ;

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_message.c						*/
/* date		: Oct 30 1995						*/
/* version	: v109							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_message						*/
/* description	: print a message on the standard output		*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void beh_message (code, str1)

int   code;
char *str1;

  {
  (void) fprintf (stdout, "BEH : ");

  switch (code)
    {
    case 3:
      (void) fprintf (stdout, "Compiling `%s` (Behaviour) ...\n", str1);
      break;
    case 13:
      (void) fprintf (stdout, "Saving '%s' in a vhdl file (vbe)\n", str1);
      break;
    case 14:
      (void) fprintf (stdout, "Builting '%s' ...\n", str1);
      break;
    default:
      (void) fprintf (stdout, "%s\n", str1);
    }
  }


/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbeaux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbeaux						*/
/* description	: delete a BEAUX structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, freeExpr					*/
/* ###--------------------------------------------------------------### */

struct beaux *beh_rmvbeaux (listbeaux, beauxname, mode)

struct beaux *listbeaux;	/* list of beaux containing the object	*/
char         *beauxname;	/* name of the BEAUX to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct beaux  headaux;
  struct beaux *ptlastaux;
  struct beaux *ptbeaux;

  if (listbeaux != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headaux.NEXT = listbeaux;
    headaux.NAME = NULL     ;
    headaux.ABL  = NULL     ;
    ptbeaux      = &headaux;
    while ((ptbeaux != NULL) && (ptbeaux->NAME != beauxname))
      {
      ptlastaux = ptbeaux;
      ptbeaux   = ptbeaux->NEXT;
      }

    if (ptbeaux != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbeaux->ABL != NULL)
          (void) fprintf (stderr,"BEH_warning : beaux `%s` not empty\n",
                          beauxname);
        }
      else
        freeExpr (ptbeaux->ABL);

      ptlastaux->NEXT = ptbeaux->NEXT;
      mbkfree (ptbeaux);
      }

    listbeaux = headaux.NEXT;
    }

  return(listbeaux);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbebus.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbebus						*/
/* description	: delete a BEBUS structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, beh_frebiabl, beh_frebinode			*/
/* ###--------------------------------------------------------------### */

struct bebus *beh_rmvbebus (listbebus, bebusname, mode)

struct bebus *listbebus;	/* list of bebus containing the object	*/
char         *bebusname;	/* name of the BEBUS to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bebus  headbus;
  struct bebus *ptlastbus;
  struct bebus *ptbebus;

  if (listbebus != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headbus.NEXT   = listbebus;
    headbus.NAME   = NULL     ;
    headbus.BIABL  = NULL     ;
    headbus.BINODE = NULL     ;
    ptbebus        = &headbus;
    while ((ptbebus != NULL) && (ptbebus->NAME != bebusname))
      {
      ptlastbus = ptbebus;
      ptbebus   = ptbebus->NEXT;
      }

    if (ptbebus != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbebus->BIABL != NULL) || (ptbebus->BINODE != NULL))
          (void) fprintf (stderr,"BEH_warning : bebus `%s` not empty\n",
                          bebusname);
        }
      else
        {
        beh_frebiabl  (ptbebus->BIABL);
        beh_frebinode (ptbebus->BINODE);
        }

      ptlastbus->NEXT = ptbebus->NEXT;
      mbkfree (ptbebus);
      }

    listbebus = headbus.NEXT;
    }

  return(listbebus);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbebux.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbebux						*/
/* description	: delete a BEBUX structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, beh_frebiabl, beh_frebinode			*/
/* ###--------------------------------------------------------------### */

struct bebux *beh_rmvbebux (listbebux, bebuxname, mode)

struct bebux *listbebux;	/* list of bebux containing the object	*/
char         *bebuxname;	/* name of the BEBUX to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bebux  headbux;
  struct bebux *ptlastbux;
  struct bebux *ptbebux;

  if (listbebux != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headbux.NEXT   = listbebux;
    headbux.NAME   = NULL     ;
    headbux.BIABL  = NULL     ;
    headbux.BINODE = NULL     ;
    ptbebux        = &headbux;
    while ((ptbebux != NULL) && (ptbebux->NAME != bebuxname))
      {
      ptlastbux = ptbebux;
      ptbebux   = ptbebux->NEXT;
      }

    if (ptbebux != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbebux->BIABL != NULL) || (ptbebux->BINODE != NULL))
          (void) fprintf (stderr,"BEH_warning : bebux `%s` not empty\n",
                          bebuxname);
        }
      else
        {
        beh_frebiabl  (ptbebux->BIABL);
        beh_frebinode (ptbebux->BINODE);
        }

      ptlastbux->NEXT = ptbebux->NEXT;
      mbkfree (ptbebux);
      }

    listbebux = headbux.NEXT;
    }

  return(listbebux);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbeder.c					*/
/* date		: Mar 22 1995						*/
/* version	: v100							*/
/* authors	: Laurent VUILLEMIN, Julien DUNOYER			*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbeder						*/
/* description	: delete a BERIN structure and return the pointer of	*/
/*		  the next object.					*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct beder *beh_rmvbeder (listbeder, bederindex)

struct beder *listbeder;	/* list of beder containing the object	*/
short         bederindex;	/* index of the BEDER to be deleted	*/

  {
  struct beder  headder;
  struct beder *ptlastder;
  struct beder *ptbeder;

  if (listbeder != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headder.NEXT    = listbeder;
    headder.INDEX=0;
    ptbeder      = &headder;
    while ((ptbeder != NULL) && (ptbeder->INDEX	 != bederindex))
      {
      ptlastder = ptbeder;
      ptbeder   = ptbeder->NEXT;
      }

    if (ptbeder != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/* ###------------------------------------------------------### */


      ptlastder->NEXT = ptbeder->NEXT;
      mbkfree (ptbeder);
      }

    listbeder = headder.NEXT;
    }

  return(listbeder);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbefig.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbefig						*/
/* description	: delete a BEFIG structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N.	*/
/* called func.	: mbkfree     , beh_frebereg, beh_frebemsg,		*/
/*		  beh_freberin, beh_frebeout, beh_frebebus,		*/
/*		  beh_frebeaux, beh_frebebux, beh_frebepor,		*/
/*		  beh_frebegen, getptype				*/
/* ###--------------------------------------------------------------### */

struct befig *beh_rmvbefig (listbefig, befigname, mode)

struct befig *listbefig;	/* list of befig containing the object	*/
char         *befigname;	/* name of the BEFIG to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct befig  headfig;
  struct befig *ptlastfig;
  struct befig *ptbefig;
  struct ptype *ptptype;

  if (listbefig != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headfig.NEXT  = listbefig;
    headfig.NAME  = NULL     ;
    headfig.BEREG = NULL     ;
    headfig.BEMSG = NULL     ;
    headfig.BERIN = NULL     ;
    headfig.BEAUX = NULL     ;
    headfig.BEBUX = NULL     ;
    headfig.BEDLY = NULL     ;
    headfig.BEPOR = NULL     ;
    headfig.USER  = NULL     ;
    ptbefig       = &headfig;
    while ((ptbefig != NULL) && (ptbefig->NAME != befigname))
      {
      ptlastfig = ptbefig;
      ptbefig   = ptbefig->NEXT;
      }

    if (ptbefig != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbefig->BEREG != NULL) || (ptbefig->BEMSG != NULL) ||
            (ptbefig->BERIN != NULL) || (ptbefig->BEOUT != NULL) ||
            (ptbefig->BEBUS != NULL) || (ptbefig->BEAUX != NULL) ||
            (ptbefig->BEBUX != NULL) || (ptbefig->BEDLY != NULL) ||
            (ptbefig->BEPOR != NULL) || (ptbefig->USER  != NULL))

          (void) fprintf (stderr,"BEH_warning : befig `%s` not empty\n",
                          befigname);
        }
      else
        {
        beh_frebereg (ptbefig->BEREG);
        beh_frebemsg (ptbefig->BEMSG);
        beh_freberin (ptbefig->BERIN);
        beh_frebeout (ptbefig->BEOUT);
        beh_frebebus (ptbefig->BEBUS);
        beh_frebeaux (ptbefig->BEAUX);
        beh_frebeaux (ptbefig->BEDLY);
        beh_frebebux (ptbefig->BEBUX);
        beh_frebepor (ptbefig->BEPOR);
        if ((ptptype = getptype (ptbefig->USER,BEH_GENERIC)) != NULL)
          beh_frebegen (ptptype->DATA);
        }

      ptlastfig->NEXT = ptbefig->NEXT;
      mbkfree (ptbefig);
      }

    listbefig = headfig.NEXT;
    }

  return(listbefig);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbegen.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbegen						*/
/* description	: delete a BEGEN structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct begen *beh_rmvbegen (listbegen, begenname, mode)

struct begen *listbegen;	/* list of begen containing the object	*/
char         *begenname;	/* name of the BEGEN to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct begen  headgen;
  struct begen *ptlastgen;
  struct begen *ptbegen;

  if (listbegen != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headgen.NEXT  = listbegen;
    headgen.NAME  = NULL     ;
    headgen.VALUE = NULL     ;
    ptbegen       = &headgen;
    while ((ptbegen != NULL) && (ptbegen->NAME != begenname))
      {
      ptlastgen = ptbegen;
      ptbegen   = ptbegen->NEXT;
      }

    if (ptbegen != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbegen->VALUE != NULL)
          (void) fprintf (stderr,"BEH_warning : begen `%s` not empty\n",
                          begenname);
        }
      else
        mbkfree (ptbegen->VALUE);

      ptlastgen->NEXT = ptbegen->NEXT;
      mbkfree (ptbegen);
      }

    listbegen = headgen.NEXT;
    }

  return(listbegen);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbemsg.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbemsg						*/
/* description	: delete a BEMSG structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, freeExpr					*/
/* ###--------------------------------------------------------------### */

struct bemsg *beh_rmvbemsg (listbemsg, bemsglabl, mode)

struct bemsg *listbemsg;	/* list of bemsg containing the object	*/
char         *bemsglabl;	/* label of the BEMSG to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bemsg  headmsg;
  struct bemsg *ptlastmsg;
  struct bemsg *ptbemsg;

  if (listbemsg != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headmsg.NEXT  = listbemsg;
    headmsg.LABEL = NULL     ;
    headmsg.ABL   = NULL     ;
    ptbemsg       = &headmsg;
    while ((ptbemsg != NULL) && (ptbemsg->LABEL != bemsglabl))
      {
      ptlastmsg = ptbemsg;
      ptbemsg   = ptbemsg->NEXT;
      }

    if (ptbemsg != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbemsg->ABL != NULL)
          (void) fprintf (stderr,"BEH_warning : bemsg `%s` not empty\n",
                          bemsglabl);
        }
      else
        freeExpr (ptbemsg->ABL);

      ptlastmsg->NEXT = ptbemsg->NEXT;
      mbkfree (ptbemsg);
      }

    listbemsg = headmsg.NEXT;
    }

  return(listbemsg);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbeout.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbeout						*/
/* description	: delete a BEOUT structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, freeExpr					*/
/* ###--------------------------------------------------------------### */

struct beout *beh_rmvbeout (listbeout, beoutname, mode)

struct beout *listbeout;	/* list of beout containing the object	*/
char         *beoutname;	/* name of the BEOUT to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct beout  headout;
  struct beout *ptlastout;
  struct beout *ptbeout;

  if (listbeout != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headout.NEXT = listbeout;
    headout.NAME = NULL     ;
    headout.ABL  = NULL     ;
    ptbeout      = &headout;
    while ((ptbeout != NULL) && (ptbeout->NAME != beoutname))
      {
      ptlastout = ptbeout;
      ptbeout   = ptbeout->NEXT;
      }

    if (ptbeout != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if (ptbeout->ABL != NULL)
          (void) fprintf (stderr,"BEH_warning : beout `%s` not empty\n",
                          beoutname);
        }
      else
        freeExpr (ptbeout->ABL);

      ptlastout->NEXT = ptbeout->NEXT;
      mbkfree (ptbeout);
      }

    listbeout = headout.NEXT;
    }

  return(listbeout);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbepor.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbepor						*/
/* description	: delete a BEPOR structure and return the pointer of	*/
/*		  the next object.					*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct bepor *beh_rmvbepor (listbepor, beporname)

struct bepor *listbepor;	/* list of bepor containing the object	*/
char         *beporname;	/* name of the BEPOR to be deleted	*/

  {
  struct bepor  headpor;
  struct bepor *ptlastpor;
  struct bepor *ptbepor;

  if (listbepor != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headpor.NEXT = listbepor;
    headpor.NAME = NULL     ;
    ptbepor      = &headpor;
    while ((ptbepor != NULL) && (ptbepor->NAME != beporname))
      {
      ptlastpor = ptbepor;
      ptbepor   = ptbepor->NEXT;
      }

    if (ptbepor != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/* ###------------------------------------------------------### */

      ptlastpor->NEXT = ptbepor->NEXT;
      mbkfree (ptbepor);
      }

    listbepor = headpor.NEXT;
    }

  return(listbepor);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvbereg.c					*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvbereg						*/
/* description	: delete a BEREG structure and return the pointer of	*/
/*		  the next object. A warning is printed out if the	*/
/*		  object to be deleted is not empty when the mode is N	*/
/* called func.	: mbkfree, beh_frebiabl, beh_frebinode			*/
/* ###--------------------------------------------------------------### */

struct bereg *beh_rmvbereg (listbereg, beregname, mode)

struct bereg *listbereg;	/* list of bereg containing the object	*/
char         *beregname;	/* name of the BEREG to be deleted	*/
char          mode;		/* recursive delete or not (Y or N)	*/

  {
  struct bereg  headreg;
  struct bereg *ptlastreg;
  struct bereg *ptbereg;

  if (listbereg != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headreg.NEXT   = listbereg;
    headreg.NAME   = NULL     ;
    headreg.BIABL  = NULL     ;
    headreg.BINODE = NULL     ;
    ptbereg        = &headreg;
    while ((ptbereg != NULL) && (ptbereg->NAME != beregname))
      {
      ptlastreg = ptbereg;
      ptbereg   = ptbereg->NEXT;
      }

    if (ptbereg != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/*    If the object has been found check the mode and, if asked	*/
	/* delete pointed objects recursively.				*/
	/* ###------------------------------------------------------### */

      if (mode == 'N')
        {
        if ((ptbereg->BIABL != NULL) || (ptbereg->BINODE != NULL))
          (void) fprintf (stderr,"BEH_warning : bereg `%s` not empty\n",
                          beregname);
        }
      else
        {
        beh_frebiabl  (ptbereg->BIABL);
        beh_frebinode (ptbereg->BINODE);
        }

      ptlastreg->NEXT = ptbereg->NEXT;
      mbkfree (ptbereg);
      }

    listbereg = headreg.NEXT;
    }

  return(listbereg);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_rmvberin.c					*/
/* date		: Sep 20 1994						*/
/* version	: v107							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_rmvberin						*/
/* description	: delete a BERIN structure and return the pointer of	*/
/*		  the next object.					*/
/* called func.	: mbkfree						*/
/* ###--------------------------------------------------------------### */

struct berin *beh_rmvberin (listberin, berinname)

struct berin *listberin;	/* list of berin containing the object	*/
char         *berinname;	/* name of the BERIN to be deleted	*/

  {
  struct berin  headrin;
  struct berin *ptlastrin;
  struct berin *ptberin;

  if (listberin != NULL)
    {

	/* ###------------------------------------------------------### */
	/*    Search the object to be deleted				*/
	/* ###------------------------------------------------------### */

    headrin.NEXT    = listberin;
    headrin.NAME    = NULL;
    headrin.OUT_REF = NULL;
    headrin.MSG_REF = NULL;
    headrin.AUX_REF = NULL;
    headrin.BUX_REF = NULL;
    headrin.BUS_REF = NULL;
    headrin.REG_REF = NULL;
    headrin.DLY_REF = NULL;
    ptberin      = &headrin;
    while ((ptberin != NULL) && (ptberin->NAME != berinname))
      {
      ptlastrin = ptberin;
      ptberin   = ptberin->NEXT;
      }

    if (ptberin != NULL)
      {

	/* ###------------------------------------------------------### */
	/*    If the object doesn't exist return the list without	*/
	/* modification.						*/
	/* ###------------------------------------------------------### */

      freechain (ptberin->OUT_REF);
      freechain (ptberin->MSG_REF);
      freechain (ptberin->AUX_REF);
      freechain (ptberin->BUX_REF);
      freechain (ptberin->BUS_REF);
      freechain (ptberin->REG_REF);
      freechain (ptberin->DLY_REF);

      ptlastrin->NEXT = ptberin->NEXT;
      mbkfree (ptberin);
      }

    listberin = headrin.NEXT;
    }

  return(listberin);
  }

/* ###--------------------------------------------------------------### */
/* 									*/
/* file		: beh_toolbug.c						*/
/* date		: Sep  3 1993						*/
/* version	: v106							*/
/* authors	: Pirouz BAZARGAN SABET					*/
/* content	: low-level function					*/
/*									*/
/* ###--------------------------------------------------------------### */


/* ###--------------------------------------------------------------### */
/* function	: beh_toolbug						*/
/* description	: print an error message on the standard error output	*/
/* called func.	: none							*/
/* ###--------------------------------------------------------------### */

void beh_toolbug (code, str1, str2, nbr1)

int   code;
char *str1;
char *str2;
int   nbr1;

  {
  fprintf (stderr,"Fatal error %d executing `%s`: ", code, str1);
  switch (code)
    {
    case 1:
      fprintf (stderr,"unknown operator\n");
      break;
    case 2:
      fprintf (stderr,"cannot create empty atom\n");
      break;
    case 3:
      fprintf (stderr,"cannot build NOT of empty expression\n");
      break;
    case 4:
      fprintf (stderr,"cannot combine empty expressions\n");
      break;
    case 5:
      fprintf (stderr,"cannot find terminal\n");
      break;
    case 10:
      fprintf (stderr,"decompiler called on empty figure\n");
      break;
    case 15 :
      fprintf (stderr,"illegal bit string value : `%c`\n",nbr1);
      break;
    case 16 :
      fprintf (stderr,"the same expression cannot be used twice\n");
      break;
    case 19:
      fprintf (stderr,"empty guard expression: `%s`\n",str2);
      break;
    case 20:
      fprintf (stderr,"empty waveform expression: `%s`\n",str2);
      break;
    case 100 :
      fprintf (stderr, "illegal use of attribute STABLE\n");
      break;
    case 101 :
      fprintf (stderr, "unknown terminal operand `%s`\n", str2);
      break;
    case 102 :
      fprintf (stderr, "unknown operator `%d`\n", nbr1);
      break;
    case 103 :
      fprintf (stderr, "empty expression\n");
      break;
    default:
      fprintf (stderr, "BUG\n");
    }
  EXIT (1);
  }
